package de.lmu.ifi.pms.cep.compilation;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.java.util.collection.array.Arrays;



public class StackCompiler<SourceProgramType, DestinationProgramType> extends Compiler<SourceProgramType, DestinationProgramType>{
	
	protected final Compiler<?, ?>[] compilerStack;

	public StackCompiler(final Class<SourceProgramType> sourceProgramType, final Class<DestinationProgramType> destinationProgramType, final Compiler<?, ?>... compilerStack) {
		super(sourceProgramType, destinationProgramType);
		this.compilerStack = Arrays.clone(compilerStack);
		if(compilerStack.length <= 0){
			if(!destinationProgramType.isAssignableFrom(sourceProgramType)){
				throw new IllegalArgumentException("Empty compiler stack, but source type does not correspond to destination type.");
			}
		}
		else{
			if(this.compilerStack[0].sourceProgramType.isAssignableFrom(sourceProgramType)){
				if(this.destinationProgramType.isAssignableFrom(this.compilerStack[this.compilerStack.length - 1].destinationProgramType)){
					for(int compiler = 1; compiler < this.compilerStack.length; ++compiler){
						if(!this.compilerStack[compiler].sourceProgramType.isAssignableFrom(this.compilerStack[compiler - 1].destinationProgramType)){
							throw new IllegalArgumentException("Destination type of compiler " + (compiler - 1) + "does not correspond to source type of compiler " + compiler + ".");
						}
					}
				}
				else{
					throw new IllegalArgumentException("Last compiler has invalid destination type.");
				}
			}
			else{
				throw new IllegalArgumentException("First compiler has invalid source type.");
			}
		}
	}

	@SuppressWarnings({ "unchecked", "rawtypes" })
	@Override
	public DestinationProgramType compile(final SourceProgramType sourceProgram){
		Object destinationProgram = sourceProgram;
		for(int compiler = 0; compiler < this.compilerStack.length; ++compiler){
			destinationProgram = ((Compiler)this.compilerStack[compiler]).compile(destinationProgram);
		}
		assert this.destinationProgramType.isAssignableFrom(destinationProgram.getClass());
		return (DestinationProgramType) destinationProgram;
	}
	
	protected static final Logger logger = LoggerFactory.getLogger(StackCompiler.class);
	
	@Override
	protected Logger logger(){
		return logger;
	}
}
