grammar Dura;

options {
  output=AST;
}

tokens {
  PROGRAM;
  PACKAGE;
  PACKAGE_NAME;
  DESCRIPTION;
  PREAMBLE;
  DEFINITIONS;
  
  SCHEMA;
  TUPLE_CONSTRAINTS;
  RELATION_CONSTRAINTS;
  
  TYPEDEF;
  CONSTDEF;
  MODIFIERS;
  EVENT;
  STATEFUL_OBJECT;
  ACTION;

  EVENT_QUERY;
  STATE_QUERY;
  
  FOR;
  ACTION_EXECUTION;
  EXECUTION_CONSTRAINTS;
  SUCCEEDS_ON;
  FAILS_ON;
  
  QUERY;
  HEAD;
  BODY;
   
  IDENTIFIERS;

  SUPPLEMENT;
  AGGREGATION;
  GROUPING;
  CONDITION;
  
  DETECT;
  DO;
  LET;
  
  CONJUNCTION;
  DISJUNCTION;
  CONCURRENT;
  NEGATION;
  EXISTS;

  ARITHMETIC_OP;
  INTERVAL_OP;
  AGGREGATION_OP;
  RELATIVE_TIMER_OP;
  CLOCK_FUNCTION;
  
  TIME_RELATION;
  ARITHMETIC_RELATION;
  UNIFICATION;
  
  PATH;
  TERM;
  VARIABLE;
  TYPED_VARIABLE;
  IDENTIFIER;
  CONSTANT;
  CONSTRAINT;

  BASIC_TYPE;
  COMPOSITE_TYPE;
  TIME_UNIT;
  DURATION;
  
  ANNOTATION;
  
  
  /* DuraC */
  
  PROGRAM;
  MODULE;
  MODULE_NAME;
  LOAD;
  IMPORT;
  EXPORT;
  DESCRIPTION;
  PREAMBLE;
  DEFINITIONS;
  
  SCHEMA;
  TUPLE_CONSTRAINTS;
  RELATION_CONSTRAINTS;
  
  TYPEDEF;
  CONSTDEF;
  MODIFIERS;
  EVENT;
  STATEFUL_OBJECT;
  ACTION;

  EVENT_QUERY;
  STATE_QUERY;
  
  FOR;
  ACTION_EXECUTION;
  EXECUTION_CONSTRAINTS;
  SUCCEEDS_ON;
  FAILS_ON;
  
  QUERY;
  HEAD;
  BODY;
   
  IDENTIFIERS;

  SUPPLEMENT;
  AGGREGATION;
  GROUPING;
  CONDITION;
  
  DETECT;
  DO;
  LET;
  
  CONJUNCTION;
  DISJUNCTION;
  CONCURRENT;
  NEGATION;
  EXISTS;

  ARITHMETIC_OP;
  INTERVAL_OP;
  AGGREGATION_OP;
  RELATIVE_TIMER_OP;
  CLOCK_FUNCTION;
  
  TIME_RELATION;
  ARITHMETIC_RELATION;
  UNIFICATION;
  
  PATH;
  TERM;
  VARIABLE;
  TYPED_VARIABLE;
  IDENTIFIER;
  CONSTANT;
  CONSTRAINT;

  BASIC_TYPE;
  COMPOSITE_TYPE;
  TIME_UNIT;
  DURATION;
  BOOLEAN;
  
  ANNOTATION;
}

@header {
package de.lmu.ifi.pms.cep.dura.compiler.antlr;
}

@lexer::header {
package de.lmu.ifi.pms.cep.dura.compiler.antlr;
}

program
  :  preamble (def+=eventDefinition | def+=stateDefinition | def+=actionDefinition | def+=reactiveRule)+
  ->  ^(PROGRAM ^(PACKAGE preamble ^(DEFINITIONS $def+)))
  ;

preamble
  : pckg description (constDefinition | typeDefinition)*
  ->  ^(PREAMBLE pckg description constDefinition* typeDefinition*)
  ;

pckg
  : ('PACKAGE' path)?
  ->  ^(PACKAGE_NAME path?)
  ;
  
description
  : ('DESCRIPTION' STRING)?
  ->  ^(DESCRIPTION STRING?)
  ;

/** 
 * constraints: 
 *   all schemaDefinition labels are pairwise distinct
 *   type name ID is unique
 *   no cyclic definitions
 */
typeDefinition 
  : 'TYPE' ID 'IS' basicType typeSupplement 'END'
  ->  ^(TYPEDEF[$ID] basicType typeSupplement)
  | 'TYPE' ID 'IS' schemaDefinition typeSupplement 'END'
  ->  ^(TYPEDEF[$ID] schemaDefinition typeSupplement)
  | 'TYPE' ID 'IS' '{' schemaDefinition (',' schemaDefinition)* '}' typeSupplement 'END'  
  ->  ^(TYPEDEF[$ID] schemaDefinition+ typeSupplement)
  ;

/** constraint: const name ID is unique */
constDefinition
  : 'CONST' ID 'IS' constTerm 'END'      ->  ^(CONSTDEF[$ID] constTerm)
  ;

/**
 * constraints:
 *   event definition schemaDefinition needs to be unique
 *   schema of derived events need to comply with schmaDenfinition
 */
eventDefinition
  : modifiers 'EVENT' schemaDefinition schemaSupplement ('WITH' eventSpecification*)? 'END'
  ->  ^(EVENT modifiers ^(SCHEMA schemaDefinition schemaSupplement) eventSpecification*)
  ;

/** constraint: stateful object definition schemaDefinition is unique */
stateDefinition
  : modifiers 'STATEFUL OBJECT' schemaDefinition schemaSupplement 'END'
  ->  ^(STATEFUL_OBJECT modifiers ^(SCHEMA schemaDefinition schemaSupplement))
  ;

/** constraint: action definition schemaDefinition is unique */
actionDefinition
  : modifiers 'ACTION' schemaDefinition schemaSupplement ((succeedsOn failsOn?) | (failsOn succeedsOn?))? ('WITH' actionSpecification)? 'END'
  ->  ^(ACTION modifiers ^(SCHEMA schemaDefinition schemaSupplement) ^(SUCCEEDS_ON succeedsOn?) ^(FAILS_ON failsOn?) actionSpecification?)
  ;
  
actionSpecification
  : 'FOR' i='action' ID? ':' term 'DO' actionComposition 'END'
  -> ^(FOR ^(BODY ^(IDENTIFIER[$i] ID?) term) ^(HEAD actionComposition))
  ;
  
actionComposition
  : 'concurrent' '{' flatAction (',' flatAction)* '}'
    (
      (executionConstraints ((succeedsOn failsOn?) | (failsOn succeedsOn?))?)
    | (succeedsOn ((failsOn executionConstraints?) | (executionConstraints failsOn?))?)
    | (failsOn ((succeedsOn executionConstraints?) | (executionConstraints succeedsOn?))?)
    )?
  -> ^(CONCURRENT flatAction+ ^(EXECUTION_CONSTRAINTS executionConstraints?) ^(SUCCEEDS_ON succeedsOn?) ^(FAILS_ON failsOn?))
  | flatAction
  -> ^(CONCURRENT flatAction ^(EXECUTION_CONSTRAINTS) ^(SUCCEEDS_ON) ^(FAILS_ON))
  ;
  
executionConstraints
  : 'where' '{' atomicConditionFormula (',' atomicConditionFormula)* '}'
  -> atomicConditionFormula+
  ;
  
succeedsOn
  : 'succeeds on' '{' eventQuery '}'
  -> eventQuery
  ;
  
failsOn
  : 'fails on' '{' eventQuery '}'
  -> eventQuery
  ;
  
flatAction
  : i='action' ID? ':' constructTerm
  -> ^(ACTION_EXECUTION ^(IDENTIFIER[$i] ID?) constructTerm)
  | i='action' ID? ':' actionComposition
  -> ^(ACTION_EXECUTION ^(IDENTIFIER[$i] ID?) actionComposition)
  ;

/** constraint: query is range restricted with respect to term */
eventSpecification
  : 'DETECT' constructTerm ('group by' '{' binding (',' binding)* '}')? 'ON' eventQuery 'END'
  ->  ^(DETECT ^(HEAD constructTerm (^(AGGREGATION ^(GROUPING binding+)))?) ^(BODY eventQuery))
  ;

/** constraint: action is range restricted with respect to query */
// reactiveRule
  // : 'ON' eventQuery 'DO' action 'END'
  // ->  ^(DO ^(BODY eventQuery) ^(HEAD action))
  // ; 
reactiveRule
  : 'ON' eventQuery 'DO' limitedActionComposition 'END'
  ->  ^(DO ^(BODY eventQuery) ^(HEAD limitedActionComposition))
  ;
  
limitedActionComposition
  : 'concurrent' '{' flatAction (',' flatAction)* '}' executionConstraints?
  -> ^(CONCURRENT flatAction+ ^(EXECUTION_CONSTRAINTS executionConstraints?) ^(SUCCEEDS_ON) ^(FAILS_ON))
  | flatAction
  -> ^(CONCURRENT flatAction ^(EXECUTION_CONSTRAINTS) ^(SUCCEEDS_ON) ^(FAILS_ON))
  ;

/** constaint: at least one event query in every disjunct */
eventQuery
  : 'and' '{' flatEventQuery (',' flatEventQuery)* '}' querySupplement ->  ^(CONJUNCTION flatEventQuery+ querySupplement)
  | 'or' '{' flatEventQuery (',' flatEventQuery)* '}' querySupplement  ->  ^(DISJUNCTION flatEventQuery+ querySupplement)
  | flatEventQuery                                                     ->  ^(CONJUNCTION flatEventQuery ^(SUPPLEMENT))
  ;

flatEventQuery
  : 'not' '{' atomicEventQuery '}' querySupplement     ->  ^(NEGATION atomicEventQuery querySupplement)
  | 'not' atomicEventQuery                             ->  ^(NEGATION atomicEventQuery ^(SUPPLEMENT))
  | 'exists' '{' atomicEventQuery '}' querySupplement  ->  ^(EXISTS atomicEventQuery querySupplement)
  | 'exists' atomicEventQuery                          ->  ^(EXISTS atomicEventQuery ^(SUPPLEMENT))
  | atomicEventQuery
  ;

complexSubquery
  : 'and' '{' flatEventQuery (',' flatEventQuery)* '}' querySupplement ->  ^(CONJUNCTION flatEventQuery+ querySupplement)
  | 'or' '{' flatEventQuery (',' flatEventQuery)* '}' querySupplement  ->  ^(DISJUNCTION flatEventQuery+ querySupplement)
  ;
  
atomicEventQuery
  : i='event' j=ID? ':' (k='action' l=ID? '$') term querySupplement ->  ^(EVENT_QUERY ^(IDENTIFIER[$i] $j?) ^(QUERY ^(IDENTIFIER[$k] $l?) term) querySupplement)
  | i='event' j=ID? ':' term querySupplement ->  ^(EVENT_QUERY ^(IDENTIFIER[$i] $j?) ^(QUERY term) querySupplement)
  | i='state' ID? ':' term querySupplement ->  ^(STATE_QUERY ^(IDENTIFIER[$i] ID?) ^(QUERY term) querySupplement)
  
  // We don't need the SUPPLEMENT node, but for consistency and for not getting tempted to use its lack as criterion 
  // for a complex subquery (if we need one we will just use special tokens) it is included.
  | i='event' ID? ':' complexSubquery ->  ^(EVENT_QUERY ^(IDENTIFIER[$i] ID?) ^(QUERY complexSubquery) ^(SUPPLEMENT))
  ;

action
  : 'concurrent' '{' atomicAction (',' atomicAction)* '}' ->  ^(CONCURRENT atomicAction+)
  | atomicAction                                          ->  ^(CONCURRENT atomicAction+)
  ;

/** constraint: action term is actually defined */  
atomicAction
  : 'action' ID? ':' constructTerm ->  ^(ACTION_EXECUTION constructTerm)
  ;

querySupplement
  : (supplement+=where | supplement+=let | supplement+=grouping)*
  ->  ^(SUPPLEMENT $supplement*)
  ;

actionSupplement
  : where*           // where is too generic here, only temporal relation can be stated in the where part of an expression. var X - var Y <= 0 may be a kind of temporal relation for event queries, but not for action definitions
  ->  ^(SUPPLEMENT where*)
  ;

typeSupplement
  : tupleConstraints?
  -> ^(SUPPLEMENT tupleConstraints?)
  ;

schemaSupplement
  : tupleConstraints? relationConstraints?
  ->  ^(SUPPLEMENT tupleConstraints? relationConstraints?)
  ;

let
  : 'let' '{' unification (',' unification)* '}'
  ->  ^(LET unification+)
  ;

unification
  : typedDataVariable '=' expr
  -> ^(UNIFICATION typedDataVariable expr)
  ;

grouping
  : 'group by' '{' binding (',' binding)* '}'
  ->  ^(AGGREGATION ^(GROUPING binding+))
  
  | 'group by' '{' binding (',' binding)* '}' 'aggregate' '{' aggregation (',' aggregation)* '}'
  ->  ^(AGGREGATION ^(GROUPING binding+) aggregation+)
  ;

binding
  : dataVariable
  | identifier
  ;

aggregation
  : typedDataVariable '=' aggregationOp '(' dataVariable ')'
  -> ^(UNIFICATION typedDataVariable ^(aggregationOp dataVariable))
  ;

/** constraint: no variables occur in mathFormula */    
tupleConstraints
  : 'where' '{' atomicMathFormula (',' atomicMathFormula)* '}'
  ->  ^(TUPLE_CONSTRAINTS atomicMathFormula+)
  ;

relationConstraints
  : 'constraints' '{' relationConstraintFormula (',' relationConstraintFormula)* '}' 
  -> ^(RELATION_CONSTRAINTS relationConstraintFormula+)
  ;

relationConstraintFormula
  : 'and' '{' relationConstraintFormula (',' relationConstraintFormula)* '}'  ->  ^(CONJUNCTION relationConstraintFormula+)
  | 'or' '{'relationConstraintFormula (',' relationConstraintFormula)* '}'    ->  ^(DISJUNCTION relationConstraintFormula+)
  | relationConstraint
  ;

relationConstraint
  : (c='bound' | c='primary-key') '(' path ')'  ->  ^(CONSTRAINT[$c] path)
  | (c='foreign-key') '(' path ',' path ')'     ->  ^(CONSTRAINT[$c] path+) 
  ;

where
  : 'where' 'and' '{' conditionFormula (',' conditionFormula)* '}'  -> ^(CONDITION ^(CONJUNCTION conditionFormula+))
  | 'where' 'or' '{' conditionFormula (',' conditionFormula)* '}'   -> ^(CONDITION ^(DISJUNCTION conditionFormula+))  
  | 'where' '{' conditionFormula (',' conditionFormula)* '}'        -> ^(CONDITION ^(CONJUNCTION conditionFormula+))
  ;

conditionFormula
  : 'and' '{' conditionFormula (',' conditionFormula)* '}'  ->  ^(CONJUNCTION conditionFormula+)
  | 'or' '{'conditionFormula (',' conditionFormula)* '}'    ->  ^(DISJUNCTION conditionFormula+)
  | atomicConditionFormula
  ;
  
atomicConditionFormula
  :  atomicIntervalFormula
  |  atomicMathFormula
  ;

/** constraints: 
  *   either both or none of the expressions is of type duration 
  */
atomicMathFormula 
  : expr arithmeticRelation expr ->  ^(arithmeticRelation expr expr)
  ;

atomicIntervalFormula
  : '{' timeInterval ',' timeInterval 
    (                     '}' op='apart-by' duration
    | (',' timeInterval)* '}' op='within' duration
    )
  ->  ^(TIME_RELATION[$op] ^(IDENTIFIERS timeInterval+) duration)
  
  | timeInterval
    ( intervalRelation timeInterval ->  ^(intervalRelation timeInterval timeInterval)
    | timepointRelation timePoint   ->  ^(timepointRelation timeInterval timePoint)   
    )
  ;

/** constraint: variable is of type timeinterval */
timeInterval
  : dataVariable
  | identifier
  | relativeTimerOp '(' timeInterval ',' duration ')'  ->  ^(relativeTimerOp timeInterval duration)
  ;
  
/** constraint: variable is of type timepoint */
timePoint
  : dataVariable
  | intervalOp '(' timeInterval ')'  ->  ^(intervalOp timeInterval) 
  ;

/*  If you change this rule remember to change the atom rule as well.
    For details see comments in atom rule.
*/
/** constraint: all timeunits are are pairwise distinct */
duration
  : time+
  ->  ^(DURATION time+)
  ;

time
  : NUMBER timeUnit
  ->  ^(timeUnit NUMBER)
  ;

/** 
 * constraints: 
 *   all term labels are pairwise distinct
 *   no cyclic definitions
 */
term
  :  label '{' '}'                  ->  label
  |  label '{' termLeaf '}'         ->  ^(label termLeaf)
  |  label '{' term (',' term)* '}' ->  ^(label term+)
  ;

/** constraints: 
 *    variables are unified with basic types
 *    constants are unified with basic types
 *    identifiers are unified with identifier types
 */ 
termLeaf
  : dataVariable
  | identifier
  | constant
  | duration
  | STRING
  | NUMBER
  ;

/**
 * constraints:
 *   all constTerm labels are pairwise distinct
 *   no cyclic definitions
 */
constTerm
  : STRING
  | NUMBER
  | constant
  | duration
  | label '{' '}'                             ->  label
  | label '{' constTerm (',' constTerm)* '}'  ->  ^(label constTerm+)
  ;


internalSchemaDefinition
  : internalLabel '{' '}'                                                           ->  internalLabel
  | internalLabel '{' basicType '}'                                                 ->  ^(internalLabel basicType)
  | internalLabel '{' compositeType '}'                                             ->  ^(internalLabel compositeType)
  | internalLabel '{' internalSchemaDefinition (',' internalSchemaDefinition)* '}'  ->  ^(internalLabel internalSchemaDefinition+)
  ;
  
internalLabel
  : aggregationOp  ->  TERM[$aggregationOp.text] 
  | intervalOp     ->  TERM[$intervalOp.text]
  | ID
  ->  TERM[$ID]
  | ID d='$' (s='initiated' | s='failed' | s='succeeded' )
  ->  TERM[$ID.text + $d.text + $s.text]
  | INTERNAL_ID
  ->  TERM[$INTERNAL_ID]
  | INTERNAL_ID d='$' (s='initiated' | s='failed' | s='succeeded' )
  ->  TERM[$INTERNAL_ID.text + $d.text + $s.text]
  | (i='action' | i='event' | i='state' | i='succeeded' | i='failed' | i='initiated')
  ->  TERM[$i]  // This "list" of allowed tokens is not complete.
  ;
  
INTERNAL_ID
  : '_' ('a'..'z' | 'A'..'Z') ('a'..'z' | 'A'..'Z' | '0'..'9' | '-' | '_')*
  ;

/** 
 * constraints: 
 *   all schemaDefinition labels are pairwise distinct
 *   no cyclic definitions
 */
schemaDefinition
  : label '{' '}'                                           ->  label
  | label '{' basicType '}'                                 ->  ^(label basicType)
  | label '{' compositeType '}'                             ->  ^(label compositeType)
  | label '{' schemaDefinition (',' schemaDefinition)* '}'  ->  ^(label schemaDefinition+)
  ;

constructTerm
  :  label '{' '}'                                    ->  label
  |  label '{' expr '}'                               ->  ^(label expr)
  |  label '{' constructTerm (',' constructTerm)* '}' ->  ^(label constructTerm+)
  ;
  
modifiers
  : modifier*
  ->  ^(MODIFIERS modifier*)
  ;

modifier
  : 'input'
  | 'output'
  | 'log'
  ;
  
expr  
  : mathExpr
  | identifier
  ;
    
mathExpr
  : (multExpr -> multExpr)
    ( (op='+' | op='-') multExpr
      ->  ^(ARITHMETIC_OP[$op] $mathExpr multExpr)
    )*
  ;
  
multExpr
  : (powExpr -> powExpr)
    ( (op='*' | op='/') powExpr
      ->  ^(ARITHMETIC_OP[$op] $multExpr powExpr)
    )*
  ;

powExpr
  : (atom -> atom)
    ( op='^' atom
      ->  ^(ARITHMETIC_OP[$op] $powExpr atom)
    )*
  ;

/**
 * constraints:
 *   constants used in aritmetic expressions need to be a number
 */
atom
  : '(' mathExpr ')'                
  ->  mathExpr
  | (op='least' | op='greatest') '(' mathExpr (',' mathExpr)* ')'  
  ->  ^(ARITHMETIC_OP[$op] mathExpr+)
  | intervalOp '(' timeInterval ')' 
  ->  ^(intervalOp timeInterval)
  | clockFunction
  | dataVariable
  | constant
  | op='-' atom
  -> ^(ARITHMETIC_OP[$op] atom)
  | aggregationOp '(' dataVariable ')'
  -> ^(aggregationOp dataVariable)
  | path
  
  /*  Here it would be more desirable to have two rules, one for NUMBER one for duration.
      However this causes parsing issues that only global backtracking seems to solve.
      
      Neither predicates nor local backtracking seem to help.
      
      To avoid backtracking these two rules have been collapsed into one.
      Keep this in mind when changing either this or the duration rule.
  */
  | (NUMBER ->  NUMBER)
    ( timeUnit time*
      ->  ^(DURATION ^(timeUnit NUMBER) time*)
    )?
    
  | STRING
  ;

path
  : label ('.' label)*  ->  ^(PATH label+)
  ;

clockFunction
  : 'system' '.' (fn='now' | fn='tuple_id_seq') '(' ')'  ->  CLOCK_FUNCTION[$fn]
  ;

label
  : aggregationOp     ->  TERM[$aggregationOp.text] 
  | intervalOp        ->  TERM[$intervalOp.text]
  | ID
  ->  TERM[$ID]
  | ID d='$' (s='initiated' | s='failed' | s='succeeded' )
  ->  TERM[$ID.text + $d.text + $s.text]
  | (i='action' | i='event' | i='state' | i='succeeded' | i='failed' | i='initiated')
  ->  TERM[$i]  // This "list" of allowed tokens is not complete.
  ;

type
  : basicType
  | compositeType
  ;

basicType 
  : (t='string' | t='int' | t='long' | t='double' | t='float' | t='boolean' | t='identifier' | t='timestamp' | t='duration')  
  ->  BASIC_TYPE[$t]
  ;

/** constraint: composite type ID is actually defined */
compositeType 
  : ID
  ->  COMPOSITE_TYPE[$ID]
  ;

identifier
  : (i='event' | i='state' | i='action') ID
  ->  ^(IDENTIFIER[$i] ID)
  ;

dataVariable
  : 'var' ID
  ->  ^(VARIABLE ID)
  ;

typedDataVariable
  : 'var' type ID
  ->  ^(TYPED_VARIABLE type ID)
  ;

/** constraint: constant ID is actually defined */
constant
  : 'const' ID
  ->  ^(CONSTANT ID)
  ;

intervalRelation
  : (rel='before' | rel='contains' | rel='overlaps' | rel='after' | rel='during' | rel='overlapped-by'
      | rel='starts' | rel='finishes' | rel='meets' | rel='started-by' | rel='finished-by'
      | rel='met-by' | rel='equals' | rel='while' | rel='valid-during')
  ->  TIME_RELATION[$rel]
  ;

timepointRelation
  : rel='valid-at'  -> TIME_RELATION[$rel]
  ;

relativeTimerOp 
  : (op='extend' | op='shorten' | op='extend-begin' | op='shorten-begin' | op='shift-forward'
      | op='shift-backward' | op='from-end' | op='from-end-backward' | op='from-start'
      | op='from-start-backward' | op='from-begin' |  op='from-begin-backward')
  ->  RELATIVE_TIMER_OP[$op]
  ;

intervalOp 
  : (op='begin' | op='end')
  ->  INTERVAL_OP[$op]
  ;

aggregationOp 
  : (op='max' | op='min' | op='mean' | op='avg' | op='count')
  ->  AGGREGATION_OP[$op]
  ;

timeUnit
  : (unit='day' | unit= 'days' | unit='hour' | unit='hours' | unit='min' | unit='sec' | unit='ms')  
  ->  TIME_UNIT[$unit]
 ;

arithmeticRelation 
  : (rel='<' | rel='<=' | rel='=' | rel='!=' | rel='>' | rel='>=')
  ->  ARITHMETIC_RELATION[$rel]
  ;


/* Tokens */

ID
  : ('a'..'z' | 'A'..'Z') ('a'..'z' | 'A'..'Z' | '0'..'9' | '-' | '_')*
  ;

NUMBER
  : ('0'..'9')+ '.' ('0'..'9')* Exponent?
  | ('0'..'9')+ Exponent?
  ;

fragment
Exponent
  : ('e' | 'E') ('+' | '-')? ('0'..'9')+ 
  ;

STRING 
  : '"' (EscapeSequence | ~('\\' | '"'))* '"'
  ;

fragment
EscapeSequence
  : '\\' ('b' | 't' | 'n' | 'f' | 'r' | '\"' | '\'' | '\\')
  ;

COMMENT
  : '//' ~('\n' | '\r')* '\r'? '\n' {$channel=HIDDEN;}
  | '/*' (options {greedy=false;} : .)* '*/' {$channel=HIDDEN;}
  ;

WS
  : (' ' | '\t' | '\r' | '\n') {$channel=HIDDEN;}
  ;