package de.lmu.ifi.pms.cep.dura.compiler.dura;

import java.io.File;
import java.io.IOException;

import org.antlr.runtime.ANTLRFileStream;
import org.antlr.runtime.CommonTokenStream;
import org.antlr.runtime.RecognitionException;
import org.antlr.runtime.tree.CommonTree;
import org.antlr.runtime.tree.CommonTreeNodeStream;

import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraLexer;
import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraParser;
import de.lmu.ifi.pms.cep.dura.compiler.dura.desugaring.Desugarer;
import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCAstValidation;


public class DuraCompiler {
	public static CommonTree parse(final ANTLRFileStream stream) throws RecognitionException {
		DuraLexer lexer = new DuraLexer(stream);
		CommonTokenStream tokens = new CommonTokenStream(lexer);
		DuraParser parser = new DuraParser(tokens);
		
		return (CommonTree) parser.program().getTree();
	}
	
	public static void compile(String filename) throws IOException, RecognitionException {
		ANTLRFileStream stream = new ANTLRFileStream(filename);
		CommonTree tree = parse(stream);
		compile(tree);
	}
	
	public static void compile(CommonTree tree) throws IOException, RecognitionException {
		Desugarer.desugar(tree);
		
		System.out.println();
		
		CommonTreeNodeStream nodes = new CommonTreeNodeStream(tree);
		
		DuraCAstValidation walker = new DuraCAstValidation(nodes);
        try {
            walker.program();
        } catch (RecognitionException e) {
            e.printStackTrace();
        }
	}
	
	public static void main(String[] args) throws Exception {
		// String filename = new File("src/emili/dura/compiler/durac/antlr/DuraC.example").getAbsolutePath();
		String filename = new File("src/de/lmu/ifi/pms/cep/dura/compiler/dura/Dura.example").getAbsolutePath();
		compile(filename);
	}
}
