package de.lmu.ifi.pms.cep.dura.compiler.dura.desugaring;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;
import java.util.Stack;

import org.antlr.runtime.CommonToken;
import org.antlr.runtime.tree.CommonTree;
import org.antlr.runtime.tree.Tree;

import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraParser;


public class HiddenBindingsAfterGrouping extends Transformation {

	@Override
	public void apply(CommonTree tree) {
		for (CommonTree eventSpecification : collectEventSpecificationNodes(tree)) {
			String tempTreeString = null;
			CommonTree annotation = null;
			if (EMIT_ANNOTATIONS) {
				tempTreeString = eventSpecification.toStringTree();
					
				annotation = new CommonTree(new CommonToken(DuraParser.ANNOTATION, "ANNOTATION"));
				annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Result of transformation " + this.getClass().getSimpleName() + ".\"")));
			}
			
			
			CommonTree head = (CommonTree) eventSpecification.getChild(0);
			assert(head.getType() == DuraParser.HEAD);
			
			CommonTree headTerm = (CommonTree) head.getChild(0);
			
			CommonTree composition = (CommonTree) eventSpecification.getChild(1).getChild(0);
			CommonTree supplement = (CommonTree) composition.getChild(composition.getChildCount() - 1);
			
			List<CommonTree> groupings = collectGroupings(supplement);
			
			List<CommonTree> suppRefBindings = new ArrayList<CommonTree>();
			if (supplement.getChildCount() > 0) {
				for (Object c : supplement.getChildren()) {
					CommonTree suppChild = (CommonTree) c;
					if (suppChild.getType() != DuraParser.AGGREGATION) {
						// Attention: Typed variables are not considered bindings here. They are pseudo bindings.
						//            That is why this collection works here.
						suppRefBindings.addAll(collectBindings(suppChild));
					}
				}
			}
			
			if (!suppRefBindings.isEmpty() && !groupings.isEmpty()) {
				for (CommonTree b : suppRefBindings) {
					CommonTree pseudoBinding = findFirstPseudoBindingOccurrence(b, composition);
					
					if (pseudoBinding != null && pseudoBinding != b) {
						for (CommonTree g : groupings) {
							if (isAfter(g, pseudoBinding) && isAfter(b, g)) {
								CommonTree recoveryBinding = pseudoBinding;
								if (pseudoBinding.getType() == DuraParser.TYPED_VARIABLE) {
									recoveryBinding = copyTree(recoveryBinding);
									
									recoveryBinding.getToken().setType(DuraParser.VARIABLE);
									recoveryBinding.getToken().setText("VARIABLE");
									recoveryBinding.deleteChild(0);
								}
								
								recoverBindingInGrouping(recoveryBinding, g);
								
								if (EMIT_ANNOTATIONS) {
									annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Binding " + recoveryBinding.toStringTree() + " recovered in " + g.toStringTree() + " of deductive rule " + tempTreeString + ".\"")));
								}
							}
						}
					}
				}
			}
			
			
			List<CommonTree> headTermBindings = collectBindings(headTerm);
			if (!headTermBindings.isEmpty() && !groupings.isEmpty()) {
				for (CommonTree b : headTermBindings) {
					// Either typed variable (pseudo), variable or identifier.
					CommonTree pseudoBinding = findFirstPseudoBindingOccurrence(b, composition);
					
					if (pseudoBinding != null) {
						for (CommonTree g : groupings) {
							if (isAfter(g, pseudoBinding)) {
								CommonTree recoveryBinding = pseudoBinding;
								if (pseudoBinding.getType() == DuraParser.TYPED_VARIABLE) {
									recoveryBinding = copyTree(recoveryBinding);
									
									recoveryBinding.getToken().setType(DuraParser.VARIABLE);
									recoveryBinding.getToken().setText("VARIABLE");
									recoveryBinding.deleteChild(0);
								}
								
								recoverBindingInGrouping(recoveryBinding, g);
								
								if (EMIT_ANNOTATIONS) {
									annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Binding " + recoveryBinding.toStringTree() + " recovered in " + g.toStringTree() + " of deductive rule " + tempTreeString + ".\"")));
								}
							}
						}
					}
				}
			}
			
			if (EMIT_ANNOTATIONS) {
				if (annotation.getChildCount() > 1) {
					eventSpecification.getParent().addChild(annotation);
				}
			}
		}
	}
	
	private void recoverBindingInGrouping(CommonTree recoveryBinding, CommonTree grouping) {
		if (grouping.getChildCount() > 0) {
			for (Object c : grouping.getChildren()) {
				if (treesEqual((CommonTree) c, recoveryBinding)) {
					return;
				}
			}
		}
		
		System.out.println("Binding " + recoveryBinding.toStringTree() + " recovered in " + grouping.toStringTree() + "!");
		
		grouping.addChild(copyTree(recoveryBinding));
	}
	
	// Pseudo because of possibly typed variables.
	private CommonTree findFirstPseudoBindingOccurrence(CommonTree of, CommonTree in) {
		Stack<CommonTree> nodeStack = new Stack<CommonTree>();
		nodeStack.push(in);
		
		while(!nodeStack.isEmpty()) {
			CommonTree node = nodeStack.pop();
			
			if (node.getType() == DuraParser.VARIABLE) {
				if (of.getType() == DuraParser.VARIABLE && treesEqual(of, node)) {
					return node;
				}
				
				continue;
			} else if (node.getType() == DuraParser.TYPED_VARIABLE) {
				if (of.getType() == DuraParser.VARIABLE &&
						((CommonTree) node.getChild(1)).getText().equals(((CommonTree) of.getChild(0)).getText())) {
					return node;
				}
				
				continue;
			} else if (node.getType() == DuraParser.IDENTIFIER) {
				if (of.getType() == DuraParser.IDENTIFIER && treesEqual(of, node)) {
					return node;
				}
				
				continue;
			}
			
			if (node.getChildCount() > 0) {
				// Because we need to find the first occurrence we have to push the first child last.
				for (int i = node.getChildCount() - 1; i >= 0; i--) {
					nodeStack.push((CommonTree) node.getChild(i));
				}
			}
		}
		
		return null;
	}
	
	private boolean isAfter(CommonTree treeA, CommonTree treeB) {
		Tree tempA = treeA.getParent();
		int treeADepth = 0;
		while (tempA != null) {
			tempA = tempA.getParent();
			treeADepth++;
		}
		
		Tree tempB = treeB.getParent();
		int treeBDepth = 0;
		while (tempB != null) {
			tempB = tempB.getParent();
			treeBDepth++;
		}
		
		tempA = treeA;
		tempB = treeB;
		if (treeADepth > treeBDepth) {
			for (int i = 0; i < treeADepth - treeBDepth; i++) {
				tempA = tempA.getParent();
			}
		} else if (treeBDepth > treeADepth) {
			for (int i = 0; i < treeBDepth - treeADepth; i++) {
				tempB = tempB.getParent();
			}
		}
		
		while (tempA.getParent() != tempB.getParent()) {
			tempA = tempA.getParent();
			tempB = tempB.getParent();
		}
			
		return tempA.getChildIndex() > tempB.getChildIndex();
	}

	private List<CommonTree> collectGroupings(CommonTree supplement) {
		assert(supplement.getType() == DuraParser.SUPPLEMENT);
		
		ArrayList<CommonTree> nodes = new ArrayList<CommonTree>();
		
		if (supplement.getChildCount() > 0) {
			for (Object c : supplement.getChildren()) {
				CommonTree temp = (CommonTree) c; 
				if (temp.getType() == DuraParser.AGGREGATION) {
					nodes.add((CommonTree) temp.getChild(0));
				}
			}
		}
		
		return nodes;
	}
	
	private List<CommonTree> collectBindings(CommonTree tree) {		
		ArrayList<CommonTree> nodes = new ArrayList<CommonTree>();
		
		Queue<CommonTree> nodeQueue = new LinkedList<CommonTree>();
		nodeQueue.add(tree);
		
		while(!nodeQueue.isEmpty()) {
			CommonTree node = (CommonTree) nodeQueue.remove();			
			
			if (node.getType() == DuraParser.IDENTIFIER || node.getType() == DuraParser.VARIABLE) {
				nodes.add(node);
			}
			
			if (node.getChildCount() > 0) {
				for (Object c : node.getChildren()) {
					CommonTree child = (CommonTree) c;
					nodeQueue.add(child);
				}
			}
		}
			
		return nodes;
	}
}