package de.lmu.ifi.pms.cep.dura.compiler.durac.queries;

import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCLexer.CONSTANT;
import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCLexer.EVENT_QUERY;
import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCLexer.NUMBER;
import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCLexer.STRING;
import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCLexer.TERM;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.Stack;

import org.antlr.runtime.tree.Tree;

import de.lmu.ifi.pms.cep.dura.compiler.antlr.utils.DuraTree;
import de.lmu.ifi.pms.cep.dura.compiler.durac.DuraCQueryCompiler;
import de.lmu.ifi.pms.cep.dura.compiler.durac.collections.DuraSchema;
import de.lmu.ifi.pms.cep.dura.compiler.durac.collections.Preamble;
import de.lmu.ifi.pms.cep.dura.compiler.durac.collections.QueryVariables;
import de.lmu.ifi.pms.cep.dura.compiler.durac.collections.UnambiguousQueryVariables;
import de.lmu.ifi.pms.cep.dura.compiler.durac.formula_transformations.FormulaNormalization;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraCompilationException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraUnknownAttributeException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraUnspecificCompliationException;
import de.lmu.ifi.pms.cep.tsa.DTSA_BufferType;
import de.lmu.ifi.pms.cep.tsa.TSA_CompositeExpression;
import de.lmu.ifi.pms.cep.tsa.TSA_Expression;
import de.lmu.ifi.pms.cep.tsa.TSA_TemporalStream;
import de.lmu.ifi.pms.cep.tsa.TSA_TemporalStreamExpression;
import de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonOperatorID;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.ConjunctiveConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_Exception;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_RenameOperator;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_SelectionOperator;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchemaTypeID;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;

public class AtomicEventQuery extends CompositeEventQuery {
	public final DuraTree origin;
	
	public final String identifier;
	
	public AtomicEventQuery(final DuraTree query, final Preamble preamble, final DuraSchema schema) {
		this(query, new HashMap<String, String>(), preamble, schema);
	}
	
	public AtomicEventQuery(final DuraTree query, final Map<String, String> substituePrefix, final Preamble preamble, final DuraSchema schema) {
		assert(query.getType() == EVENT_QUERY);
		
		this.identifier = query.getChild(0).getChild(0).getText();
		DuraTree querySupplement = query.getChild(2);
		
		String queryStreamName = query.getChild(1).getChild(0).getText();
		
		try {
			TSA_TemporalStream queryStream = schema.streams.getStreamByName(TemporalStreamSchemaTypeID.EVENT, preamble, queryStreamName);
		
			TSA_RenameOperator addIdentiferPrefix = new TSA_RenameOperator(query.getPositionDescription(), new FixedArrayNavigableMap<String, String>(new String[]{ "" }, new String[] { this.identifier }));
			TSA_Expression renamedAttributes = new TSA_CompositeExpression(addIdentiferPrefix, new TSA_TemporalStreamExpression(queryStream.full_name, DTSA_BufferType.WORKING, queryStream.schema, queryStream.origin));
			
			QueryVariables ambiguousVariables = new QueryVariables(query, schema, preamble);

			ConditionFormula implUnifications = ambiguousVariables.compileImplicitVariableUnifications();
			UnambiguousQueryVariables variables = ambiguousVariables.dropAmbiguousMappings();

			FormulaNormalization refiner = new FormulaNormalization(variables, renamedAttributes.schema(), substituePrefix, schema.constants, preamble);

			ConditionFormula implLiterals = refiner.apply(compileImplicitLiteralBindings(query));
	
			TSA_SelectionOperator implicitSelections = new TSA_SelectionOperator(query.getPositionDescription(), implLiterals, implUnifications);
			TSA_Expression implicitConditions = new TSA_CompositeExpression(implicitSelections, renamedAttributes);
			
			CompositeEventQuery supplement = DuraCQueryCompiler.compileSupplement(querySupplement, implicitConditions, variables, schema.constants, preamble);
			
			this.origin = query;
			this.variables = supplement.getVariables();
			this.expression = supplement.getExpression();
			this.identifierMapping = supplement.getIdentifierMapping();
		} catch (TSA_Exception e) {
			throw new DuraCompilationException(e, query);
		} catch (DuraUnknownAttributeException e) {
			throw new DuraCompilationException(e.getMessage(substituePrefix), query);
		} catch (DuraUnspecificCompliationException e) {
			throw new DuraCompilationException(e, query);
		}
	}
	
	private static ConditionFormula compileImplicitLiteralBindings(final DuraTree query) {
		assert(query.getType() == EVENT_QUERY);

		Stack<String> path = new Stack<String>();
		ArrayList<ConditionFormula> conditions = new ArrayList<ConditionFormula>();

		DuraTree queryTerm = query.getChild(1).getChild(0);
		Tree identifier = query.getChild(0).getChild(0);
		
		path.push(identifier.getText());

		for (int i=0; i<queryTerm.getChildCount(); i++) {
			DuraTree child = queryTerm.getChild(i);
			
			compileImplicitLiteralBindings(child, conditions, path);
		}
		
		if (conditions.size() == 0) {
			return ConditionFormula.Top;
		} else {
			return new ConjunctiveConditionFormula(new FixedArray<ConditionFormula>(conditions)); 
		}
	}
	
	private static void compileImplicitLiteralBindings(final DuraTree term, final ArrayList<ConditionFormula> conditions, final Stack<String> path) {
		switch (term.getType()) {
		case TERM:
			path.add(term.getText());
			
			for (int i=0; i<term.getChildCount(); i++) {
				DuraTree child = term.getChild(i);
				
				compileImplicitLiteralBindings(child, conditions, path);
			}

			path.pop();
			
			break;

		case STRING:
		case NUMBER:
		case CONSTANT:
			Term<?> value = DuraCQueryCompiler.compileTerm(term);
			Term<?> attribute = Term.attr(DuraCQueryCompiler.joinPath(path));
			
			conditions.add(new BinaryComparissonFormula(BinaryComparissonOperatorID.EQUAL, attribute, value));
			
			break;
		}
	}
}
