package de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings;

import static de.lmu.ifi.pms.cep.tsa.conditions.AggregationFunctionID.AVG;
import static de.lmu.ifi.pms.cep.tsa.conditions.AggregationFunctionID.COUNT;
import static de.lmu.ifi.pms.cep.tsa.conditions.AggregationFunctionID.MAX;
import static de.lmu.ifi.pms.cep.tsa.conditions.AggregationFunctionID.MIN;
import static de.lmu.ifi.pms.cep.tsa.conditions.AggregationFunctionID.SUM;

import org.antlr.runtime.tree.Tree;

import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCLexer;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraUnexpectedTypeException;
import de.lmu.ifi.pms.cep.tsa.conditions.AggregationFunction;
import de.lmu.ifi.pms.cep.tsa.conditions.AggregationFunctionID;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.cep.tsa.schema.Type;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;


public class DuraAggregationFunctionMapping {
	private static final FixedArrayNavigableMap<String, AggregationFunctionID> fromDuraMap = new FixedArrayNavigableMap<String, AggregationFunctionID>(
			new                String[]{ "count", "min", "max", "avg", "sum" }, 
			new AggregationFunctionID[]{   COUNT,   MIN,   MAX,   AVG,   SUM }
	);
	
	public static AggregationFunctionID fromToken(final Tree token) {
		assert(token.getType() == DuraCLexer.AGGREGATION_OP);
		
		return fromDuraMap.get(token.getText());
	}
	
	public static BasicType getType(AggregationFunction<?> fn, Term<? extends Type> argument) throws DuraUnexpectedTypeException {
		
		switch (fn.ID) {
		case COUNT:
			return BasicTypeID.INT.type;

		case AVG:
			if (argument.type instanceof BasicType && DuraBasicTypesMapping.isNumber((BasicType) argument.type)) {
				return BasicTypeID.DOUBLE.type;
			} else {
				throw new DuraUnexpectedTypeException(String.format("argument of function %s expects a number type, found %s", fn.name, argument.type));
			}

		default:
			if (argument.type instanceof BasicType) {
				return (BasicType) argument.type;
			} else {
				throw new DuraUnexpectedTypeException(String.format("argument of function %s expects a basic type, found %s", fn.name, argument.type));
			}
		}
		
	}

}
