package de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings;

import static de.lmu.ifi.pms.cep.tsa.conditions.TupleFunctionID.DIVISION;
import static de.lmu.ifi.pms.cep.tsa.conditions.TupleFunctionID.GREATEST;
import static de.lmu.ifi.pms.cep.tsa.conditions.TupleFunctionID.LEAST;
import static de.lmu.ifi.pms.cep.tsa.conditions.TupleFunctionID.MINUS;
import static de.lmu.ifi.pms.cep.tsa.conditions.TupleFunctionID.PLUS;
import static de.lmu.ifi.pms.cep.tsa.conditions.TupleFunctionID.POWER;
import static de.lmu.ifi.pms.cep.tsa.conditions.TupleFunctionID.TIMES;

import java.util.ArrayList;

import org.antlr.runtime.tree.Tree;

import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCLexer;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraUnexpectedTypeException;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.conditions.TupleFunction;
import de.lmu.ifi.pms.cep.tsa.conditions.TupleFunctionID;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import static de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID.*;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;

public class DuraMathFunctionMapping {

	private static final FixedArrayNavigableMap<String, TupleFunctionID> fromDuraMap = new FixedArrayNavigableMap<String, TupleFunctionID>(
			new          String[]{      "/", "greatest", "least",   "-",  "+",   "^",    "*" }, 
			new TupleFunctionID[]{ DIVISION,   GREATEST,   LEAST, MINUS, PLUS, POWER, TIMES  }
	);
	
	public static TupleFunctionID fromToken(final Tree token) {
		assert(token.getType() == DuraCLexer.ARITHMETIC_OP);
		
		return fromDuraMap.get(token.getText());
	}
	
	public static BasicType getType(final TupleFunction<?> fn, final Term<?>[] typedArguments) throws DuraUnexpectedTypeException {
		ArrayList<BasicType> argumentTypes = new ArrayList<BasicType>(typedArguments.length);
		
		for (Term<?> argument : typedArguments) {
			if (argument.isTyped && argument.type instanceof BasicType) {
				argumentTypes.add((BasicType) argument.type);
			} else {
				throw new DuraUnexpectedTypeException(String.format("arguments of function %s must be basic types, found %s", fn.name, argument.type));
			}
		}

		//first take care of the odd cases
		outer : switch (fn.ID) {
		case PLUS:
			if (argumentTypes.get(0).ID==TIMESTAMP && argumentTypes.get(1).ID==DURATION) {
				return TIMESTAMP.type;
			} else if (argumentTypes.get(0).ID==DURATION && argumentTypes.get(1).ID==DURATION) {
				return DURATION.type;
			}
			
			break;
			
		case MINUS:
			if (argumentTypes.get(0).ID==TIMESTAMP && argumentTypes.get(1).ID==TIMESTAMP) {
				return DURATION.type;
			} else if (argumentTypes.get(0).ID==TIMESTAMP && argumentTypes.get(1).ID==DURATION) {
				return TIMESTAMP.type;
			}
			
			break;
			
		case LEAST:
		case GREATEST:
			for (BasicType type : argumentTypes) {
				if (type.ID != TIMESTAMP) {
					break outer;
				}
			}
			
			return TIMESTAMP.type;
			
		case DIVISION:
			return DOUBLE.type;
			
		default:
		}

		//then assume that all arguments have a number type and determine the "highest" type
		BasicType functionType = INT.type;
		
		for (BasicType argumentType : argumentTypes) {
			if (DuraBasicTypesMapping.isNumber(argumentType)) {
				if (argumentType.compareTo(functionType) > 0) {
					functionType = argumentType;
				}
			} else {
				throw new DuraUnexpectedTypeException(String.format("arguments of function %s must be number types, found %s", fn.name, argumentType));
			}
		}
		
		return functionType;
	}
	
	
}
