package de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings;

import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCLexer;
import de.lmu.ifi.pms.cep.dura.compiler.antlr.utils.DuraTree;
import de.lmu.ifi.pms.cep.dura.compiler.durac.DuraCQueryCompiler;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.UnexpectedTokenException;
import de.lmu.ifi.pms.cep.tsa.conditions.Attribute;
import de.lmu.ifi.pms.cep.tsa.conditions.ClockFunction;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.DefaultDefinition;
import de.lmu.ifi.pms.cep.tsa.conditions.StreamBoundConstraint;


public class DuraRelationConstraintMapping {
	
	public static ConditionFormula fromToken(final DuraTree token) {
		assert(token.getType() == DuraCLexer.CONSTRAINT);
		
		String name = token.getText();
		
		if (name.equals("bound")) {
			Attribute<?> path = (Attribute<?>) DuraCQueryCompiler.compileTerm(token.getChild(0));
			
			return new StreamBoundConstraint(path, false);//FIXME value for "monotonic" parameter correct?
		} else if (name.equals("default")) {
			DuraTree unification = token.getChild(0);
			
			Attribute<?> path = (Attribute<?>) DuraCQueryCompiler.compileTerm(unification.getChild(0));
			ClockFunction clock = (ClockFunction) DuraCQueryCompiler.compileTerm(unification.getChild(1));
			
			return new DefaultDefinition(path, clock);
		} else if (name.equals("references")) {
			//TODO: implement
			
			return ConditionFormula.Top;
		} else if (name.equals("key")) {
			//TODO: implement
			
			return ConditionFormula.Top;
		} else {
			throw new UnexpectedTokenException(token);
		}
	}
}
