package de.lmu.ifi.pms.cep.dura.compiler.durac.tsa;

import org.simpleframework.xml.Element;
import org.simpleframework.xml.ElementList;

import de.lmu.ifi.pms.cep.dura.compiler.durac.collections.TypeDefinitions;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.conditions.NamedTypeAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.NamedTypeRecord;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.conditions.TimestampAttribute;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.cep.tsa.schema.CompositeType;
import de.lmu.ifi.pms.cep.tsa.schema.NamedType;

public class TimeInterval extends NamedTypeRecord {
	public TimeInterval(final NamedTypeAttribute<CompositeType> interval) {
		this(TimestampAttribute.create(TSA_SyntaxPatterns.prefixAttribute(interval.name, interval.type.type.attributes.keys().get(0))), TimestampAttribute.create(TSA_SyntaxPatterns.prefixAttribute(interval.name, interval.type.type.attributes.keys().get(1))));
		
		assert(interval.type.equals(TypeDefinitions.TIME_INTERVAL));
	}
	
	public TimeInterval(final Term<BasicType> begin, final Term<BasicType> end) {
		super(TypeDefinitions.TIME_INTERVAL, begin, end);
		
		assert(begin.type.equals(BasicTypeID.TIMESTAMP.type));
		assert(end.type.equals(BasicTypeID.TIMESTAMP.type));
	}
	
	@Override
	public TimeInterval newTerm(final Term<?>... arguments) {
		assert(arguments.length == 2);
		
		assert(arguments[0].isTyped && arguments[0].type.equals(BasicTypeID.TIMESTAMP.type));
		assert(arguments[1].isTyped && arguments[1].type.equals(BasicTypeID.TIMESTAMP.type));
		
		return new TimeInterval((Term<BasicType>) arguments[0], (Term<BasicType>) arguments[1]);
	}
	
	TimeInterval(@Element(name="type", required=false) final NamedType<CompositeType> type, @ElementList(entry="term", inline=true, required=false, empty=false)  final java.util.ArrayList<Term<?>> arguments) {
		super(TypeDefinitions.TIME_INTERVAL, arguments);
		assert type == null;
		assert arguments.size() == 2;
		assert(arguments.get(0).isTyped && arguments.get(0).type.equals(BasicTypeID.TIMESTAMP.type));
		assert(arguments.get(1).isTyped && arguments.get(1).type.equals(BasicTypeID.TIMESTAMP.type));
	}
	
	@Override
	protected NamedType<CompositeType> getType() {
		return null;
	}
	
	public Term<BasicType> begin() {
		return (Term<BasicType>) this.arg(0);
	}
	
	public Term<BasicType> end() {
		return (Term<BasicType>) this.arg(1);
	}
}
