package de.lmu.ifi.pms.cep.event_mill.configuration;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.net.URL;
import java.text.SimpleDateFormat;
import java.util.Calendar;

import org.slf4j.LoggerFactory;

import ch.qos.logback.core.Context;

import de.lmu.ifi.pms.cep.COMMON_CONSTANTS;
import de.lmu.ifi.pms.cep.compilation.CompilerFactory;
import de.lmu.ifi.pms.cep.dura.DuraC_TSA_CompilerFactory;
import de.lmu.ifi.pms.cep.event_mill.execution.scheduler.BasicQuerySequencer;
import de.lmu.ifi.pms.cep.event_mill.execution.scheduler.QuerySequencer;
import de.lmu.ifi.pms.cep.event_mill.sql.optimization.dtsa.DTSA_RecoveryEnhacerFactory;
import de.lmu.ifi.pms.cep.tsa.DTSA_Program;
import de.lmu.ifi.pms.cep.tsa.TSA_Program;
import de.lmu.ifi.pms.cep.tsa.compiler.TSA_DTSA_CompilerFactory;
import de.lmu.ifi.pms.cep.tsa.compiler.TSA_DefaultDifferencerFactory;
import de.lmu.ifi.pms.cep.tsa.optimization.DTSA_GarbageCollectorFactory;
import de.lmu.ifi.pms.cep.tsa.optimization.DTSA_OptimizerFactory;
import de.lmu.ifi.pms.cep.tsa.optimization.TSA_OptimizerFactory;

public final class DEFAULTS {
	
	public final static String defaultConfigResourcePath = "event-mill.configuration.xml";
	public final static String defaultIOConfigResourcePath = "event-mill.io.configuration.xml";
	
	public final static long BasicSchedulerDefaultDelay = 1000;
	
	public final static QuerySequencer defaultSequencer = new BasicQuerySequencer();
	
	public final static CompilerFactory<String[], TSA_Program> defaultTSACompiler = new DuraC_TSA_CompilerFactory();
	
	public final static CompilerFactory<TSA_Program, DTSA_Program> defaultDTSACompiler = new TSA_DTSA_CompilerFactory(new TSA_OptimizerFactory[]{},new TSA_DefaultDifferencerFactory(), new DTSA_OptimizerFactory[]{new DTSA_RecoveryEnhacerFactory(), new DTSA_GarbageCollectorFactory()});
	
	public final static DTSA_OptimizerFactory defaultDTSAOptimizer(){
		return new DTSA_RecoveryEnhacerFactory();
	}
	
	public final static String logTime;
	public final static String logFileDir;
	
	static{
		String time = null;
		String path = null;
		try{
			Context loggingContext = (Context) LoggerFactory.getILoggerFactory();
			time = loggingContext.getProperty(COMMON_CONSTANTS.LOG_FILE_TIMESTAMP_PROPERTY);
			path = loggingContext.getProperty(COMMON_CONSTANTS.LOG_FILE_DIR_PROPERTY);
		}
		catch (ClassCastException e) {
		
		}
		if(time == null){
			time = new SimpleDateFormat(COMMON_CONSTANTS.LOG_FILE_TIMESTAMP_PATTERN).format(Calendar.getInstance().getTime());
		}
		if(path == null){
			path = "log/";
		}
		logTime = time;
		logFileDir = path;
	}

	public static final String _o_id = ".id";
	public static final String _o_block = ".block";
	
	private DEFAULTS(){
		
	}
	
	public static String resourceFilePath(final String resourcePath){
		return resourceFilePath(resourcePath, DEFAULTS.class);
	}
	
	public static String resourceFilePath(final String resourcePath, final Class<?> clazz){
		ClassLoader classLoader = clazz.getClassLoader();
		if(classLoader == null){
			classLoader = ClassLoader.getSystemClassLoader();
		}
		return resourceFilePath(resourcePath, clazz.getName(), classLoader);
	}
	
	public static String resourceFilePath(final String resourcePath, final String className, final ClassLoader classLoader){		
		URL url = classLoader.getResource(className.replace('.', '/')+".class");
		if(url != null){
			String protocol = url.getProtocol();
			if(protocol.equals("file")){
				File directory;
				try {
					directory = new File(url.toURI());
				} catch (URISyntaxException e) {
					throw new Error(e);
				}
				for(int nameLength = className.split("\\.").length; nameLength > 0; --nameLength){
					directory = directory.getParentFile();
				}		
				return new File(directory.getAbsolutePath() + "/" + resourcePath).getAbsolutePath();
			}
			else if(protocol.equals("jar")){
				String file = url.getFile();
				if(file.startsWith("file:/") & file.contains(".jar!")){
					try {
						URL jarURL = new URL(file.substring(0, file.indexOf('!')));
						return new File(new File(jarURL.toURI()).getParentFile().getAbsolutePath() + "/" + resourcePath).getAbsolutePath();
					} catch (URISyntaxException e) {
						throw new Error(e);
					} catch (MalformedURLException e) {
						throw new Error(e);
					}
				}
			}
		}
		throw new RuntimeException();
	}

}
