package de.lmu.ifi.pms.cep.event_mill.io;

import java.util.concurrent.ConcurrentLinkedQueue;

import de.lmu.ifi.pms.cep.event_mill.io.format.MessageFormatProvider;
import de.lmu.ifi.pms.cep.event_mill.io.format.MessageGenerator;
import de.lmu.ifi.pms.cep.event_mill.io.format.MessageReader;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;

public class SpecificMessageAdapter<MessageType1, MessageType2> implements Adapter<MessageType1, MessageType2>{
	
	protected final String messageType;
	protected final FixedArrayNavigableMap<String, BasicTypeID> types;
	
	protected final MessageFormatProvider<MessageType1> formatter1;
	protected final MessageFormatProvider<MessageType2> formatter2;
	
	protected final ConcurrentLinkedQueue<MessageReader<MessageType1>> messageReaders1;
	protected final ConcurrentLinkedQueue<MessageGenerator<MessageType2>> messageGenerators2;
	public final int numberOfAttributes;
	protected final BasicTypeID[] typeArray;
	
	public SpecificMessageAdapter(final String messageType, final FixedArrayNavigableMap<String, BasicTypeID> types, final MessageFormatProvider<MessageType1> formatter1, final MessageFormatProvider<MessageType2> formatter2){
		super();
		this.messageType = messageType;
		this.types = types;
		this.formatter1 = formatter1;
		this.formatter2 = formatter2;
		this.messageReaders1 = new ConcurrentLinkedQueue<MessageReader<MessageType1>>();
		this.messageGenerators2 = new ConcurrentLinkedQueue<MessageGenerator<MessageType2>>();
		this.numberOfAttributes = types.size();
		this.typeArray = types.values().toArray(new BasicTypeID[this.numberOfAttributes]);
	}

	@Override
	public MessageType2 adapt(final MessageType1 message) {
    	BasicTypeID attributeType;
    	
    	MessageReader<MessageType1> messageReader1 = this.messageReaders1.poll();
    	MessageGenerator<MessageType2> messageGenerator2 = this.messageGenerators2.poll();
    	if(messageReader1 == null){
    		messageReader1 = this.formatter1.messageReader(this.messageType, this.types);
    	}
    	if(messageGenerator2 == null){
    		messageGenerator2 = this.formatter2.messageGenerator(this.messageType, this.types);
    	}    	
    	
    	messageReader1.decompose(message);
    	for (int attribute = 1; attribute <= this.numberOfAttributes; ++attribute){
    		attributeType = this.typeArray[attribute - 1];
    		switch (attributeType) {
			case BOOLEAN:
				messageGenerator2.setBoolean(attribute, messageReader1.getBoolean(attribute));
				break;
			case STRING:
				messageGenerator2.setString(attribute, messageReader1.getString(attribute));
				break;
			case INT:
				messageGenerator2.setInt(attribute, messageReader1.getInt(attribute));
				break;
			case LONG:
				messageGenerator2.setLong(attribute, messageReader1.getLong(attribute));
				break;				
			case FLOAT:
				messageGenerator2.setFloat(attribute, messageReader1.getFloat(attribute));
				break;
			case DOUBLE:
				messageGenerator2.setDouble(attribute, messageReader1.getDouble(attribute));
				break;
			case TIMESTAMP:
				messageGenerator2.setTimestamp(attribute, messageReader1.getTimestamp(attribute));
				break;				
			case DURATION:
				messageGenerator2.setDuration(attribute, messageReader1.getDuration(attribute));
				break;				
			case IDENTIFIER:
				messageGenerator2.setIdentifier(attribute, messageReader1.getIdentifier(attribute));
				break;
			case REFERENCE: throw new IllegalArgumentException("Type \"reference\" not supported.");
			default:
				break;
			}
    	}
    		
    	MessageType2 convertedMessage = messageGenerator2.compose();
    		
		messageReader1.clearMessage();
		messageGenerator2.clearParameters();
		
		this.messageReaders1.add(messageReader1);
		this.messageGenerators2.add(messageGenerator2);
		
		return convertedMessage;
	}
	
}
