package de.lmu.ifi.pms.cep.event_mill.io.bus;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.event_mill.io.MessageSource;
import de.lmu.ifi.pms.cep.event_mill.io.MessageSourceSpliter;
import de.lmu.ifi.pms.cep.event_mill.io.SpecificBufferedMessageSource;
import de.lmu.ifi.pms.cep.event_mill.io.format.MessageReader;
import de.lmu.ifi.pms.cep.event_mill.io.stream.SourceStream;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;

import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;

public class BusSourceStream<MessageType> extends SourceStream {
	
	static final int initialMessageBufferCapacity = 128;
	
	protected MessageType[] messageBuffer;
	protected int messageIndex;
	protected int messageBufferLength;
	
	final MessageReader<MessageType> messageReader;
	protected final MessageSource<MessageType> messageSource;
	protected final MessageSourceSpliter<MessageType> spliter;
	protected final SpecificBufferedMessageSource<MessageType> specificMessageSource;
	
	public BusSourceStream(final String messageType, final FixedArrayNavigableMap<String, BasicTypeID> types, final MessageSourceSpliter<MessageType> spliter, final MessageBus<MessageType> bus){
		super(messageType, types);
		this.messageBuffer = null;		
		this.messageBufferLength = Integer.MIN_VALUE;
		
		this.messageReader = bus.formatter().messageReader(messageType, types);
		this.messageSource = bus.source();
		this.spliter = spliter;
		this.specificMessageSource = spliter.source(messageType);
	}

	@Override
	public boolean nextBlock() {
		this.spliter.writeBlock(this.messageSource.readBlock());	
		this.messageBuffer = this.specificMessageSource.readBlock();
		this.messageIndex = -1;
		this.messageBufferLength = this.messageBuffer.length;
		return true;//TODO check if open?
	}

	@Override
	public boolean next() {
		++this.messageIndex;
		if(this.messageIndex < this.messageBufferLength){
			this.messageReader.clearMessage();
			this.messageReader.decompose(this.messageBuffer[this.messageIndex]);
			this.messageBuffer[this.messageIndex] = null;
			return true;
		}
		else{
			this.messageBuffer = null;
			this.messageBufferLength = Integer.MIN_VALUE;
			return false;
		}
	}

	@Override
	public String getString(final int parameterIndex) {
		return this.messageReader.getString(parameterIndex);
	}

	@Override
	public boolean getBoolean(final int parameterIndex) {
		return this.messageReader.getBoolean(parameterIndex);
	}

	@Override
	public byte getByte(final int parameterIndex) {
		return this.messageReader.getByte(parameterIndex);
	}

	@Override
	public short getShort(final int parameterIndex) {
		return this.messageReader.getShort(parameterIndex);
	}

	@Override
	public int getInt(final int parameterIndex) {
		return this.messageReader.getInt(parameterIndex);
	}

	@Override
	public long getLong(final int parameterIndex) {
		return this.messageReader.getLong(parameterIndex);
	}

	@Override
	public long getTimestamp(final int parameterIndex) {
		return this.messageReader.getTimestamp(parameterIndex);
	}

	@Override
	public long getDuration(final int parameterIndex) {
		return this.messageReader.getDuration(parameterIndex);
	}

	@Override
	public long getIdentifier(final int parameterIndex) {
		return this.messageReader.getIdentifier(parameterIndex);
	}

	@Override
	public float getFloat(final int parameterIndex) {
		return this.messageReader.getFloat(parameterIndex);
	}

	@Override
	public double getDouble(final int parameterIndex) {
		return this.messageReader.getDouble(parameterIndex);
	}
	
	//logging
	protected static final Logger logger = LoggerFactory.getLogger(BusSourceStream.class);
	
	@Override
	protected Logger logger(){
		return logger;
	}

}
