package de.lmu.ifi.pms.cep.event_mill.io.bus;

import java.util.Arrays;
import java.util.concurrent.ConcurrentLinkedQueue;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.event_mill.io.MessageSink;
import de.lmu.ifi.pms.cep.event_mill.io.MessageSource;
import de.lmu.ifi.pms.cep.event_mill.io.format.MessageFormatProvider;

public class HeapBroadcastMessageBus<MessageType> extends MessageBus<MessageType> {
	
	final String identifier; 
	
	volatile ConcurrentLinkedQueue<MessageType> inputPipe;
	int sourceCount;
	//java.util.ArrayList<MessageSource<MessageType>> sources;
	
	final HeapBroadcastMessageBusProvider<MessageType> busProvider;
	
	final MessageFormatProvider<MessageType> formater;
	
	protected HeapBroadcastMessageBus(final MessageFormatProvider<MessageType> formater, final String identifier, final HeapBroadcastMessageBusProvider<MessageType> busProvider) {
		super(formater.type, busProvider);
		this.formater = formater;
		this.identifier = identifier;
		this.busProvider = busProvider;
		
		this.sourceCount = 0;
		//this.sources = new java.util.ArrayList<MessageSource<MessageType>>();
		//FIXME close sources and sinks if bus is closed. Attention: use WeakReferences 
	}

	@Override
	public MessageFormatProvider<MessageType> formatter() {
		return this.formater;
	}
	
	@Override
	public MessageSink<MessageType> sink(){		
		return new MessageSink<MessageType>(this.type) {		
			@Override
			public void writeBlock(final MessageType[] messages) {
				for(ConcurrentLinkedQueue<MessageType> pipe: HeapBroadcastMessageBus.this.busProvider.pipes.values()){
					if(pipe != HeapBroadcastMessageBus.this.inputPipe){
						pipe.addAll(Arrays.asList(messages));
					}
				}
			}
			
			@Override
			public void writeBlock(final java.util.Collection<MessageType> messages) {
				for(ConcurrentLinkedQueue<MessageType> pipe: HeapBroadcastMessageBus.this.busProvider.pipes.values()){
					if(pipe != HeapBroadcastMessageBus.this.inputPipe){
						pipe.addAll(messages);
					}
				}
			}

			@Override
			public void write(final MessageType message) {
				assert message != null;
				for(ConcurrentLinkedQueue<MessageType> pipe: HeapBroadcastMessageBus.this.busProvider.pipes.values()){
					if(pipe != HeapBroadcastMessageBus.this.inputPipe){
						pipe.offer(message);
					}
				}
			}
			
			@Override
			public void close() {

			}
		};
	}
	
	@Override
	public synchronized MessageSource<MessageType> source(){
		//FIXME allow only if not closed
		if(this.sourceCount == 0){
			this.inputPipe = new ConcurrentLinkedQueue<MessageType>();
			this.busProvider.pipes.put(this.identifier, this.inputPipe);
		}
		++this.sourceCount;
		assert this.inputPipe != null;
		
		MessageSource<MessageType> source = new MessageSource<MessageType>(this.type) {

			@Override
			public MessageType read() {
				return HeapBroadcastMessageBus.this.inputPipe.poll();
			}
			
			@Override
			public MessageType[] readBlock() {
				MessageType message = HeapBroadcastMessageBus.this.inputPipe.poll();
				if(message == null){
					return this.emptyBlock;
				}
				else{
					java.util.ArrayList<MessageType> block = new java.util.ArrayList<MessageType>();			
					while(message != null){
						block.add(message);
						message = HeapBroadcastMessageBus.this.inputPipe.poll();
					}
					return block.toArray(this.newBlock(block.size()));
				}
			}
			
			@Override
			public void close() {
				synchronized (HeapBroadcastMessageBus.this) {
					--HeapBroadcastMessageBus.this.sourceCount;
					if(HeapBroadcastMessageBus.this.sourceCount == 0){
						HeapBroadcastMessageBus.this.busProvider.pipes.remove(HeapBroadcastMessageBus.this.identifier);
						HeapBroadcastMessageBus.this.inputPipe.clear();
						HeapBroadcastMessageBus.this.inputPipe = null;						
					}
				}
			}
			
		};
		
		return source;
	}
	
	protected static final Logger logger = LoggerFactory.getLogger(HeapBroadcastMessageBus.class);
	
	@Override
	protected Logger logger() {
		return logger;
	}
	
	

}
