package de.lmu.ifi.pms.cep.event_mill.io.format;

import static de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID.*;

import org.simpleframework.xml.Attribute;
import org.simpleframework.xml.Element;

import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;

import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;

public class SemiStructuredMessageFormatProvider extends StructuredStringMessageFormatProvider {
	
	public static final String defaultIdentifier = "de.lmu.ifi.pms.cep.io.format.semi-structured";
	public static final ParameterFormatter defaultFormatter = new ParameterFormatterImpl();
	
	@Element(name="parameter-formatter", required=false)
	ParameterFormatter getFormatter(){
		if(this.formatter == defaultFormatter){
			return null;
		}
		else{
			return this.formatter;
		}
	}
	
	public SemiStructuredMessageFormatProvider() {
		this(defaultIdentifier);
	}
	
	public SemiStructuredMessageFormatProvider(final String identifier) {
		this(identifier, defaultFormatter);
	}
	
	public SemiStructuredMessageFormatProvider(@Attribute(name="iri", required=false) final String identifier, @Element(name="parameter-formatter", required=false) final ParameterFormatter formatter) {
		super(identifier, (formatter == null ? defaultFormatter : formatter) );
	}

	@Override
	public String messageType(final String message) {
		int index = message.indexOf('{');
		if(index < 0){
			throw new IllegalArgumentException("Message does not meet the format of semi structured expressions.\n" + message);
		}
		else{
			return message.substring(0, index);
		}
	}
	
	@Override
	protected void getMessageHead(final String messageType, final StringBuilder sb, final boolean hasAttributes) {
		sb.append(messageType);
		sb.append("{");
		if(hasAttributes){
			sb.append(" ");
		}
	}
	
	@Override
	protected void getMessageFoot(final String messageType, final StringBuilder sb, final boolean hasAttributes) {
		if(hasAttributes){
			sb.append(" ");
		}
		sb.append("}");
	}
	
	@Override
	protected void getAttributePrefix(final String attribute_local_name, final StringBuilder sb, final boolean hasSubAttributes, final boolean hasValue, final int depth) {
		sb.append(attribute_local_name);
		sb.append("{");
		if(hasSubAttributes | hasValue){
			sb.append(" ");
		}
	}
	
	@Override
	protected String getAttributeDelimiter(final int depth) {
		return ", ";
	}
	
	@Override
	protected void getAttributePostfix(final String attribute_local_name, final StringBuilder sb, final boolean hasSubAttributes, final boolean hasValue, final int depth) {
		if(hasSubAttributes | hasValue){
			sb.append(" ");
		}
		sb.append("}");
	}

	public static void main(final String[] args) {
		String messageType = "identity";
		FixedArrayNavigableMap<String, BasicTypeID> types = new FixedArrayNavigableMap<String, BasicTypeID>(
				new String[]{		"person.name", "person.address.street", "person.address.number"}, 
				new BasicTypeID[]{	STRING, STRING, INT}
				);
		
		MessageFormatProvider<String> formatProvider = new SemiStructuredMessageFormatProvider();
		
		MessageGenerator<String> messageGenerator = formatProvider.messageGenerator(messageType, types);
		
		messageGenerator.setString("person.name", "Hans Meier");
		messageGenerator.setString("person.address.street", "Münchner Straße");
		messageGenerator.setInt("person.address.number", 15);
		
		String message = messageGenerator.compose();
		System.out.println(message);
		messageGenerator.clearParameters();
		
		MessageReader<String> messageReader = formatProvider.messageReader(messageType, types);
		messageReader.decompose(message);
		
		messageGenerator.setString("person.name", messageReader.getString("person.name"));
		messageGenerator.setString("person.address.street", messageReader.getString("person.address.street"));
		messageGenerator.setInt("person.address.number", messageReader.getInt("person.address.number"));
		
		message = messageGenerator.compose();
		System.out.println(message);
	}
	
}
