package de.lmu.ifi.pms.cep.event_mill.io.stream;

import java.util.Arrays;

import org.simpleframework.xml.Attribute;
import org.simpleframework.xml.ElementList;
import org.simpleframework.xml.ElementListUnion;


import de.lmu.ifi.pms.java.util.collection.Collection;


public class CompositeStreamIOProvider extends StreamIOProvider {
	
	protected final StreamIOProvider[] streamIOProviders;
	
	@ElementListUnion({
		@ElementList(entry="empty-stream", inline=true, type=EmptyStreamIOProvider.class, required=false, empty=false),
		@ElementList(entry="composite-stream", inline=true, type=CompositeStreamIOProvider.class, required=false, empty=false),
		@ElementList(entry="stream", inline=true, type=StreamIOProvider.class, required=false, empty=false)
	})
	protected java.util.ArrayList<StreamIOProvider> getStreams(){
		if(this.streamIOProviders != null){
			return new java.util.ArrayList<StreamIOProvider>(java.util.Arrays.asList(this.streamIOProviders));
		}
		else{
			return null;
		}
	}
	
	public CompositeStreamIOProvider(final String identifier, final StreamIOProvider... streamIOProviders) {
		super(identifier);
		this.streamIOProviders = Arrays.copyOf(streamIOProviders, streamIOProviders.length);
	}
	
	public CompositeStreamIOProvider(final String identifier, final Collection<StreamIOProvider> streamIOProviders) {
		this(identifier, streamIOProviders.javaCollection());
	}
	
	public CompositeStreamIOProvider(@Attribute(name="iri")final String identifier, @ElementList(entry="stream", inline=true) final java.util.ArrayList<StreamIOProvider> streamIOProviders) {
		this(identifier, (java.util.Collection<StreamIOProvider>) streamIOProviders);
	}
	
	public CompositeStreamIOProvider(final String identifier, final java.util.Collection<StreamIOProvider> streamIOProviders) {
		super(identifier);
		if(streamIOProviders == null){
			this.streamIOProviders = new StreamIOProvider[0];
		}
		else{
			this.streamIOProviders = streamIOProviders.toArray(new StreamIOProvider[streamIOProviders.size()]);
		}
	}

	@Override
	public boolean accepts(final String identity) {
		for(StreamIOProvider provider: this.streamIOProviders){
			if(provider.accepts(identity)){
				return true;
			}
		}
		return false;
	}

	@Override
	public StreamIO connect(final String identity) {
		for(StreamIOProvider provider: this.streamIOProviders){
			if(provider.accepts(identity)){
				return provider.connect(identity);
			}
		}
		throw new IllegalArgumentException("Cannot connect to identity \"" + identity + "\".");
	}

}
