package de.lmu.ifi.pms.cep.event_mill.sql;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.concurrent.locks.ReentrantLock;
import java.util.regex.Matcher;

import de.lmu.ifi.pms.cep.event_mill.execution.program.ExecutableQuery;
import de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper;
import de.lmu.ifi.pms.cep.event_mill.sql.optimization.dtsa.DTSA_RecoveryEnhacer;
import de.lmu.ifi.pms.cep.tsa.DTSA_BufferType;
import de.lmu.ifi.pms.cep.tsa.DTSA_QueryType;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.time.Clock;
import de.lmu.ifi.pms.cep.tsa.time.functions.TimeFunction;

import de.lmu.ifi.pms.java.util.collection.array.Arrays;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;
import de.lmu.ifi.pms.java.util.map.fixed.FixedEnumMap;


public class SQL_ExecutableQuery extends ExecutableQuery<SQL_ExecutableProgram, SQL_StreamBuffer>
{
	
	public final int id;
	public final SQL_Translation sql_translation;
	
	public final String id_attribute;
	
	long lastRound;
	public final String round_attribute;
	//first value = id, second value = round
	protected final long[] queryInfo = new long[2];
	
	protected final int numberOfReferencedClocks;
	protected final FixedArrayNavigableMap<String, Clock> referencedClocks;
	protected final int numberOfOutputStreamBounds;
	
	protected final FixedArray<String> outputStreamBounds;
	protected long[] lastOutputStreamBoundValues;
	protected long[] nextOutputStreamBoundValues;
	protected long[] lastInputRelevanceBoundValues;
	protected long[] nextInputRelevanceBoundValues;
	protected long[] lastOutputStreamBoundWriteValues;
	protected long[] nextOutputStreamBoundWriteValues;

	protected final int numberOfInputStreamBounds;
	protected final FixedArray<String> inputStreamBounds;
	protected long[] lastInputStreamBoundAndClockMinValues;
	protected long[] nextInputStreamBoundAndClockMinValues;
	protected final long[] clockValues;

	protected final long[] outputStreamBoundClockIncreaseValues;
	protected final TimeFunction streamBoundsFunction;
	protected final TimeFunction streamBoundsClockIncreasesFunction;
	
	protected final TimeFunction relevanceBoundsFunction;
	
	protected final ReentrantLock executionLock = new ReentrantLock();
	protected boolean chrashed = false;
	protected final Object boundValuesLock = new Object();
	
	protected boolean usePreparedStatements;
	
	protected final String eval_statement;
	protected boolean areSetEvalParameters = false;
	protected PreparedStatement sql_eval_statement;
	
	protected static enum ParameterSource{
		InputLastStreamBoundValues,
		InputNowStreamBoundValues,
		OutputLastStreamBoundValues,
		OutputNowStreamBoundValues,
		ClockLastStreamBoundValues,
		ClockNowMinValues,
		ClockNowValues,
		QueryInfo
	}
	
	protected final int numberOfParameters;
	protected final ParameterSource[] parameterMappingSource;
	protected final int[] parameterMappingLocalIndex;
	
	protected final String updateStatement;
	protected boolean areSetMetaParameters = false;
	protected PreparedStatement MetaEntryUpdate;

	protected final long[] lastTupleCounts;
	protected final FixedArrayNavigableMap<String, SQL_StreamBuffer> inputs;
	
	protected FixedArrayNavigableMap<String, SQL_StreamBuffer> outputs;
	
	protected final boolean isMonotonic;
	
	protected Statement loggingStatement;
	
	public SQL_ExecutableQuery(final SQL_ExecutableProgram program, final int id, final DTSA_QueryType type, final String prog_package, final boolean isMonotonic, final String sql_execution_statement, final FixedArrayNavigableSet<String> referencedClocks, final TimeFunction minTimesFunction, final TimeFunction minTimeClockIncreasesFunction, final TimeFunction relevanceBoundsFunction, final DTSA_BufferType input_type, final FixedArrayNavigableSet<String> inputs, final FixedArrayNavigableSet<String> outputs, final long lastRound, final long[] inputStreamBoundValues, final long[] clockMinValues){
		this(program, id, type, prog_package, isMonotonic, sql_execution_statement, referencedClocks, minTimesFunction, minTimeClockIncreasesFunction, relevanceBoundsFunction, input_type, inputs, outputs, lastRound, inputStreamBoundValues, clockMinValues, program.usePreparedStatements);
	}

	public SQL_ExecutableQuery(final SQL_ExecutableProgram program, final int id, final DTSA_QueryType type, final String prog_package, final boolean isMonotonic, final String sql_execution_statement, final FixedArrayNavigableSet<String> referencedClocks, final TimeFunction minTimesFunction, final TimeFunction minTimeClockIncreasesFunction, final TimeFunction relevanceBoundsFunction, final DTSA_BufferType input_type, final FixedArrayNavigableSet<String> inputs, final FixedArrayNavigableSet<String> outputs, final long lastRound, final long[] inputStreamBoundValues, final long[] clockMinValues, final boolean usePreparedStatements)
	{
		super(type, prog_package, program);
		
		FixedEnumMap<DTSA_BufferType,FixedArrayNavigableMap<String, SQL_StreamBuffer>> availableStreamBuffers = program.StreamBuffers();
		FixedArrayNavigableMap<String, Clock> availableClocks = program.clocks;
		
		
		this.id = id;
		this.sql_translation = program.sql_translation;
		
		//FIXME retrieve name of sql attribute generically
		this.id_attribute = SQL_Translation.SQL_BasicAttributeName(DTSA_RecoveryEnhacer.tsa_recovery_query_id);
		
		this.lastRound = lastRound;
		//FIXME retrieve name of sql attribute generically
		this.round_attribute = SQL_Translation.SQL_BasicAttributeName(DTSA_RecoveryEnhacer.tsa_recovery_round);
		
		SQL_StreamBuffer[] outputBuffers = new SQL_StreamBuffer[outputs.size()];
		
		if(outputBuffers.length > 0){
			DTSA_BufferType output_type = SQL_Translation.outputBufferType(this.type);
			
			for(int index = 0; index < outputBuffers.length; ++index){
				String output = outputs.get(index);
				int streamBufferIndex = availableStreamBuffers.get(output_type).keys().indexOf(output);
				if(streamBufferIndex >= 0){
					outputBuffers[index] = availableStreamBuffers.get(output_type).values().get(streamBufferIndex);
				}
				else{
					throw new RuntimeException("Query " + this.id + "requires buffer \"" + output + "\", but \"" + output + "\" is not available in the program.");
				}
			}
		}
		
		this.outputs = new FixedArrayNavigableMap<String, SQL_StreamBuffer>(outputs, new FixedArray<SQL_StreamBuffer>(false, outputBuffers));
		
		for(SQL_StreamBuffer output: this.outputs.values()){
			output.RegisterInputQuery(this);
		}
		
		ArrayList<String> outputStreamBounds = new ArrayList<String>();
		int[] outputNumberOfPreceedingStreamBoundIdentifiers = new int[outputs.size()];
		for(int output = 0; output < outputNumberOfPreceedingStreamBoundIdentifiers.length; ++output){
			SQL_StreamBuffer outputBuffer = this.outputs.values().get(output);
			outputNumberOfPreceedingStreamBoundIdentifiers[output] = outputStreamBounds.size();
			
			outputStreamBounds.addAll(outputBuffer.stream_bounds);
		}
		this.numberOfOutputStreamBounds = outputStreamBounds.size();
		this.outputStreamBounds = new FixedArray<String>(outputStreamBounds.toArray(new String[this.numberOfOutputStreamBounds]));
		
		this.lastOutputStreamBoundValues = new long[this.numberOfOutputStreamBounds];
		this.nextOutputStreamBoundValues = new long[this.numberOfOutputStreamBounds];
		this.lastOutputStreamBoundWriteValues = new long[this.numberOfOutputStreamBounds];
		this.nextOutputStreamBoundWriteValues = new long[this.numberOfOutputStreamBounds];
		
		if(!SQL_Translation.isPossibleInputBuffersType(type, input_type)){
			throw new RuntimeException("The type of query " + this.id + " and its specified input type are incompatible.");
		}
		
		SQL_StreamBuffer[] inputBuffers = new SQL_StreamBuffer[inputs.size()];
		for(int index = 0; index < inputBuffers.length; ++index){
			String input = inputs.get(index);
			int streamBufferIndex = availableStreamBuffers.get(input_type).keys().indexOf(input);
			if(streamBufferIndex >= 0){
				inputBuffers[index] = availableStreamBuffers.get(input_type).values().get(streamBufferIndex);
			}
			else{
				throw new RuntimeException("Query " + this.id + "requires buffer \"" + input + "\", but \"" + input + "\" is not available in the program.");
			}
		}		
		this.inputs = new FixedArrayNavigableMap<String, SQL_StreamBuffer>(inputs, new FixedArray<SQL_StreamBuffer>(false, inputBuffers));
		for(SQL_StreamBuffer input: this.inputs.values()){
			input.RegisterOutputQuery(this);
		}
		this.lastTupleCounts = new long[inputs.size()];
		Arrays.fill(this.lastTupleCounts, Long.MIN_VALUE);//Will lead to at most one unnecessary query execution
		
		ArrayList<String> inputStreamBounds = new ArrayList<String>();
		int[] inputNumberOfPreceedingStreamBoundIdentifiers = new int[inputs.size()];
		for(int input = 0; input < inputNumberOfPreceedingStreamBoundIdentifiers.length; ++input){
			SQL_StreamBuffer inputBuffer = this.inputs.values().get(input);
			inputNumberOfPreceedingStreamBoundIdentifiers[input] = inputStreamBounds.size();
			
			inputStreamBounds.addAll(inputBuffer.stream_bounds);
		}
		this.numberOfInputStreamBounds = inputStreamBounds.size();
		this.inputStreamBounds = new FixedArray<String>(inputStreamBounds.toArray(new String[this.numberOfInputStreamBounds]));

		if(!(this.numberOfInputStreamBounds == inputStreamBoundValues.length)){
			throw new IllegalArgumentException();
			//TODO better exception
		}
		
		this.numberOfReferencedClocks = referencedClocks.size();
		Clock[] clocks = new Clock[this.numberOfReferencedClocks];
		for(int clock = 0; clock < this.numberOfReferencedClocks; ++ clock){
			if(availableClocks.has(referencedClocks.get(clock))){
				clocks[clock] = availableClocks.get(referencedClocks.get(clock));
			}
			else{
				throw new IllegalArgumentException("Query " + this.id + "references unknown clock: " + referencedClocks.get(clock));
				//TODO better exception
			}
		}
		this.referencedClocks = new FixedArrayNavigableMap<String, Clock>(referencedClocks, new FixedArray<Clock>(false, clocks));
		
		this.clockValues = new long[this.numberOfReferencedClocks];
		
		this.lastInputStreamBoundAndClockMinValues = new long[this.numberOfInputStreamBounds + this.numberOfReferencedClocks];
		this.nextInputStreamBoundAndClockMinValues = new long[this.lastInputStreamBoundAndClockMinValues.length];
		
		System.arraycopy(inputStreamBoundValues, 0, this.lastInputStreamBoundAndClockMinValues, 0, this.numberOfInputStreamBounds);
		System.arraycopy(clockMinValues, 0, this.lastInputStreamBoundAndClockMinValues, this.numberOfInputStreamBounds, this.numberOfReferencedClocks);
		
		this.outputStreamBoundClockIncreaseValues = new long[this.numberOfOutputStreamBounds];
		
		this.streamBoundsFunction = minTimesFunction;
		assert this.streamBoundsFunction.argumentArity() == this.numberOfInputStreamBounds + this.numberOfReferencedClocks;
		assert this.streamBoundsFunction.resultArity() == this.numberOfOutputStreamBounds: "Query " + this.id + ": The query has to compute exactly as many output stream bound values as required by its output type.";
		this.streamBoundsClockIncreasesFunction = minTimeClockIncreasesFunction;
		assert this.streamBoundsClockIncreasesFunction.argumentArity() == this.numberOfInputStreamBounds + this.numberOfReferencedClocks;
		assert this.streamBoundsClockIncreasesFunction.resultArity() == this.numberOfOutputStreamBounds;

		this.streamBoundsFunction.apply(this.lastInputStreamBoundAndClockMinValues, 0, this.lastOutputStreamBoundValues, 0);
		this.updateOutputStreamBoundWriteValues();
		
		this.lastInputRelevanceBoundValues = new long[this.numberOfInputStreamBounds];
		this.nextInputRelevanceBoundValues = new long[this.numberOfInputStreamBounds];
		
		this.relevanceBoundsFunction = relevanceBoundsFunction;
		assert this.relevanceBoundsFunction.argumentArity() == this.numberOfOutputStreamBounds;
		assert this.relevanceBoundsFunction.resultArity() == this.numberOfInputStreamBounds: "Query " + this.id + ": The query has to compute exactly as many input relevance values as required by its input types.";
		this.relevanceBoundsFunction.apply(this.lastOutputStreamBoundValues, 0, this.lastInputRelevanceBoundValues, 0);
		
		assert type != DTSA_QueryType.STATIC | (this.numberOfInputStreamBounds == 0 & this.numberOfOutputStreamBounds == 0);
		
		this.queryInfo[0] = this.id;
		this.queryInfo[1] = this.lastRound;
		
		String[] parameters = this.sql_translation.parametersOfInstanciatedSQLStatement(sql_execution_statement);
		//this.parameter_splited_eval_statement = sql_translation.parameterSplitedInstanciatedSQLStatement(sql_execution_statement);

		this.numberOfParameters = parameters.length;
		this.parameterMappingSource = new ParameterSource[this.numberOfParameters];
		this.parameterMappingLocalIndex = new int[this.numberOfParameters];
		
		for(int parameterIndex = 0; parameterIndex < this.numberOfParameters; ++parameterIndex){
			String parameter = parameters[parameterIndex];
			assert parameter.matches(SQL_Translation.parameter_qualified_regEx);
			
			Matcher parameterMatcher = SQL_Translation.parameter_qualified_Pattern.matcher(parameter);
			parameterMatcher.matches();
			
			int[] groupIndexOf = new int[]{1,6,11,16,21,23,25,27};
			if(parameterMatcher.group(groupIndexOf[ParameterSource.InputLastStreamBoundValues.ordinal()]) != null){
				int groupIndex =  groupIndexOf[ParameterSource.InputLastStreamBoundValues.ordinal()];
				assert parameterMatcher.group(groupIndex).equals("input.last");
				++groupIndex;
				
				String bufferName = parameterMatcher.group(groupIndex);
				int buffer_index = inputs.indexOf(bufferName);
				assert buffer_index >= 0;
				groupIndex += 3;
				String attributeName = parameterMatcher.group(groupIndex);
				int stream_bound_index = this.inputs.values().get(buffer_index).stream_bounds.indexOf(TSA_SyntaxPatterns.streamBoundQualifiedName(bufferName, attributeName));
				if(stream_bound_index < 0){
					throw new IllegalArgumentException("Unknown min-time \"" + parameterMatcher.group(groupIndex)+ "\"");
				}
				
				this.parameterMappingSource[parameterIndex] = ParameterSource.InputLastStreamBoundValues;
				this.parameterMappingLocalIndex[parameterIndex] = inputNumberOfPreceedingStreamBoundIdentifiers[buffer_index] + stream_bound_index;
				assert this.parameterMappingLocalIndex[parameterIndex] >= 0; 
				++groupIndex;
			}
			else if(parameterMatcher.group(groupIndexOf[ParameterSource.InputNowStreamBoundValues.ordinal()]) != null){
				int groupIndex =  groupIndexOf[ParameterSource.InputNowStreamBoundValues.ordinal()];
				assert parameterMatcher.group(groupIndex).equals("input.now");
				++groupIndex;
				
				String bufferName = parameterMatcher.group(groupIndex);
				int buffer_index = inputs.indexOf(bufferName);
				assert buffer_index >= 0;
				groupIndex += 3;
				String attributeName = parameterMatcher.group(groupIndex);
				int stream_bound_index = this.inputs.values().get(buffer_index).stream_bounds.indexOf(TSA_SyntaxPatterns.streamBoundQualifiedName(bufferName, attributeName));
				if(stream_bound_index < 0){
					throw new IllegalArgumentException("Unknown stream bound \"" + parameterMatcher.group(groupIndex)+ "\"");
				}
				
				this.parameterMappingSource[parameterIndex] = ParameterSource.InputNowStreamBoundValues;
				this.parameterMappingLocalIndex[parameterIndex] = inputNumberOfPreceedingStreamBoundIdentifiers[buffer_index] + stream_bound_index;
				assert this.parameterMappingLocalIndex[parameterIndex] >= 0;
				++groupIndex;
			}
			else if(parameterMatcher.group(groupIndexOf[ParameterSource.OutputLastStreamBoundValues.ordinal()]) != null){
				int groupIndex =  groupIndexOf[ParameterSource.OutputLastStreamBoundValues.ordinal()];
				assert parameterMatcher.group(groupIndex).equals("output.last");
				++groupIndex;
				
				String bufferName = parameterMatcher.group(groupIndex);
				int buffer_index = outputs.indexOf(bufferName);
				assert buffer_index >= 0;
				groupIndex += 3;
				String attributeName = parameterMatcher.group(groupIndex);
				int stream_bound_index = this.outputs.values().get(buffer_index).stream_bounds.indexOf(TSA_SyntaxPatterns.streamBoundQualifiedName(bufferName, attributeName));
				if(stream_bound_index < 0){
					throw new IllegalArgumentException("Unknown stream bound \"" + parameterMatcher.group(groupIndex)+ "\"");
				}
				
				this.parameterMappingSource[parameterIndex] = ParameterSource.OutputLastStreamBoundValues;
				this.parameterMappingLocalIndex[parameterIndex] = outputNumberOfPreceedingStreamBoundIdentifiers[buffer_index] + stream_bound_index;
				assert this.parameterMappingLocalIndex[parameterIndex] >= 0;
				++groupIndex;
			}
			else if(parameterMatcher.group(groupIndexOf[ParameterSource.OutputNowStreamBoundValues.ordinal()]) != null){
				int groupIndex =  groupIndexOf[ParameterSource.OutputNowStreamBoundValues.ordinal()];
				assert parameterMatcher.group(groupIndex).equals("output.now");
				++groupIndex;
				
				String bufferName = parameterMatcher.group(groupIndex);
				int buffer_index = outputs.indexOf(bufferName);
				assert buffer_index >= 0;
				groupIndex += 3;
				String attributeName = parameterMatcher.group(groupIndex);
				int stream_bound_index = this.outputs.values().get(buffer_index).stream_bounds.indexOf(TSA_SyntaxPatterns.streamBoundQualifiedName(bufferName, attributeName));
				if(stream_bound_index < 0){
					throw new IllegalArgumentException("Unknown stream bound \"" + parameterMatcher.group(groupIndex)+ "\"");
				}
				
				this.parameterMappingSource[parameterIndex] = ParameterSource.OutputNowStreamBoundValues;
				this.parameterMappingLocalIndex[parameterIndex] = outputNumberOfPreceedingStreamBoundIdentifiers[buffer_index] + stream_bound_index;
				assert this.parameterMappingLocalIndex[parameterIndex] >= 0;
				++groupIndex;
			}
			else if(parameterMatcher.group(groupIndexOf[ParameterSource.ClockLastStreamBoundValues.ordinal()]) != null){
				int groupIndex =  groupIndexOf[ParameterSource.ClockLastStreamBoundValues.ordinal()];
				assert parameterMatcher.group(groupIndex).equals("clock.min.last");
				++groupIndex;
				
				int clock_index = this.referencedClocks.keys().indexOf(parameterMatcher.group(groupIndex));
				if(clock_index < 0){
					throw new IllegalArgumentException("Unknown clock \"" + parameterMatcher.group(groupIndex)+ "\"");
				}
				
				this.parameterMappingSource[parameterIndex] = ParameterSource.ClockLastStreamBoundValues;
				this.parameterMappingLocalIndex[parameterIndex] = this.numberOfInputStreamBounds + clock_index;
				++groupIndex;
			}
			else if(parameterMatcher.group(groupIndexOf[ParameterSource.ClockNowMinValues.ordinal()]) != null){
				int groupIndex =  groupIndexOf[ParameterSource.ClockNowMinValues.ordinal()];
				assert parameterMatcher.group(groupIndex).equals("clock.min.now");
				++groupIndex;
				
				int clock_index = this.referencedClocks.keys().indexOf(parameterMatcher.group(groupIndex));
				if(clock_index < 0){
					throw new IllegalArgumentException("Unknown clock \"" + parameterMatcher.group(groupIndex)+ "\"");
				}
				
				this.parameterMappingSource[parameterIndex] = ParameterSource.ClockNowMinValues;
				this.parameterMappingLocalIndex[parameterIndex] = this.numberOfInputStreamBounds + clock_index;
				++groupIndex;
			}
			else if(parameterMatcher.group(groupIndexOf[ParameterSource.ClockNowValues.ordinal()]) != null){
				int groupIndex =  groupIndexOf[ParameterSource.ClockNowValues.ordinal()];
				assert parameterMatcher.group(groupIndex).equals("clock.now");
				++groupIndex;
				
				int clock_index = this.referencedClocks.keys().indexOf(parameterMatcher.group(groupIndex));
				if(clock_index < 0){
					throw new IllegalArgumentException("Unknown clock \"" + parameterMatcher.group(groupIndex)+ "\"");
				}
				
				this.parameterMappingSource[parameterIndex] = ParameterSource.ClockNowValues;
				this.parameterMappingLocalIndex[parameterIndex] = clock_index;
				++groupIndex;
			}
			else if(parameterMatcher.group(groupIndexOf[ParameterSource.QueryInfo.ordinal()]) != null){
				int groupIndex =  groupIndexOf[ParameterSource.QueryInfo.ordinal()];
				assert parameterMatcher.group(groupIndex).equals("query");
				++groupIndex;
				
				int index;
				if(parameterMatcher.group(groupIndex).equals("id")){
					index = 0;
				}
				else{
					assert parameterMatcher.group(groupIndex).equals("round");
					index = 1;
				}
				
				this.parameterMappingSource[parameterIndex] = ParameterSource.QueryInfo;
				this.parameterMappingLocalIndex[parameterIndex] = index;
				++groupIndex;
			}
			else{
				throw new Error();
			}
		}
		
		this.eval_statement = sql_execution_statement;		
		this.updateStatement = this.sql_translation.sqlUpdateQueryMetaEntry(this.program.meta_schema, this.id, this.numberOfOutputStreamBounds, this.numberOfInputStreamBounds, this.numberOfReferencedClocks);
		
		this.isMonotonic = isMonotonic;		
		
		this.program.register(this);
		
		this.usePreparedStatements = usePreparedStatements;
		this.initStatements();
	}
	
	protected void atStartInitLock(){
		this.executionLock.lock();
	}
	
	protected void atStartInitClockMinValues(final long[] clockValues){		
		assert this.executionLock.isHeldByCurrentThread();

		for(int referencedClock = 0; referencedClock < this.referencedClocks.size(); ++referencedClock){
			assert this.program.clocks.has(this.referencedClocks.keys().get(referencedClock));
			this.lastInputStreamBoundAndClockMinValues[this.numberOfInputStreamBounds + referencedClock] = clockValues[this.program.clocks.keys().indexOf(this.referencedClocks.keys().get(referencedClock))];
		}
		
		this.streamBoundsClockIncreasesFunction.apply(this.lastInputStreamBoundAndClockMinValues, 0, this.outputStreamBoundClockIncreaseValues, 0);
	}
	
	protected void atStartInitStreamBoundValues(){
		assert this.executionLock.isHeldByCurrentThread();
		if(this.type.isEvalQuery){
			Arrays.fill(this.lastOutputStreamBoundValues, SQL_CONSTANTS.BIGINT.MAX_VALUE);
			Arrays.fill(this.lastOutputStreamBoundWriteValues, SQL_CONSTANTS.BIGINT.MAX_VALUE);
		}
		else{assert this.type.isGarbageQuery;
			Arrays.fill(this.lastOutputStreamBoundValues, SQL_CONSTANTS.BIGINT.MIN_VALUE);
			Arrays.fill(this.lastOutputStreamBoundWriteValues, SQL_CONSTANTS.BIGINT.MIN_VALUE);
		}
	}
	
	protected boolean atStartInitPropagateStreamBoundValues(final long[] clockValues){
		assert this.executionLock.isHeldByCurrentThread();
		//update input stream bound values
		boolean changed;
		
		if(this.type.isEvalQuery){
			int startIndex = 0;
			for(int inputIndex = 0; inputIndex < this.inputs.size(); ++inputIndex){
				SQL_StreamBuffer input = this.inputs.values().get(inputIndex);
				System.arraycopy(input.atStartInitWriteValues(clockValues), 0, this.lastInputStreamBoundAndClockMinValues, startIndex, input.numberOfStreamBounds);
				startIndex += input.numberOfStreamBounds;
			}
			
			//compute output stream bound values
			this.streamBoundsFunction.apply(this.lastInputStreamBoundAndClockMinValues, 0, this.lastOutputStreamBoundValues, 0);
			
			//combine with clock increases
			this.outputStreamBoundWriteValues(this.lastOutputStreamBoundValues, this.outputStreamBoundClockIncreaseValues, this.nextOutputStreamBoundWriteValues);
			
			//check if increased
			changed = false;
			for(int streamBound = 0; streamBound < this.lastOutputStreamBoundWriteValues.length; ++streamBound){
				if(this.nextOutputStreamBoundWriteValues[streamBound] < this.lastOutputStreamBoundWriteValues[streamBound]){
					changed = true;
					break;
				}
			}
			
			//make next to last
			long[] lastOutputStreamBoundWriteValues = this.lastOutputStreamBoundWriteValues;
			this.lastOutputStreamBoundWriteValues = this.nextOutputStreamBoundWriteValues;
			this.nextOutputStreamBoundWriteValues = lastOutputStreamBoundWriteValues;
		}
		else{assert this.type.isGarbageQuery;
			changed = false;
		}
		
		return changed;
	}
	
	protected void outputStreamBoundWriteValues(final long[] outputStreamBoundValues, final long[] outputStreamBoundValuesClockIncreases, final long[] outputStreamBoundWriteValues){
		long outputMinTime;
		long outputMinTimeClockIncrease;
		assert outputStreamBoundValues.length == this.numberOfOutputStreamBounds;
		assert outputStreamBoundValuesClockIncreases.length == this.numberOfOutputStreamBounds;
		assert outputStreamBoundWriteValues.length == this.numberOfOutputStreamBounds;
		assert this.type != DTSA_QueryType.STATIC | outputStreamBoundWriteValues.length == 0;
		//FIXME INIT queries after first and only execution return PlusInfty
		for(int outputMinTimeIndex = 0; outputMinTimeIndex < this.numberOfOutputStreamBounds; ++outputMinTimeIndex){
			outputMinTime = outputStreamBoundValues[outputMinTimeIndex];
			outputMinTimeClockIncrease = outputStreamBoundValuesClockIncreases[outputMinTimeIndex];
			outputStreamBoundWriteValues[outputMinTimeIndex] = (outputMinTime < outputMinTimeClockIncrease ? outputMinTimeClockIncrease: outputMinTime);
		}
	}
	
	protected void atStartInitSetMetaEntry() throws SQLException{
		this.updateMetaEntry(-1, this.lastOutputStreamBoundWriteValues, this.lastInputStreamBoundAndClockMinValues);
	}
	
	protected void atStartInitUnlock(){
		this.executionLock.unlock();
	}
	
	
	protected void initStatements(){
		this.executionLock.lock();
		try{
			int tryCount = 2;
			
			while(true){
				Connection connection = this.program.bufferStore.connection();
				try
				{
					this.sql_eval_statement = this.sql_translation.prepareStatement(connection, this.eval_statement, this.usePreparedStatements);
					this.MetaEntryUpdate = this.sql_translation.prepareStatement(connection, this.updateStatement, this.usePreparedStatements);
					this.areSetEvalParameters = false;
					this.areSetMetaParameters = false;
					
					if(this.logger().isDebugEnabled()){
						if(this.loggingStatement != null){
							try{this.loggingStatement.close();} catch (SQLException e){}
						}
						this.loggingStatement = connection.createStatement();
					}
					break;
				} catch (SQLException e)
				{
					if(SQL_EXCEPTIONS.isConnectionFailure(e)){
						this.debug("QUERY " + this.id + " - Connection failure. Trying again.", e);
						this.program.bufferStore.reconnect(connection);
					}
					else if(SQL_EXCEPTIONS.isMonetDBBug(e) & tryCount > 0){
						this.warn("QUERY " + this.id + " - MonetDB bug. Trying to reconnect.", e);
						this.program.bufferStore.reconnect(connection);
						--tryCount;
					}
					else{
						this.error("QUERY " + this.id, e);
						throw new RuntimeException(e);
						//TODO better exception
					}
				}
			}
		}
		finally{
			this.executionLock.unlock();
		}
	}

	@Override
	public FixedArrayNavigableMap<String, SQL_StreamBuffer> inputs()
	{
		return this.inputs;
	}

	@Override
	public FixedArrayNavigableMap<String, SQL_StreamBuffer> outputs()
	{
		return this.outputs;
	}
	
	protected void updateClockMinValues(final long[] inputStreamBoundAndClockMinValues){
		for(int clock = 0; clock < this.numberOfReferencedClocks; ++clock){
			inputStreamBoundAndClockMinValues[this.numberOfInputStreamBounds + clock] = this.referencedClocks.values().get(clock).min();	
		}
	}
	
	protected void updateOutputStreamBoundClockIncreaseValues(){
		this.executionLock.lock();
		try{
			this.updateClockMinValues(this.lastInputStreamBoundAndClockMinValues);
			this.streamBoundsClockIncreasesFunction.apply(this.lastInputStreamBoundAndClockMinValues, 0, this.outputStreamBoundClockIncreaseValues, 0);
		}
		finally{
			this.executionLock.unlock();
		}
	}

	@Override
	public void execute()
	{
		this.executionLock.lock();
		try{
			
			long currentRound = this.lastRound + 1;
			
			this.info("QUERY {} (execution {}) - Start", this.id, currentRound);			
				
				long[] currentInputStreamBoundAndClockMinValues;
				long[] currentOutputStreamBoundValues;
				long[] currentOutputStreamBoundWriteValues;
				long[] currentInputRelevanceBoundValues;
				long tupleNumber;
	
				currentInputStreamBoundAndClockMinValues = this.nextInputStreamBoundAndClockMinValues;
				currentOutputStreamBoundValues = this.nextOutputStreamBoundValues;
				currentOutputStreamBoundWriteValues = this.nextOutputStreamBoundWriteValues;			
				currentInputRelevanceBoundValues = this.nextInputRelevanceBoundValues;
	
				//get current values of input stream bounds of query
				boolean mightComputeNewTuples = false;
				int startIndex = 0;
				for(int inputIndex = 0; inputIndex < this.inputs.size(); ++inputIndex){
					SQL_StreamBuffer input = this.inputs.values().get(inputIndex);
					synchronized (input) {
						long[] boundValues;
						if(this.type.isEvalQuery){
							boundValues = input.streamBoundWriteValues();
						}
						else{assert this.type.isGarbageQuery;
							boundValues = input.streamBoundReadValues();
						}
						System.arraycopy(boundValues, 0, currentInputStreamBoundAndClockMinValues, startIndex, input.numberOfStreamBounds);//update inputMinTimes
						long inputTupleCount = input.tupleCount(); 
						if(inputTupleCount > this.lastTupleCounts[inputIndex]){//test if input contains any new tuple compared with the last execution of the query
							this.lastTupleCounts[inputIndex] = inputTupleCount;
							mightComputeNewTuples = true;
						}
					}
					startIndex += input.numberOfStreamBounds;
				}
				assert startIndex == this.numberOfInputStreamBounds;
				
				//get current values of clock min times
				this.updateClockMinValues(currentInputStreamBoundAndClockMinValues);
				
				if(!this.isMonotonic){//FIXME or !hasMonotonicStreamBounds
					mightComputeNewTuples = true;
					//FIXME
				}
				if(this.type == DTSA_QueryType.STATIC | this.type == DTSA_QueryType.INIT | this.type == DTSA_QueryType.INPUT){
					mightComputeNewTuples = true;
					//FIXME					
				}
				//compute current values of output stream bounds for query
				this.streamBoundsFunction.apply(currentInputStreamBoundAndClockMinValues, 0, currentOutputStreamBoundValues, 0);
				
				//compute current values of input stream bounds for query
				this.relevanceBoundsFunction.apply(currentOutputStreamBoundValues, 0, currentInputRelevanceBoundValues, 0);
				
				
				{//update query info
					this.queryInfo[1] = currentRound;
				}
				
				if(this.logger().isDebugEnabled()){
					this.debug("QUERY {} (execution {})\n" + this.valuesToString(this.lastInputStreamBoundAndClockMinValues, currentInputStreamBoundAndClockMinValues, currentOutputStreamBoundValues), this.id, Long.valueOf(currentRound));
				}
				if(this.logger().isTraceEnabled()){
					
					for(SQL_StreamBuffer input: this.inputs().values()){
						while(true){
							try{
								ResultSet tuples = this.loggingStatement.executeQuery(this.sql_translation.sqlReadBufferContent(input.schema, input.table_name));
								String tuplesTable = SQL_QueryBuildHelper.toString(tuples, -1);
								if(this.type.isEvalQuery){
									this.trace("QUERY {} (execution {}) - Snapshot of table \"{}\".\"{}\" before the execution of the eval statement:\n{}", this.id, currentRound, input.schema, input.table_name, tuplesTable);
								}
								else{assert this.type.isGarbageQuery;
									assert this.outputs.size() == 0;
									assert this.inputs.size() == 1;
									this.trace("QUERY {} (execution {}) - Snapshot of table \"{}\".\"{}\" before the execution of the garbage collection statement:\n{}", this.id, currentRound, input.schema, input.table_name, tuplesTable);
								}
								break;
							}
							catch(SQLException e){
								this.trace("QUERY " + this.id + " (execution " + currentRound + ") - SQLSTATE: {} - MESSAGE: {}", e.getSQLState(), e.getMessage());
								if(SQL_EXCEPTIONS.isConnectionFailure(e)){
									this.debug("QUERY " + this.id + " (execution " + currentRound + ") - Connection failure. Trying again.", e);
									this.program.bufferStore.reconnect(this.sql_eval_statement.getConnection());
									this.initStatements();
								}
								else if(SQL_EXCEPTIONS.isRollback(e)){
									this.debug("QUERY " + this.id + " (execution " + currentRound + ") - SQL debug table content statement rolled back. Trying again.", e);
								}
								else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
									this.debug("QUERY " + this.id + " (execution " + currentRound + ") - Handle of prepared SQL eval statement is invalid. Repreparing statement and trying again.", e);
									this.initStatements();
									//TODO only init the affected statement?
								}
								else{
									this.chrashed = true;
									throw e;
								}
							}				
						}
					}
					
				}
				
				if(mightComputeNewTuples){
					
					
					{//update clocks
						for(int clock = 0; clock < this.numberOfReferencedClocks; ++clock){
							this.clockValues[clock] = this.referencedClocks.values().get(clock).now();
						}								
					}
					if(this.type.isEvalQuery){
						this.debug("QUERY {} (execution {}) - Executing eval statement.", this.id, currentRound);
					}
					else{assert this.type.isGarbageQuery;
						this.debug("QUERY {} (execution {}) - Executing garbage collection statement.", this.id, currentRound);
						
						if(this.logger().isDebugEnabled()){
							assert this.outputs.size() == 0;
							assert this.inputs.size() == 1;
							
							SQL_StreamBuffer input = this.inputs.values().get(0);
							
							while(true){
								try{
									SimulatedPreparedStatement logSQL = new SimulatedPreparedStatement(this.program.bufferStore.driver, this.loggingStatement, this.eval_statement.replaceFirst("DELETE FROM", "SELECT * FROM"));
									for(int parameter = 0; parameter < this.numberOfParameters; ++parameter){						
										logSQL.setLong(parameter + 1, currentInputStreamBoundAndClockMinValues[this.parameterMappingLocalIndex[parameter]]);
									}
									ResultSet newTuples = this.loggingStatement.executeQuery(logSQL.executableStatement());
									String oldTuplesTable = SQL_QueryBuildHelper.toString(newTuples, -1);
									this.debug("QUERY {} (execution {}) - Garbage collection will delete the following tuples from table \"{}\".\"{}\":\n{}", this.id, currentRound, input.schema, input.table_name, oldTuplesTable);
									break;
								}
								catch(SQLException e){
									this.trace("QUERY " + this.id + " (execution " + currentRound + ") - SQLSTATE: {} - MESSAGE: {}", e.getSQLState(), e.getMessage());
									if(SQL_EXCEPTIONS.isConnectionFailure(e)){
										this.debug("QUERY " + this.id + " (execution " + currentRound + ") - Connection failure. Trying again.", e);
										this.program.bufferStore.reconnect(this.sql_eval_statement.getConnection());
										this.initStatements();
									}
									else if(SQL_EXCEPTIONS.isRollback(e)){
										this.debug("QUERY " + this.id + " (execution " + currentRound + ") - SQL debug table content statement rolled back. Trying again.", e);
									}
									else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
										this.debug("QUERY " + this.id + " (execution " + currentRound + ") - Handle of prepared SQL eval statement is invalid. Repreparing statement and trying again.", e);
										this.initStatements();
										//TODO only init the affected statement?
									}
									else{
										this.chrashed = true;
										throw e;
									}
								}				
							}
						}
							
					}					
					
					while(true){
						try{
							if(!this.areSetEvalParameters){
								long[] parameterSource;
								for(int parameter = 0; parameter < this.numberOfParameters; ++parameter){
			
									switch(this.parameterMappingSource[parameter]){
										case InputLastStreamBoundValues:
											parameterSource = this.lastInputStreamBoundAndClockMinValues;
											break;
										case InputNowStreamBoundValues:
											parameterSource = currentInputStreamBoundAndClockMinValues;
											break;
										case OutputLastStreamBoundValues:
											parameterSource = this.lastOutputStreamBoundValues;
											break;
										case OutputNowStreamBoundValues:
											parameterSource = currentOutputStreamBoundValues;
											break;
										case ClockLastStreamBoundValues:
											parameterSource = this.lastInputStreamBoundAndClockMinValues;
											break;
										case ClockNowMinValues:
											parameterSource = currentInputStreamBoundAndClockMinValues;
											break;
										case ClockNowValues:
											parameterSource = this.clockValues;
											break;
										case QueryInfo:
											parameterSource = this.queryInfo;
											break;
										default:
											throw new Error();
									}
									
									this.sql_eval_statement.setLong(parameter + 1, parameterSource[this.parameterMappingLocalIndex[parameter]]);
								}
								this.areSetEvalParameters = true;
							}
							
							tupleNumber = this.sql_eval_statement.executeUpdate();
							this.areSetEvalParameters = false;
							
							assert tupleNumber >= 0;
							break;
						}
						catch(SQLException e){
							this.trace("QUERY " + this.id + " (execution " + currentRound + ") - SQLSTATE: {} - MESSAGE: {}", e.getSQLState(), e.getMessage());
							if(SQL_EXCEPTIONS.isConnectionFailure(e)){
								this.debug("QUERY " + this.id + " (execution " + currentRound + ") - Connection failure. Trying again.", e);
								this.program.bufferStore.reconnect(this.sql_eval_statement.getConnection());
								this.initStatements();
							}
							else if(SQL_EXCEPTIONS.isRollback(e)){
								this.debug("QUERY " + this.id + " (execution " + currentRound + ") - SQL statement rolled back. Trying again.", e);
							}
							else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
								this.debug("QUERY " + this.id + " (execution " + currentRound + ") - Handle of prepared SQL eval statement is invalid. Repreparing statement and trying again.", e);
								this.initStatements();
								//TODO only init the affected statement?
							}
							else{
								this.chrashed = true;
								throw e;
							}
						}				
					}
					
					if(this.type.isEvalQuery){
						this.debug("QUERY {} (execution {}) - Executed eval statement.", this.id, currentRound);
					}
					else{assert this.type.isGarbageQuery;
						this.debug("QUERY {} (execution {}) - Executed garbage collection statement.", this.id, currentRound);
					}
						
					try{ this.sql_eval_statement.clearParameters(); } catch(SQLException e){}
					
					if(this.logger().isDebugEnabled()){
						
						if(this.type.isEvalQuery){
							assert this.outputs.size() == 1;
							SQL_StreamBuffer output = this.outputs.values().get(0);
							
							this.debug("QUERY {} (execution {}) - Evaluation inserted {} tuples into table \"{}\".\"{}\".", this.id, currentRound, tupleNumber, output.schema, output.table_name);
							
							while(true){
								try{
									ResultSet newTuples = this.loggingStatement.executeQuery(this.sql_translation.sqlReadBufferContent(output.schema, output.table_name, this.id_attribute, this.round_attribute, this.id, currentRound));
									String newTuplesTable = SQL_QueryBuildHelper.toString(newTuples, -1);
									this.debug("QUERY {} (execution {}) - Evaluation inserted the following tuples into table \"{}\".\"{}\":\n{}", this.id, currentRound, output.schema, output.table_name, newTuplesTable);
									break;
								}
								catch(SQLException e){
									this.trace("QUERY " + this.id + " (execution " + currentRound + ") - SQLSTATE: {} - MESSAGE: {}", e.getSQLState(), e.getMessage());
									if(SQL_EXCEPTIONS.isConnectionFailure(e)){
										this.debug("QUERY " + this.id + " (execution " + currentRound + ") - Connection failure. Trying again.", e);
										this.program.bufferStore.reconnect(this.sql_eval_statement.getConnection());
										this.initStatements();
									}
									else if(SQL_EXCEPTIONS.isRollback(e)){
										this.debug("QUERY " + this.id + " (execution " + currentRound + ") - SQL debug table content statement rolled back. Trying again.", e);
									}
									else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
										this.debug("QUERY " + this.id + " (execution " + currentRound + ") - Handle of prepared SQL eval statement is invalid. Repreparing statement and trying again.", e);
										this.initStatements();
										//TODO only init the affected statement?
									}
									else{
										this.chrashed = true;
										throw e;
									}
								}				
							}
						}
						else{assert this.type.isGarbageQuery;				
							assert this.outputs.size() == 0;
							assert this.inputs.size() == 1;
							
							SQL_StreamBuffer input = this.inputs.values().get(0);
							
							this.debug("QUERY {} (execution {}) - Garbage collection deleted {} tuples from table \"{}\".\"{}\".", this.id, currentRound, tupleNumber, input.schema, input.table_name);
						}
						
					}
					
				}
				else{
					this.debug("QUERY {} (execution {}) - Execution of eval statement is omited because no new tuples could have been produced.", this.id, currentRound);
					tupleNumber = 0;
				}
				
				if(this.logger().isTraceEnabled()){
					
					if(this.type.isEvalQuery){
						assert this.outputs.size() == 1;
						SQL_StreamBuffer output = this.outputs.values().get(0);
					
						while(true){
							try{
								ResultSet tuples = this.loggingStatement.executeQuery(this.sql_translation.sqlReadBufferContent(output.schema, output.table_name));
								String tuplesTable = SQL_QueryBuildHelper.toString(tuples, -1);
								this.trace("QUERY {} (execution {}) - Snapshot of table \"{}\".\"{}\" after the execution of the eval statement:\n{}", this.id, currentRound, output.schema, output.table_name, tuplesTable);
								break;
							}
							catch(SQLException e){
								this.trace("QUERY " + this.id + " (execution " + currentRound + ") - SQLSTATE: {} - MESSAGE: {}", e.getSQLState(), e.getMessage());
								if(SQL_EXCEPTIONS.isConnectionFailure(e)){
									this.debug("QUERY " + this.id + " (execution " + currentRound + ") - Connection failure. Trying again.", e);
									this.program.bufferStore.reconnect(this.sql_eval_statement.getConnection());
									this.initStatements();
								}
								else if(SQL_EXCEPTIONS.isRollback(e)){
									this.debug("QUERY " + this.id + " (execution " + currentRound + ") - SQL debug table content statement rolled back. Trying again.", e);
								}
								else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
									this.debug("QUERY " + this.id + " (execution " + currentRound + ") - Handle of prepared SQL eval statement is invalid. Repreparing statement and trying again.", e);
									this.initStatements();
									//TODO only init the affected statement?
								}
								else{
									this.chrashed = true;
									throw e;
								}
							}				
						}
					
					}
					else{assert this.type.isGarbageQuery;
						assert this.outputs.size() == 0;
						assert this.inputs.size() == 1;
						
						SQL_StreamBuffer input = this.inputs.values().get(0);

						while(true){
							try{
								ResultSet tuples = this.loggingStatement.executeQuery(this.sql_translation.sqlReadBufferContent(input.schema, input.table_name));
								String tuplesTable = SQL_QueryBuildHelper.toString(tuples, -1);
								this.trace("QUERY {} (execution {}) - Snapshot of table \"{}\".\"{}\" after the execution of the garbage collection statement:\n{}", this.id, currentRound, input.schema, input.table_name, tuplesTable);
								break;
							}
							catch(SQLException e){
								this.trace("QUERY " + this.id + " (execution " + currentRound + ") - SQLSTATE: {} - MESSAGE: {}", e.getSQLState(), e.getMessage());
								if(SQL_EXCEPTIONS.isConnectionFailure(e)){
									this.debug("QUERY " + this.id + " (execution " + currentRound + ") - Connection failure. Trying again.", e);
									this.program.bufferStore.reconnect(this.sql_eval_statement.getConnection());
									this.initStatements();
								}
								else if(SQL_EXCEPTIONS.isRollback(e)){
									this.debug("QUERY " + this.id + " (execution " + currentRound + ") - SQL debug table content statement rolled back. Trying again.", e);
								}
								else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
									this.debug("QUERY " + this.id + " (execution " + currentRound + ") - Handle of prepared SQL eval statement is invalid. Repreparing statement and trying again.", e);
									this.initStatements();
									//TODO only init the affected statement?
								}
								else{
									this.chrashed = true;
									throw e;
								}
							}				
						}
					}
					
				}
				
				//access to new outputMinTimes is still protected by execution lock 
				assert this.executionLock.isHeldByCurrentThread();
				
				this.updateOutputStreamBoundWriteValues(currentOutputStreamBoundValues, currentOutputStreamBoundWriteValues);
				
				this.debug("QUERY {} (execution {})\n\tOutputMinWriteTimes: {}", this.id, currentRound, currentOutputStreamBoundWriteValues);
				
				this.updateMetaEntry(currentRound, this.lastOutputStreamBoundWriteValues, currentInputStreamBoundAndClockMinValues);
				
				synchronized (this.boundValuesLock) {
					this.lastRound = currentRound;
					this.nextInputStreamBoundAndClockMinValues = this.lastInputStreamBoundAndClockMinValues;
					this.lastInputStreamBoundAndClockMinValues = currentInputStreamBoundAndClockMinValues;
					
					this.nextInputRelevanceBoundValues = this.lastInputRelevanceBoundValues;
					this.lastInputRelevanceBoundValues = currentInputRelevanceBoundValues;
					
					this.nextOutputStreamBoundValues = this.lastOutputStreamBoundValues;
					this.lastOutputStreamBoundValues = currentOutputStreamBoundValues;
				
					this.nextOutputStreamBoundWriteValues = this.lastOutputStreamBoundWriteValues;
					this.lastOutputStreamBoundWriteValues = currentOutputStreamBoundWriteValues;
				}

				
				if(this.type.isEvalQuery){
					assert this.outputs.size() == 1;
					this.outputs.values().get(0).notifyWrite(tupleNumber);			
					for(SQL_StreamBuffer input: this.inputs.values()){
						input.notifyRead();
					}
				}
				else{assert this.type.isGarbageQuery;
					assert this.outputs.size() == 0;
					assert this.inputs.size() == 1;
					for(SQL_StreamBuffer input: this.inputs.values()){
						input.notifyDelete(tupleNumber);
					}
				}
				
				//TODO better log message
				this.info("QUERY {} (execution {}) - Finished", this.id, currentRound);			
				this.progLogger.info("QUERY {} (execution {}) - Finished", this.id, currentRound);

		}
		catch (SQLException e)
		{			
			//FIXME throw exeption out of query and catch there? ExecutionException?
			//TODO more differentiated reaction
			//TODO better log message
			this.error("QUERY " + this.id, e);
			this.chrashed = true;
			throw new RuntimeException(e);
		}
		catch(RuntimeException e){
			this.error("QUERY " + this.id, e);
			this.chrashed = true;
			throw e;
		}
		catch(Error e){
			this.error("QUERY " + this.id, e);
			this.chrashed = true;
			throw e;
		}
		finally{//FIXME executes finally also when catch was sucessful?
			this.executionLock.unlock();
		}
	}
	
	protected void updateMetaEntry(final long round, final long[] outputStreamBoundValues, final long[] inputStreamBoundAndClockMinValues) throws SQLException {
		this.debug("QUERY {} (execution {}) - Updating meta entry.", this.id, round);
		int updateCount;
		while(true){
			try{
				this.sql_translation.sqlSetUpdateQueryMetaEntry(this.MetaEntryUpdate, round, outputStreamBoundValues, inputStreamBoundAndClockMinValues, this.numberOfInputStreamBounds);
				this.areSetMetaParameters = true;
				
				updateCount = this.MetaEntryUpdate.executeUpdate();
				this.areSetMetaParameters = false;
				
				assert updateCount >= 0;
				break;
			}
			catch(SQLException e){
				this.trace("QUERY " + this.id + " (execution " + round + ") - SQLSTATE: {} - MESSAGE: {}", e.getSQLState(), e.getMessage());
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("QUERY " + this.id + " (execution " + round + ") - Connection failure. Trying again.", e);
					this.program.bufferStore.reconnect(this.MetaEntryUpdate.getConnection());
					this.initStatements();
				}
				else if(SQL_EXCEPTIONS.isRollback(e)){
					this.debug("QUERY " + this.id + " (execution " + round + ") - Meta entry update rolled back. Trying again.", e);
				}
				else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
					this.debug("QUERY " + this.id + " (execution " + round + ") - Handle of prepared meta entry update statement is invalid. Repreparing statement and trying again.", e);
					this.initStatements();
					//TODO only init the affected statement?
				}
				else{
					this.chrashed = true;
					throw e;
				}
			}				
		}
		
		this.debug("QUERY {} (execution {}) - Updated meta entry.", this.id, round);
		try{ this.MetaEntryUpdate.clearParameters(); } catch(SQLException e){}
	}
	
	protected void updateOutputStreamBoundWriteValues(final long[] outputStreamBoundValues, final long[] outputStreamBoundWriteValues){
		this.executionLock.lock();
		try{//only perform update if there is at least one clock that increses the min times (same for buffer -> only if one input query ... else only when input query has send notify about execution)
			this.updateOutputStreamBoundClockIncreaseValues();
			
			this.outputStreamBoundWriteValues(outputStreamBoundValues, this.outputStreamBoundClockIncreaseValues, outputStreamBoundWriteValues);
		}
		finally{
			this.executionLock.unlock();
		}
	}
	
	protected void updateOutputStreamBoundWriteValues(){
		this.executionLock.lock();
		try{			
			synchronized (this.boundValuesLock) {
				this.updateOutputStreamBoundWriteValues(this.lastOutputStreamBoundValues, this.lastOutputStreamBoundWriteValues);
			}
		}
		finally{
			this.executionLock.unlock();
		}
	}

	/**
	 * @return
	 */
	protected long[] outputStreamBoundWriteValues()
	{
		boolean locked = this.executionLock.tryLock();
		try{
			if(locked & !this.chrashed){//currently not executing
				this.updateOutputStreamBoundWriteValues();
			}
			long[] outputMinWriteTimes = new long[this.numberOfOutputStreamBounds];
			synchronized (this.boundValuesLock) {
				System.arraycopy(this.lastOutputStreamBoundWriteValues, 0, outputMinWriteTimes, 0, this.numberOfOutputStreamBounds);	
			}
			return outputMinWriteTimes;
		}
		finally{
			if(locked){
				this.executionLock.unlock();
			}
		}
	}
	
	/**
	 * unit of relevance values
	 */
	
	public long[] inputStreamBoundReadValues()
	{
		long[] inputStreamBoundReadValues = new long[this.numberOfInputStreamBounds];
		synchronized (this.boundValuesLock) {
			System.arraycopy(this.lastInputRelevanceBoundValues, 0, inputStreamBoundReadValues, 0, this.numberOfInputStreamBounds);
		}
		return inputStreamBoundReadValues;
	}
	
	/**
	 * end of unit
	 */
	
	protected String valuesToString(final long[] lastInputStreamBoundAndClockMinValues, final long[] currentInputStreamBoundAndClockMinValues, final long[] currentOutputStreamBoundValues){
		StringBuilder sb = new StringBuilder();
		this.writeValues(sb, lastInputStreamBoundAndClockMinValues, currentInputStreamBoundAndClockMinValues, currentOutputStreamBoundValues);
		return sb.toString();
	}
	
	protected void writeValues(final StringBuilder sb, final long[] lastInputStreamBoundAndClockMinValues, final long[] currentInputStreamBoundAndClockMinValues, final long[] currentOutputStreamBoundValues){
		writeValues(sb, new FixedArray<String>("LastInputStreamBoundValues", "CurrentClockMinValues"), this.inputStreamBounds, java.util.Arrays.copyOf(lastInputStreamBoundAndClockMinValues, this.numberOfInputStreamBounds), java.util.Arrays.copyOf(currentInputStreamBoundAndClockMinValues, this.numberOfInputStreamBounds));
		sb.append("\n");
		writeValues(sb, "CurrentClockMinValues", this.referencedClocks.keys(), java.util.Arrays.copyOfRange(currentInputStreamBoundAndClockMinValues, this.numberOfInputStreamBounds, currentInputStreamBoundAndClockMinValues.length));
		sb.append("\n");
		SQL_ExecutableQuery.writeValues(sb, "CurrentOutputStreamBoundValues", this.outputStreamBounds, currentOutputStreamBoundValues);
		sb.append("\n");
	}
	
	public static String whitespace(final int count)
	{
		return repeatString(" ", count);
	}
	
	public static String repeatString(final String s, final int count)
	{
		int finalLength = s.length() * count;
		String result = s;
		while(result.length() < finalLength){
			result+=result;
		}
		return result.substring(0, finalLength);
	}
	
	public static void writeValues(final StringBuilder sb, final String rowName, final FixedArray<String> colNames, final long[] values){
		writeValues(sb, new FixedArray<String>(rowName), colNames, values);
	}
	
	public static void writeValues(final StringBuilder sb, final String[] rowNames, final String[] colNames, final long[]... values){
		writeValues(sb, new FixedArray<String>(false, rowNames), new FixedArray<String>(false, colNames), values);
	}
	
	public static void writeValues(final StringBuilder sb, final FixedArray<String> rowNames, final FixedArray<String> colNames, final long[]... values){
		assert rowNames.size() == values.length;
		final int cols = colNames.size();
		if(cols > 0 | rowNames.size() > 0){
			int maxColNameLength = 0;
			int[] columnLength = new int[colNames.size() + 1];
			int maxColSize = 0;
			
			{			
			String colName;
			for (int col = 0; col < cols; col++) {
				colName = colNames.get(col);
				if(colName.length() > maxColNameLength){
					maxColNameLength = colName.length();					
				}
				if(colName.length() > 22){
					columnLength[col] = colName.length();
				}
				else{
					columnLength[col] = 22;
				}
				if(columnLength[col] > maxColSize){
					maxColSize = columnLength[col];
				}
			}
			if(maxColNameLength > maxColSize){
				maxColSize = maxColNameLength;
			}
			}
						
			int maxRowNameLength = 0;
			for (String rowName: rowNames) {
				if(maxRowNameLength < rowName.length()){
					maxRowNameLength = rowName.length();
				}
			}
			
			String whitespace = whitespace((maxColSize > maxRowNameLength ? maxColSize : maxRowNameLength));
			
			sb.append(whitespace.substring(0, maxRowNameLength));
			sb.append(" | ");
			if(cols > 0){
				String colName;
				for (int col = 0; col < cols; col++) {
					colName = colNames.get(col);
					sb.append(whitespace.substring(0, columnLength[col] - colName.length()));
					sb.append(colName);
					
					sb.append(" | ");
				}
				sb.setLength(sb.length() - " ".length());
			}
			sb.append("\n");
			
			sb.append(repeatString("-", maxRowNameLength));
			sb.append("-+-");
			for (int col = 0; col < cols; col++) {
				sb.append(repeatString("-", columnLength[col]));
				sb.append("-+-");
			}
			sb.setLength(sb.length() - "-".length());
			sb.append("\n");
			
			
			for (int row = 0; row < values.length; row++) {
				assert values[row].length == colNames.size();
				String rowName = rowNames.get(row);
				sb.append(whitespace.substring(0, maxRowNameLength - rowName.length()));
				sb.append(rowName);
				sb.append(" | ");
				
				if(cols > 0){
					for (int col = 0; col < cols; col++) {
						long value = values[row][col];
						String valueString;
	
						valueString = Long.toString(value);
	
						sb.append(whitespace.substring(0, columnLength[col] - valueString.length()));
						sb.append(valueString);
						
						sb.append(" | ");
					}
					sb.setLength(sb.length() - " ".length());
				}
				sb.append("\n");
			}
		}		
	}
	
	public void writeTo(final StringBuilder sb){
		sb.append("\n");
		sb.append(this.type.toString());
		
		if(this.type.isEvalQuery){
			assert this.outputs.size() == 1;
			
			sb.append("\nINSERT INTO ");
			sb.append(this.outputs.values().get(0).full_name);
			sb.append("\nFROM ");
			
			String inputs = this.inputs.keys().toString();
			sb.append(inputs.substring(1, inputs.length()-1));
			sb.append("\n\n");
		}
		else{assert this.type.isGarbageQuery;
			assert this.outputs.size() == 0;
			assert this.inputs.size() == 1;
		
			sb.append("\nDELETE FROM ");
			sb.append(this.inputs.values().get(0).full_name);
		}
		
		writeValues(sb, new FixedArray<String>("LastInputStreamBoundValues", "NextInputStreamBoundValues"), this.inputStreamBounds, java.util.Arrays.copyOf(this.lastInputStreamBoundAndClockMinValues, this.numberOfInputStreamBounds), java.util.Arrays.copyOf(this.nextInputStreamBoundAndClockMinValues, this.numberOfInputStreamBounds));
		sb.append("\n");
		writeValues(sb, "LastClockMinValues", this.referencedClocks.keys(), java.util.Arrays.copyOfRange(this.lastInputStreamBoundAndClockMinValues, this.numberOfInputStreamBounds, this.lastInputStreamBoundAndClockMinValues.length));
		sb.append("\n");
		long[] outputStreamBoundWriteValues = new long[this.numberOfOutputStreamBounds];
		this.outputStreamBoundWriteValues(this.lastOutputStreamBoundValues, this.outputStreamBoundClockIncreaseValues, outputStreamBoundWriteValues);
		long[] outputStreamBoundClockIncreaseVaules = this.outputStreamBoundClockIncreaseValues;
		SQL_ExecutableQuery.writeValues(sb, new FixedArray<String>("LastOutputStreamBoundValues", "OutputStreamBoundClockIncreaseValues", "OutputStreamBoundWriteValues", "NextOutputStreamBoundValues", "NextOutputStreamBoundWriteValues", "LastOutputStreamBoundWriteValues"), this.outputStreamBounds, this.lastOutputStreamBoundValues, outputStreamBoundClockIncreaseVaules, outputStreamBoundWriteValues, this.nextOutputStreamBoundValues, this.nextOutputStreamBoundWriteValues, this.lastOutputStreamBoundWriteValues);

		sb.append(this.eval_statement);
	}
	
	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		this.writeTo(sb);
		return sb.toString();
	}

}
