package de.lmu.ifi.pms.cep.event_mill.sql;

import org.simpleframework.xml.Attribute;
import org.simpleframework.xml.Element;
import org.simpleframework.xml.ElementList;
import org.simpleframework.xml.ElementMap;
import org.simpleframework.xml.Path;
import org.simpleframework.xml.Root;
import org.simpleframework.xml.Serializer;
import org.simpleframework.xml.core.Persister;
import org.simpleframework.xml.stream.Format;

import de.lmu.ifi.pms.cep.tsa.DTSA_BufferType;
import de.lmu.ifi.pms.cep.tsa.DTSA_QueryType;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;

import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;
import de.lmu.ifi.pms.java.util.string.StringBuilder;


@Root(name="query")
public class SQL_Query_Spec {

	@Element(name="prog_package", required=false)
	public final String prog_package;

	@Element(name="description", required=false)
	public final String description;
	
	@Element(name="type")
	public final DTSA_QueryType type;
	
	@Attribute(name="monotonic")
	public final boolean isMonotonic;
	
	public final FixedArrayNavigableSet<String> referencedClocks;
	public final int numberOfReferencedClocks;
	@ElementList(name="referenced-clocks", entry="clock", inline=false)
	java.util.ArrayList<String> getReferencedClocks(){
		return new java.util.ArrayList<String>(this.referencedClocks.javaCollection());
	}
	
	@Element(name="input-type")
	public final DTSA_BufferType inputBuffersType;
	public final FixedArrayNavigableMap<String, String> inputs;
	
	@Path(value="inputs")
	@ElementMap(entry="input", key="alias", value="buffer", inline=true)
	java.util.HashMap<String, String> getInputs(){
		return new java.util.HashMap<String,String>(this.inputs.javaMap());
	}

	public final FixedArrayNavigableSet<String> inputBuffers;
	
	
	public final FixedArrayNavigableSet<String> outputBuffers;
	
	@ElementList(entry="output", inline=true, required=false)
	java.util.ArrayList<String> getOutputs(){
		return new java.util.ArrayList<String>(this.outputBuffers.javaCollection());
	}
	
	public final FixedArrayNavigableMap<String, Term<?>> streamBoundFunctions;
	@ElementList(name="streamBoundFunctions", entry="function", inline=false)
	java.util.ArrayList<TimeSpec> getStreamBoundFunctions(){
		return TimeSpec.fromMap(this.streamBoundFunctions);
	}
	
	public final FixedArrayNavigableMap<String, Term<?>> relevanceBoundFunctions;
	@ElementList(name="relevanceBoundFunctions", entry="function", inline=false)
	java.util.ArrayList<TimeSpec> getRelevanceBoundFunctions(){
		return TimeSpec.fromMap(this.relevanceBoundFunctions);
	}
	
	public final FixedArrayNavigableMap<String, Term<?>> streamBoundFunctionClockIncreases;
	@ElementList(name="streamBoundFunctionClockIncreases", entry="function", inline=false)
	java.util.ArrayList<TimeSpec> getStreamBoundFunctionClockIncreases(){
		return TimeSpec.fromMap(this.streamBoundFunctionClockIncreases);
	}

	public final int numberOfStreamBounds;
	
	@Element(name="query", data=true)
	public final String SQL_PreparedQuery;
	
	SQL_Query_Spec(final String prog_package, final String description, final DTSA_QueryType type, final boolean isMonotonic,
			final DTSA_BufferType input_type,
			final FixedArrayNavigableMap<String, String> inputs,
			final FixedArrayNavigableSet<String> outputs,
			final FixedArrayNavigableMap<String, Term<?>> streamBoundFunctions,
			final FixedArrayNavigableMap<String, Term<?>> streamBoundFunctionClockIncreases,
			final FixedArrayNavigableMap<String, Term<?>> relevanceBoundFunctions,
			final FixedArrayNavigableSet<String> referencedClocks,
			final String SQL_PreparedQuery) {
		super();
		this.prog_package = (prog_package == null ? "" : prog_package);
		if(!TSA_SyntaxPatterns.isValidPackageName(this.prog_package)){
			throw new IllegalArgumentException("Invalid package name: " + this.prog_package);
		}
		
		this.description = (description == null ? "" : description);
		this.type = type;
		this.isMonotonic = isMonotonic;
	
		this.inputBuffersType = input_type;
		
		this.inputs = inputs;
		this.inputBuffers = new FixedArrayNavigableSet<String>(this.inputs.values());
		
		this.outputBuffers = outputs;
		this.streamBoundFunctions = streamBoundFunctions;
		this.streamBoundFunctionClockIncreases = streamBoundFunctionClockIncreases;
		this.relevanceBoundFunctions = relevanceBoundFunctions;
		if(streamBoundFunctions.size() != this.streamBoundFunctionClockIncreases.size()){
			throw new IllegalArgumentException();
			//TODO better exception
		}
		this.numberOfStreamBounds = this.streamBoundFunctions.size();
		this.SQL_PreparedQuery = SQL_PreparedQuery;
		
		this.referencedClocks = referencedClocks;
		this.numberOfReferencedClocks = this.referencedClocks.size();
	}
	
	SQL_Query_Spec(	@Element(name="prog_package", required=false) final String prog_package, 
			@Element(name="description", required=false) final String description,
			@Element(name="type") final DTSA_QueryType type,
			@Attribute(name="monotonic") final boolean isMonotonic,
			@Element(name="input-type") final DTSA_BufferType input_type,
			@Path(value="inputs")@ElementMap(entry="input", key="alias", value="buffer", inline=true) final java.util.HashMap<String, String> inputs, 
			@ElementList(entry="output", inline=true, required=false) final java.util.ArrayList<String> outputs,
			@ElementList(name="streamBoundFunctions", entry="function", inline=false) final java.util.ArrayList<TimeSpec> streamBoundFunctions,
			@ElementList(name="streamBoundFunctionClockIncreases", entry="function", inline=false) final java.util.ArrayList<TimeSpec> streamBoundFunctionClockIncreases,
			@ElementList(name="relevanceBoundFunctions", entry="function", inline=false) final java.util.ArrayList<TimeSpec> relevanceBoundFunctions,
			@ElementList(name="referenced-clocks", entry="clock", inline=false) final
			java.util.ArrayList<String> referencedClocks,
			@Element(name="query", data=true) final String SQL_PreparedQuery) {
		this(prog_package, description, type, isMonotonic,
			input_type,
			new FixedArrayNavigableMap<String, String>(inputs),
			(outputs != null ? new FixedArrayNavigableSet<String>(outputs) : new FixedArrayNavigableSet<String>()),
			TimeSpec.toMap(streamBoundFunctions),
			TimeSpec.toMap(streamBoundFunctionClockIncreases),
			TimeSpec.toMap(relevanceBoundFunctions),
			new FixedArrayNavigableSet<String>(referencedClocks),
			SQL_PreparedQuery
			);
	}
	
	@SuppressWarnings("unchecked")
	public SQL_Query_Spec(final String prog_package, final String description, final DTSA_QueryType type, final boolean isMonotonic,
			final DTSA_BufferType input_type,
			final FixedArrayNavigableMap<String, String> inputs,
			final String output,
			final FixedArrayNavigableMap<String, Term<?>> streamBoundFunctions,
			final FixedArrayNavigableMap<String, Term<?>> streamBoundFunctionClockIncreases,
			final FixedArrayNavigableMap<String, Term<?>> relevanceBoundFunctions,
			final String SQL_PreparedQuery,
			final SQL_Translation sql_translation) {
		this(prog_package, description, type, isMonotonic, input_type, inputs, new FixedArrayNavigableSet<String>(output), streamBoundFunctions, streamBoundFunctionClockIncreases, relevanceBoundFunctions, FixedArrayNavigableSet.merge(sql_translation.referencedClocks(streamBoundFunctions), sql_translation.referencedClocks(streamBoundFunctionClockIncreases), sql_translation.referencedClocks(SQL_PreparedQuery)), SQL_PreparedQuery);
		assert type == DTSA_QueryType.STATIC | type == DTSA_QueryType.INIT | type == DTSA_QueryType.INPUT | type == DTSA_QueryType.WORKING | type == DTSA_QueryType.OUTPUT | type == DTSA_QueryType.LOG;
	}
	
	@SuppressWarnings("unchecked")
	public SQL_Query_Spec(final String prog_package, final String description, final DTSA_QueryType type, final boolean isMonotonic,
			final DTSA_BufferType input_type,
			final FixedArrayNavigableMap<String, String> inputs,
			final FixedArrayNavigableMap<String, Term<?>> streamBoundFunctions,
			final FixedArrayNavigableMap<String, Term<?>> streamBoundFunctionClockIncreases,
			final FixedArrayNavigableMap<String, Term<?>> relevanceBoundFunctions,
			final String SQL_PreparedQuery,
			final SQL_Translation sql_translation) {
		this(prog_package, description, type, isMonotonic, input_type, inputs, new FixedArrayNavigableSet<String>(), streamBoundFunctions, streamBoundFunctionClockIncreases, relevanceBoundFunctions, FixedArrayNavigableSet.merge(sql_translation.referencedClocks(streamBoundFunctions), sql_translation.referencedClocks(streamBoundFunctionClockIncreases), sql_translation.referencedClocks(SQL_PreparedQuery)), SQL_PreparedQuery);
		assert type.isGarbageQuery;
	}

	@Override
	public String toString() {
		Serializer serializer = new Persister();
		StringBuilder result = new StringBuilder();
		try {
			serializer.write(this, result);
			return result.toString();
			//return serializer.read(FixedArray.class, xml_buffer_specs);
		} catch (Exception e) {
			throw new IllegalArgumentException(e);
		}
	}
	
	@SuppressWarnings("unchecked")
	public static FixedArray<SQL_Query_Spec> deserialize(final String xml_buffer_specs){
		Serializer serializer = new Persister(new Format("<?xml version=\"1.0\" encoding=\"UTF-8\"?>"));
		try {
			return new FixedArray<SQL_Query_Spec>(serializer.read(java.util.ArrayList.class, xml_buffer_specs));
		} catch (Exception e) {
			throw new IllegalArgumentException(e);
		}
	}
	
	public static String serialize(final FixedArray<SQL_Query_Spec> buffer_specs){
		Serializer serializer = new Persister(new Format("<?xml version=\"1.0\" encoding=\"UTF-8\"?>"));
		StringBuilder result = new StringBuilder();
		try {
			serializer.write(new java.util.ArrayList<SQL_Query_Spec>(buffer_specs.javaCollection()), result);
			return result.toString();
			//return serializer.read(FixedArray.class, xml_buffer_specs);
		} catch (Exception e) {
			throw new IllegalArgumentException(e);
		}
	}
}
