package de.lmu.ifi.pms.cep.event_mill.sql;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.SQLFeatureNotSupportedException;
import java.sql.SQLWarning;
import java.sql.Statement;

import de.lmu.ifi.pms.cep.Version;

public class SQL_StatementWrapper implements Statement {
	
	protected final Statement statement;

	public SQL_StatementWrapper(final Statement statement) {
		super();
		this.statement = statement;
	}
	
	public SQL_StatementWrapper(final Connection connection) throws SQLException{
		this(connection.createStatement());
	}

	@Override
	public <T> T unwrap(final Class<T> iface) throws SQLException {
		return this.statement.unwrap(iface);
	}

	@Override
	public boolean isWrapperFor(final Class<?> iface) throws SQLException {
		return this.statement.isWrapperFor(iface);
	}

	@Override
	public ResultSet executeQuery(final String sql) throws SQLException {
		return this.statement.executeQuery(sql);
	}

	@Override
	public int executeUpdate(final String sql) throws SQLException {
		return this.statement.executeUpdate(sql);
	}

	@Override
	public void close() throws SQLException {
		this.statement.close();
	}

	@Override
	public int getMaxFieldSize() throws SQLException {
		return this.statement.getMaxFieldSize();
	}

	@Override
	public void setMaxFieldSize(final int max) throws SQLException {
		this.statement.setMaxFieldSize(max);
	}

	@Override
	public int getMaxRows() throws SQLException {
		return this.statement.getMaxRows();
	}

	@Override
	public void setMaxRows(final int max) throws SQLException {
		this.statement.setMaxRows(max);
	}

	@Override
	public void setEscapeProcessing(final boolean enable) throws SQLException {
		this.statement.setEscapeProcessing(enable);
	}

	@Override
	public int getQueryTimeout() throws SQLException {
		return this.statement.getQueryTimeout();
	}

	@Override
	public void setQueryTimeout(final int seconds) throws SQLException {
		this.statement.setQueryTimeout(seconds);
	}

	@Override
	public void cancel() throws SQLException {
		this.statement.cancel();
	}

	@Override
	public SQLWarning getWarnings() throws SQLException {
		return this.statement.getWarnings();
	}

	@Override
	public void clearWarnings() throws SQLException {
		this.statement.clearWarnings();
	}

	@Override
	public void setCursorName(final String name) throws SQLException {
		this.statement.setCursorName(name);
	}

	@Override
	public boolean execute(final String sql) throws SQLException {
		return this.statement.execute(sql);
	}

	@Override
	public ResultSet getResultSet() throws SQLException {
		return this.statement.getResultSet();
	}

	@Override
	public int getUpdateCount() throws SQLException {
		return this.statement.getUpdateCount();
	}

	@Override
	public boolean getMoreResults() throws SQLException {
		return this.statement.getMoreResults();
	}

	@Override
	public void setFetchDirection(final int direction) throws SQLException {
		this.statement.setFetchDirection(direction);
	}

	@Override
	public int getFetchDirection() throws SQLException {
		return this.statement.getFetchDirection();
	}

	@Override
	public void setFetchSize(final int rows) throws SQLException {
		this.statement.setFetchSize(rows);
	}

	@Override
	public int getFetchSize() throws SQLException {
		return this.statement.getFetchSize();
	}

	@Override
	public int getResultSetConcurrency() throws SQLException {
		return this.statement.getResultSetConcurrency();
	}

	@Override
	public int getResultSetType() throws SQLException {
		return this.statement.getResultSetType();
	}

	@Override
	public void addBatch(final String sql) throws SQLException {
		this.statement.addBatch(sql);
	}

	@Override
	public void clearBatch() throws SQLException {
		this.statement.clearBatch();
	}

	@Override
	public int[] executeBatch() throws SQLException {
		return this.statement.executeBatch();
	}

	@Override
	public Connection getConnection() throws SQLException {
		return this.statement.getConnection();
	}

	@Override
	public boolean getMoreResults(final int current) throws SQLException {
		return this.statement.getMoreResults(current);
	}

	@Override
	public ResultSet getGeneratedKeys() throws SQLException {
		return this.statement.getGeneratedKeys();
	}

	@Override
	public int executeUpdate(final String sql, final int autoGeneratedKeys) throws SQLException {
		return this.statement.executeUpdate(sql, autoGeneratedKeys);
	}

	@Override
	public int executeUpdate(final String sql, final int[] columnIndexes) throws SQLException {
		return this.statement.executeUpdate(sql, columnIndexes);
	}

	@Override
	public int executeUpdate(final String sql, final String[] columnNames) throws SQLException {
		return this.statement.executeUpdate(sql, columnNames);
	}

	@Override
	public boolean execute(final String sql, final int autoGeneratedKeys) throws SQLException {
		return this.statement.execute(sql, autoGeneratedKeys);
	}

	@Override
	public boolean execute(final String sql, final int[] columnIndexes) throws SQLException {
		return this.statement.execute(sql, columnIndexes);
	}

	@Override
	public boolean execute(final String sql, final String[] columnNames) throws SQLException {
		return this.statement.execute(sql, columnNames);
	}

	@Override
	public int getResultSetHoldability() throws SQLException {
		return this.statement.getResultSetHoldability();
	}

	@Override
	public boolean isClosed() throws SQLException {
		return this.statement.isClosed();
	}

	@Override
	public void setPoolable(final boolean poolable) throws SQLException {
		this.statement.setPoolable(poolable);
	}

	@Override
	public boolean isPoolable() throws SQLException {
		return this.statement.isPoolable();
	}

	public void closeOnCompletion() throws SQLException {
		if(Version.Java.greaterOrEqual1dot7){
			this.statement.closeOnCompletion();
		}
		else{
			throw new SQLFeatureNotSupportedException();
		}
	}

	public boolean isCloseOnCompletion() throws SQLException {
		if(Version.Java.greaterOrEqual1dot7){
			return this.statement.isCloseOnCompletion();
		}
		else{
			throw new SQLFeatureNotSupportedException();
		}
	}

}
