package de.lmu.ifi.pms.cep.event_mill.sql.compiler;

import org.simpleframework.xml.Attribute;
import org.simpleframework.xml.Element;
import org.simpleframework.xml.ElementList;

import de.lmu.ifi.pms.cep.compilation.Compiler;
import de.lmu.ifi.pms.cep.compilation.CompilerFactory;
import de.lmu.ifi.pms.cep.event_mill.sql.SQL_Program_Spec;
import de.lmu.ifi.pms.cep.tsa.DTSA_Program;
import de.lmu.ifi.pms.cep.tsa.optimization.DTSA_Optimizer;
import de.lmu.ifi.pms.cep.tsa.optimization.DTSA_OptimizerFactory;

import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;


public class DTSA_SQL_CompilerFactory extends CompilerFactory<DTSA_Program, SQL_Program_Spec>{
	
	public static final String defaultIdentifier = "de.lmu.ifi.pms.cep.event_mill.sql.compiler";
	
	
	public final FixedArray<DTSA_OptimizerFactory> preWriteOptimizers;
	
	@ElementList(inline=true, entry="optimizer", type=DTSA_OptimizerFactory.class, required=false)
	java.util.ArrayList<DTSA_OptimizerFactory> getOptimizers(){
		return new java.util.ArrayList<DTSA_OptimizerFactory>(this.preWriteOptimizers.javaCollection());
	}
	
	@Element(name="sql-writer")
	public final DTSA_SQL_WriterFactory dtsa_SQL_Writer;
	
	public DTSA_SQL_CompilerFactory(final DTSA_SQL_WriterFactory dtsa_SQL_Writer) {
		this(defaultIdentifier, dtsa_SQL_Writer);
	}
	
	public DTSA_SQL_CompilerFactory(final String identifier, final DTSA_SQL_WriterFactory dtsa_SQL_Writer) {
		this(identifier, new FixedArray<DTSA_OptimizerFactory>(), dtsa_SQL_Writer);
	}
	
	public DTSA_SQL_CompilerFactory(@Attribute(name="iri", required=true) final String identifier, @ElementList(inline=true, entry="optimizer", type=DTSA_OptimizerFactory.class, required=false) final java.util.ArrayList<DTSA_OptimizerFactory> preWriteOptimizers, @Element(name="sql-writer") final DTSA_SQL_WriterFactory dtsa_SQL_Writer) {
		this(identifier, (preWriteOptimizers == null ? new FixedArray<DTSA_OptimizerFactory>() : new FixedArray<DTSA_OptimizerFactory>(preWriteOptimizers)), dtsa_SQL_Writer);
	}
	
	public DTSA_SQL_CompilerFactory(final String identifier, final DTSA_OptimizerFactory[] preWriteOptimizers, final DTSA_SQL_WriterFactory dtsa_SQL_Writer) {
		this(identifier, new FixedArray<DTSA_OptimizerFactory>(preWriteOptimizers), dtsa_SQL_Writer);
	}
	
	public DTSA_SQL_CompilerFactory(final DTSA_OptimizerFactory[] preWriteOptimizers, final DTSA_SQL_WriterFactory dtsa_SQL_Writer) {
		this(defaultIdentifier, preWriteOptimizers, dtsa_SQL_Writer);
	}
	
	public DTSA_SQL_CompilerFactory(final String identifier, final FixedArray<DTSA_OptimizerFactory> preWriteOptimizers, final DTSA_SQL_WriterFactory dtsa_SQL_Writer) {
		super(DTSA_Program.class, SQL_Program_Spec.class, identifier);
		this.preWriteOptimizers = preWriteOptimizers;
		this.dtsa_SQL_Writer = dtsa_SQL_Writer;
	}

	@Override
	public Compiler<DTSA_Program, SQL_Program_Spec> make() {
		final DTSA_Optimizer[] preWriteOptimizers = new DTSA_Optimizer[this.preWriteOptimizers.size()];
		for(int optimizer = 0; optimizer < preWriteOptimizers.length; ++optimizer){
			preWriteOptimizers[optimizer] = this.preWriteOptimizers.get(optimizer).make();
		}
		return new DTSA_SQL_Compiler(preWriteOptimizers, this.dtsa_SQL_Writer.make());
	}
		
}
