package de.lmu.ifi.pms.cep.event_mill.sql.compiler;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.event_mill.sql.SQL_Buffer_Spec;
import de.lmu.ifi.pms.cep.event_mill.sql.SQL_Program_Spec;
import de.lmu.ifi.pms.cep.event_mill.sql.SQL_Query_Spec;
import de.lmu.ifi.pms.cep.event_mill.sql.SQL_Translation;
import de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_ExpressionFactory;
import de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_ExpressionFactoryDirectInput;
import de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_ExpressionFactoryInput;
import de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryFactory;
import de.lmu.ifi.pms.cep.tsa.DTSA_BufferType;
import de.lmu.ifi.pms.cep.tsa.DTSA_CompositeExpression;
import de.lmu.ifi.pms.cep.tsa.DTSA_Expression;
import de.lmu.ifi.pms.cep.tsa.DTSA_Program;
import de.lmu.ifi.pms.cep.tsa.DTSA_Query;
import de.lmu.ifi.pms.cep.tsa.DTSA_QueryType;
import de.lmu.ifi.pms.cep.tsa.DTSA_StreamBuffer;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.conditions.TypedAttribute;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_Exception;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_RuntimeException;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_SchemaException;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_BasicOperator;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_ComplexOperator;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_GroupingAndAggregationOperator;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_ImbeddingOperator;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_ProjectionOperator;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_RenameOperator;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_SelectionOperator;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchema;

import de.lmu.ifi.pms.java.util.adaption.CollectionAdapter;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;


public class DTSA_SQL_WriterImpl extends DTSA_SQL_Writer {
	
	public final SQL_Translation sql_translation;
	
	public DTSA_SQL_WriterImpl(final SQL_Translation sql_translation) {
		this.sql_translation = sql_translation;
	}

	@Override
	public SQL_Program_Spec compile(final DTSA_Program dtsa_program) {
		
		ArrayList<SQL_Buffer_Spec> buffer_specs = new ArrayList<SQL_Buffer_Spec>();
		
		for(DTSA_BufferType buffer_type: dtsa_program.streamBuffers.keys()){
			if(buffer_type != DTSA_BufferType.BLACK_HOLE_INPUT & buffer_type != DTSA_BufferType.BLACK_HOLE_WORKING){
				for(DTSA_StreamBuffer dtsa_streamBuffer: dtsa_program.streamBuffers.get(buffer_type).values()){
					try{
						buffer_specs.add(this.writeBuffer(dtsa_streamBuffer));
					} catch (TSA_Exception e) {
						throw new TSA_RuntimeException(e);
					}
				}
			}
		}
		
		ArrayList<SQL_Query_Spec> query_specs = new ArrayList<SQL_Query_Spec>();
		
		for(DTSA_QueryType query_type: dtsa_program.queries.keys()){
			for(DTSA_Query dtsa_query: dtsa_program.queries.get(query_type)){
				try{
					query_specs.add(this.write(dtsa_query));
				} catch (TSA_Exception e) {
					throw new TSA_RuntimeException(e);
				}
			}
			
		}

		return new SQL_Program_Spec(dtsa_program.namespace, dtsa_program.simple_name, dtsa_program.description, buffer_specs, query_specs);
	}
	
	public SQL_Buffer_Spec writeBuffer(final DTSA_StreamBuffer dtsa_streamBuffer) throws TSA_Exception{		
		return new SQL_Buffer_Spec(dtsa_streamBuffer.prog_package, dtsa_streamBuffer.simple_name, dtsa_streamBuffer.description, dtsa_streamBuffer.type, dtsa_streamBuffer.schema.basicAttributes, dtsa_streamBuffer.streamBounds.keys());
	}
	
	@SuppressWarnings("unchecked")
	public SQL_Query_Spec write(final DTSA_Query dtsa_query) throws TSA_Exception{
		SQL_ExpressionFactory expression = this.write(dtsa_query.expression);
		SQL_QueryFactory query = new SQL_QueryFactory(dtsa_query.output.full_name, dtsa_query.output.type, dtsa_query.output.schema.basicAttributes.keys(), expression);
		String sqlQuery = "\n" + query.toString() + "\n";
		if(dtsa_query.type.isEvalQuery){
			return new SQL_Query_Spec(dtsa_query.prog_package, dtsa_query.description, dtsa_query.type, dtsa_query.isMonotonic, dtsa_query.input_type, query.aliasMapping, dtsa_query.output.full_name, (FixedArrayNavigableMap<String, Term<?>>)(FixedArrayNavigableMap<String, ?>) dtsa_query.streamBoundFunctions, (FixedArrayNavigableMap<String, Term<?>>)(FixedArrayNavigableMap<String, ?>) dtsa_query.streamBoundClockIncreaseFunctions, (FixedArrayNavigableMap<String, Term<?>>)(FixedArrayNavigableMap<String, ?>) dtsa_query.relevanceBoundFunctions, sqlQuery, this.sql_translation);
		}
		else if(dtsa_query.type.isGarbageQuery){
			return new SQL_Query_Spec(dtsa_query.prog_package, dtsa_query.description, dtsa_query.type, dtsa_query.isMonotonic, dtsa_query.input_type, query.aliasMapping, new FixedArrayNavigableMap<String, Term<?>>(), new FixedArrayNavigableMap<String, Term<?>>(), (FixedArrayNavigableMap<String, Term<?>>)(FixedArrayNavigableMap<String, ?>) dtsa_query.relevanceBoundFunctions, sqlQuery, this.sql_translation);
		}
		else{
			throw new Error("Unexpected query type: " + dtsa_query.type);
		}
	}
	
	public SQL_ExpressionFactory write(final DTSA_Expression dtsa_expression) throws TSA_Exception{
		if(dtsa_expression instanceof DTSA_StreamBuffer){
			return new SQL_ExpressionFactory(this.sql_translation, this.writeInput((DTSA_StreamBuffer)dtsa_expression));
		}
		else if(dtsa_expression instanceof DTSA_CompositeExpression){
			return this.write((DTSA_CompositeExpression)dtsa_expression);
		}
		else{
			throw new Error("Programming error - Unconsidered case " + dtsa_expression.getClass().getName());
		}
	}
	
	public SQL_ExpressionFactoryDirectInput writeInput(final DTSA_StreamBuffer dtsa_streamBuffer){
		return new SQL_ExpressionFactoryDirectInput(dtsa_streamBuffer.full_name, dtsa_streamBuffer.type, dtsa_streamBuffer.schema().basicAttributes.keys());
	}
	
	public SQL_ExpressionFactory write(final DTSA_CompositeExpression dtsa_compositeExpression) throws TSA_Exception{
		SQL_ExpressionFactory expression = null;
		switch (dtsa_compositeExpression.operator.ID) {
		case BASIC: 
		{
			switch (((TSA_BasicOperator)dtsa_compositeExpression.operator).ID) {
			case CROSSPRODUCT: 
			{				
				assert dtsa_compositeExpression.subExpressions.size() > 0;
				DTSA_Expression dtsa_subExpression = dtsa_compositeExpression.subExpressions.get(0);
				expression = this.write(dtsa_subExpression);
				
				ArrayList<SQL_ExpressionFactoryInput> furtherInputs = new ArrayList<SQL_ExpressionFactoryInput>();
				for(int subExpressionIndex = 1; subExpressionIndex < dtsa_compositeExpression.subExpressions.size(); ++subExpressionIndex){
					dtsa_subExpression = dtsa_compositeExpression.subExpressions.get(subExpressionIndex);
					if(dtsa_subExpression instanceof DTSA_StreamBuffer){
						furtherInputs.add(this.writeInput((DTSA_StreamBuffer)dtsa_subExpression));
					}
					else{
						furtherInputs.add(this.write(dtsa_subExpression).toSubquery());
					}
				}
				
				FixedArray<SQL_ExpressionFactoryInput> furtherInputsFixed = new FixedArray<SQL_ExpressionFactoryInput>(furtherInputs);
				if(expression.canAddInputs()){
					expression.addInputs(furtherInputsFixed);
				}
				else{
					expression = new SQL_ExpressionFactory(this.sql_translation, expression.toSubquery());
					assert expression.canAddInputs();
					expression.addInputs(furtherInputsFixed);
				}
				break;
			}
			case UNION:
			{
				assert dtsa_compositeExpression.subExpressions.size() > 0;
				DTSA_Expression dtsa_subExpression = dtsa_compositeExpression.subExpressions.get(0);
				expression = this.write(dtsa_subExpression);

				ArrayList<SQL_ExpressionFactoryInput> furtherInputs = new ArrayList<SQL_ExpressionFactoryInput>();
				for(int subExpressionIndex = 1; subExpressionIndex < dtsa_compositeExpression.subExpressions.size(); ++subExpressionIndex){
					dtsa_subExpression = dtsa_compositeExpression.subExpressions.get(subExpressionIndex);
					if(dtsa_subExpression instanceof DTSA_StreamBuffer){
						furtherInputs.add(this.writeInput((DTSA_StreamBuffer)dtsa_subExpression));
					}
					else{
						furtherInputs.add(this.write(dtsa_subExpression).toSubquery());
					}
				}
				
				FixedArray<SQL_ExpressionFactoryInput> furtherInputsFixed = new FixedArray<SQL_ExpressionFactoryInput>(furtherInputs);
				if(expression.canAddUnionAll()){
					expression.addUnionAll(furtherInputsFixed);
				}
				else{
					expression = new SQL_ExpressionFactory(this.sql_translation, expression.toSubquery());
					assert expression.canAddUnionAll();
					expression.addUnionAll(furtherInputsFixed);
				}
				break;
			}
			case GROUPING:
			{
				//FIXME
				throw new Error("Not implemented case");
				//break;
			}
			case IMBEDDING:
			{
				TSA_ImbeddingOperator operator = (TSA_ImbeddingOperator) dtsa_compositeExpression.operator;
				assert dtsa_compositeExpression.subExpressions.size() == 1;
				DTSA_Expression dtsa_subExpression = dtsa_compositeExpression.subExpressions.get(0);
				SQL_ExpressionFactory subExpression = this.write(dtsa_subExpression);
					//TypedAttributes primarily sorted by name
					FixedArrayNavigableSet<String> newAttributeNames = new FixedArrayNavigableSet<String>(false, false, new CollectionAdapter<TypedAttribute<?>, String>(operator.simpleDefinitions.keys()){
						@Override
						public String adapt(final TypedAttribute<?> source) {
							return source.name;
						}
					}.toArray(new String[operator.simpleDefinitions.size()])); 
				FixedArrayNavigableMap<String, Term<?>> newAttributes = new FixedArrayNavigableMap<String, Term<?>>(newAttributeNames, operator.simpleDefinitions.values());
				if(subExpression.canAddDefinitions(newAttributes)){
					subExpression.addDefinitions(newAttributes);
					expression = subExpression;
				}
				else{
					expression = new SQL_ExpressionFactory(this.sql_translation, subExpression.toSubquery());
					expression.addDefinitions(newAttributes);
				}
				break;
			}
			case PROJECTION:
			{ 
				TSA_ProjectionOperator operator = (TSA_ProjectionOperator) dtsa_compositeExpression.operator;
				assert dtsa_compositeExpression.subExpressions.size() == 1;
				DTSA_Expression dtsa_subExpression = dtsa_compositeExpression.subExpressions.get(0);
				SQL_ExpressionFactory subExpression = this.write(dtsa_subExpression);
				FixedArrayNavigableSet<String> basicRetainedAttributes = TemporalStreamSchema.simpleRetain(dtsa_subExpression.schema().basicAttributes.keys(),operator.attributes); 
				if(subExpression.canAddProjection()){
					subExpression.addProjection(basicRetainedAttributes);
					expression = subExpression;
				}
				else{
					expression = new SQL_ExpressionFactory(this.sql_translation, subExpression.toSubquery());
					expression.addProjection(basicRetainedAttributes);
				}
				break;
			}
			case SELECTION:
			{
				TSA_SelectionOperator operator = (TSA_SelectionOperator) dtsa_compositeExpression.operator;
				assert dtsa_compositeExpression.subExpressions.size() == 1;
				DTSA_Expression dtsa_subExpression = dtsa_compositeExpression.subExpressions.get(0);
				SQL_ExpressionFactory subExpression = this.write(dtsa_subExpression); 
				if(subExpression.canAddSelection()){
					subExpression.addSelection(operator.conditions);
					expression = subExpression;
				}
				else{
					expression = new SQL_ExpressionFactory(this.sql_translation, subExpression.toSubquery());
					expression.addSelection(operator.conditions);
				}
				break;
			}	
			case SETDIFFERENCE:
			{
				assert dtsa_compositeExpression.subExpressions.size() == 2;
				DTSA_Expression dtsa_positiveSubExpression = dtsa_compositeExpression.subExpressions.get(0);
				SQL_ExpressionFactory poitiveExpression = this.write(dtsa_positiveSubExpression);
				DTSA_Expression dtsa_negativeSubExpression = dtsa_compositeExpression.subExpressions.get(1);
				SQL_ExpressionFactoryInput negativeInput;
				if(dtsa_negativeSubExpression instanceof DTSA_StreamBuffer){
					negativeInput = this.writeInput((DTSA_StreamBuffer)dtsa_negativeSubExpression);
				}
				else{
					negativeInput = this.write(dtsa_negativeSubExpression).toSubquery();
				}
				
				if(poitiveExpression.canAddSetDifference()){
					poitiveExpression.addSetDifference(negativeInput);
					expression = poitiveExpression;
				}
				else{
					expression = new SQL_ExpressionFactory(this.sql_translation, poitiveExpression.toSubquery());
					expression.addSetDifference(negativeInput);;
				}
				break;
			}
			}
			break;
		}
		case COMPLEX: 
		{
			switch (((TSA_ComplexOperator)dtsa_compositeExpression.operator).ID) {
			case RENAME: {
				TSA_RenameOperator operator = (TSA_RenameOperator) dtsa_compositeExpression.operator;
				assert dtsa_compositeExpression.subExpressions.size() == 1;
				DTSA_Expression dtsa_subExpression = dtsa_compositeExpression.subExpressions.get(0);
				SQL_ExpressionFactory subExpression = this.write(dtsa_subExpression);
				FixedArrayNavigableMap<String, String> simpleRename;
				try {
					simpleRename = TemporalStreamSchema.simpleMapping(dtsa_subExpression.schema().basicAttributes.keys(), operator.mapping);
				} catch (TSA_SchemaException e) {
					throw new TSA_Exception(operator.origin, e);
				}
				if(subExpression.canAddRename()){
					subExpression.addRename(simpleRename);
					expression = subExpression;
				}
				else{
					expression = new SQL_ExpressionFactory(this.sql_translation, subExpression.toSubquery());
					expression.addRename(simpleRename);
				}
				break;
			}
			case GROUPING_AND_AGGREGATION: {
				TSA_GroupingAndAggregationOperator operator = (TSA_GroupingAndAggregationOperator) dtsa_compositeExpression.operator;
				assert dtsa_compositeExpression.subExpressions.size() == 1;
				DTSA_Expression dtsa_subExpression = dtsa_compositeExpression.subExpressions.get(0);
				SQL_ExpressionFactory subExpression = this.write(dtsa_subExpression);
				FixedArrayNavigableSet<String> basicGroupingAttributes = TemporalStreamSchema.simpleRetain(dtsa_subExpression.schema().basicAttributes.keys(),operator.grouping_attributes);
					//TypedAttributes primarily sorted by name
					FixedArrayNavigableSet<String> aggregationAttributeNames = new FixedArrayNavigableSet<String>(false, false, new CollectionAdapter<TypedAttribute<?>, String>(operator.definitionsOfAggregationAttributes.keys()){
						@Override
						public String adapt(final TypedAttribute<?> source) {
							return source.name;
						}
					}.toArray(new String[operator.definitionsOfAggregationAttributes.size()])); 
				FixedArrayNavigableMap<String, Term<?>> aggregationAttributes = new FixedArrayNavigableMap<String, Term<?>>(aggregationAttributeNames, operator.definitionsOfAggregationAttributes.values());				
				if(subExpression.canAddGrouping()){
					subExpression.addGrouping(basicGroupingAttributes);
					assert subExpression.canAddAggregation();
					subExpression.addAggregation(aggregationAttributes);
					expression = subExpression;
				}
				else{
					expression = new SQL_ExpressionFactory(this.sql_translation, subExpression.toSubquery());
					expression.addGrouping(basicGroupingAttributes);
					assert expression.canAddAggregation();
					expression.addAggregation(aggregationAttributes);
				}
				break;
			}
			case DISTINCT_PROJECTION:{
				//FIXME
				throw new Error("Not implemented case");
			}
			case ANTI_SEMI_JOIN:{
				//FIXME
				throw new Error("Not implemented case");
			}
			case SEMI_JOIN:{
				//FIXME
				throw new Error("Not implemented case");
			}
				//	default:
				//		break;
			}
			break;
		}
		case COMPOSITE: 
		{
			//FIXME
			throw new Error("Not implemented case");
			//break;
		}
			//	default:
			//		break;
		}
		assert expression != null;
		return expression;
	}
	
	protected static final Logger logger = LoggerFactory.getLogger(DTSA_SQL_WriterImpl.class);
	
	@Override
	protected Logger logger(){
		return logger;
	}

}
