package de.lmu.ifi.pms.cep.event_mill.sql.creation;

import static de.lmu.ifi.pms.cep.tsa.conditions.Term.attr;

import de.lmu.ifi.pms.cep.event_mill.sql.SQL_Translation;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.conditions.AggregationFunction;
import de.lmu.ifi.pms.cep.tsa.conditions.AtomicTerm;
import de.lmu.ifi.pms.cep.tsa.conditions.Attribute;
import de.lmu.ifi.pms.cep.tsa.conditions.BasicTypeValue;
import de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.BooleanTerm;
import de.lmu.ifi.pms.cep.tsa.conditions.BooleanValue;
import de.lmu.ifi.pms.cep.tsa.conditions.CastFunction;
import de.lmu.ifi.pms.cep.tsa.conditions.ClockFunction;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.ConjunctiveConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.ConstantLiteral;
import de.lmu.ifi.pms.cep.tsa.conditions.DisjunctiveConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.DoubleValue;
import de.lmu.ifi.pms.cep.tsa.conditions.DurationValue;
import de.lmu.ifi.pms.cep.tsa.conditions.FloatValue;
import de.lmu.ifi.pms.cep.tsa.conditions.Function;
import de.lmu.ifi.pms.cep.tsa.conditions.IdentifierValue;
import de.lmu.ifi.pms.cep.tsa.conditions.IntValue;
import de.lmu.ifi.pms.cep.tsa.conditions.LongValue;
import de.lmu.ifi.pms.cep.tsa.conditions.NegatedConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.Parameter;
import de.lmu.ifi.pms.cep.tsa.conditions.ReferenceValue;
import de.lmu.ifi.pms.cep.tsa.conditions.StringValue;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.conditions.TimestampValue;
import de.lmu.ifi.pms.cep.tsa.conditions.TupleFunction;
import de.lmu.ifi.pms.cep.tsa.conditions.Value;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;

import de.lmu.ifi.pms.java.math.Math;
import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.adaption.CollectionAdapter;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;

public class SQL_ExpressionFactory {
	
	public final SQL_Translation sql_translation;
	
	enum FactoryState{
		EMPTY, SINGLE_INPUT, CROSSPRODUCT, SELECTION, GROUPING, AGGREGATION, HAVING, PROJECTION_IMBED, EXCEPT, UNION_ALL
	}
	
	FactoryState state;
	
	FixedArrayNavigableMap<String, Term<?>> output = null;
	
	ConditionFormula having = ConditionFormula.Top;
	
	FixedArrayNavigableSet<String> grouping = null;
	boolean hasGrouping = false;
	
	ConditionFormula selection = ConditionFormula.Top;
	
	FixedArrayNavigableSet<String> inputSchema = new FixedArrayNavigableSet<String>();
	
	FixedArray<SQL_ExpressionFactoryInput> inputs = new FixedArray<SQL_ExpressionFactoryInput>();
	int numberOfDirectInputs = 0;
	int numberOfSubqueries = 0;
	
	public FixedArrayNavigableSet<String> schema() {
		if(FactoryState.AGGREGATION.ordinal() <= this.state.ordinal() & this.state.ordinal() < FactoryState.EXCEPT.ordinal() ){
			return this.output.keys();
		}
		else{
			return this.preOutputSchema();
		}
	}
	
	protected FixedArrayNavigableSet<String> preOutputSchema() {
		if(this.hasGrouping & FactoryState.GROUPING.ordinal() <= this.state.ordinal() & this.state.ordinal() < FactoryState.EXCEPT.ordinal() ){
			return this.grouping;
		}
		else{
			return this.preGroupingSchema();
		}
	}
	
	protected FixedArrayNavigableSet<String> preGroupingSchema() {
		if(this.state.ordinal() > FactoryState.EMPTY.ordinal()){
			return this.inputSchema;
		}
		else{
			throw new IllegalStateException(this.state.toString());
			//TODO better exception
		}
	}
	
	public SQL_ExpressionFactory(final SQL_Translation sql_translation){
		this.sql_translation = sql_translation;
		this.state = FactoryState.EMPTY;
	}
	
	public SQL_ExpressionFactory(final SQL_Translation sql_translation, final SQL_ExpressionFactoryInput... inputs) {
		this(sql_translation, new FixedArray<SQL_ExpressionFactoryInput>(inputs));
	}
	
	public SQL_ExpressionFactory(final SQL_Translation sql_translation, final FixedArray<SQL_ExpressionFactoryInput> inputs) {
		this(sql_translation);
		this.addInputs(inputs);
	}
	
	public boolean canAddProjection(){
		return this.state.ordinal() <= FactoryState.PROJECTION_IMBED.ordinal();
	}
	
	public void checkAddProjection(final FixedArrayNavigableSet<String> retain){
		if(this.canAddProjection()){
			if(retain.size() <= 0){
				throw new IllegalArgumentException("Projection must preserve at least one attribute.");
			}
			else if(!this.schema().containsAll(retain)){
				throw new IllegalArgumentException("Projection contains undefined attributes.");
			}
		}
		else{
			throw new IllegalStateException(this.state.toString());
		}
	}

	public void addProjection(final FixedArrayNavigableSet<String> retain){
		this.checkAddProjection(retain);
		//initialize output if neccessary
		if(this.output == null){
			this.initOutput();
		}
		Term<?>[] retainedDefinitions = new Term[retain.size()];
		int attribute = 0;
		for (int retainedAttribute = 0; retainedAttribute < retainedDefinitions.length; retainedAttribute++) {
			while(retain.get(retainedAttribute).compareTo(this.output.keys().get(attribute)) > 0){
				++attribute;
			}
			retainedDefinitions[retainedAttribute] = this.output.values().get(attribute);
			++attribute;
		}
		this.output = new FixedArrayNavigableMap<String, Term<?>>(retain, new FixedArray<Term<?>>(false, retainedDefinitions));
		this.state = FactoryState.PROJECTION_IMBED;
	}

	public boolean canAddDefinitions(final FixedArrayNavigableMap<String, Term<?>> newAttributes){
		if(FactoryState.SINGLE_INPUT.ordinal() <= this.state.ordinal() & this.state.ordinal() <= FactoryState.PROJECTION_IMBED.ordinal()){
			int numberOfNewAttributes = newAttributes.size();
			//FIXME pro
			for(int newAttribute = 0; newAttribute < numberOfNewAttributes; ++newAttribute){
				Term<?> definition = newAttributes.values().get(newAttribute);
				if(definition instanceof Attribute || definition instanceof TupleFunction || definition instanceof ConditionFormula || definition instanceof ClockFunction || definition instanceof Value<?>){//FIXME Constants to another case, only Values (but for the first time also Literals)
					if(this.schema().containsAll(definition.AttributeNames())){
						if(!this.preOutputSchema().containsAll(definition.AttributeNames())){
							return false;
						}
					}
				}
			}
			return true;
		}
		else{
			return false;
		}
	}
	
	public void checkAddDefinitions(final FixedArrayNavigableMap<String, Term<?>> newAttributes){
		if(this.canAddDefinitions(newAttributes)){
			int numberOfNewAttributes = newAttributes.size();
			//FIXME pro
			for(int newAttribute = 0; newAttribute < numberOfNewAttributes; ++newAttribute){
				String newAttributeName = newAttributes.keys().get(newAttribute);
				Term<?> definition = newAttributes.values().get(newAttribute);
				if(!TSA_SyntaxPatterns.isValidAttributeName(newAttributeName)){
					throw new IllegalArgumentException("Invalid attribute name: " + newAttributeName);
				}
				if(this.schema().contains(newAttributeName)){
					throw new IllegalArgumentException("Schema already contains attribute with name: " + newAttributeName);
				}
				if(definition instanceof AggregationFunction){
					if(!this.canAddAggregation()){
						throw new IllegalStateException();
					}
					else{
						FixedArrayNavigableSet<String> usedAttributes = definition.AttributeNames(); 
						if(!this.preGroupingSchema().containsAll(usedAttributes)){
							throw new IllegalArgumentException("Aggregation uses attributes that are not contained in the schema.");
						} 
						
						//none of the used Attributes is contained in the grouping
						for(String usedAttribute: usedAttributes){
							if(this.grouping.contains(usedAttribute)){
								throw new IllegalArgumentException("Aggregation uses grouping attributes.");
							}
						}
					}
					//TODO check that argument only one Attribute
				}
				else if(definition instanceof Attribute || definition instanceof TupleFunction || definition instanceof ConditionFormula || definition instanceof ClockFunction || definition instanceof Parameter || definition instanceof Value<?>){//FIXME Constants to another case, only Values (but for the first time also Literals)
					if(!this.schema().containsAll(definition.AttributeNames())){
						throw new IllegalArgumentException("Definition uses undefined attributes.");
					}
					if(!this.preOutputSchema().containsAll(definition.AttributeNames())){
						throw new IllegalArgumentException("Definition uses undefined attributes.");
					}
					//TODO check that only TupleFunctions
					
				}
				else{
					throw new IllegalArgumentException("Illegal definition: " + definition);
				}
			}
		}
		else{
			throw new IllegalStateException();
		}
	}
	
	public void addDefinitions(final FixedArrayNavigableMap<String, Term<?>> newAttributes){		
		this.checkAddDefinitions(newAttributes);
		//initialize output if neccessary
		if(this.output == null){
			this.initOutput();
		}
		this.output = FixedArrayNavigableMap.merge(this.output, newAttributes);
		this.state = FactoryState.PROJECTION_IMBED;
	}
	
	public boolean canAddRename(){
		return FactoryState.SINGLE_INPUT.ordinal() <= this.state.ordinal() & this.state.ordinal() <= FactoryState.PROJECTION_IMBED.ordinal();
	}
	
	public void checkAddRename(final FixedArrayNavigableMap<String, String> simpleRename){
		if(this.canAddRename()){
			for(String newName: simpleRename.keys()){
				if(!TSA_SyntaxPatterns.isValidAttributeName(newName)){
					throw new IllegalArgumentException("Invalid argument name: " + newName); 
				}
			}
			FixedArrayNavigableSet<String> schema = this.schema();
			for(String oldName: simpleRename.values()){
				if(!schema.contains(oldName)){
					throw new IllegalArgumentException("Rename uses undefined attribute: " + oldName); 
				}
			}
		}
		else{
			throw new IllegalStateException(this.state.toString());
		}
	}
	
	public void addRename(final FixedArrayNavigableMap<String, String> simpleRename){
		this.checkAddRename(simpleRename);
		//rename: newName -> oldName 

		//initialize output if neccessary
		if(this.output == null){
			this.initOutput();
		}
		int numberOfAttributes = simpleRename.size();
		final Term<?>[] reorderedDefinitions = new Term<?>[numberOfAttributes];
		for(int attribute = 0; attribute < numberOfAttributes; ++attribute){
			String oldName = simpleRename.values().get(attribute);
			int oldNameIndex = this.output.keys().indexOf(oldName);
			if(oldNameIndex >= 0){
				reorderedDefinitions[attribute] = this.output.values().get(oldNameIndex);
			}
			else{
				throw new IllegalArgumentException("Invalid mapping to non existing attribute: " + oldName);
			}
		}
		this.output = new FixedArrayNavigableMap<String, Term<?>>(simpleRename.keys(), new FixedArray<Term<?>>(reorderedDefinitions));
		this.state = FactoryState.PROJECTION_IMBED;
	}
	
	protected void initOutput(){
		assert this.output == null;
		FixedArrayNavigableSet<String> outputNames = this.preOutputSchema();
		FixedArray<Attribute<?>> outputTerms = new FixedArray<Attribute<?>>(new CollectionAdapter<String, Attribute<?>>(outputNames){
			@Override
			public Attribute<?> adapt(final String source) {
				return attr(source);
			}
		});
		this.output = new FixedArrayNavigableMap<String, Term<?>>(outputNames, outputTerms);
	}
	
	public boolean canAddAggregation(){
		return FactoryState.GROUPING.ordinal() <= this.state.ordinal() & this.state.ordinal() <= FactoryState.AGGREGATION.ordinal();
	}
	
	public void checkAddAggregation(final FixedArrayNavigableMap<String, Term<?>> newAttributes){
		if(this.canAddAggregation()){
			for(Term<?> definition: newAttributes.values()){
				if(!(definition instanceof AggregationFunction)){
					throw new IllegalArgumentException(definition.toString() + "is not an aggregation function." );
				}
			}
			this.checkAddDefinitions(newAttributes);
		}
		else{
			throw new IllegalStateException();
			//TODO better exception
		}
	}

	public void addAggregation(final FixedArrayNavigableMap<String, Term<?>> newAttributes){
		this.checkAddAggregation(newAttributes);
		//initialize output if neccessary
		this.addDefinitions(newAttributes);
		this.state = FactoryState.AGGREGATION;
	}

	public boolean canAddGrouping(){
		return FactoryState.SINGLE_INPUT.ordinal() <= this.state.ordinal() & this.state.ordinal() < FactoryState.GROUPING.ordinal();
	}
	
	public void checkAddGrouping(final FixedArrayNavigableSet<String> grouping){
		if(this.canAddGrouping()){
			if(!this.schema().containsAll(grouping)){
				throw new IllegalArgumentException("Schema does not contain all attributes contained in the grouping.");
			}
		}
		else{
			throw new IllegalStateException();
			//TODO better exception
		}
	}
	
	public void addGrouping(final FixedArrayNavigableSet<String> grouping){
		this.checkAddGrouping(grouping);
		this.grouping = grouping;
		this.hasGrouping = true;
		this.state = FactoryState.GROUPING;
	}
	
	public boolean canAddSelection(){
		return FactoryState.SINGLE_INPUT.ordinal() <= this.state.ordinal() & this.state.ordinal() < FactoryState.PROJECTION_IMBED.ordinal();
	}
	
	public void checkAddSelection(final ConditionFormula selection){
		if(!this.canAddSelection()){
			throw new IllegalStateException();
			//TODO better exception
		}
		if(!this.schema().containsAll(selection.AttributeNames())){
			throw new IllegalArgumentException("Selection condition contains attributes that are not contained in the schema.");
			//TODO better exception
		}
	}
	
	public void addSelection(final FixedArray<ConditionFormula> selection){
		this.addSelection(new ConjunctiveConditionFormula(selection));
	}
	
	public void addSelection(ConditionFormula selection){
		this.checkAddSelection(selection);
		if(this.state.ordinal() <= FactoryState.SELECTION.ordinal()){
			this.selection = new ConjunctiveConditionFormula(this.selection, selection).flaten();
			this.state = FactoryState.SELECTION;
		}
		else{
			selection = selection.replaceAtomicTerm(new Adapter<AtomicTerm<?>, Term<?>>(){
				
				@Override
				public Term<?> adapt(final AtomicTerm<?> source) {
					if(source instanceof Attribute<?>){
						Attribute<?> attribute = (Attribute<?>) source;
						if(!SQL_ExpressionFactory.this.grouping.contains(attribute.name)){
							assert SQL_ExpressionFactory.this.state.ordinal() == FactoryState.AGGREGATION.ordinal();
							assert SQL_ExpressionFactory.this.output != null;
							return SQL_ExpressionFactory.this.output.get(attribute.name);
						}
					}
					return source;
				}
				
			});
			
			this.having = new ConjunctiveConditionFormula(this.having, selection).flaten();
			this.state = FactoryState.HAVING;
		}
	}
	
	public boolean canAddSetDifference(){
		return this.state == FactoryState.SINGLE_INPUT;
	}
	
	public void checkAddSetDifference(final SQL_ExpressionFactoryInput exceptInput){
		if(this.canAddSetDifference()){
			assert this.inputs.size() == 1;
			if(FixedArrayNavigableSet.symmetricDifference(this.inputs.get(0).schema(), exceptInput.schema()).size() > 0){
				throw new IllegalArgumentException("Attributes of positive input and negative input do not match.");
				//TODO better exception
			}
		}
		else{
			throw new IllegalStateException();
			//TODO better exception
		}
	}
	
	public void addSetDifference(final SQL_ExpressionFactoryInput exceptInput){
		this.checkAddSetDifference(exceptInput);
		assert this.inputs.size() == 1;
		this.uncheckedAddInputs(new FixedArray<SQL_ExpressionFactoryInput>(exceptInput));
		this.state = FactoryState.EXCEPT;
	}
	
	public boolean canAddUnionAll(){
		return this.state == FactoryState.SINGLE_INPUT | this.state == FactoryState.UNION_ALL;
	}
	
	public void checkAddUnionAll(final FixedArray<SQL_ExpressionFactoryInput> newInputs){
		if(this.canAddUnionAll()){
			SQL_ExpressionFactoryInput firstInput = (this.inputs.size() == 0 ? newInputs.get(0) : this.inputs.get(0));
			for(int inputIndex = 0; inputIndex < newInputs.size(); ++inputIndex){
				SQL_ExpressionFactoryInput input = newInputs.get(inputIndex);
				if(FixedArrayNavigableSet.symmetricDifference(firstInput.schema(), input.schema()).size() > 0){
					throw new IllegalArgumentException("Attributes of input 1 and input " + (inputIndex + 1) + " do not match.");
					//TODO better exception
				}
			}
		}
		else{
			throw new IllegalStateException();
			//TODO better exception
		}
	}
	
	public void addUnionAll(final FixedArray<SQL_ExpressionFactoryInput> newInputs){
		this.checkAddUnionAll(newInputs);
		this.uncheckedAddInputs(new FixedArray<SQL_ExpressionFactoryInput>(newInputs));
		if(this.inputs.size() > 1){
			this.state = FactoryState.UNION_ALL;	
		}
	}

	public boolean canAddInputs(){
			return this.state.ordinal() <= FactoryState.CROSSPRODUCT.ordinal();
	}
	
	public void checkAddInputs(final FixedArray<SQL_ExpressionFactoryInput> inputs){
		if(this.canAddInputs()){
			if(inputs.size() > 0){
				int numberOfNewAttributes = 0;
				for (SQL_ExpressionFactoryInput input : inputs) {
					numberOfNewAttributes += input.schema().size();
				}
				//check names
				FixedArrayNavigableSet<String> mergedNewInputNames = FixedArrayNavigableSet.merge(new CollectionAdapter<SQL_ExpressionFactoryInput,FixedArrayNavigableSet<String>>(inputs){
					@Override
					public FixedArrayNavigableSet<String> adapt(final SQL_ExpressionFactoryInput source){
						return source.schema(); 
					}			
				});
				
				FixedArrayNavigableSet<String> mergedInputNames = FixedArrayNavigableSet.merge(this.inputSchema, mergedNewInputNames);
				//better: merge arrays without dublicate elimination. then do the elimination run as check
				
				if(mergedInputNames.size() != this.inputSchema.size() + numberOfNewAttributes){
					throw new IllegalArgumentException();
					//TODO better exception
				}
			}
			else{
				throw new IllegalArgumentException();
				//TODO better exception
			}
		}
		else{
			throw new IllegalStateException();
			//TODO better exception
		}
	}
	
	public void addInputs(final FixedArray<SQL_ExpressionFactoryInput> newInputs){
		if(newInputs.size() > 0){
			this.checkAddInputs(newInputs);

			this.uncheckedAddInputs(newInputs);
		}
		else{
			throw new IllegalArgumentException();
			//TODO better exception
		}
	}
	
	@SuppressWarnings("unchecked")
	public void uncheckedAddInputs(final FixedArray<SQL_ExpressionFactoryInput> newInputs){
		if(newInputs.size() > 0){

			assert this.having == ConditionFormula.Top;
			int numberOfNewAttributes = 0;
			int numberOfNewDirectInputs = 0;
			int numberOfNewSubQueries = 0;
			for (SQL_ExpressionFactoryInput input : newInputs) {
				numberOfNewAttributes += input.schema().size();
				numberOfNewDirectInputs += input.numberOfDirectInputs(true);
				numberOfNewSubQueries += input.numberOfSubqueries(true);
			}
			//check names
			FixedArrayNavigableSet<String> mergedNewInputNames = FixedArrayNavigableSet.merge(new CollectionAdapter<SQL_ExpressionFactoryInput,FixedArrayNavigableSet<String>>(newInputs){
				@Override
				public FixedArrayNavigableSet<String> adapt(final SQL_ExpressionFactoryInput source){
					return source.schema(); 
				}			
			});
			
			FixedArrayNavigableSet<String> mergedInputNames = FixedArrayNavigableSet.merge(this.inputSchema, mergedNewInputNames);

			this.inputs = FixedArray.concat(this.inputs, newInputs);				
			this.numberOfDirectInputs += numberOfNewDirectInputs;
			this.numberOfSubqueries += numberOfNewSubQueries;
			this.inputSchema = mergedInputNames;

			if(this.inputs.size() == 1){
				this.state = FactoryState.SINGLE_INPUT;
			}
			else{assert this.inputs.size() > 1;
				this.state = FactoryState.CROSSPRODUCT;
			}
		}
	}
	
	
	void writeOutput(final StringBuilder sqlQuery, final int indentDepth){
		if(this.state.ordinal() < FactoryState.EXCEPT.ordinal()){
			SQL_QueryBuildHelper.indentTab(sqlQuery, indentDepth);
			sqlQuery.append("SELECT ");
			if(this.output == null){
				FixedArrayNavigableSet<String> attributes = this.schema();
				final int numberOfAttributes = attributes.size();
				assert numberOfAttributes > 0;
				for(int attribute = 0; attribute < numberOfAttributes; ++attribute){
					SQL_QueryBuildHelper.newLineAndIndentTab(sqlQuery, indentDepth + 1);
					sqlQuery.append("\"");
					assert TSA_SyntaxPatterns.isValidAttributeName(attributes.get(attribute));
					//FIXME long attribute names
					sqlQuery.append(SQL_Translation.SQL_BasicAttributeName(attributes.get(attribute)));
					sqlQuery.append("\"");
					sqlQuery.append(",");
				}
				sqlQuery.setLength(sqlQuery.length() - ",".length());
				//sqlQuery.append("*");
			}
			else{
				final int numberOfOutputAttributes = this.output.size();
				assert numberOfOutputAttributes > 0;
				for(int attribute = 0; attribute < numberOfOutputAttributes; ++attribute){
					SQL_QueryBuildHelper.newLineAndIndentTab(sqlQuery, indentDepth + 1);
					this.writeTerm(this.output.values().get(attribute), true, sqlQuery, indentDepth + 1);
					sqlQuery.append(" AS ");
					sqlQuery.append("\"");
					assert TSA_SyntaxPatterns.isValidAttributeName(this.output.keys().get(attribute));
					//FIXME long attribute names
					sqlQuery.append(SQL_Translation.SQL_BasicAttributeName(this.output.keys().get(attribute)));
					sqlQuery.append("\"");
					sqlQuery.append(",");
				}
				sqlQuery.setLength(sqlQuery.length() - ",".length());
			}
		}
	}
	
	public int numberOfDirectInputs(){
		if(this.inputs != null){
			return this.numberOfDirectInputs;
		}
		else{
			throw new IllegalStateException();
			//TODO better exception
		}
	}
	
	public int numberOfSubqueries(){
		if(this.inputs != null){
			return this.numberOfSubqueries;
		}
		else{
			throw new IllegalStateException();
			//TODO better exception
		}
	}
	
	protected int formatingLength(){
		return Math.max(Integer.toString(this.numberOfDirectInputs()).length(),Integer.toString(this.numberOfSubqueries()).length());
	}
	
	public FixedArrayNavigableMap<String, String> aliasMapping(){
		return this.aliasMapping(0, this.formatingLength());
	}
	
	public FixedArrayNavigableMap<String, String> aliasMapping(int directInputNumber, final int formatingLength){
		if(this.inputs != null){
			ArrayList<FixedArrayNavigableMap<String, String>> aliasMappings = new ArrayList<FixedArrayNavigableMap<String,String>>(this.inputs.size());
			for(SQL_ExpressionFactoryInput input: this.inputs){
				aliasMappings.add(input.aliasMapping(directInputNumber, formatingLength));
				directInputNumber += input.numberOfDirectInputs(true);
			}
			return FixedArrayNavigableMap.merge(aliasMappings);
		}
		else{
			throw new IllegalStateException();
			//TODO better exception
		}
	}
	
	void writeInput(final StringBuilder sqlQuery, int directInputNumber, int subqueryNumber, final int formatingLength, final int indentDepth){
		if(this.state.ordinal() < FactoryState.EXCEPT.ordinal()){
			assert this.inputs.size() > 0;
			SQL_QueryBuildHelper.newLineAndIndentTab(sqlQuery, indentDepth);
			sqlQuery.append("FROM \n");
			for(int input = 0; input < this.inputs.size(); ++input){
				this.inputs.get(input).writeTo(sqlQuery, directInputNumber, subqueryNumber, formatingLength, indentDepth + 1, true);
				directInputNumber += this.inputs.get(input).numberOfDirectInputs(true);
				subqueryNumber += this.inputs.get(input).numberOfSubqueries(true);
				sqlQuery.append(",\n");
			}
			sqlQuery.setLength(sqlQuery.length() - ",\n".length());//remove last ", "
		}
	}
	
	void writeSetDifference(final StringBuilder sqlQuery, int directInputNumber, int subqueryNumber, final int formatingLength, final int indentDepth){
		if(this.state == FactoryState.EXCEPT){
			assert this.inputs.size() == 2;			
			this.inputs.get(0).writeTo(sqlQuery, directInputNumber, subqueryNumber, formatingLength, indentDepth + 1, false);
			directInputNumber += this.inputs.get(0).numberOfDirectInputs(false);
			subqueryNumber += this.inputs.get(0).numberOfSubqueries(false);
			SQL_QueryBuildHelper.newLineAndIndentTab(sqlQuery, indentDepth);
			if(this.sql_translation.driver.supportsExceptCorresponding()){
				sqlQuery.append("EXCEPT CORRESPONDING\n");
				this.inputs.get(1).writeTo(sqlQuery, directInputNumber, subqueryNumber, formatingLength, indentDepth + 1, false);
			}
			else{
				sqlQuery.append("EXCEPT\n");
				//FIXME maybe check that attributes of input relations are in lexicografical order, or introduce reodering
				this.inputs.get(1).writeTo(sqlQuery, directInputNumber, subqueryNumber, formatingLength, indentDepth + 1, false);
			}
		}
	}
	
	void writeUnionAll(final StringBuilder sqlQuery, int directInputNumber, int subqueryNumber, final int formatingLength, final int indentDepth){
		if(this.state == FactoryState.UNION_ALL){
			assert this.inputs.size() > 0;
			SQL_QueryBuildHelper.newLineAndIndentTab(sqlQuery, indentDepth);
			for(int input = 0; input < this.inputs.size(); ++input){
				this.inputs.get(input).writeTo(sqlQuery, directInputNumber, subqueryNumber, formatingLength, indentDepth + 1, true);
				directInputNumber += this.inputs.get(input).numberOfDirectInputs(true);
				subqueryNumber += this.inputs.get(input).numberOfSubqueries(true);
				sqlQuery.append("\nUNION\n");
			}
			sqlQuery.setLength(sqlQuery.length() - "\nUNION\n".length());//remove last "\nUNION\n"
		}
	}
	
	void writeSelection(final StringBuilder sqlQuery, final int indentDepth){
		if(this.selection != ConditionFormula.Top){
			this.selection = this.selection.flaten();
			if(this.selection != ConditionFormula.Top){
				SQL_QueryBuildHelper.newLineAndIndentTab(sqlQuery, indentDepth);
				sqlQuery.append("WHERE ");
				this.writeCondition(this.selection, false, sqlQuery, indentDepth);
			}
		}
	}
	
	void writeGrouping(final StringBuilder sqlQuery, final int indentDepth){
		if(this.hasGrouping){
			assert this.inputs.size() > 0;
			final int numberOfGroupingAttributes = this.grouping.size();
			assert numberOfGroupingAttributes > 0;
			SQL_QueryBuildHelper.newLineAndIndentTab(sqlQuery, indentDepth);
			sqlQuery.append("GROUP BY  ");
			sqlQuery.append("\"");
			assert TSA_SyntaxPatterns.isValidAttributeName(this.grouping.get(0));
			sqlQuery.append(SQL_Translation.SQL_BasicAttributeName(this.grouping.get(0)));
			sqlQuery.append("\"");
			for(int attribute = 1; attribute < numberOfGroupingAttributes; ++attribute){
				sqlQuery.append(", ");
				sqlQuery.append("\"");
				assert TSA_SyntaxPatterns.isValidAttributeName(this.grouping.get(attribute));
				sqlQuery.append(SQL_Translation.SQL_BasicAttributeName(this.grouping.get(attribute)));
				sqlQuery.append("\"");
			}
		}
	}
	
	void writeHaving(final StringBuilder sqlQuery, final int indentDepth){
		if(this.having != ConditionFormula.Top){
			assert this.inputs.size() > 0;
			this.having = this.having.flaten();
			if(this.having != ConditionFormula.Top){
				SQL_QueryBuildHelper.newLineAndIndentTab(sqlQuery, indentDepth);
				sqlQuery.append("HAVING ");
				this.writeCondition(this.having, false, sqlQuery, indentDepth);
				sqlQuery.append(" ");
			}
		}
	}
	
	@Override
	public final String toString() {
		StringBuilder sqlQuery = new StringBuilder();
		this.writeTo(sqlQuery);
		sqlQuery.append(";");
		return sqlQuery.toString();
	}
	
	public void writeTo(final StringBuilder sqlQuery, final int directInputNumber, final int subqueryNumber, final int formatingLength, final int indentDepth){
		if(this.state != FactoryState.EMPTY){
			if(this.state.ordinal() < FactoryState.EXCEPT.ordinal()){
				this.writeOutput(sqlQuery, indentDepth);
				this.writeInput(sqlQuery, directInputNumber, subqueryNumber, formatingLength, indentDepth);
				this.writeSelection(sqlQuery, indentDepth);
				this.writeGrouping(sqlQuery, indentDepth);
				this.writeHaving(sqlQuery, indentDepth);
			}
			else if(this.state == FactoryState.EXCEPT){
				this.writeSetDifference(sqlQuery, directInputNumber, subqueryNumber, formatingLength, indentDepth);
			}
			else if(this.state == FactoryState.UNION_ALL){
				
			}
			else{
				throw new Error("Unimplemented case factory state \"" + this.state.name() + "\"");
			}
			sqlQuery.append("\n");
		}
		else{
			throw new IllegalStateException(this.state.toString());
		}
	}
	
	public void writeTo(final StringBuilder sqlQuery){
		this.writeTo(sqlQuery, 0, 0, this.formatingLength(), 0);
	}
	
	void writeCondition(final ConditionFormula condition, final boolean nested, final StringBuilder sqlQuery, final int indentDepth){
		assert condition == condition.flaten();
		switch (condition.ID) {
		case TOP:{
			sqlQuery.append("TRUE");
			break;
		}
		case BOTTOM:{
			sqlQuery.append("FALSE");
			break;
		}
		case ATOMIC:
		{
			if(condition instanceof BinaryComparissonFormula){
				BinaryComparissonFormula binaryComparisson = (BinaryComparissonFormula) condition;
				this.writeTerm(binaryComparisson.firstTerm(), true, sqlQuery, indentDepth);
				
				switch (binaryComparisson.operator) {
				case GREATER:
				{
					sqlQuery.append(" > ");
					break;
				}
				case GREATER_EQUAL:
				{
					sqlQuery.append(" >= ");
					break;
				}
				case EQUAL:
				{
					sqlQuery.append(" = ");
					break;
				}
				case LESS_EQUAL:
				{
					sqlQuery.append(" <= ");
					break;
				}
				case LESS:
				{
					sqlQuery.append(" < ");
					break;
				}
				case NOT_EQUAL:
				{
					sqlQuery.append(" <> ");
					break;
				}
				default: throw new Error("Unimplemented operator: " + binaryComparisson.operator.toString());
				}
				
				this.writeTerm(binaryComparisson.secondTerm(), true, sqlQuery, indentDepth);
			}
			else if(condition instanceof BooleanTerm){
				//FIXME on monetdb fix
				sqlQuery.append("( ");
				this.writeTerm(((BooleanTerm)condition).term, true, sqlQuery, indentDepth);
				sqlQuery.append(" = true )");
			}
			else{
				//FIXME better exception
				throw new IllegalArgumentException();
			}
			
			break;
		}
		case CONJUNCTIVE:
		{
			ConjunctiveConditionFormula conjunctiveCondition = (ConjunctiveConditionFormula) condition;
			assert conjunctiveCondition.subformulas().size() >= 2;
			FixedArray<ConditionFormula> subFormulas = conjunctiveCondition.subformulas();
			int numberOfSubFormulas = subFormulas.size();
			
			if(nested){
				SQL_QueryBuildHelper.newLineAndIndentTab(sqlQuery, indentDepth);
				sqlQuery.append("(   ");
			}
			
			this.writeCondition(subFormulas.get(0), true, sqlQuery, indentDepth + 1);
			
			for(int subformula = 1; subformula < numberOfSubFormulas; ++subformula){
				SQL_QueryBuildHelper.newLineAndIndentTab(sqlQuery, indentDepth);
				if(nested){
					sqlQuery.append("AND ");
				}
				else{
					sqlQuery.append("  AND ");
				}
				this.writeCondition(subFormulas.get(subformula), true, sqlQuery, indentDepth + 1);
			}
			
			if(nested){
				SQL_QueryBuildHelper.newLineAndIndentTab(sqlQuery, indentDepth);
				sqlQuery.append(")");
			}
			break;
		}
		case DISJUNCTIVE:
		{
			DisjunctiveConditionFormula disjunctiveCondition = (DisjunctiveConditionFormula) condition;
			assert disjunctiveCondition.subformulas().size() >= 2;
			FixedArray<ConditionFormula> subFormulas = disjunctiveCondition.subformulas();
			int numberOfSubFormulas = subFormulas.size();
			
			if(nested){
				SQL_QueryBuildHelper.newLineAndIndentTab(sqlQuery, indentDepth);
				sqlQuery.append("(   ");
			}
			
			this.writeCondition(subFormulas.get(0), true, sqlQuery, indentDepth + 1);
			
			for(int subformula = 1; subformula < numberOfSubFormulas; ++subformula){
				SQL_QueryBuildHelper.newLineAndIndentTab(sqlQuery, indentDepth);
				if(nested){
					sqlQuery.append("OR ");
				}
				else{
					sqlQuery.append("   OR ");
				}
				this.writeCondition(subFormulas.get(subformula), true, sqlQuery, indentDepth + 1);
			}
			
			if(nested){
				SQL_QueryBuildHelper.newLineAndIndentTab(sqlQuery, indentDepth);
				sqlQuery.append(")");
			}
			break;
		}
		case NEGATED:
		{
			NegatedConditionFormula negatedCondition = (NegatedConditionFormula) condition;
			
			sqlQuery.append("NOT ( ");
			
			this.writeCondition(negatedCondition.subformula(), false, sqlQuery, indentDepth);		
						
			sqlQuery.append(" )");

			break;
		}
		}
	}
	
	void writeTerm(final Term<?> term, final boolean nested, final StringBuilder sqlQuery, final int indentDepth){
		switch (term.ID) {
		case ATOMIC:
		{
			AtomicTerm<?> atomicTerm = (AtomicTerm<?>) term;
			switch (atomicTerm.ID) {
			case ATTRIBUTE:
			{
				Attribute<?> attribute = (Attribute<?>) atomicTerm;
				sqlQuery.append("\"");
				sqlQuery.append(SQL_Translation.SQL_BasicAttributeName(attribute.name));
				sqlQuery.append("\"");
				break;
			}
			case VARIABLE:
			{
				throw new IllegalArgumentException();
				//FIXME better Exception
				//break;
			}
			case CONSTANT:
			{
				if(atomicTerm instanceof BasicTypeValue){
					BasicTypeValue basicValue = (BasicTypeValue) atomicTerm;
					switch (basicValue.type.ID) {
					case BOOLEAN: 
					{
						sqlQuery.append(((BooleanValue)basicValue).value);
						break;
					}
					case DOUBLE:
					{
						sqlQuery.append(Double.toString(((DoubleValue)basicValue).value));
						break;
					}
					case DURATION:
					{
						sqlQuery.append(((DurationValue)basicValue).value);
						break;
					}
					case FLOAT:
					{
						sqlQuery.append(Float.toString(((FloatValue)basicValue).value));
						break;
					}
					case IDENTIFIER:
					{
						sqlQuery.append(((IdentifierValue)basicValue).value);
						break;
					}
					case INT:
					{
						sqlQuery.append(((IntValue)basicValue).value);
						break;
					}
					case LONG:
					{
						sqlQuery.append(((LongValue)basicValue).value);
						break;
					}
					case REFERENCE:
					{
						sqlQuery.append(((ReferenceValue)basicValue).value);
						break;
					}
					case STRING:
					{
						sqlQuery.append("'");
						sqlQuery.append(this.sql_translation.escape(((StringValue)basicValue).value));
						sqlQuery.append("'");
						break;
					}
					case TIMESTAMP:
					{
						sqlQuery.append(((TimestampValue)basicValue).value);
						break;
					}
					}
				}
				else if(atomicTerm instanceof ConstantLiteral){
					sqlQuery.append("'");
					sqlQuery.append(this.sql_translation.escape(((ConstantLiteral)atomicTerm).literal));
					sqlQuery.append("'");
					//FIXME remove this case when all constant values (numbers) can be typed by the DURA compiler
				}
				else{
					throw new IllegalArgumentException();
					//FIXME better Exception
					//break;
				}
				break;
			}
			case MIN_TIME:
			{
				//TODO better exception
				throw new IllegalArgumentException("Illegal case " + atomicTerm.ID);
				//break;
			}
			case PARAMETER:
			{
				Parameter parameter = (Parameter) atomicTerm;
				sqlQuery.append("CAST(/*'");
				sqlQuery.append(parameter.name);
				sqlQuery.append("'*/? AS ");
				sqlQuery.append(this.sql_translation.SQLType(parameter.type.ID));
				sqlQuery.append(")");
				break;
			}
			default:
				throw new Error("Programming error: Uncovered case " + atomicTerm.ID);
			}
			
			break;
		}
		case RECORD:
		{
			throw new IllegalArgumentException();
			//FIXME better Exception
			//break;
		}
		case TUPLE:
		{
			throw new IllegalArgumentException();
			//FIXME better Exception
			//break;
		}
		case FUNCTION:
		{
			switch (((Function<?>) term).ID) {
			case TUPLE: {
				TupleFunction<?> function = (TupleFunction<?>) term;
				switch (function.ID) {
				case PLUS:	
				{	
					if(nested){
						sqlQuery.append("( ");
					}
					
					this.writeTerm(function.arg(0), true, sqlQuery, indentDepth);
					
					sqlQuery.append(" + ");
					
					this.writeTerm(function.arg(1), true, sqlQuery, indentDepth);
					
					if(nested){
						sqlQuery.append(" )");
					}
					break;
				}
				case MINUS:	
				{	
					if(nested){
						sqlQuery.append("( ");
					}
					
					this.writeTerm(function.arg(0), true, sqlQuery, indentDepth);
					
					sqlQuery.append(" - ");
					
					this.writeTerm(function.arg(1), true, sqlQuery, indentDepth);
					
					if(nested){
						sqlQuery.append(" )");
					}
					break;
				}
				case TIMES:	
				{	
					if(nested){
						sqlQuery.append("( ");
					}
					
					this.writeTerm(function.arg(0), true, sqlQuery, indentDepth);
					
					sqlQuery.append(" * ");
					
					this.writeTerm(function.arg(1), true, sqlQuery, indentDepth);
					
					if(nested){
						sqlQuery.append(" )");
					}
					break;
				}
				case DIVISION:	
				{	
					if(nested){
						sqlQuery.append("( ");
					}
					
					this.writeTerm(function.arg(0), true, sqlQuery, indentDepth);
					
					sqlQuery.append(" / ");
					
					this.writeTerm(function.arg(1), true, sqlQuery, indentDepth);
					
					if(nested){
						sqlQuery.append(" )");
					}
					break;
				}
				case POWER:	
				{	
					if(nested){
						sqlQuery.append("( ");
					}
					
					this.writeTerm(function.arg(0), true, sqlQuery, indentDepth);
					
					sqlQuery.append(" ^ ");
					
					this.writeTerm(function.arg(1), true, sqlQuery, indentDepth);
					
					if(nested){
						sqlQuery.append(" )");
					}
					break;
				}
				case LEAST:	
				{	
					sqlQuery.append("LEAST");
					sqlQuery.append("( ");
					if(function.arity() > 0){
						this.writeTerm(function.arg(0), (function.arity() == 1), sqlQuery, indentDepth);
						for(int arg = 1; arg < function.arity(); ++arg){
							sqlQuery.append(", ");
							this.writeTerm(function.arg(arg), true, sqlQuery, indentDepth);
						}
					}
					sqlQuery.append(" )");
					break;
				}
				case GREATEST:	
				{	
					sqlQuery.append("GREATEST");
					sqlQuery.append("( ");
					if(function.arity() > 0){
						this.writeTerm(function.arg(0), (function.arity() == 1), sqlQuery, indentDepth);
						for(int arg = 1; arg < function.arity(); ++arg){
							sqlQuery.append(", ");
							this.writeTerm(function.arg(arg), true, sqlQuery, indentDepth);
						}
					}
					sqlQuery.append(" )");
					break;
				}
				case USER:	
				{	
					throw new IllegalArgumentException();
					//FIXME better Exception
					//break;
				}
				}
				break;
			}
			case CLOCK: {
				//FIXME translation should be database independent
				ClockFunction function = (ClockFunction) term;
				if(function.type.ID == BasicTypeID.IDENTIFIER){//FIXME more flexible
					sqlQuery.append(this.sql_translation.driver.sqlGetSequenceNextValueFunction("/*'$seq'*/\"meta_schema\".\"" + function.name + "\"/*'$'*/"));
				}
				else{
					sqlQuery.append("CAST(/*'clock.now:");
					sqlQuery.append(function.name);
					sqlQuery.append("'*/? AS ");
					sqlQuery.append(this.sql_translation.SQLType(function.type.ID));
					sqlQuery.append(")");
				}
				break;
			}
			case CAST: {
				//FIXME translation should be database independent
				//FIXME implement
				CastFunction function = (CastFunction) term;
				if(function.type instanceof BasicType){
					if(function.arg().type instanceof BasicType && ((BasicType)function.arg().type).ID == BasicTypeID.STRING){//FIXME more flexible
						sqlQuery.append("CAST(");
						this.writeTerm(function.arg(), (function.arity() == 1), sqlQuery, 0);
						sqlQuery.append(" AS ");
						sqlQuery.append(this.sql_translation.SQLType(((BasicType)function.type).ID));
						sqlQuery.append(" )");
					}
					else{
						this.writeTerm(function.arg(), (function.arity() == 1), sqlQuery, indentDepth);
					}
					break;
				}
				else{
					throw new IllegalArgumentException("Cast is not supported: " + function.toString());
				}
			}
			case AGGREGATION: {
				AggregationFunction<?> function = (AggregationFunction<?>) term;
				switch (function.ID) {
				case AVG: {
					sqlQuery.append("GREATEST");
					break;
				}
				case COUNT:
				{
					sqlQuery.append("COUNT");
					break;
				}
				case MAX:
				{	
					sqlQuery.append("MAX");
					break;
				}
				case MIN:
				{
					sqlQuery.append("MIN");
					break;
				}
				case SUM:
				{
					sqlQuery.append("SUM");
					break;
				}
				case USER:
				{
					throw new IllegalArgumentException();
					//FIXME better Exception
					//break;
				}
				//	default:
				//		break;
				}
				sqlQuery.append("( ");
				assert function.arity() == 1;
				this.writeTerm(function.arg(0), false, sqlQuery, indentDepth);
				sqlQuery.append(" )");	
				break;
			}			
			//	default:
			//		break;
			}
			break;
		}
		case BOOLEAN_FUNCTION:
		{
			ConditionFormula conditionFormula = (ConditionFormula) term;
			sqlQuery.append("( ");
				this.writeCondition(conditionFormula, true, sqlQuery, indentDepth);
			sqlQuery.append(" )");
		}
		}
	}
	
	public SQL_ExpressionFactoryInput toSubquery(){
		return new SQL_ExpressionFactorySubqueryInput(this);
	}
	
}
