package de.lmu.ifi.pms.cep.event_mill.sql.creation;

import de.lmu.ifi.pms.cep.event_mill.sql.SQL_Translation;
import de.lmu.ifi.pms.cep.tsa.DTSA_BufferType;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;

import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;

public class SQL_ExpressionFactoryDirectInput extends SQL_ExpressionFactoryInput {
	
	public final String full_name;
	public final DTSA_BufferType buffer_type;
	
	protected final FixedArrayNavigableSet<String> schema;
	
	
	
	public SQL_ExpressionFactoryDirectInput(final String full_name, final DTSA_BufferType buffer_type, final FixedArrayNavigableSet<String> schema) {
		super();
		if(TSA_SyntaxPatterns.isValidBufferFullName(full_name)){
			this.full_name = full_name;
			if(buffer_type == DTSA_BufferType.INPUT | buffer_type == DTSA_BufferType.WORKING){
				this.buffer_type = buffer_type;
			}
			else{
				throw new IllegalArgumentException("Only tables in input and working schema may serve as input.");
			}
			this.schema = schema;
		}
		else{
			throw new IllegalArgumentException();
			//TODO better exception
		}
	}

	@Override
	public FixedArrayNavigableSet<String> schema() {
		return this.schema;
	}
	
	@Override
	public int numberOfDirectInputs(final boolean writeAlias){
		return (writeAlias ? 1 : 0);
	}
	
	@Override
	public int numberOfSubqueries(final boolean writeAlias){
		return 0;
	}
	
	@Override
	protected String alias(final int directInputNumber, final int formatingLength){
		return SQL_Translation.SQL_BasicTableName(String.format("input%0" + formatingLength + "d.%s", directInputNumber, this.full_name));
	}
	
	@Override
	public FixedArrayNavigableMap<String, String> aliasMapping(final int directInputNumber, final int formatingLength){
		return new FixedArrayNavigableMap<String, String>(new String[]{this.alias(directInputNumber, formatingLength)}, new String[]{this.full_name});
	}
	
	@Override
	public void writeTo(final StringBuilder sqlQuery, final int directInputNumber, final int subqueryNumber, final int formatingLength, final int indentDepth, final boolean writeAlias) {
		SQL_QueryBuildHelper.indentTab(sqlQuery, indentDepth);
		sqlQuery.append("/*'$table'*/");
		sqlQuery.append("\"");
		switch (this.buffer_type) {
		case INPUT: 
		{
			sqlQuery.append("input_schema");
			break;
		}
		case WORKING:
		{
			sqlQuery.append("working_schema");
			break;
		}
		case OUTPUT:
		{
			throw new IllegalArgumentException("Tables in output schema may not serve as input.");
			//TODO better exception
		}
		case LOG:
		{
			throw new IllegalArgumentException("Tables in log schema may not serve as input.");
			//TODO better exception
		}
		case BLACK_HOLE_INPUT:
		{
			throw new IllegalArgumentException("Black hole tables may not serve as input.");
			//TODO better exception
		}
		case BLACK_HOLE_WORKING:
		{
			throw new IllegalArgumentException("Black hole tables may not serve as input.");
			//TODO better exception
		}
		}
		
		sqlQuery.append("\".\"");
		sqlQuery.append(SQL_Translation.SQL_BasicTableName(this.full_name));
		sqlQuery.append("\"");
		sqlQuery.append("/*'$'*/");
		
		if(writeAlias){
			sqlQuery.append(" AS ");
			
			sqlQuery.append("\"");
			sqlQuery.append(this.alias(directInputNumber, formatingLength));
			sqlQuery.append("\"");
		}
	}

}
