package de.lmu.ifi.pms.cep.event_mill.sql.creation;

import de.lmu.ifi.pms.cep.event_mill.sql.SQL_Translation;
import de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_ExpressionFactory.FactoryState;
import de.lmu.ifi.pms.cep.tsa.DTSA_BufferType;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;

import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;

public class SQL_QueryFactory {
	
	public final String full_name;
	public final DTSA_BufferType buffer_type;
	
	protected final FixedArrayNavigableSet<String> schema;
	
	public final SQL_ExpressionFactory expression;
	
	public final FixedArrayNavigableMap<String, String> aliasMapping;
	
	public SQL_QueryFactory(final String output_full_name, final DTSA_BufferType output_buffer_type, final FixedArrayNavigableSet<String> output_schema, final SQL_ExpressionFactory expression) {
		super();
		if(TSA_SyntaxPatterns.isValidBufferFullName(output_full_name)){
			this.full_name = output_full_name;
			if(output_buffer_type == DTSA_BufferType.WORKING | output_buffer_type == DTSA_BufferType.OUTPUT | output_buffer_type == DTSA_BufferType.LOG | output_buffer_type == DTSA_BufferType.BLACK_HOLE_INPUT | output_buffer_type == DTSA_BufferType.BLACK_HOLE_WORKING){
				this.buffer_type = output_buffer_type;
			}
			else{
				throw new IllegalArgumentException("Only tables in working, output and log schema or formal black hole buffers for garbage collection may serve as output.");
			}
			this.schema = output_schema;
			if(this.schema.containsAll(expression.schema())){
				this.expression = expression;
			}
			else{
				throw new IllegalArgumentException("Expression derives attributes that are not contained in the schema of the output buffer.");
			}
			this.aliasMapping = expression.aliasMapping();
		}
		else{
			throw new IllegalArgumentException();
			//TODO better exception
		}
	}
	
	@Override
	public final String toString() {
		StringBuilder sqlQuery = new StringBuilder();
		this.writeTo(sqlQuery);
		return sqlQuery.toString();
	}
	
	public int numberOfDirectInputs(){
		return this.expression.numberOfDirectInputs();
	}
	
	public int numberOfSubqueries(){
		return this.expression.numberOfSubqueries();
	}
	
	public void writeTo(final StringBuilder sqlQuery){
		if(this.buffer_type == DTSA_BufferType.WORKING | this.buffer_type == DTSA_BufferType.OUTPUT | this.buffer_type == DTSA_BufferType.LOG){
			sqlQuery.append("INSERT INTO ");
		}
		else if(this.buffer_type == DTSA_BufferType.BLACK_HOLE_INPUT | this.buffer_type == DTSA_BufferType.BLACK_HOLE_WORKING){
			sqlQuery.append("DELETE FROM ");
		}
		else{
			throw new Error("Unexpected case");
		}
		sqlQuery.append("/*'$table'*/");
		sqlQuery.append("\"");
		switch (this.buffer_type) {
		case INPUT: 
		{
			throw new IllegalArgumentException("Tables in input schema may not serve as input.");
		}
		case WORKING:
		{
			sqlQuery.append("working_schema");
			break;
		}
		case OUTPUT:
		{
			sqlQuery.append("output_schema");
			break;
		}
		case LOG:
		{
			sqlQuery.append("log_schema");
			break;
		}
		case BLACK_HOLE_INPUT:
		{
			sqlQuery.append("input_schema");
			break;
		}
		case BLACK_HOLE_WORKING:
		{
			sqlQuery.append("working_schema");
			break;
		}
		}
		
		sqlQuery.append("\".\"");
		sqlQuery.append(SQL_Translation.SQL_BasicTableName(this.full_name));
		sqlQuery.append("\"");
		sqlQuery.append("/*'$'*/");
		sqlQuery.append("\n");
		
		if(this.buffer_type == DTSA_BufferType.WORKING | this.buffer_type == DTSA_BufferType.OUTPUT | this.buffer_type == DTSA_BufferType.LOG){
			
			sqlQuery.append("( ");
			
			FixedArrayNavigableSet<String> expressionSchema = this.expression.schema();
			assert expressionSchema.size() > 0;
			
			
			for(String attribute: expressionSchema){
				sqlQuery.append("\"");
				sqlQuery.append(SQL_Translation.SQL_BasicAttributeName(attribute));
				sqlQuery.append("\"");
				sqlQuery.append(", ");
			}
			sqlQuery.setLength(sqlQuery.length() - ", ".length());//remove ",\n" that has been appended last
			
			sqlQuery.append(" )");
			sqlQuery.append("\n");
			
			this.expression.writeTo(sqlQuery);		
		}
		else if(this.buffer_type == DTSA_BufferType.BLACK_HOLE_INPUT | this.buffer_type == DTSA_BufferType.BLACK_HOLE_WORKING){
			
			if(this.expression.inputs.size() == 1){
				SQL_ExpressionFactoryDirectInput input;
				if(this.expression.inputs.get(0) instanceof SQL_ExpressionFactoryDirectInput){
					input = (SQL_ExpressionFactoryDirectInput) this.expression.inputs.get(0);
				}
				else{
					assert this.expression.inputs.get(0) instanceof SQL_ExpressionFactorySubqueryInput;
					SQL_ExpressionFactory subExpression = ((SQL_ExpressionFactorySubqueryInput)this.expression.inputs.get(0)).subQuery;
					if(subExpression.inputs.size() == 1 & subExpression.state==FactoryState.PROJECTION_IMBED){
						if(subExpression.inputs.get(0) instanceof SQL_ExpressionFactoryDirectInput){
							//FIXME check that subExpression soley consists of an projection
							input = (SQL_ExpressionFactoryDirectInput) subExpression.inputs.get(0);
						}
						else{
							throw new RuntimeException("Unimplemented case.");
						}
					}
					else{
						throw new RuntimeException("Unimplemented case.");
					}
				}
				
				DTSA_BufferType expectedInputBufferType;
				if(this.buffer_type == DTSA_BufferType.BLACK_HOLE_INPUT){
					expectedInputBufferType = DTSA_BufferType.INPUT;
				}
				else if(this.buffer_type == DTSA_BufferType.BLACK_HOLE_WORKING){
					expectedInputBufferType = DTSA_BufferType.WORKING;
				}
				else{
					throw new RuntimeException("Unimplemented case.");
				}
				
				if(input.buffer_type == expectedInputBufferType & input.full_name.equals(this.full_name)){
					if(this.expression.state == FactoryState.SELECTION){
						
						this.expression.writeSelection(sqlQuery, 0);
						sqlQuery.append("\n");
						
						return;
					}
				}			
			}

			throw new RuntimeException("Unimplemented case.");

			
			
		}
		else{
			throw new Error("Unexpected case");
		}
		
		
		sqlQuery.append(";");
	}

}
