package de.lmu.ifi.pms.cep.examples.hello_world;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.util.Iterator;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.event_mill.configuration.IOConfiguration;
import de.lmu.ifi.pms.cep.event_mill.console.AbstractConsoleThread;
import de.lmu.ifi.pms.cep.event_mill.io.format.StringPatternGenerator;
import de.lmu.ifi.pms.cep.event_mill.io.stream.MessageSourceStream;
import de.lmu.ifi.pms.cep.event_mill.io.stream.StreamIO;


public class GreetPersonConsole extends AbstractConsoleThread{
	
	StreamIO io;
	
	MessageSourceStream<String> greetMessages;
	long delay = 1000;
	
	boolean running = false;
	boolean quit = false;

	public GreetPersonConsole(final Iterator<String> immediateCommands, final BufferedReader inputStream, final PrintStream outputStream) {
		super("GreetPersonConsole", immediateCommands, inputStream, outputStream);
		
		try {
			this.io = IOConfiguration.read().streamProvider("de.lmu.ifi.pms.cep.event-mill.prototype").connect("de.lmu.ifi.pms.cep.event-mill.prototype:example.hello-world:test");
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
		
	}

	@Override
	public void run(){
		
//		try {
			
			String command;
			this.outputStream.println("Welcome,");
			this.outputStream.println("Type \"start\" when you are ready.");
			this.outputStream.println("(For leaving type \"quit\")");
			
			while(!this.quit){
			while(!this.running){
				command = this.nextCommand();
				if(command.equalsIgnoreCase("quit")){
					this.quit = true;
					
					this.outputStream.println("Bye!");
					break;
				}
				else if(command.equalsIgnoreCase("start")){	
					this.outputStream.println("Be polite!");
					if(this.greetMessages == null){
						this.greetMessages = new MessageSourceStream<String>("example.hello-world.greet-person", this.io, new StringPatternGenerator("Greet %{person}", this.io.sourceTypes("example.hello-world.greet-person")));
					}
					this.running = true;
				}
				else{
					this.outputStream.println("What do you mean?");
		
				}
			}
			
			long earliestNewBlockRead = System.currentTimeMillis() - this.delay;
			long currentTime;
			long delay;
			while(this.running){
				try {
					if(this.inputStream.ready()){
						command = this.nextCommand();
						if(command.equalsIgnoreCase("quit")){
							this.running = false;
							this.quit = true;
							
							this.outputStream.println("Bye!");
							continue;
						}
						else if(command.equalsIgnoreCase("stop")){
							this.running = false;
							
							continue;
						}
						else{
							this.outputStream.println("What do you mean?");

						}
					}
				} catch (IOException e) {
					this.error("Problem with input stream", e);
					e.printStackTrace();
					throw new RuntimeException(e);
				}
				
				
				currentTime = System.currentTimeMillis();
				delay = earliestNewBlockRead - currentTime;
				while(delay > 0){
					try {
						Thread.sleep(delay);
					} catch (InterruptedException e) {

					}
					currentTime = System.currentTimeMillis();
					delay = earliestNewBlockRead - currentTime;
				}
				this.greetMessages.nextBlock();
				
				earliestNewBlockRead = currentTime + this.delay;
				
				while(this.greetMessages.next()){
		        	this.outputStream.println(this.greetMessages.current());
		        	this.outputStream.flush();
				}
			}
			}
//		
//		} catch (Interruo e) {
//			
//		}
		if(this.greetMessages != null){
			this.greetMessages.close();
		}
		System.exit(0);
	}
	
	//logging
	protected static final Logger logger = LoggerFactory.getLogger(GreetPersonConsole.class);
	
	@Override
	protected Logger logger(){
		return logger;
	}
	
	public static void main(final String[] args) {
		BufferedReader inputStream = new BufferedReader(new InputStreamReader(System.in));
		PrintStream outputStream = System.out;
		
		new GreetPersonConsole(java.util.Arrays.asList(args).iterator(), inputStream, outputStream).start();
	}
	
}
