package de.lmu.ifi.pms.cep.tsa;

import java.util.Comparator;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import de.lmu.ifi.pms.cep.event_mill.sql.SQL_Translation;




public final class TSA_SyntaxPatterns {

	private TSA_SyntaxPatterns(){
		
	}
	
	public final static String ProgramNamespaceRegEx = "(?:[a-zA-Z_](?:[a-zA-Z0-9_[-]]*[a-zA-Z0-9_])?(?:[.][a-zA-Z_](?:[a-zA-Z0-9_[-]]*[a-zA-Z0-9_])?)*)?";
	public final static Pattern ProgramNamespacePattern = Pattern.compile(ProgramNamespaceRegEx);
	
	public final static String ProgramNameRegEx = "[a-zA-Z_](?:[a-zA-Z0-9_[-]]*[a-zA-Z0-9_])?";
	public final static Pattern ProgramNamePattern = Pattern.compile(ProgramNameRegEx);
	
	public final static String ProgramFullNameRegEx = "(?:(" + ProgramNamespaceRegEx + ")[.])?(" + ProgramNameRegEx + ")";
	public final static Pattern ProgramFullNamePattern = Pattern.compile(ProgramFullNameRegEx);
	
	public final static String program_instance_name_RegEx = "(?:[a-zA-Z_](?:[a-zA-Z0-9_[-]]*[a-zA-Z0-9_])?)?";
	public final static Pattern program_instance_name_Pattern = Pattern.compile(program_instance_name_RegEx);
	public final static String program_instance_full_name_RegEx = ProgramFullNameRegEx + "[:]" + "(" + program_instance_name_RegEx + ")";
	public final static Pattern program_instance_full_name_Pattern = Pattern.compile(program_instance_full_name_RegEx);
	
	public final static String BufferNameRegEx = "[.]?[a-zA-Z_](?:[a-zA-Z0-9_[-][$]]*[a-zA-Z0-9_])?";
	public final static Pattern BufferNamePattern = Pattern.compile(BufferNameRegEx);
	
	public final static String SimplePackageNameRegEx = "[.]?(?:[a-zA-Z_](?:[a-zA-Z0-9_[-]]*[a-zA-Z0-9_])?)";
	public final static String PackageNameRegEx = "(?:" + SimplePackageNameRegEx + "[.]" + ")" + "*";
	public final static Pattern PackageNamePattern = Pattern.compile(PackageNameRegEx);
	
	public final static String BufferFullNameRegEx = "(" + PackageNameRegEx + ")(" + BufferNameRegEx + ")";
	public final static Pattern BufferFullNamePattern = Pattern.compile(BufferFullNameRegEx);
	
	//public final static String AliasBufferNameRegEx = BufferNameRegEx;
	//public final static Pattern AliasBufferNamePattern = Pattern.compile(BufferNameRegEx);
	
	public final static String AttributeSimpleNameRegEx = "[.]?(?:[a-zA-Z_](?:[a-zA-Z0-9_[-]]*[a-zA-Z0-9_])?)";
	public final static Pattern AttributeSimpleNamePattern = Pattern.compile(AttributeSimpleNameRegEx);
	
	public final static String AttributeNameRegEx = "(?:" + AttributeSimpleNameRegEx + "[.]" + ")" + "*" + AttributeSimpleNameRegEx;
	public final static Pattern AttributeNamePattern = Pattern.compile(AttributeNameRegEx);
	
	public final static String QualifiedStreamBoundNameRegEx = "(" + BufferFullNameRegEx + "):(" + AttributeNameRegEx + ")";
	public final static Pattern QualifiedStreamBoundNamePattern = Pattern.compile(QualifiedStreamBoundNameRegEx);
	
	//public final static String LocalMinTimeNameRegEx = "(" + AliasBufferNameRegEx + ")[.](" + MinTimeNameRegEx + ")";
	//public final static Pattern LocalMinTimeNamePattern = Pattern.compile(LocalMinTimeNameRegEx);
	
	public final static String ClockNameRegEx = AttributeNameRegEx;
	public final static Pattern ClockNamePattern = Pattern.compile(ClockNameRegEx);
	
	public static String programFullName(final String namespace, final String simple_name){
		assert namespace.matches(ProgramNamespaceRegEx);
		assert simple_name.matches(ProgramNameRegEx);
		if(namespace.equals("")){
			return simple_name;
		}
		else{
			return namespace + "." + simple_name;
		}
	}
	
	public static String[] programFullName(final String full_name){
		assert full_name.matches(ProgramFullNameRegEx);
		
		Matcher decoder = ProgramFullNamePattern.matcher(full_name);
		if(decoder.matches()){
			return new String[]{(decoder.group(1) == null ? "" : decoder.group(1)),decoder.group(2)};
		}
		else{
			throw new IllegalArgumentException("Invalid name: " + full_name);
		}
	}
	
	public static int compareProgramFullName(final String full_name1, final String full_name2){
		String[] split_full_name1 = programFullName(full_name1);
		String[] split_full_name2 = programFullName(full_name2);
		int cmp = split_full_name1[0].compareTo(split_full_name2[0]);
		if(cmp == 0){
			cmp = split_full_name1[1].compareTo(split_full_name2[1]);	
		}
		return cmp;
	}
	
	public final static Comparator<String> programFullNameComparator = new Comparator<String>() {
		
		@Override
		public int compare(final String full_name1, final String full_name2) {
			return compareProgramFullName(full_name1, full_name2);
		}
		
	};
	
	public static boolean isValidBufferFullName(final String name){
		return BufferFullNamePattern.matcher(name).matches();
	}
	
	public static String bufferFullName(final String prog_package, final String simple_name){
		assert prog_package.matches(PackageNameRegEx);
		assert simple_name.matches(BufferNameRegEx);
		
		return prog_package + simple_name;
	}
	
	public static String[] bufferFullName(final String full_name){
		assert full_name.matches(BufferFullNameRegEx);
		
		Matcher matcher = BufferFullNamePattern.matcher(full_name);
		matcher.matches();
		return new String[]{matcher.group(1), matcher.group(2)};
	}
	
	
	public static int compareBufferFullName(final String full_name1, final String full_name2){
		String[] split_full_name1 = bufferFullName(full_name1);
		String[] split_full_name2 = bufferFullName(full_name2);
		int cmp = split_full_name1[0].compareTo(split_full_name2[0]);
		if(cmp == 0){
			cmp = split_full_name1[1].compareTo(split_full_name2[1]);	
		}
		return cmp;
	}
	
	public final static Comparator<String> bufferFullNameComparator = new Comparator<String>() {
		
		@Override
		public int compare(final String full_name1, final String full_name2) {
			return compareBufferFullName(full_name1, full_name2);
		}
		
	};
	
	public static String bufferQualifiedName(final String prog_full_name, final String buffer_full_name){
		assert prog_full_name.matches(program_instance_full_name_RegEx);
		assert buffer_full_name.matches(BufferFullNameRegEx);
		
		return prog_full_name + ":" + buffer_full_name;
	}
	
	public static boolean isValidAttributeSimpleName(final String simple_name){
		return AttributeSimpleNamePattern.matcher(simple_name).matches();
	}
	
	public static boolean isValidAttributeName(final String name){
		return AttributeNamePattern.matcher(name).matches();
	}
	
	public static boolean isValidAttributeName(final String[] splitName){
		for(String name: splitName){
			if(!isValidAttributeSimpleName(name)){
				return false;
			}
		}
		return true;
	}
	
	public static String attributeName(final String... name){
		StringBuilder sb = new StringBuilder(64);
		for(String simpleName: name){
			sb.append(simpleName);
			sb.append('.');
		}
		sb.setLength(sb.length() - 1);//remove last dot
		return sb.toString();
	}
	
	public static String[] splitAttributeName(final String name){
		assert name.matches(AttributeNameRegEx) : "Invalid attribute name: " + name;		
		
		String[] splitName = name.split("(?<!\\A|[.])[.]");
		assert isValidAttributeName(splitName) : "Implementation error: Invalid split.";
		return splitName;
	}

	public static String attributeQualifiedName(final String buffer_full_name, final String attribute_name){
		assert buffer_full_name.matches(BufferFullNameRegEx);
		assert attribute_name.matches(AttributeNameRegEx);
		
		return buffer_full_name + ":" + attribute_name;
	}
	
	public static boolean isSuperAttribute(final String attributeName1, final String attributeName2){
		assert isValidAttributeName(attributeName2);
		if(isValidAttributeName(attributeName1)){
			assert isValidAttributeName(attributeName2);
			return attributeName1.equals(attributeName2) | attributeName2.startsWith(makeAttributeNamePrefix(attributeName1));
		}
		else{
			return false;
		}
	}
	
	public static void makeAttributeNamePrefix(final StringBuilder attributeName){
		assert isValidAttributeName(attributeName.toString());
		attributeName.append('.');
	}
	
	public static String makeAttributeNamePrefix(final String attributeName){
		assert isValidAttributeName(attributeName);
		return attributeName + '.';
	}
	
	public static String prefixAttribute(final String attributeName1, final String attributeName2){
		assert isValidAttributeName(attributeName2);
		if(attributeName1 == null || attributeName1.length() == 0){
			return attributeName2;
		}
		else{assert isValidAttributeName(attributeName1);
			return makeAttributeNamePrefix(attributeName1) + attributeName2;
		}
	}
	
	public static void prefixAttribute(final StringBuilder attributeName1, final String attributeName2){
		assert isValidAttributeName(attributeName2);
		if(attributeName1.length() == 0){
			attributeName1.append(attributeName2);			
		}
		else{assert isValidAttributeName(attributeName1.toString());
			makeAttributeNamePrefix(attributeName1); 
			attributeName1.append(attributeName2);
		}
	}
	
	public static String attributeSuffix(final String attributeName1, final String attributeName2){
		assert isSuperAttribute(attributeName1, attributeName2);
		return attributeName2.substring(makeAttributeNamePrefix(attributeName2).length(), attributeName2.length());
	}
	
	public static boolean isValidClockName(final String name){
		return ClockNamePattern.matcher(name).matches();
	}
	
	public static boolean isValidParameterName(final String name){
		return SQL_Translation.parameter_qualified_Pattern.matcher(name).matches();
	}
	
	public static boolean isValidPackageName(final String name){
		return PackageNamePattern.matcher(name).matches();
	}
	
	public static boolean isValidBufferName(final String name){
		return BufferNamePattern.matcher(name).matches();
	}
	
	public static String streamBoundQualifiedName(final String buffer_full_name, final String minTime_name){
		assert buffer_full_name.matches(BufferFullNameRegEx);
		assert minTime_name.matches(AttributeNameRegEx);
		
		return buffer_full_name + ":" + minTime_name;
	}
	
	public static boolean isValidStreamBoundQualifiedName(final String name){
		return QualifiedStreamBoundNamePattern.matcher(name).matches();
	}
	
	public static String[] minTimeQualifiedName(final String qualified_minTime_name){
		assert qualified_minTime_name.matches(QualifiedStreamBoundNameRegEx);
		Matcher matcher = QualifiedStreamBoundNamePattern.matcher(qualified_minTime_name);
		matcher.matches();
		return new String[]{matcher.group(1), matcher.group(4)};
	}
	
	/*public static String minTimeLocalName(final String buffer_alias_name, final String attribute_name){
		assert buffer_alias_name.matches(AliasBufferNameRegEx);
		assert attribute_name.matches(MinTimeNameRegEx);
		
		return buffer_alias_name + "." + attribute_name;
	}*/
	
	/*public static String[] minTimeLocalName(final String local_attribute_name){
		assert local_attribute_name.matches(LocalMinTimeNameRegEx);
		Matcher matcher = LocalMinTimeNamePattern.matcher(local_attribute_name);
		matcher.matches();
		return new String[]{matcher.group(1), matcher.group(2)};
	}*/

	public static String programInstFullName(final String namespace, final String simple_name, final String instance){
		return programInstFullName(programFullName(namespace, simple_name), instance);
	}
	
	public static String programInstFullName(final String program_full_name, final String instance){
		assert isValidProgramFullName(instance);
		return program_full_name + ":" + instance;
	}
	
	public static boolean isValidNamespace(final String namespace){
		return ProgramNamespacePattern.matcher(namespace).matches();
	}
	
	public static boolean isValidProgramSimpleName(final String simple_name){
		return ProgramNamePattern.matcher(simple_name).matches();
	}
	
	public static boolean isValidProgramFullName(final String instance){
		return ProgramFullNamePattern.matcher(instance).matches();
	}
	
	public static boolean isValidProgramInstanceName(final String instance){
		return program_instance_name_Pattern.matcher(instance).matches();
	}

	public static String[] programInstFullName(final String full_name){
		Matcher decoder = program_instance_full_name_Pattern.matcher(full_name);
		if(decoder.matches()){
			return new String[]{(decoder.group(1) == null ? "" : decoder.group(1)),decoder.group(2),decoder.group(3)};
		}
		else{
			throw new IllegalArgumentException("Invalid name: " + full_name);
		}
	}
	
	
	public static int compareProgramInstFullName(final String full_name1, final String full_name2){
		String[] split_full_name1 = programInstFullName(full_name1);
		String[] split_full_name2 = programInstFullName(full_name2);
		int cmp = split_full_name1[0].compareTo(split_full_name2[0]);
		if(cmp == 0){
			cmp = split_full_name1[1].compareTo(split_full_name2[1]);
			if(cmp == 0){
				cmp = split_full_name1[2].compareTo(split_full_name2[2]);	
			}
		}
		return cmp;
	}
	
	public final static Comparator<String> programInstFullNameComparator = new Comparator<String>() {
		
		@Override
		public int compare(final String full_name1, final String full_name2) {
			return compareProgramInstFullName(full_name1, full_name2);
		}
		
	};
	
}
