package de.lmu.ifi.pms.cep.tsa.conditions;

import org.simpleframework.xml.Element;
import org.simpleframework.xml.Root;
import org.simpleframework.xml.core.Replace;
import org.simpleframework.xml.core.Resolve;

import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.schema.AType;
import de.lmu.ifi.pms.cep.tsa.schema.NoType;
import de.lmu.ifi.pms.cep.tsa.schema.Type;


@Root(name="attribute")
public class Attribute<TypeClass extends AType> extends AtomicTerm<TypeClass> {
	
	@org.simpleframework.xml.Attribute(name="name")
	public final String name;
	
	@Element(name="type", required=false)
	protected final TypeClass getType(){
		if(this.type == NoType.instance){
			return null;
		}
		else{
			assert this.type instanceof Type;
			return this.type;
		}
	}
	
	@SuppressWarnings("unused")
	@Resolve
	private Attribute<?> resolve(){
		if(this.type == null){
			return new UntypedAttribute(this.name);
		}
		else{
			assert this.type != NoType.instance;
			return TypedAttribute.create(this.name, (Type)this.type);
		}
	}
	
	@Replace
	Attribute<TypeClass> replace(){
		return new Attribute<TypeClass>(this.name, this.type);
	}

	Attribute(@org.simpleframework.xml.Attribute(name="name")final String name, @Element(name="type", required=false)final TypeClass type) {
		super(AtomicTermID.ATTRIBUTE, type);
		if(!TSA_SyntaxPatterns.isValidAttributeName(name)){
			throw new IllegalArgumentException("Invalid attribute name: " + name);
		}
		this.name = name;
	}
	
	public Attribute<TypeClass> rename(final String newName){
		throw new AbstractMethodError();
	}
	
	@Override
	protected int compareToImpl(final Term<?> o) {
		int cmp;
		if(o instanceof Attribute<?>){
			cmp = this.name.compareTo(((Attribute<?>)o).name);
			if(cmp == 0){
				cmp = super.compareToImpl(o);
			}
		}
		else{
			cmp = super.compareToImpl(o);
			assert cmp != 0;
		}
		return cmp;
	}
	
	@Override
	protected boolean equalsImpl(final Object o) {
		if(this.getClass() == o.getClass()){
			return this.name.equals(((Attribute<?>)o).name);
		}
		else{
			return false;
		}
	}
	
	@Override
	public int hashCode() {
		return super.hashCode() ^ this.name.hashCode();
	}

	
	@Override
	public String toString() {
		return "attr " + this.name;
	}
}
