package de.lmu.ifi.pms.cep.tsa.conditions;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;

import org.simpleframework.xml.Root;
import org.simpleframework.xml.core.Resolve;

import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;

import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.adaption.CollectionAdapter;
import de.lmu.ifi.pms.java.util.collection.AbstractCollection;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.collection.fixed.empty.Empty;
import de.lmu.ifi.pms.java.util.filter.Filter;
import de.lmu.ifi.pms.java.util.tupel.Single;


@Root(name="condition")
public abstract class ConditionFormula extends Term<BasicType>
{
	
	public static final class TopC extends ConditionFormula {
	
		private TopC() {
			super(ConditionFormulaID.TOP, true, true);
		}
		
		@Override
		public int DNFsize(){
			return 1;
		}
		
		@Override
		protected AbstractCollection<ConditionFormula> DNFiteration(final boolean positivePolarity) {
			if(positivePolarity){
				return new Single<ConditionFormula, ConditionFormula>(this);
			}
			else{
				return Bottom.DNFiteration(true);
			}
		}

		@Override
		public DisjunctiveConditionFormula DNF() {
			return new DisjunctiveConditionFormula(new ConjunctiveConditionFormula());
		}
		
		@Override
		public AbstractCollection<ConjunctiveConditionFormula> DNFiteration() {
			return new Single<ConjunctiveConditionFormula,ConjunctiveConditionFormula>(new ConjunctiveConditionFormula());
		}
		
		@Override
		public int CNFsize() {
			return 0;
		}
		
		@Override
		protected AbstractCollection<ConditionFormula> CNFiteration(final boolean positivePolarity) {
			if(positivePolarity){
				return Empty.instance();
			}
			else{
				return Bottom.CNFiteration(true);
			}
		}

		@Override
		public ConjunctiveConditionFormula CNF() {
			return new ConjunctiveConditionFormula(new DisjunctiveConditionFormula(this));
		}
		
		@Override
		public AbstractCollection<DisjunctiveConditionFormula> CNFiteration() {
			return new Single<DisjunctiveConditionFormula,DisjunctiveConditionFormula>(new DisjunctiveConditionFormula(this));
		}
		
		@Override
		public ConditionFormula replaceAtomicTerm(final Adapter<? super AtomicTerm<?>, ? extends Term<?>> replacement) {
			return this;
		}
		
		@Override
		public ConditionFormula replaceTerm(final Adapter<Term<?>, ? extends Term<?>> replacement) {
			return this;
		}
		
		@Override
		protected void collect(final Filter<Term<?>> filter, final ArrayList<Term<?>> formulas) {
	
		}
		
		@Override
		public boolean containsConjunction() {
			return false;
		}
		
		@Override
		public boolean containsDisjunction() {
			return false;
		}
		
		@Override
		public boolean containsNegation() {
			return false;
		}
		
		@Override
		protected TopC normalizeImpl() {
			return this;
		}

		@Override
		protected TopC flatenImpl() {
			return this;
		}

		@Override
		protected TopC focus(final Filter<? super AtomicConditionFormula> topic, final boolean positivePolarity) {
			return this;
		}

		@Override
		public TopC replaceAtomicFormula(final Adapter<AtomicConditionFormula, ? extends ConditionFormula> replacement) {
			return this;
		}
		
		@Override
		protected boolean equalsImpl(final Object o) {
			assert this != o;
			return false;
		}
		
		@Override
		public void writeTo(final StringBuilder sb, final int indentDepth) {
			indentTab(sb, indentDepth);
			sb.append(this.ID);
		}
		
		@Resolve
		TopC resolve(){
			assert Top != null;
			return Top;
		}
	
	}

	public static final class BottomC extends ConditionFormula {
	
		private BottomC() {
			super(ConditionFormulaID.BOTTOM, true, true);
		}
		
		@Override
		public int DNFsize(){
			return 0;
		}
		
		@Override
		protected AbstractCollection<ConditionFormula> DNFiteration(final boolean positivePolarity) {
			if(positivePolarity){
				return Empty.instance();
			}
			else{
				return Top.DNFiteration(true);
			}
		}

		@Override
		public DisjunctiveConditionFormula DNF() {
			return new DisjunctiveConditionFormula(new ConjunctiveConditionFormula(this));
		}
		
		@Override
		public AbstractCollection<ConjunctiveConditionFormula> DNFiteration() {
			return Empty.instance();
		}
		
		@Override
		public int CNFsize() {
			return 1;
		}
		
		@Override
		protected AbstractCollection<ConditionFormula> CNFiteration(final boolean positivePolarity) {
			if(positivePolarity){
				return new Single<ConditionFormula, ConditionFormula>(this);
			}
			else{
				return Top.CNFiteration(true);
			}
		}

		@Override
		public ConjunctiveConditionFormula CNF() {
			return new ConjunctiveConditionFormula(new DisjunctiveConditionFormula());
		}
		
		@Override
		public AbstractCollection<DisjunctiveConditionFormula> CNFiteration() {
			return new Single<DisjunctiveConditionFormula,DisjunctiveConditionFormula>(new DisjunctiveConditionFormula());
		}
		
		@Override
		public ConditionFormula replaceAtomicTerm(final Adapter<? super AtomicTerm<?>, ? extends Term<?>> replacement) {
			return this;
		}
		
		
		@Override
		public ConditionFormula replaceTerm(final Adapter<Term<?>, ? extends Term<?>> replacement) {
			return this;
		}
		
		@Override
		protected void collect(final Filter<Term<?>> filter, final ArrayList<Term<?>> formulas) {
	
		}
		
		@Override
		public boolean containsConjunction() {
			return false;
		}
		
		@Override
		public boolean containsDisjunction() {
			return false;
		}
		
		@Override
		public boolean containsNegation() {
			return false;
		}
		
		@Override
		protected BottomC normalizeImpl() {
			return this;
		}

		@Override
		protected BottomC flatenImpl() {
			return this;
		}

		@Override
		protected BottomC focus(final Filter<? super AtomicConditionFormula> topic, final boolean positivePolarity) {
			return this;
		}

		@Override
		public BottomC replaceAtomicFormula(final Adapter<AtomicConditionFormula, ? extends ConditionFormula> replacement) {
			return this;
		}
		
		@Override
		protected boolean equalsImpl(final Object o) {
			assert this != o;
			return false;
		}
		
		@Override
		public void writeTo(final StringBuilder sb, final int indentDepth) {
			indentTab(sb, indentDepth);
			sb.append(this.ID);
		}
		
		@Resolve
		BottomC resolve(){
			return Bottom;
		}
	
	}

	@SuppressWarnings("hiding")
	public final ConditionFormulaID ID;
	public final boolean isNormalized;
	public final boolean isFlatened;
	public final static ConditionFormula.BottomC Bottom = new ConditionFormula.BottomC();
	public final static ConditionFormula.TopC Top = new ConditionFormula.TopC();

	ConditionFormula(final ConditionFormulaID ID, final boolean isNormalized, final boolean isFlatened) {
		super(TermID.BOOLEAN_FUNCTION, BasicTypeID.BOOLEAN.type);
		this.ID = ID;
		this.isNormalized = isNormalized;
		this.isFlatened = isFlatened;
		assert !(isFlatened & !isNormalized);
	}
	
	/*
	@Override
	@SuppressWarnings("unchecked")
	public final FixedArrayNavigableSet<Attribute<?>> Variables(){
		return new FixedArrayNavigableSet<Attribute<?>>((AbstractCollection<Attribute<?>>)(AbstractCollection<?>)this.collectTerms(VariableFilter.instance));
	}

	@Override
	public FixedArrayNavigableSet<String> VariableNames(){
		return new FixedArrayNavigableSet<String>(new CollectionAdapter<Term<?>, String>(this.collectTerms(VariableFilter.instance)){
			
			@Override
			public String adapt(final Term<?> source) {
				return ((Variable<?>) source).name;
			}
			
		});
	}
	
	@Override
	@SuppressWarnings("unchecked")
	public final FixedArrayNavigableSet<Attribute<?>> Attributes(){
		return new FixedArrayNavigableSet<Attribute<?>>((AbstractCollection<Attribute<?>>)(AbstractCollection<?>)this.collectTerms(AttributeFilter.instance));
	}
	
	@Override
	public FixedArrayNavigableSet<String> AttributeNames(){
		return new FixedArrayNavigableSet<String>(new CollectionAdapter<Term<?>, String>(this.collectTerms(AttributeFilter.instance)){
			
			@Override
			public String adapt(final Term<?> source) {
				return ((Attribute<?>) source).name;
			}
			
		});
	}*/
	
	public ConditionFormula normalize(){
		if(this.isNormalized){
			return this;
		}
		else{
			return this.normalizeImpl();
		}
	}
	
	protected abstract ConditionFormula normalizeImpl();
	
	public ConditionFormula flaten(){
		if(this.isFlatened){
			return this;
		}
		else if(this.isNormalized){
			return this.flatenImpl();
		}
		else{
			return this.normalize().flaten();
		}
	}
	
	protected abstract ConditionFormula flatenImpl();
	
	public DisjunctiveConditionFormula DNF(){
		AbstractCollection<ConjunctiveConditionFormula> conjunctions = this.DNFiteration();
		return new DisjunctiveConditionFormula(conjunctions);
	}
	
	public AbstractCollection<ConjunctiveConditionFormula> DNFiteration(){
		return new CollectionAdapter<ConditionFormula, ConjunctiveConditionFormula>(this.DNFiteration(true)){

			@Override
			public ConjunctiveConditionFormula adapt(final ConditionFormula source) {
				ConditionFormula flatenedSource = source.flaten();
				switch (flatenedSource.ID) {
				case ATOMIC: case NEGATED: case BOTTOM: case TOP:
					return new ConjunctiveConditionFormula(source);
				default:
					return (ConjunctiveConditionFormula) flatenedSource;
				}
			}
			
		};
	}
	
	public abstract int DNFsize();

	protected abstract AbstractCollection<ConditionFormula> DNFiteration(boolean positivePolarity);
	
	public ConjunctiveConditionFormula CNF(){
		AbstractCollection<DisjunctiveConditionFormula> disjunctions = this.CNFiteration();
		return new ConjunctiveConditionFormula(disjunctions);
	}
	
	public AbstractCollection<DisjunctiveConditionFormula> CNFiteration(){
		return new CollectionAdapter<ConditionFormula, DisjunctiveConditionFormula>(this.CNFiteration(true)){

			@Override
			public DisjunctiveConditionFormula adapt(final ConditionFormula source) {
				ConditionFormula flatenedSource = source.flaten();
				switch (flatenedSource.ID) {
				case ATOMIC: case NEGATED:
					return new DisjunctiveConditionFormula(source);
				default:
					return (DisjunctiveConditionFormula) flatenedSource;				
				}
			}
			
		};
	}
	
	public abstract int CNFsize();
	
	protected abstract AbstractCollection<ConditionFormula> CNFiteration(boolean positivePolarity);
	
	public abstract boolean containsNegation();
	public abstract boolean containsConjunction();
	public abstract boolean containsDisjunction();
	
	public ConditionFormula focus(final Filter<? super AtomicConditionFormula> topic){
		return this.focus(topic, true);
	}
	
	protected abstract ConditionFormula focus(Filter<? super AtomicConditionFormula> topic, boolean positivePolarity);
	
	public abstract ConditionFormula replaceAtomicFormula(Adapter<AtomicConditionFormula, ? extends ConditionFormula> replacement);
	
	@Override
	public abstract ConditionFormula replaceAtomicTerm(Adapter<? super AtomicTerm<?>, ? extends Term<?>> replacement);
	
	public abstract ConditionFormula replaceTerm(Adapter<Term<?>, ? extends Term<?>> replacement);
	
	public AbstractCollection<ConditionFormula> collectFormulas(final Filter<? super ConditionFormula> filter){
		ArrayList<ConditionFormula> formulas = new ArrayList<ConditionFormula>();
		this.collectFormulas(filter, formulas);
		return formulas;
	}
	
	protected void collectFormulas(final Filter<? super ConditionFormula> filter, final ArrayList<ConditionFormula> formulas){
		if(filter.accepts(this)){
			formulas.add(this);
		}
	}
	
	@Override
	protected abstract void collect(Filter<Term<?>> filter, ArrayList<Term<?>> terms);
	
	@Override
	public abstract void writeTo(final StringBuilder sb, final int indentDepth);
	
	@Override
	public void writeTo(final StringBuilder sb){
		this.writeTo(sb, 0);
	}
	
	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		this.writeTo(sb);
		return sb.toString();
	}
	
	public static ConjunctiveConditionFormula and(final ConditionFormula... formulas){
		return new ConjunctiveConditionFormula(formulas);
	}
	
	public static ConjunctiveConditionFormula and(final Collection<? extends ConditionFormula> formulas){
		return new ConjunctiveConditionFormula(formulas);
	}
	
	public static DisjunctiveConditionFormula or(final ConditionFormula... formulas){
		return new DisjunctiveConditionFormula(formulas);
	}
	
	public static DisjunctiveConditionFormula or(final Collection<? extends ConditionFormula> formulas){
		return new DisjunctiveConditionFormula(formulas);
	}
	
	public static NegatedConditionFormula not(final ConditionFormula formula){
		return new NegatedConditionFormula(formula);
	}

	
}
