package de.lmu.ifi.pms.cep.tsa.conditions;

import org.simpleframework.xml.Root;

import de.lmu.ifi.pms.cep.tsa.schema.AType;
import de.lmu.ifi.pms.cep.tsa.schema.NamedType;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;


@Root(name="tuple")
public abstract class Record<TypeClass extends AType> extends CompositeTerm<TypeClass> {
	
	public final FixedArrayNavigableMap<String, Term<?>> arguments;
	
	Record(final TypeClass type, final FixedArrayNavigableSet<String> argumentNames, final java.util.ArrayList<Term<?>> arguments) {
		this(type, argumentNames, false, arguments.toArray(new Term<?>[arguments.size()]));
	}
	
	public Record(final TypeClass type, final FixedArrayNavigableSet<String> argumentNames, final Term<?>... arguments) {
		this(type, argumentNames, true, arguments);
	}
	
	protected Record(final TypeClass type, final FixedArrayNavigableSet<String> argumentNames, final boolean copy, final Term<?>... arguments) {
		super(TermID.RECORD, type, copy, arguments);
		if(argumentNames.size() == arguments.length){
			this.arguments = new FixedArrayNavigableMap<String, Term<?>>(argumentNames, new FixedArray<Term<?>>(this.argumentsArray));
		}
		else{
			throw new IllegalArgumentException("Number of argument terms differs from number of specified arguments.");
		}
	}
	
	@Override
	public abstract Record<TypeClass> newTerm(final Term<?>... arguments);

	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append("{");
		if(this.arguments.size() > 0){
			for(int argument = 0; argument < this.arguments.size(); ++argument){
				sb.append(this.arguments.keys().get(argument));
				sb.append("{ ");
				sb.append(this.arguments.values().get(argument));
				sb.append(" }");
				sb.append(", ");
			}
			sb.setLength(sb.length() - ", ".length());
		}
		sb.append("}");
		if(this.type instanceof NamedType<?>){
			sb.append(" : ");
			sb.append(((NamedType<?>)this.type).name);
		}
		return sb.toString();
	}
	
	public Term<?> arg(final String name) {
		return this.arguments.get(name);
	}
	
}
