package de.lmu.ifi.pms.cep.tsa.conditions;

import org.simpleframework.xml.Attribute;
import org.simpleframework.xml.Element;
import org.simpleframework.xml.Root;

import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;


@Root(name="stream-bound")
public class StreamBound extends AtomicTerm<BasicType> {
	
	@Attribute(name="buffer")
	public final String buffer_name;
	@Attribute(name="name")
	public final String name;
	public final String qualified_name;
	
	@Element(name="type", required=true)
	protected final BasicType getType(){
		return this.type;
	}
	
	public StreamBound(final String buffer_name, final String name, final BasicTypeID typeID) {
		this(buffer_name, name, typeID.type);
	}

	public StreamBound(@Attribute(name="buffer") final String buffer_name, @Attribute(name="name") final String name, @Element(name="type", required=true) final BasicType type) {
		super(AtomicTermID.MIN_TIME, type);
		//FIXME
		if(!TSA_SyntaxPatterns.isValidBufferFullName(buffer_name)){
			throw new IllegalArgumentException("Invalid name for buffer: " + buffer_name);
		}
		if(!TSA_SyntaxPatterns.isValidAttributeName(name)){
			throw new IllegalArgumentException("Invalid name for stream bound: " + name);
		}
		this.buffer_name = buffer_name;
		this.name = name;
		this.qualified_name = TSA_SyntaxPatterns.streamBoundQualifiedName(buffer_name, name);
		//TODO allow only IDENTIFIER and TIMESTAMP?
	}
	
	@Override
	protected int compareToImpl(final Term<?> o) {
		int cmp;
		if(o instanceof StreamBound){
			StreamBound sbi = (StreamBound) o;
			cmp = this.qualified_name.compareTo(sbi.qualified_name);
			if(cmp == 0){
				cmp = this.type.compareTo(sbi.type);
				if(cmp == 0){
					cmp = super.compareToImpl(o);
				}
			}
		}
		else{
			cmp = super.compareToImpl(o);
			assert cmp != 0;
		}
		return cmp;
	}
	
	@Override
	protected boolean equalsImpl(final Object o) {
		if(this.getClass() == o.getClass()){
			return this.qualified_name.equals(((StreamBound)o).qualified_name) && this.type.equals(((StreamBound)o).type);
		}
		else{
			return false;
		}
	}
	
	@Override
	public int hashCode() {
		return super.hashCode() ^ this.qualified_name.hashCode();
	}

	
	@Override
	public String toString() {
		return this.qualified_name;
	}
}
