package de.lmu.ifi.pms.cep.tsa.conditions;

import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.schema.CompositeType;
import de.lmu.ifi.pms.cep.tsa.schema.NamedType;
import de.lmu.ifi.pms.cep.tsa.schema.Type;

public class TypedVariable<TypeClass extends Type> extends Variable<TypeClass> {

	protected TypedVariable(final String name, final TypeClass type) {
		super(name, type);
	}
	
	@SuppressWarnings("unchecked")
	public static <TypeClass extends Type> TypedVariable<TypeClass> create(final String name, final TypeClass type){
		switch (type.ID) {
		case BASIC:
			return (TypedVariable<TypeClass>) BasicTypeVariable.create(name, (BasicType)type);
		case COMPOSITE:
			return (TypedVariable<TypeClass>) DataTypeVariable.create(name, (CompositeType)type);
		case NAMED:
			return (TypedVariable<TypeClass>) NamedTypeVariable.create(name, (NamedType<?>)type);		
		default:
			throw new Error("Implementation erroe, missing case: " + type.ID);
		}
	}

	@Override
	public TypedVariable<TypeClass> rename(final String newName){
		throw new AbstractMethodError();
	}
	
	
	@Override
	protected int compareToImpl(final Term<?> o) {
		int cmp = super.compareToImpl(o);
		if(cmp == 0){
			cmp = this.type.compareTo(((TypedVariable<?>)o).type);
		}
		return cmp;
	}
	
	@Override
	protected boolean equalsImpl(final Object o) {
		if(super.equalsImpl(o)){
			return this.type.equals(((TypedVariable<?>)o).type);
		}
		else{
			return false;
		}
	}
	
	@Override
	public int hashCode() {
		return super.hashCode() ^ this.type.hashCode();
	}
	
	@Override
	public String toString() {
		return super.toString() + " : " + this.type.toString();
	}
}
