package de.lmu.ifi.pms.cep.tsa.conditions.adaptor;

import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.conditions.AtomicTerm;
import de.lmu.ifi.pms.cep.tsa.conditions.Attribute;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_SchemaException;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchema;
import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;

public class RenameAttributes implements Adapter<AtomicTerm<?>, AtomicTerm<?>>{
	
	public final FixedArrayNavigableMap<String, String> mapping;
	
	public RenameAttributes(final FixedArrayNavigableMap<String, String> mapping) throws TSA_SchemaException {
		super();
		TemporalStreamSchema.checkNameMappingStatic(mapping);
		this.mapping = mapping;
	}

	@Override
	public AtomicTerm<?> adapt(final AtomicTerm<?> source) {
		if(this.mapping.isEmpty()){
			return source;
		}
		else{
			if(source instanceof Attribute<?>){
				Attribute<?> attribute = (Attribute<?>) source;
				if(this.mapping.has("")){
					assert this.mapping.size() == 1;
					String newPrefix = this.mapping.values().get(0);
					if(newPrefix.equals("")){
						return source;
					}
					else{
						String newName = newPrefix + "." + attribute.name;//FIXME more generic against changes in TSA_Syntaxpatterns
						return attribute.rename(newName);
					}
				}
				else{
					int index = this.mapping.keys().indexOf(attribute.name);
					if(index >= 0){
						return attribute.rename(this.mapping.values().get(index));
					}
					else{
						int insertionIndex = -(index + 1);
						if(0 < insertionIndex){
							String potentialPrefixAttribute = this.mapping.keys().get(insertionIndex - 1);
							if(TSA_SyntaxPatterns.isSuperAttribute(potentialPrefixAttribute, attribute.name)){
								String newPrefix = this.mapping.values().get(insertionIndex);
								String replace;
								if(newPrefix.equals("")){
									replace = TSA_SyntaxPatterns.makeAttributeNamePrefix(potentialPrefixAttribute);
								}
								else{
									replace = potentialPrefixAttribute;
								}							
								String newName = attribute.name.replaceFirst(replace, newPrefix);
								return attribute.rename(newName);
							}
							else{
								return source;
							}
						}
						else{
							return source;
						}
					}
				}
			}
			else{
				return source;
			}
		}
	}

}
