package de.lmu.ifi.pms.cep.tsa.operators;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;
import static de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula.and;

import org.simpleframework.xml.Attribute;
import org.simpleframework.xml.ElementList;

import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.adaptor.PrefixVariables;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_Exception;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_SchemaException;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchema;
import de.lmu.ifi.pms.cep.tsa.schema.Type;

import de.lmu.ifi.pms.java.util.adaption.CollectionAdapter;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.fixed.FixedCollection;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;



public class TSA_CrossproductOperator extends TSA_BasicOperator
{
	@Attribute(name="input-arity")
	protected final int inputArity;
	
	public TSA_CrossproductOperator(final String origin, final int inputArity){
		this(new FixedArray<String>(false, origin), inputArity);
	}
	
	public TSA_CrossproductOperator(@ElementList(entry="origin", inline=true) final java.util.ArrayList<String> origin
			, @Attribute(name="input-arity")final int inputArity){
		this(new FixedArray<String>(origin), inputArity);
	}

	public TSA_CrossproductOperator(final FixedCollection<String> origin, @Attribute(name="input-arity")final int inputArity)
	{
		super(TSA_BasicOperatorID.CROSSPRODUCT, origin);
		if(inputArity >= 2){
			this.inputArity = inputArity;
		}
		else{
			throw new IllegalArgumentException("The crossproduct expects a minimum arity of 2. (" + inputArity + ")");
		}
	}
	
	@Override
	public int inputArity() {
		return this.inputArity;
	}
	
	@Override
	public TemporalStreamSchema apply(final Collection<TemporalStreamSchema> inputs) throws TSA_Exception {
		this.checkInputs(inputs);
		
		FixedArrayNavigableSet<String> attributeNames = FixedArrayNavigableSet.merge(new CollectionAdapter<TemporalStreamSchema, FixedArrayNavigableSet<String>>(inputs){
			@Override
			public FixedArrayNavigableSet<String> adapt(final TemporalStreamSchema source) {
				return source.attributes.keys();
			}
		});
		
		Type[] attributeTypes = new Type[attributeNames.size()];
		for(TemporalStreamSchema input: inputs){
			for(int attribute = 0; attribute < input.attributes.size(); ++attribute){
				attributeTypes[attributeNames.indexOf(input.attributes.keys().get(attribute))] = input.attributes.values().get(attribute);
			}			
		}
		
		ConditionFormula[] tupleConstraints = new ConditionFormula[inputs.size()];
		ConditionFormula[] relationConstraints = new ConditionFormula[inputs.size()];
		int inputIndex = 0;
		for(TemporalStreamSchema input: inputs){
			PrefixVariables prefixer = new PrefixVariables("$" + inputIndex);
			tupleConstraints[inputIndex] = and(input.tupleConstraints()).replaceAtomicTerm(prefixer);
			relationConstraints[inputIndex] = and(input.relationConstraints()).replaceAtomicTerm(prefixer);
			++inputIndex;
		}
		
		try {
			return new TemporalStreamSchema(new FixedArrayNavigableMap<String, Type>(attributeNames, new FixedArray<Type>(false, attributeTypes)), new FixedArray<ConditionFormula>(false, tupleConstraints), new FixedArray<ConditionFormula>(false, relationConstraints), false);
		} catch (TSA_SchemaException e) {
			throw new TSA_Exception(this.origin, e);
		}
	}
	
	@Override
	public void checkInputs(final Collection<TemporalStreamSchema> inputs) throws TSA_Exception {
		super.checkInputs(inputs);
		
		int numberOfAttributes = 0;
		for(TemporalStreamSchema input: inputs){
			numberOfAttributes += input.basicAttributes.size();
		}
		
		//attributes must be disjunct
		//FIXME consider prefixes!!
		FixedArrayNavigableSet<String> attributeNames = FixedArrayNavigableSet.merge(new CollectionAdapter<TemporalStreamSchema, FixedArrayNavigableSet<String>>(inputs){
			@Override
			public FixedArrayNavigableSet<String> adapt(final TemporalStreamSchema source) {
				return source.basicAttributes.keys();
			}
		});
		
		if(numberOfAttributes > attributeNames.size()){
			throw new TSA_Exception(this.origin, "Attributes are not disjunct.");
			//TODO better exception
		}
	}
	
	@Override
	public boolean isMonotonic(Collection<TemporalStreamSchema> inputs){
		return true;
	}
	
	@Override
	public void check(final Collection<TemporalStreamSchema> inputs) throws TSA_Exception {
		this.checkInputs(inputs);
	}
	
	@Override
	public void writeTo(final StringBuilder sb, final int indentDepthFirstLine, final int indentDepth) {
		indentTab(sb, indentDepthFirstLine);
		sb.append(this.ID.toString());
	}

}
