package de.lmu.ifi.pms.cep.tsa.operators;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;
import static de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula.*;
import static de.lmu.ifi.pms.cep.tsa.conditions.Term.var;
import static de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID.*;

import org.simpleframework.xml.ElementList;

import de.lmu.ifi.pms.cep.tsa.conditions.AtomicTerm;
import de.lmu.ifi.pms.cep.tsa.conditions.Attribute;
import de.lmu.ifi.pms.cep.tsa.conditions.BasicTypeAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.BasicTypeVariable;
import de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonOperatorID;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.TypedAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.Variable;
import de.lmu.ifi.pms.cep.tsa.conditions.adaptor.PrefixVariables;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_Exception;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_SchemaException;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchema;
import de.lmu.ifi.pms.cep.tsa.time.constraints.SimpleBoundTemporalFormula;

import de.lmu.ifi.pms.java.lang.iteration.Iterator;
import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.collection.fixed.FixedCollection;
import de.lmu.ifi.pms.java.util.map.Entry;

public class TSA_SetDifferenceOperator extends TSA_BasicOperator{
	
	public TSA_SetDifferenceOperator(final String origin){
		this(new FixedArray<String>(origin));
	}
	
	public TSA_SetDifferenceOperator(@ElementList(entry="origin", inline=true) final java.util.ArrayList<String> origin){
		this(new FixedArray<String>(origin));
	}
	
	
	public TSA_SetDifferenceOperator(final FixedCollection<String> origin)
	{
		super(TSA_BasicOperatorID.SETDIFFERENCE, origin);
	}
	
	@Override
	public int inputArity() {
		return 2;
	}

	@Override
	public TemporalStreamSchema apply(final Collection<TemporalStreamSchema> inputs) throws TSA_Exception {
		this.checkInputs(inputs);
		assert inputs.size() == 2;
		Iterator<TemporalStreamSchema> inputIterator = inputs.iterator();
		TemporalStreamSchema positiveInput = inputIterator.next();
		TemporalStreamSchema negativeInput = inputIterator.next();
		
		
		//prefix variables to ensure free names
		final Adapter<AtomicTerm<?>, AtomicTerm<?>> prefixerPosive = new PrefixVariables("$$+");
		final Adapter<AtomicTerm<?>, AtomicTerm<?>> prefixerNegative = new PrefixVariables("$$-");
		
				
		//convert removed attributes to variables
		final Adapter<AtomicTerm<?>, AtomicTerm<?>> removedAttributeReplacer = new Adapter<AtomicTerm<?>, AtomicTerm<?>>(){
			@Override
			public AtomicTerm<?> adapt(final AtomicTerm<?> source) {
				if(source instanceof Attribute){
					Attribute<?> attribute = (Attribute<?>) source;
					if(attribute instanceof TypedAttribute<?>){
						return var("$" + attribute.name, ((TypedAttribute<?>)attribute).type);
					}
					else{
						return var("$" + attribute.name);
					}
				}
				return source;
			}
		};
		
		final ArrayList<Attribute<?>> identifierAttributes = new ArrayList<Attribute<?>>();
		final ArrayList<Attribute<?>> timestampAttributes = new ArrayList<Attribute<?>>();
		
		for(Entry<String, BasicType> attribute: positiveInput.basicAttributes()){
			if(attribute.value().ID == IDENTIFIER){
				identifierAttributes.add(BasicTypeAttribute.create(attribute.key(), attribute.value()));
			}
			if(attribute.value().ID == TIMESTAMP){
				timestampAttributes.add(BasicTypeAttribute.create(attribute.key(), attribute.value()));
			}
		}
		
		FixedArray<ConditionFormula> relationConstraints1 = and(positiveInput.relationConstraints()).replaceAtomicTerm(prefixerPosive).subformulas();
		FixedArray<ConditionFormula> relationConstraints2 = and(negativeInput.relationConstraints()).replaceAtomicTerm(prefixerNegative).replaceAtomicTerm(removedAttributeReplacer).subformulas();
		
		final ArrayList<Variable<?>> streamBoundIdentifierVariables2 = new ArrayList<Variable<?>>();
		final ArrayList<Variable<?>> streamBoundTimestampVariables2 = new ArrayList<Variable<?>>();
		
		for(Variable<?> streamBoundVariable2: and(relationConstraints2).Variables()){
			assert streamBoundVariable2 instanceof BasicTypeVariable;
			if(streamBoundVariable2.type == BasicTypeID.IDENTIFIER.type){
				streamBoundIdentifierVariables2.add(streamBoundVariable2);
			}
			if(streamBoundVariable2.type == BasicTypeID.TIMESTAMP.type){
				streamBoundTimestampVariables2.add(streamBoundVariable2);
			}
		}
		
		SimpleBoundTemporalFormula[] dontCareFormulas1 = new SimpleBoundTemporalFormula[
		                      streamBoundIdentifierVariables2.size() +
		                      streamBoundTimestampVariables2.size()
		                      ];
		
		BinaryComparissonFormula[] doCareFormulas2 = new BinaryComparissonFormula[
		                                              		                      identifierAttributes.size() +
		                                              		                      timestampAttributes.size()
		                                              		                      ];
		
		int dontCareIndex = 0;
		for(Variable<?> sbVariable2: streamBoundIdentifierVariables2){
			dontCareFormulas1[dontCareIndex] = new SimpleBoundTemporalFormula(IDENTIFIER, sbVariable2.name, false);
			++dontCareIndex;
		}
		for(Variable<?> sbVariable2: streamBoundTimestampVariables2){
			dontCareFormulas1[dontCareIndex] = new SimpleBoundTemporalFormula(TIMESTAMP, sbVariable2.name, false);
			++dontCareIndex;
		}
		
		
		int doCareIndex = 0;
		
		for(Attribute<?> attribute: identifierAttributes){
			doCareFormulas2[doCareIndex] = new BinaryComparissonFormula(BinaryComparissonOperatorID.EQUAL, removedAttributeReplacer.adapt(attribute), attribute);
			++doCareIndex;
		}
		for(Attribute<?> attribute: timestampAttributes){
			doCareFormulas2[doCareIndex] = new BinaryComparissonFormula(BinaryComparissonOperatorID.EQUAL, removedAttributeReplacer.adapt(attribute), attribute);
			++doCareIndex;
		}
		
		@SuppressWarnings("unchecked")
		FixedArray<ConditionFormula> tupleConstraints = FixedArray.concat(
				and(positiveInput.tupleConstraints()).replaceAtomicTerm(prefixerPosive).subformulas(), 
				new FixedArray<ConditionFormula>(or(
						and(dontCareFormulas1), 
						and(and(doCareFormulas2), and(negativeInput.tupleConstraints()).replaceAtomicTerm(prefixerNegative).replaceAtomicTerm(removedAttributeReplacer))						
				)));
		
		@SuppressWarnings("unchecked")
		FixedArray<ConditionFormula> relationConstraints = FixedArray.concat(relationConstraints1, relationConstraints2);
		
		try {
			return TemporalStreamSchema.create(positiveInput.ID, positiveInput.attributes, tupleConstraints, relationConstraints, false);
		} catch (TSA_SchemaException e) {
			throw new TSA_Exception(this.origin, e);
		}
	}
	
	@Override
	public boolean isMonotonic(final Collection<TemporalStreamSchema> inputs) {
		assert inputs.size() == 2;
		Iterator<TemporalStreamSchema> inputsIterator = inputs.iterator();
		inputsIterator.next();//skip first input
		if(inputsIterator.next().isStatic){
			return true;
		}
		else{
			return false;
		}
	}
	
	@Override
	public void checkInputs(final Collection<TemporalStreamSchema> inputs) throws TSA_Exception {
		super.checkInputs(inputs);
		assert inputs.size() == 2;
		Iterator<TemporalStreamSchema> inputIterator = inputs.iterator();
		TemporalStreamSchema positiveInput = inputIterator.next();
		TemporalStreamSchema negativeInput = inputIterator.next();
		
		if(FixedArrayNavigableSet.symmetricDifference(positiveInput.basicAttributes.keys(), negativeInput.basicAttributes.keys()).size() > 0){
			throw new TSA_Exception(this.origin, "Attributes of positive and negative input do not match.");
		}
		
		int numberOfAttributes = positiveInput.basicAttributes.size();
		for(int attribute = 0; attribute < numberOfAttributes; ++attribute){
			if(!positiveInput.basicAttributes.values().get(attribute).equals(negativeInput.basicAttributes.values().get(attribute))){
				String attributeName = positiveInput.basicAttributes.keys().get(attribute);
				assert attributeName.equals(negativeInput.basicAttributes.keys().get(attribute));
				throw new TSA_Exception(this.origin, "The types for attribute \"" + attributeName + "\" do not match.");
			}
		}
		
		//FIXME check temporal relations in first input (?) -> check?
	}
	
	@Override
	public void check(final Collection<TemporalStreamSchema> inputs) {
		//FIXME implement
	}
	
	@Override
	public void writeTo(final StringBuilder sb, final int indentDepthFirstLine, final int indentDepth) {
		indentTab(sb, indentDepthFirstLine);
		sb.append(this.ID.toString());
	}
}
