package de.lmu.ifi.pms.cep.tsa.schema;

import org.simpleframework.xml.Element;
import org.simpleframework.xml.ElementMap;

import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.ConjunctiveConditionFormula;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_SchemaException;

import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.map.Map;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;

public class StateStreamSchema extends TemporalStreamSchema {
	
	public StateStreamSchema(final FixedArrayNavigableMap<String, ? extends Type> attributes, final FixedArray<ConditionFormula> tupleConstraints, final FixedArray<ConditionFormula> relationConstraints, final boolean addTypeTupleConstraints) throws TSA_SchemaException {
		super(TemporalStreamSchemaTypeID.STATE, attributes, tupleConstraints, relationConstraints, addTypeTupleConstraints);
	}

	public StateStreamSchema(final Map<String, ? extends Type, ?> attributes, final FixedArray<ConditionFormula> tupleConstraints, final FixedArray<ConditionFormula> relationConstraints) throws TSA_SchemaException {
		this(new FixedArrayNavigableMap<String, Type>(attributes), tupleConstraints, relationConstraints, true);
	}
	
	protected StateStreamSchema(
			  @ElementMap(name="attributes") final java.util.HashMap<String, ? extends Type> attributes
			, @Element(name="tuple-constraints") final ConjunctiveConditionFormula tupleConstraints
			, @Element(name="relation-constraints") final ConjunctiveConditionFormula relationConstraints
		) throws TSA_SchemaException {
		this(new FixedArrayNavigableMap<String, Type>(attributes), tupleConstraints.subformulas() , relationConstraints.subformulas(), false);
	}
	
	protected StateStreamSchema(final Map<String, ? extends Type, ?> attributes, final FixedArray<ConditionFormula> tupleConstraints, final FixedArray<ConditionFormula> relationConstraints, final boolean addTypeTupleConstraints) throws TSA_SchemaException {
		this(new FixedArrayNavigableMap<String, Type>(attributes), tupleConstraints, relationConstraints, addTypeTupleConstraints);
	}

	@Override
	public StateStreamSchema newStreamSchema(final FixedArrayNavigableMap<String, ? extends Type> attributes, final FixedArray<ConditionFormula> tupleConstraints, final FixedArray<ConditionFormula> relationConstraints, final boolean addTypeTupleConstraints) throws TSA_SchemaException{
		return new StateStreamSchema(attributes, tupleConstraints, relationConstraints, addTypeTupleConstraints);
	}
	
}
