package de.lmu.ifi.pms.cep.tsa.schema;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;
import static de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula.*;
import static de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormulaID.TOP;
import static de.lmu.ifi.pms.cep.tsa.conditions.Term.*;
import static de.lmu.ifi.pms.java.util.filter.AbstractFilter.notFilter;

import org.simpleframework.xml.Element;
import org.simpleframework.xml.ElementMap;
import org.simpleframework.xml.Root;

import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.conditions.AtomicConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.AtomicStreamBoundFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.AtomicTerm;
import de.lmu.ifi.pms.cep.tsa.conditions.AtomicTermID;
import de.lmu.ifi.pms.cep.tsa.conditions.Attribute;
import de.lmu.ifi.pms.cep.tsa.conditions.BasicTypeAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonOperatorID;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormulaID;
import de.lmu.ifi.pms.cep.tsa.conditions.ConjunctiveConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.DataTypeAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.DataTypeRecord;
import de.lmu.ifi.pms.cep.tsa.conditions.DefaultDefinition;
import de.lmu.ifi.pms.cep.tsa.conditions.DisjunctiveConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.NamedTypeAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.NamedTypeRecord;
import de.lmu.ifi.pms.cep.tsa.conditions.Record;
import de.lmu.ifi.pms.cep.tsa.conditions.StreamBoundConstraint;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.conditions.TypedAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.UntypedAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.adaptor.PrefixAttributes;
import de.lmu.ifi.pms.cep.tsa.conditions.adaptor.PrefixVariables;
import de.lmu.ifi.pms.cep.tsa.conditions.adaptor.RenameAttributes;
import de.lmu.ifi.pms.cep.tsa.conditions.filter.AttributeFilter;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_SchemaException;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_TypeException;
import de.lmu.ifi.pms.cep.tsa.time.AtomicStreamBoundFormulaFilter;
import de.lmu.ifi.pms.cep.tsa.time.StreamBoundConstraintFilter;

import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.adaption.CollectionAdapter;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.array.Arrays;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.filter.AbstractFilter;
import de.lmu.ifi.pms.java.util.map.Map;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;

@Root(name="schema")
public class TemporalStreamSchema{

	public final TemporalStreamSchemaTypeID ID;
	
	public final FixedArrayNavigableMap<String, BasicType> basicAttributes;
	
	public final FixedArrayNavigableMap<String, Type> attributes;
	
	protected final CompositeType type;
	
	public final Adapter<AtomicTerm<?>, AtomicTerm<?>> attributeTyper;
	
	protected Adapter<AtomicTerm<?>, AtomicTerm<?>> attributeTyper(){
		return new Adapter<AtomicTerm<?>, AtomicTerm<?>>() {
			
			@Override
			public AtomicTerm<?> adapt(final AtomicTerm<?> source) {
				if(source.ID == AtomicTermID.ATTRIBUTE & !source.isTyped){
					UntypedAttribute attribute = (UntypedAttribute) source;
					try {
						return TemporalStreamSchema.this.getAttribute(attribute.name);
					} catch (TSA_TypeException e) {}
				}
				return source;
			}
			
		};
	}
	
	public final boolean isStatic;
	
	@ElementMap(name="attributes", entry="attribute", key="name", value="type")
	java.util.HashMap<String, Type> getAttributes(){
		return new java.util.HashMap<String, Type>(this.attributes.javaMap());
	}
	
	final ConjunctiveConditionFormula tupleConstraints;
	
	@Element(name="tuple-constraints")
	ConjunctiveConditionFormula getTupleConstraintFormula(){
		return this.tupleConstraints;
	}
	
	final ConjunctiveConditionFormula relationConstraints;
	
	@Element(name="relation-constraints")
	ConjunctiveConditionFormula getRelationConstraintFormula(){
		return this.relationConstraints;
	}
	
	public final FixedArrayNavigableMap<String, Term<? extends Type>> implicitAttributeSpecifications;

	public static final AbstractFilter<ConditionFormula> streamBoundFilter = new AtomicStreamBoundFormulaFilter();

	public static final AbstractFilter<ConditionFormula> streamBoundConstraintFilter = new StreamBoundConstraintFilter();
	
/*	@ElementMap(entry="implicit-attribute-specifications", inline=true, key="min-time", attribute=true, required=false)
	final java.util.HashMap<String, Term<? extends Type>> getImplicitAttributeSpecifications(){
		return new java.util.HashMap<String, Term<? extends Type>>(this.implicitAttributeSpecifications.javaMap());
	}*/
	
	protected TemporalStreamSchema(final TemporalStreamSchemaTypeID ID, final Map<String, ? extends Type, ?> attributes, final FixedArray<ConditionFormula> tupleConstraints, final FixedArray<ConditionFormula> relationConstraints) throws TSA_SchemaException {
		this(ID, attributes, tupleConstraints, relationConstraints, true);
	}
	
	protected TemporalStreamSchema(final Map<String, ? extends Type, ?> attributes, final FixedArray<ConditionFormula> tupleConstraints, final FixedArray<ConditionFormula> relationConstraints, final boolean addTypeTupleConstraints) throws TSA_SchemaException {
		this(TemporalStreamSchemaTypeID.GENERAL, attributes, tupleConstraints, relationConstraints);
	}
	
	protected TemporalStreamSchema(final TemporalStreamSchemaTypeID ID, final Map<String, ? extends Type, ?> attributes, final FixedArray<ConditionFormula> tupleConstraints, final FixedArray<ConditionFormula> relationConstraints, final boolean addTypeTupleConstraints) throws TSA_SchemaException {
		this(ID, new FixedArrayNavigableMap<String, Type>(attributes), tupleConstraints, relationConstraints, addTypeTupleConstraints);
	}
	
	public TemporalStreamSchema(final Map<String, ? extends Type, ?> attributes, final Collection<ConditionFormula> tupleConstraints, final Collection<ConditionFormula> relationConstraints) throws TSA_SchemaException {
		this(TemporalStreamSchemaTypeID.GENERAL, attributes, new FixedArray<ConditionFormula>(tupleConstraints), new FixedArray<ConditionFormula>(relationConstraints));
	}
	
	public TemporalStreamSchema(final FixedArrayNavigableMap<String, ? extends Type> attributes, final FixedArray<ConditionFormula> tupleConstraints, final FixedArray<ConditionFormula> relationConstraints) throws TSA_SchemaException {
		this(attributes, tupleConstraints, relationConstraints, true);
	}
	
	public TemporalStreamSchema(final FixedArrayNavigableMap<String, ? extends Type> attributes, final FixedArray<ConditionFormula> tupleConstraints, final FixedArray<ConditionFormula> relationConstraints, final boolean addTypeTupleConstraints) throws TSA_SchemaException {
		this(TemporalStreamSchemaTypeID.GENERAL, attributes, tupleConstraints, relationConstraints, addTypeTupleConstraints);
	}
	
	public static final Adapter<AtomicTerm<?>, Term<?>> complexAttributeEliminator = new Adapter<AtomicTerm<?>, Term<?>>() {
		
		@Override
		public Term<?> adapt(final AtomicTerm<?> source) {
			if(source instanceof DataTypeAttribute){
				DataTypeAttribute attribute = (DataTypeAttribute) source;			
				return this.record(attribute.name, attribute.type);
			}
			if(source instanceof NamedTypeAttribute<?>){
				NamedTypeAttribute<?> attribute = (NamedTypeAttribute<?>) source;
				if(attribute.type.type.ID == TypeID.COMPOSITE){
					return this.record(attribute.name, attribute.type);	
				}
			}
			return source;
		}
		
		@SuppressWarnings("unchecked")
		public Term<?> record(final String name, final Type type){
			
			NamelessType nameLessType = (type.ID == TypeID.NAMED ? ((NamedType<?>)type).type : (NamelessType) type);
			if(nameLessType.ID == TypeID.BASIC){
				return attr(name, type);
			}
			else if(nameLessType.ID == TypeID.COMPOSITE){
				CompositeType compositeType = (CompositeType) nameLessType;
				Term<?>[] arguments = new Term<?>[compositeType.attributes.size()];
				for(int arg = 0; arg < arguments.length; ++arg){
					arguments[arg] = this.record(TSA_SyntaxPatterns.prefixAttribute(name, compositeType.attributes.keys().get(arg)), compositeType.attributes.values().get(arg));
				}		
				if(type == compositeType){
					return new DataTypeRecord(compositeType, arguments);
				}
				else{
					assert type instanceof NamedType<?>;
					return new NamedTypeRecord(((NamedType<CompositeType>)type), arguments);
				}
			}
			else{
				throw new Error("Unexpected case.");
			}
		}
		
	};
	
	public static final Adapter<AtomicConditionFormula, ConditionFormula> normalizeEqual = new Adapter<AtomicConditionFormula, ConditionFormula>() {
		
		@Override
		public ConditionFormula adapt(final AtomicConditionFormula source) {
			if(source instanceof BinaryComparissonFormula){
				BinaryComparissonFormula bcf = (BinaryComparissonFormula) source;
				if(bcf.operator == BinaryComparissonOperatorID.EQUAL | bcf.operator == BinaryComparissonOperatorID.NOT_EQUAL){
					//FIXME activate if(bcf.firstTerm().type.equals(bcf.secondTerm().type)){
						if(bcf.firstTerm() instanceof Record<?> & bcf.secondTerm() instanceof Record<?>){
							Record<?> record1 = (Record<?>) bcf.firstTerm();
							Record<?> record2 = (Record<?>) bcf.secondTerm();
							assert FixedArrayNavigableSet.symmetricDifference(record1.arguments.keys(), record2.arguments.keys()).size() == 0;
							
							ConditionFormula[] componentFormulas = new ConditionFormula[record1.arguments.size()];
							for(int argument = 0; argument < componentFormulas.length; ++argument){
								componentFormulas[argument] = this.adapt(new BinaryComparissonFormula(bcf.operator, record1.arg(argument), record2.arg(argument)));
							}
							if(bcf.operator == BinaryComparissonOperatorID.EQUAL){
								return and(componentFormulas).flaten();
							}
							else{assert bcf.operator == BinaryComparissonOperatorID.NOT_EQUAL;
								return or(componentFormulas).flaten();
							}
						}
					/*}
					else{
						return Bottom;
					}*/
					}
			}
			return source;
		}
		
	};
	
	protected FixedArrayNavigableMap<String, Term<? extends Type>> implicitAttributeSpecifications() throws TSA_SchemaException{
		ConditionFormula implicitSpecificationsFormula = and(this.relationConstraints()).focus(DefaultDefinition.implicitAttributeValuesSpecifications).flaten();
		if(implicitSpecificationsFormula.ID == ConditionFormulaID.TOP){
			return new FixedArrayNavigableMap<String, Term<? extends Type>>();
		}
		if(implicitSpecificationsFormula.ID == ConditionFormulaID.BOTTOM){
			throw new TSA_SchemaException("Relation constraints are inconsistent.");
		}
		if(implicitSpecificationsFormula.ID == ConditionFormulaID.ATOMIC){
			implicitSpecificationsFormula = and(implicitSpecificationsFormula);
		}
		if(implicitSpecificationsFormula.ID == ConditionFormulaID.CONJUNCTIVE){
			ConjunctiveConditionFormula conjunctiveImplicitSpecificationsFormula = (ConjunctiveConditionFormula) implicitSpecificationsFormula;
			ArrayList<String> implicitAttributeNames = new ArrayList<String>(conjunctiveImplicitSpecificationsFormula.subformulas().size());
			ArrayList<Term<? extends Type>> definitions = new ArrayList<Term<? extends Type>>(conjunctiveImplicitSpecificationsFormula.subformulas().size());
			
			for (ConditionFormula subFormula: conjunctiveImplicitSpecificationsFormula.subformulas()) {
				DefaultDefinition definition = (DefaultDefinition) subFormula;
				if(this.hasAttribute(definition.attribute.name)){
					if(definition.attribute.isTyped){
						if(!definition.attribute.type.equals(this.typeOf(definition.attribute.name))){
							throw new TSA_SchemaException("The type of the attribute \"" + definition.attribute.name + "\" in its implicit definition is incompatible with the type specified in the schema.");
						}
					}
					if(definition.definition.isTyped){
						if(definition.definition.type.equals(this.typeOf(definition.attribute.name))){
							implicitAttributeNames.add(definition.attribute.name);
							definitions.add((Term<? extends Type>) definition.definition);
						}
						else{
							throw new TSA_SchemaException("The type of the implicit definition of attribute \"" + definition.attribute.name + "\" is incompatible with the type specified in the schema.");
						}
					}
					else{
						throw new TSA_SchemaException("The implicit definition of attribute \"" + definition.attribute.name + "\" must be typed.");
					}
				}
				else{
					throw new TSA_SchemaException("Attribute \"" + definition.attribute.name + "\" is not contained in the schema.");
				}
			}
			
			return new FixedArrayNavigableMap<String, Term<? extends Type>>(implicitAttributeNames, definitions);
		}
		else{
			throw new TSA_SchemaException("Invalid implicit attribute definitions");
		}
	}
	
	protected TemporalStreamSchema(
			  @ElementMap(name="attributes") final java.util.HashMap<String, ? extends Type> attributes
			, @Element(name="tuple-constraints") final ConjunctiveConditionFormula tupleConstraints
			, @Element(name="relation-constraints") final ConjunctiveConditionFormula relationConstraints
		) throws TSA_SchemaException {
		this(new FixedArrayNavigableMap<String, Type>(attributes), tupleConstraints.subformulas() , relationConstraints.subformulas(), false);
	}

	@SuppressWarnings("unchecked")
	protected TemporalStreamSchema(final TemporalStreamSchemaTypeID ID, final FixedArrayNavigableMap<String, ? extends Type> attributes, FixedArray<ConditionFormula> tupleConstraints, final FixedArray<ConditionFormula> relationConstraints, final boolean addTypeTupleConstraints) throws TSA_SchemaException {		
		this.ID = ID;
		this.checkAttributeNames(attributes.keys());
		
		this.type = type(attributes);
		this.attributeTyper = this.attributeTyper();
		this.attributes = this.type.attributes;
		
		this.basicAttributes = this.type.basicAttributes();
		
		this.checkTupleConstraints(tupleConstraints);
		this.checkRelationConstraints(relationConstraints);
		
		if(addTypeTupleConstraints){
			tupleConstraints = FixedArray.concat(tupleConstraints, this.type.tupleConstraints());
		}
		ConditionFormula normalizedTupleConstraints = and(tupleConstraints).replaceAtomicTerm(TemporalStreamSchema.complexAttributeEliminator).replaceAtomicTerm(this.attributeTyper).replaceAtomicFormula(TemporalStreamSchema.normalizeEqual).normalize();
		this.tupleConstraints = (normalizedTupleConstraints.ID == ConditionFormulaID.CONJUNCTIVE ? (ConjunctiveConditionFormula) normalizedTupleConstraints : and(normalizedTupleConstraints));
		ConditionFormula normalizedRelationConstraints = and(relationConstraints).replaceAtomicTerm(TemporalStreamSchema.complexAttributeEliminator).replaceAtomicTerm(this.attributeTyper).replaceAtomicFormula(TemporalStreamSchema.normalizeEqual).normalize();
		this.relationConstraints = (normalizedRelationConstraints.ID == ConditionFormulaID.CONJUNCTIVE ? (ConjunctiveConditionFormula) normalizedRelationConstraints : and(normalizedRelationConstraints));
		
		this.implicitAttributeSpecifications = this.implicitAttributeSpecifications();
		
		this.isStatic = and(this.relationConstraints()).focus(AtomicStreamBoundFormulaFilter.instance).flaten().ID == TOP & and(this.relationConstraints()).focus(StreamBoundConstraintFilter.instance).flaten().ID == TOP;

	}
	
	protected static CompositeType type(final FixedArrayNavigableMap<String, ? extends Type> attributes) throws TSA_TypeException {
		CompositeType type = CompositeType.get(attributes);//create (invalid) empty composite Type
		return type;
	}
	
	/*
	protected static FixedArrayNavigableMap<String, BasicType> basicAttributes(final FixedArrayNavigableMap<String, ? extends Type> attributes){
		ArrayList<FixedArrayNavigableMap<String, BasicType>> basicAttributes = new ArrayList<FixedArrayNavigableMap<String,BasicType>>();
		final int numberOfAttributes = attributes.size();
		for(int attribute = 0; attribute < numberOfAttributes; ++attribute){
			basicAttributes.add(attributes.values().get(attribute).basicAttributes(attributes.keys().get(attribute)));
		}
		
		FixedArrayNavigableSet<String> attributeNames = FixedArrayNavigableSet.<String>merge(new CollectionAdapter<FixedArrayNavigableMap<String, BasicType>, FixedArrayNavigableSet<String>>(basicAttributes){
						@Override
						public FixedArrayNavigableSet<String> adapt(final FixedArrayNavigableMap<String, BasicType> source) {
							return source.keys();
						}
		});
		FixedArray<BasicType> attributeTypes = FixedArray.concat(new CollectionAdapter<FixedArrayNavigableMap<String, BasicType>, FixedArray<BasicType>>(basicAttributes){
						@Override
						public FixedArray<BasicType> adapt(final FixedArrayNavigableMap<String, BasicType> source) {
							return source.values();
						}
		});
		
		return new FixedArrayNavigableMap<String, BasicType>(attributeNames, attributeTypes);
	}*/
	
	void checkAttributeNames(final FixedArrayNavigableSet<String> names) throws TSA_SchemaException{
		final int numberOfNames = names.size();		
		String last = "...";
		assert !TSA_SyntaxPatterns.isValidAttributeName(last);
		for(int index = 0; index < numberOfNames; ++index){
			String name = names.get(index);
			if(!TSA_SyntaxPatterns.isValidAttributeName(name)){
				throw new TSA_SchemaException("Illegal attribute name: " + name );
				//TODO better exception
			}
			if(TSA_SyntaxPatterns.isSuperAttribute(last, name)){
				throw new TSA_SchemaException("Illegal attribute names: " + last + " is prefix of " + name );
			}
			last = name;
		}
	}
	
	void checkTupleConstraints(final FixedArray<ConditionFormula> tupleConstraints) throws TSA_SchemaException{
		assert this.basicAttributes != null;
		//may contain variables
		for(ConditionFormula constraint: tupleConstraints){
			for(Term<?> term: constraint.collect(AttributeFilter.instance)){
				Attribute<?> attribute = (Attribute<?>) term;
				//check that only available attributes are used
				if(!this.hasAttribute(attribute.name)){
					throw new TSA_SchemaException("Constraint " + constraint + " uses unavailable attribute " + attribute.name); 
				}				
				//FIXME check that attributes have same/compatible type
			}
			//check that types of the same variable are compatible
		}
	}
	
	void checkRelationConstraints(final FixedArray<ConditionFormula> relationConstraints) throws TSA_SchemaException{
		assert this.basicAttributes != null;
		//may contain variables
		for(ConditionFormula constraint: relationConstraints){
			for(Term<?> term: constraint.collect(AttributeFilter.instance)){
				Attribute<?> attribute = (Attribute<?>) term;
				//check that only available attributes are used
				if(!this.hasAttribute(attribute.name)){
					throw new TSA_SchemaException("Constraint " + constraint + " uses unavailable attribute " + attribute.name); 
				}				
				//FIXME check that attributes have same/compatible type
			}
			//check that types of the same variable are compatible
		}
	}
	
	public boolean hasAttribute(final String name){
		try{
			this.typeOf(name);
			return true;
		}
		catch (TSA_TypeException e) {
			return false;
		}
	}
	
	public boolean hasAttributes(final Collection<String> names){
		for(String name: names){
			if(!this.hasAttribute(name)){
				return false;
			}
		}
		return true;
	}
	
	public TypedAttribute<?> getAttribute(final String name)  throws TSA_TypeException{
		return TypedAttribute.create(name, this.typeOf(name));
	}
	
	public Type typeOf(final String name)  throws TSA_TypeException{
		return this.type.typeOf(name);
	}
	
	public FixedArrayNavigableMap<String, BasicType> basicAttributes(){
		return this.basicAttributes;
	}
	
	public FixedArray<ConditionFormula> tupleConstraints(){
		return this.tupleConstraints.subformulas();
	}
	
	public FixedArray<ConditionFormula> relationConstraints(){
		return this.relationConstraints.subformulas();		
	}
	
	public ConditionFormula streamBoundFormula() throws TSA_SchemaException{
		ConditionFormula streamBoundFormula = and(this.relationConstraints).focus(streamBoundFilter).flaten();
		checkExpressionStreamBoundFormula(streamBoundFormula);
		return streamBoundFormula;
	}
	
	public ConditionFormula streamBoundFormulaFromConstraints(final String buffer_full_name) throws TSA_SchemaException{
		return TemporalStreamSchema.streamBoundFormulaFromConstraints(buffer_full_name, this.relationConstraints());
	}
	
	public TemporalStreamSchema convertStreamBoundConstraintsToStreamBoundFormulas(final String buffer_full_name) throws TSA_SchemaException{
		@SuppressWarnings("unchecked")
		FixedArray<ConditionFormula> relationConstraints = FixedArray.concat(
				and(this.relationConstraints()).focus(notFilter(TemporalStreamSchema.streamBoundConstraintFilter)).subformulas(),
				new FixedArray<ConditionFormula>(this.streamBoundFormulaFromConstraints(buffer_full_name))
			);
		return this.newStreamSchema(this.attributes, this.tupleConstraints(), relationConstraints, false);
	}
	
	public static void checkRetainStatic(final FixedArray<String> attributes)  throws TSA_SchemaException{
		for(String attribute: attributes){
			if(!(attribute.equals("") | TSA_SyntaxPatterns.isValidAttributeName(attribute))){
				throw new TSA_SchemaException("Invalid attribute name: " + attribute);
			}
		}
	}
	
	public static FixedArrayNavigableSet<String> normalizeforRetain(final FixedArray<String> attributes){
		if(attributes.size() > 0){
			String[] sortedAttributes = new String[attributes.size()];
			attributes.copyTo(sortedAttributes, 0);
			if(!(attributes instanceof FixedArrayNavigableSet)){
				java.util.Arrays.sort(sortedAttributes);
			}
			
			int readIndex = 1;
			int writeIndex = 1;
			String last = sortedAttributes[0];
			while(readIndex < sortedAttributes.length){
				//prefix is lexicographically smaller than the full String
				String current = sortedAttributes[readIndex];
				if(!TSA_SyntaxPatterns.isSuperAttribute(last, current)){
					sortedAttributes[writeIndex] = current;
					++writeIndex;
					last = current;
				}				
				++readIndex;
			}
			
			return new FixedArrayNavigableSet<String>(false, false, java.util.Arrays.copyOf(sortedAttributes, writeIndex));
		}
		else{
			return new FixedArrayNavigableSet<String>();
		}
	}
	
	public static FixedArrayNavigableSet<String> simpleRetain(final FixedArrayNavigableSet<String> attributes, FixedArrayNavigableSet<String> retainedAttributes){
		retainedAttributes = normalizeforRetain(retainedAttributes);
		
		//if a complex attribute is kept, then a prefix of its name is in the projection attributes and thus all its sub attributes are kept, too
		ArrayList<String> remainingAttributeNames = new ArrayList<String>();
		
		int attributeIndex = 0;
		final int mappingSize = retainedAttributes.size();
		//Note: both the keys of mapping and attributes are sorted lexicographically			
		for(int mappingIndex = 0; mappingIndex < mappingSize; ++mappingIndex){
			String currentKey = retainedAttributes.get(mappingIndex);
			while(attributeIndex < attributes.size() && currentKey.compareTo(attributes.get(attributeIndex)) >  0){
				++attributeIndex;
			}
			while(attributeIndex < attributes.size() && TSA_SyntaxPatterns.isSuperAttribute(currentKey, attributes.get(attributeIndex))){
				remainingAttributeNames.add(attributes.get(attributeIndex));
				++attributeIndex;
			}				
		}
		return new FixedArrayNavigableSet<String>(false, false, remainingAttributeNames.toArray(new String[remainingAttributeNames.size()]));
	}
	
	public TemporalStreamSchema retain(FixedArrayNavigableSet<String> attributes) throws TSA_SchemaException{
		checkRetainStatic(attributes);
		attributes = normalizeforRetain(attributes);

		if(attributes.contains("")){
			assert attributes.size() == 1;
			return this;
		}
		else{
			//check existence of attributes
			for(String attribute: attributes){
				if(!this.hasAttribute(attribute)){
					throw new TSA_SchemaException("There is no attribute \"" + attribute + "\" in this schema.");
				}
			}
			
			final CompositeType newType = this.type.retain(attributes);
			
			//prefix variables to ensure free names
			final Adapter<AtomicTerm<?>, AtomicTerm<?>> prefixer = new PrefixVariables("$");
			//convert removed attributes to variables
			final FixedArrayNavigableSet<String> removedAttributeNamesIndex = FixedArrayNavigableSet.difference(this.basicAttributes.keys(), newType.basicAttributes().keys());
			final Adapter<AtomicTerm<?>, AtomicTerm<?>> removedAttributeReplacer = new Adapter<AtomicTerm<?>, AtomicTerm<?>>(){
				@Override
				public AtomicTerm<?> adapt(final AtomicTerm<?> source) {
					if(source instanceof Attribute){
						Attribute<?> attribute = (Attribute<?>) source;
						if(removedAttributeNamesIndex.contains(attribute.name)){
							if(attribute instanceof TypedAttribute<?>){
								return var("$" + attribute.name, ((TypedAttribute<?>)attribute).type);
							}
							else{
								return var("$" + attribute.name);
							}
						}
					}
					return source;
				}
			};
			
			AbstractFilter<AtomicConditionFormula> removedAttributesDefaultConstraintFilter = new AbstractFilter<AtomicConditionFormula>() {
				
				@Override
				public boolean accepts(final AtomicConditionFormula element) {
					if(element instanceof DefaultDefinition){
						DefaultDefinition defaultDefinition = (DefaultDefinition) element;
						if(!newType.hasAttribute(defaultDefinition.attribute.name)){
							return true;
						}
						else{
							assert newType.hasAttribute(defaultDefinition.attribute.name);
							try {
								if(!newType.typeOf(defaultDefinition.attribute.name).equals(defaultDefinition.attribute.type)){
									return true;
								}
							} catch (TSA_TypeException e) {
								throw new Error(e);
							}
						}
					}
					return false;
				}
			};
			
			FixedArray<ConditionFormula> tupleConstraints = and(this.tupleConstraints()).replaceAtomicTerm(prefixer).replaceAtomicTerm(removedAttributeReplacer).subformulas();
			
			FixedArray<ConditionFormula> relationConstraints = and(this.relationConstraints()).focus(notFilter(removedAttributesDefaultConstraintFilter)).replaceAtomicTerm(prefixer).replaceAtomicTerm(removedAttributeReplacer).subformulas();
			 
			return this.newStreamSchema(newType.attributes, tupleConstraints, relationConstraints, false);
		}
		
	}
	
	public static void checkNameMappingStatic(final FixedArrayNavigableMap<String, String> mapping) throws TSA_SchemaException{
		if(!mapping.isEmpty()){
			//check keys
			if(mapping.keys().first().equals("")){
				if(mapping.size() > 1){
					throw new TSA_SchemaException("Mapping contains rename for \"\". No further renames are allowed.");
				}
			}
			else{
				String lastKey = "...";
				assert !TSA_SyntaxPatterns.isValidAttributeName(lastKey);
				for(String key: mapping.keys()){
					if(!TSA_SyntaxPatterns.isValidAttributeName(key)){
						throw new TSA_SchemaException("Key is not a valid attribute name: " + key);
					}
					if(TSA_SyntaxPatterns.isSuperAttribute(lastKey, key)){
						throw new TSA_SchemaException("Mapping contains rename for \"" + lastKey + "\". It may not contain a further mapping for \"" + key + "\".");
					}
					else{
						lastKey = key;
					}
				}
			}
			//check mappings
			for(String rename: mapping.values()){
				if(!(rename.equals("") | TSA_SyntaxPatterns.isValidAttributeName(rename))){
					throw new TSA_SchemaException("Value is not a valid mapping: " + rename);
				}
			}
		}
	}
	
	public static FixedArrayNavigableMap<String, String> simpleMapping(final FixedArrayNavigableSet<String> attributeNames, final FixedArrayNavigableMap<String, String> mapping) throws TSA_SchemaException{
		//mapping: oldPrefix -> newPrefix
		//simpleMapping: newName -> oldName
		if(mapping.isEmpty()){
			return new FixedArrayNavigableMap<String, String>(attributeNames, attributeNames);
		}
		else{
			checkNameMappingStatic(mapping);
			if(mapping.keys().contains("")){
				assert mapping.size() == 1;
				final String rename = mapping.values().get(0);
				if(rename.equals("")){
					return new FixedArrayNavigableMap<String, String>(attributeNames, attributeNames);
				}
				else{
					String prefix = TSA_SyntaxPatterns.makeAttributeNamePrefix(rename);
					
					final String[] newAttributeNames = new String[attributeNames.size()];
					for(int attribute = 0; attribute < newAttributeNames.length; ++attribute){
						newAttributeNames[attribute] = prefix + attributeNames.get(attribute);
					}
					
					//order remains the same, thus newAttributeNames is ordered
					return new FixedArrayNavigableMap<String, String>(new FixedArrayNavigableSet<String>(false, false, newAttributeNames), attributeNames);
				}
			}
			else{assert !mapping.keys().contains("");
				
				
				int numberOfAttributes = attributeNames.size();
				final String[] newAttributeNames = new String[numberOfAttributes];
				
				int attributeIndex = 0;
				final int mappingSize = mapping.size();
				//Note: both the keys of mapping and attributes are sorted lexicographically			
				for(int mappingIndex = 0; mappingIndex < mappingSize; ++mappingIndex){
					String currentKey = mapping.keys().get(mappingIndex);
					String currentMapping = mapping.values().get(mappingIndex);
					while(attributeIndex < numberOfAttributes && currentKey.compareTo(attributeNames.get(attributeIndex)) >  0){
						String currentAttributeName = attributeNames.get(attributeIndex);
						newAttributeNames[attributeIndex] = currentAttributeName;
						++attributeIndex;
					}
					while(attributeIndex < numberOfAttributes && TSA_SyntaxPatterns.isSuperAttribute(currentKey, attributeNames.get(attributeIndex))){
						String currentAttributeName = attributeNames.get(attributeIndex);
						String replace;
						if(currentMapping.equals("")){
							replace = TSA_SyntaxPatterns.makeAttributeNamePrefix(currentKey);
						}
						else{
							replace = currentKey;
						}							
						String newName = currentAttributeName.replaceFirst(replace, currentMapping);
						newAttributeNames[attributeIndex] = newName;
						++attributeIndex;
					}				
				}
				while(attributeIndex < numberOfAttributes){
					newAttributeNames[attributeIndex] = attributeNames.get(attributeIndex);
					++attributeIndex;
				}
				//Note: In that case the newAttributeNames may not be sorted
				int[] sourcePositionsNew = new int[numberOfAttributes];
				final String[] newSortedAttributeNames = java.util.Arrays.copyOf(newAttributeNames, numberOfAttributes);
				Arrays.sort(newSortedAttributeNames, sourcePositionsNew, true);			
				//check for name clashes
				String lastAttributeName = "...";
				assert !TSA_SyntaxPatterns.isValidAttributeName(lastAttributeName);
				for(int attribute = 0; attribute < numberOfAttributes; ++attribute){
					String attributeName = newSortedAttributeNames[attribute];
					if(TSA_SyntaxPatterns.isSuperAttribute(lastAttributeName, attributeName)){
						throw new TSA_SchemaException("Rename causes name clash: " + lastAttributeName + ", " + attributeName);
					}
					else{
						lastAttributeName = attributeName;
					}
				}
				
				final String[] oldAttributeNames = new String[numberOfAttributes];
				for(int attribute = 0; attribute < numberOfAttributes; ++attribute){
					oldAttributeNames[attribute] = attributeNames.get(sourcePositionsNew[attribute]);
				}

				
				return new FixedArrayNavigableMap<String, String>(new FixedArrayNavigableSet<String>(false, false, newSortedAttributeNames), new FixedArray<String>(false, oldAttributeNames));
			}
		}
	}
	
	public TemporalStreamSchema rename(final FixedArrayNavigableMap<String, String> mapping) throws TSA_SchemaException{
		//oldPrefix -> newPrefix
		if(mapping.isEmpty()){
			return this;
		}
		else{
			checkNameMappingStatic(mapping);
			if(mapping.keys().contains("")){
				assert mapping.size() == 1;
				final String rename = mapping.values().get(0);
				if(rename.equals("")){
					return this;
				}
				else{
					
					CompositeType newType = CompositeType.get(rename, this.type);
					
					FixedArray<ConditionFormula> newStreamTupleConstraints = and(this.tupleConstraints()).replaceAtomicTerm(new PrefixAttributes(rename)).subformulas();
					
					FixedArray<ConditionFormula> newRelationConstraints = and(this.relationConstraints()).replaceAtomicTerm(new PrefixAttributes(rename)).subformulas();
										
					return this.newStreamSchema(newType.attributes, newStreamTupleConstraints, newRelationConstraints, false);
					
					/*
					FixedArrayNavigableMap<String, Type> attributes = this.type.attributes; 
					FixedArrayNavigableSet<String> attributeNames = attributes.keys();
					final String[] newAttributeNames = new String[attributeNames.size()];
					for(int attribute = 0; attribute < newAttributeNames.length; ++attribute){
						newAttributeNames[attribute] = TSA_SyntaxPatterns.prefixAttribute(rename, attributeNames.get(attribute)) ;
					}
					
					return this.newStreamSchema(new FixedArrayNavigableMap<String, Type>(new FixedArrayNavigableSet<String>(false, false, newAttributeNames), attributes.values()), newStreamTupleConstraints, newRelationConstraints, false);
					*/
				}
			}
			else{assert !mapping.keys().contains("");
			
			//TODO pre check for collisions
			
			FixedArrayNavigableSet<String> affectedBasicAttributeNames = simpleRetain(this.basicAttributes.keys(), mapping.keys());
			FixedArrayNavigableSet<String> unchangedBasicAttributeNames = FixedArrayNavigableSet.difference(this.basicAttributes.keys(), affectedBasicAttributeNames);
			
			Type[] renamedAttributeTypes = new Type[mapping.size()];
			for(int attribute = 0; attribute < renamedAttributeTypes.length; ++attribute){
				renamedAttributeTypes[attribute] = this.typeOf(mapping.keys().get(attribute));
			}
			
			FixedArrayNavigableMap<String, Type> renamedAttributes = new FixedArrayNavigableMap<String, Type>(mapping.values(), new FixedArray<Type>(false, renamedAttributeTypes));
			
			CompositeType renamedType;
			if(unchangedBasicAttributeNames.size() > 0){
				renamedType = this.type.retain(unchangedBasicAttributeNames).extend(renamedAttributes);
			}
			else{
				renamedType = CompositeType.get(renamedAttributes);
			}
			
			FixedArray<ConditionFormula> newStreamTupleConstraints = and(this.tupleConstraints()).replaceAtomicTerm(new RenameAttributes(mapping)).subformulas();
			
			FixedArray<ConditionFormula> newRelationConstraints = and(this.relationConstraints()).replaceAtomicTerm(new RenameAttributes(mapping)).subformulas();
												
			return this.newStreamSchema(renamedType.attributes, newStreamTupleConstraints, newRelationConstraints, false);
			}
		}
	}
	
	@SuppressWarnings("unchecked")
	public TemporalStreamSchema restrict(final FixedArray<ConditionFormula> conditions) throws TSA_SchemaException{
		if(conditions.isEmpty()){
			return this;
		}
		else{
			for(Attribute<?> attribute: new ConjunctiveConditionFormula(conditions).Attributes()){
				if(!this.hasAttribute(attribute.name)){
					throw new TSA_SchemaException("Condition contains attribute that is not contained in the schema: " + attribute.name);
				}
				/*FIXME activate
				if(!attribute.type.equals(this.typeOf(attribute.name))){
					throw new TSA_SchemaException("The type of the attribute in the condition is incompatible to the type of the attribute in the schema: " + attribute.name);
				}*/
			}
			if(and(conditions).Variables().size() > 0){
				throw new TSA_SchemaException("Condition may not contain variables: " + and(conditions).VariableNames());
			}			
			
			//@SuppressWarnings("unchecked")
			//FixedArray<ConditionFormula> newTupleConstraints = and(conditions).replaceAtomicTerm(this.complexAttributeEliminator).replaceAtomicFormula(TemporalStreamSchema.this.normalizeEqual).subformulas();
			return this.newStreamSchema(this.type.attributes, FixedArray.concat(this.tupleConstraints(), conditions), this.relationConstraints(), false);
		}
	}
	
	@SuppressWarnings("unchecked")
	public TemporalStreamSchema constraint(final FixedArray<ConditionFormula> conditions) throws TSA_SchemaException{
		if(conditions.isEmpty()){
			return this;
		}
		else{
			for(Attribute<?> attribute: new ConjunctiveConditionFormula(conditions).Attributes()){
				if(!this.hasAttribute(attribute.name)){
					throw new TSA_SchemaException("Condition contains attribute that is not contained in the schema: " + attribute.name);
				}
				/*FIXME activate
				if(!attribute.type.equals(this.typeOf(attribute.name))){
					throw new TSA_SchemaException("The type of the attribute in the condition is incompatible to the type of the attribute in the schema: " + attribute.name);
				}*/
			}
			if(and(conditions).Variables().size() > 0){
				throw new TSA_SchemaException("Condition may not contain variables: " + and(conditions).VariableNames());
			}
			
			//@SuppressWarnings("unchecked")
			//FixedArray<ConditionFormula> newRelationConstraints = and(conditions).replaceAtomicTerm(this.complexAttributeEliminator).subformulas();
			return this.newStreamSchema(this.type.attributes, this.tupleConstraints(), FixedArray.concat(this.relationConstraints(), conditions), false);
		}
	}

	protected static FixedArrayNavigableMap<String, BasicType> basicAttributes(final FixedArrayNavigableMap<String, ? extends Type> attributes) throws TSA_TypeException{
		ArrayList<FixedArrayNavigableMap<String, BasicType>> basicAttributes = new ArrayList<FixedArrayNavigableMap<String,BasicType>>();
		final int numberOfAttributes = attributes.size();
		for(int attribute = 0; attribute < numberOfAttributes; ++attribute){
			basicAttributes.add(attributes.values().get(attribute).basicAttributes(attributes.keys().get(attribute)));
		}
		
		FixedArrayNavigableSet<String> attributeNames = FixedArrayNavigableSet.<String>merge(new CollectionAdapter<FixedArrayNavigableMap<String, BasicType>, FixedArrayNavigableSet<String>>(basicAttributes){
						@Override
						public FixedArrayNavigableSet<String> adapt(final FixedArrayNavigableMap<String, BasicType> source) {
							return source.keys();
						}
		});
		FixedArray<BasicType> attributeTypes = FixedArray.concat(new CollectionAdapter<FixedArrayNavigableMap<String, BasicType>, FixedArray<BasicType>>(basicAttributes){
						@Override
						public FixedArray<BasicType> adapt(final FixedArrayNavigableMap<String, BasicType> source) {
							return source.values();
						}
		});
		
		return new FixedArrayNavigableMap<String, BasicType>(attributeNames, attributeTypes);
	}
	
	public TemporalStreamSchema extend(final FixedArrayNavigableMap<String, ? extends Type> newAttributes) throws TSA_SchemaException{
		if(newAttributes.isEmpty()){
			return this;
		}
		else{
					
			FixedArrayNavigableMap<String, BasicType> newBasicAttributes = basicAttributes(newAttributes);
			//check for name collisions
			for(String basicAtributeName: newBasicAttributes.keys()){
				if(this.hasAttribute(basicAtributeName)){
					throw new TSA_SchemaException("Name clash: Schema already contains attribute with name " + basicAtributeName); 
				}
			}
			
			int numberOfNewAttributes = newAttributes.size();
			
			@SuppressWarnings("unchecked")
			FixedArray<ConditionFormula>[] combinedTupleConstraints = (FixedArray<ConditionFormula>[]) new FixedArray<?>[numberOfNewAttributes + 1];
			for(int attribute = 0; attribute < numberOfNewAttributes; ++attribute){
				combinedTupleConstraints[attribute] = newAttributes.values().get(attribute).tupleConstraints(newAttributes.keys().get(attribute));
			}
			combinedTupleConstraints[numberOfNewAttributes] = this.tupleConstraints();
			FixedArray<ConditionFormula> newTupleConstraints = new FixedArray<ConditionFormula>(new CollectionAdapter<ConditionFormula, ConditionFormula>(FixedArray.concat(combinedTupleConstraints)){
				@Override
				public ConditionFormula adapt(final ConditionFormula source) {
					return source.replaceAtomicTerm(TemporalStreamSchema.complexAttributeEliminator);
				}
			});
			
			FixedArray<ConditionFormula> newRelationConstraints = and(this.relationConstraints()).subformulas();//does not contain complex attributes
			
			//TODO more efficient constructor
			return this.newStreamSchema(this.type.extend(newAttributes).attributes, newTupleConstraints, newRelationConstraints, false);
		}
	}
	
	protected TemporalStreamSchema newStreamSchema(final FixedArrayNavigableMap<String, ? extends Type> attributes, final FixedArray<ConditionFormula> tupleConstraints, final FixedArray<ConditionFormula> relationConstraints, final boolean addTypeTupleConstarints) throws TSA_SchemaException{
		return new TemporalStreamSchema(attributes, tupleConstraints, relationConstraints, addTypeTupleConstarints);
	}
	
	@Override
	public boolean equals(final Object obj) {
		if(obj == this){
			return true;
		}
		else{
			if(obj instanceof TemporalStreamSchema){
				TemporalStreamSchema other = (TemporalStreamSchema) obj;
				if(this.ID == other.ID & this.type.equals(other.type)){
					//FIXME check constraints
					return true;
				}
			}
		}
		return false;	
	}
	
	@Override
	public int hashCode() {
		return super.hashCode();
		//FIXME implement
	}
	
	public void writeTo(final StringBuilder sb, final int indentDepth, final boolean indentFirstLine, final String name, final boolean finish){
		if(indentFirstLine){
			indentTab(sb, indentDepth);
		}
		sb.append(this.ID.toString());
		sb.append("\n");
		indentTab(sb, indentDepth + 1);
		sb.append(name);
		sb.append("{\n");
		for(int attribute = 0; attribute < this.basicAttributes.size(); ++attribute){
			indentTab(sb, indentDepth + 2);
			sb.append(this.basicAttributes.keys().get(attribute));
			sb.append("{");
			sb.append(this.basicAttributes.values().get(attribute));
			sb.append("}");
			sb.append(",\n");
		}
		sb.setLength(sb.length() - ",\n".length());
		sb.append("\n");
		indentTab(sb, indentDepth + 1);
		sb.append("}\n");
		if(this.tupleConstraints().size() > 0){
			indentTab(sb, indentDepth + 1);
			sb.append("where {\n");
			for(int constraint = 0; constraint < this.tupleConstraints().size(); ++constraint){
				this.tupleConstraints().get(constraint).writeTo(sb, indentDepth + 2);
				sb.append(",\n");
			}
			sb.setLength(sb.length() - ",\n".length());
			sb.append("\n");
			indentTab(sb, indentDepth + 1);
			sb.append("}\n");
		}
		if(this.relationConstraints().size() > 0){
			indentTab(sb, indentDepth + 1);
			sb.append("constraints {\n");
			for(int constraint = 0; constraint < this.relationConstraints().size(); ++constraint){
				this.relationConstraints().get(constraint).writeTo(sb, indentDepth + 2);
				sb.append(",\n");
			}
			sb.setLength(sb.length() - ",\n".length());
			sb.append("\n");
			indentTab(sb, indentDepth + 1);
			sb.append("}\n");			
		}
		if(finish){
			indentTab(sb, indentDepth);
			sb.append("END");
		}
	}
	
	public void writeTo(final StringBuilder sb){
		this.writeTo(sb, 0, false, "", true);
	}
	
	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		this.writeTo(sb);
		return sb.toString();
	}
	
	public static boolean isValidBufferStreamBoundFormula(final ConditionFormula streamBoundFormula){
		final ConditionFormula flatenedStreamBoundFormula = streamBoundFormula.flaten();
		if(flatenedStreamBoundFormula.ID == ConditionFormulaID.TOP){
			return true;
		}
		if(flatenedStreamBoundFormula.ID == ConditionFormulaID.ATOMIC){
			assert flatenedStreamBoundFormula instanceof AtomicStreamBoundFormula;
			return true;
		}
		if(flatenedStreamBoundFormula.ID == ConditionFormulaID.DISJUNCTIVE){
			DisjunctiveConditionFormula disjunction = (DisjunctiveConditionFormula) flatenedStreamBoundFormula;
			for(ConditionFormula subFormula: disjunction.subformulas()){
				if(subFormula.ID != ConditionFormulaID.ATOMIC){
					return false;
				}
			}
			return true;
		}
		return false;
	}

	public static void checkBufferStreamBoundFormula(final ConditionFormula streamBoundFormula) throws TSA_SchemaException{
		if(!TemporalStreamSchema.isValidBufferStreamBoundFormula(streamBoundFormula)){
			throw new TSA_SchemaException("Invalid stream bound formula for stream buffer.");
		}
	}

	public static boolean isValidExpressionStreamBoundFormula(final ConditionFormula streamBoundFormula){
		final ConditionFormula flatenedStreamBoundFormula = streamBoundFormula.flaten();
		if(TemporalStreamSchema.isValidBufferStreamBoundFormula(flatenedStreamBoundFormula)){
			return true;
		}
		if(streamBoundFormula.ID == ConditionFormulaID.CONJUNCTIVE){
			ConjunctiveConditionFormula conjunction = (ConjunctiveConditionFormula) flatenedStreamBoundFormula;
			for(ConditionFormula subFormula: conjunction.subformulas()){
				if(!TemporalStreamSchema.isValidBufferStreamBoundFormula(subFormula)){
					return false;
				}
			}
			return true;
		}
		return false;
	}

	public static ConditionFormula streamBoundFormulaFromConstraints(final String buffer_full_name, final FixedArray<ConditionFormula> relationConstraints) throws TSA_SchemaException{
		ConditionFormula streamBoundConstraints = and(relationConstraints).focus(streamBoundConstraintFilter).flaten();
		ConditionFormula streamBoundFormula;
		if(streamBoundConstraints.ID == ConditionFormulaID.TOP){
			streamBoundFormula = Top;
		}
		else{
			Adapter<AtomicConditionFormula, ConditionFormula> streamBoundAdapter = new Adapter<AtomicConditionFormula, ConditionFormula>() {
				
				@Override
				public ConditionFormula adapt(final AtomicConditionFormula source) {
					if (source instanceof StreamBoundConstraint) {
						StreamBoundConstraint streamBoundConstraint = (StreamBoundConstraint) source;
						assert streamBoundConstraint.atomicTerm instanceof BasicTypeAttribute;
						return not(new AtomicStreamBoundFormula(buffer_full_name, (BasicTypeAttribute) streamBoundConstraint.atomicTerm, streamBoundConstraint.isMonotonic));
					}
					return source;
				}
				
			};
			streamBoundFormula =  not(streamBoundConstraints.replaceAtomicFormula(streamBoundAdapter)).flaten();
		}
		
		TemporalStreamSchema.checkBufferStreamBoundFormula(streamBoundFormula);
		return streamBoundFormula;
	}

	public static void checkExpressionStreamBoundFormula(final ConditionFormula streamBoundFormula) throws TSA_SchemaException{
		if(!TemporalStreamSchema.isValidExpressionStreamBoundFormula(streamBoundFormula)){
			throw new TSA_SchemaException("Invalid stream bound formula.");
		}
	}

	public static TemporalStreamSchema create(final TemporalStreamSchemaTypeID ID, final Map<String, ? extends Type, ?> attributes, final FixedArray<ConditionFormula> tupleConstraints, final FixedArray<ConditionFormula> relationConstraints) throws TSA_SchemaException{
		return ID.schema(attributes, tupleConstraints, relationConstraints);
	}
	
	public static TemporalStreamSchema create(final TemporalStreamSchemaTypeID ID, final Map<String, ? extends Type, ?> attributes, final FixedArray<ConditionFormula> tupleConstraints, final FixedArray<ConditionFormula> relationConstraints, final boolean addTypeTupleConstraints) throws TSA_SchemaException{
		return ID.schema(attributes, tupleConstraints, relationConstraints, addTypeTupleConstraints);
	}

}
