package de.lmu.ifi.pms.cep.tsa.schema;

import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_SchemaException;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.map.Map;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayMap;

public enum TemporalStreamSchemaTypeID {
	
	GENERAL{
		
		@Override
		public TemporalStreamSchema schema(final Map<String, ? extends Type, ?> attributes, final FixedArray<ConditionFormula> tupleConstraints, final FixedArray<ConditionFormula> relationConstraints, final boolean addTypeTupleConstraints) throws TSA_SchemaException
		{
			return new TemporalStreamSchema(attributes, tupleConstraints, relationConstraints, addTypeTupleConstraints);
		}
		
	},
	
	EVENT{

		@Override
		public EventStreamSchema schema(final Map<String, ? extends Type, ?> attributes, final FixedArray<ConditionFormula> tupleConstraints, final FixedArray<ConditionFormula> relationConstraints, final boolean addTypeTupleConstraints) throws TSA_SchemaException
		{
			return new EventStreamSchema(attributes, tupleConstraints, relationConstraints, addTypeTupleConstraints);
		}
		
	},
	
	STATE{

		@Override
		public StateStreamSchema schema(final Map<String, ? extends Type, ?> attributes, final FixedArray<ConditionFormula> tupleConstraints, final FixedArray<ConditionFormula> relationConstraints, final boolean addTypeTupleConstraints) throws TSA_SchemaException
		{
			return new StateStreamSchema(attributes, tupleConstraints, relationConstraints, addTypeTupleConstraints);
		}
	
	},
	
	ACTION{

		@Override
		public ActionStreamSchema schema(final Map<String, ? extends Type, ?> attributes, final FixedArray<ConditionFormula> tupleConstraints, final FixedArray<ConditionFormula> relationConstraints, final boolean addTypeTupleConstraints) throws TSA_SchemaException
		{
			return new ActionStreamSchema(attributes, tupleConstraints, relationConstraints, addTypeTupleConstraints);
		}
	
	};
	
	TemporalStreamSchemaTypeID(){
		
	}
	
	public TemporalStreamSchema schema(final Map<String, ? extends Type, ?> attributes, final FixedArray<ConditionFormula> tupleConstraints, final FixedArray<ConditionFormula> relationConstraints) throws TSA_SchemaException{ 
		return this.schema(attributes, tupleConstraints, relationConstraints, true);
	}
	
	public abstract TemporalStreamSchema schema(Map<String, ? extends Type, ?> attributes, final FixedArray<ConditionFormula> tupleConstraints, final FixedArray<ConditionFormula> relationConstraints, final boolean addTypeTupleConstraints) throws TSA_SchemaException;
	
	public TemporalStreamSchema schema(final String[] names, final Type[] types, final ConditionFormula[] tupleConstraints, final ConditionFormula[] relationConstraints) throws TSA_SchemaException
	{
		return this.schema(new FixedArrayMap<String, Type>(names, types), new FixedArray<ConditionFormula>(false, tupleConstraints), new FixedArray<ConditionFormula>(false, relationConstraints));
	}

}
