package de.lmu.ifi.pms.cep.tsa.time.functions;

import org.simpleframework.xml.Element;
import org.simpleframework.xml.ElementArray;


public class MapTimeFunction extends TimeFunction {
	
	@Element(name="arity")
	final int arity;
	
	@ElementArray(name="mapping")
	final int[] mapping;

	public MapTimeFunction(@Element(name="arity") final int argumentArity, @ElementArray(name="mapping") final int[] mapping) {
		super();		
		this.arity = argumentArity;
		this.mapping = mapping;
		
		//check
		for(int index = 0; index < mapping.length; ++index){
			if(mapping[index] < 0){
				throw new IllegalArgumentException("Illegal mapping: mapping " + mapping[index] + " is smaller than 0.");
			}
			if(mapping[index] >= argumentArity){
				throw new IllegalArgumentException("Illegal mapping: mapping " + mapping[index] + " is greater than the argument arity " + argumentArity);
			}
		}
	}

	@Override
	public void apply(final long[] arguments, final int argStart, final long[] results, final int resultStart) {
		assert argStart >= 0;
		assert resultStart >= 0;
		assert arguments.length >= argStart + this.argumentArity();
		assert results.length >= resultStart + this.resultArity(); 
		
		for(int index = 0; index < this.mapping.length; ++index){
			assert this.mapping[index] >= 0;
			assert this.mapping[index] + argStart <= arguments.length;
			results[index + resultStart] = arguments[this.mapping[index] + argStart];
		}
	}

	@Override
	public final int argumentArity() {
		return this.arity;
	}

	@Override
	public final int resultArity() {
		return this.mapping.length;
	}

}
