package de.lmu.ifi.pms.cep.tsa.time.functions;

import org.simpleframework.xml.Root;
import org.simpleframework.xml.Serializer;
import org.simpleframework.xml.core.Persister;

import de.lmu.ifi.pms.cep.COMMON_CONSTANTS;
import de.lmu.ifi.pms.java.util.string.StringBuilder;


@Root(name="time_function")
public abstract class TimeFunction {
	
	public abstract void apply(long[] arguments, int argStart, long[] results, int resultStart);
	
	public abstract int argumentArity();
	
	public abstract int resultArity();
	
	public String serialize(){
		Persister persister = new Persister(COMMON_CONSTANTS.SERIALIZATION_FORMAT);
		StringBuilder sb = new StringBuilder();
		try {
			persister.write(new ParallelCompositeTimeFunction(this), sb);
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
		return sb.toString();
	}
	
	public static TimeFunction deserialize(final String timeFunction){
		Persister persister = new Persister(COMMON_CONSTANTS.SERIALIZATION_FORMAT);
		try {
			return persister.read(ParallelCompositeTimeFunction.class, timeFunction).subFunctions[0];
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}
	
	@Override
	public String toString() {
		Persister persister = new Persister();
		StringBuilder sb = new StringBuilder();
		try {
			persister.write(new ParallelCompositeTimeFunction(this), sb);
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
		return sb.toString();
	}
	
	@SuppressWarnings("unused")
	public static void main(final String[] args) {
		
		TimeFunction add = new AddDuration(1000);
		TimeFunction min = new Min(2);
		TimeFunction identity = IdentityTimeFunction.instance; 
		TimeFunction max = new Max(3);
		TimeFunction all = new ParallelCompositeTimeFunction(add, min, identity);
		TimeFunction compose = new SequencialCompositeTimeFunction(max, all);
		TimeFunction map = new MapTimeFunction(1, new int[]{0,0});
		TimeFunction compose2 = new SequencialCompositeTimeFunction(map, compose);
		
		TimeFunction emptyParallelTimeFunction = new ParallelCompositeTimeFunction();
		
		TimeFunction timeFunction = emptyParallelTimeFunction;
		
		Serializer serializer = new Persister(COMMON_CONSTANTS.SERIALIZATION_FORMAT);
		
		StringBuilder sb = new StringBuilder();
		
		try {
			serializer.write(timeFunction, sb);
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		String serializedFunction = sb.toString();
		
		System.out.println(serializedFunction);
		
		try {
			TimeFunction deserializedTimeFunction = serializer.read(timeFunction.getClass(), serializedFunction);
			System.in.read();
		} catch (Exception e) {
			e.printStackTrace();
		}
	
		System.exit(0);
	}

}

