package de.lmu.ifi.pms.java.util;

public class SynchronizedBuffer<Type> extends Buffer<Type> {

	protected final Object lock;
	
	public SynchronizedBuffer(final int capacity, final Object lock) {
		super(capacity);
		this.lock = lock;
	}

	public SynchronizedBuffer(final int capacity, final Class<Type> type, final Object lock) {
		super(capacity, type);
		this.lock = lock;
	}

	public SynchronizedBuffer(final int capacity, final Class<Type> type) {
		super(capacity, type);
		this.lock = this;
	}

	public SynchronizedBuffer(final int capacity) {
		super(capacity);
		this.lock = this;
	}
	
	@Override
	protected void setCapacity(final int capacity, final Class<Type> type) {
		synchronized(this.lock){
			super.setCapacity(capacity, type); //TODO implement
		}
	}
	
	@Override
	protected void setCapacity(final int capacity) {
		synchronized(this.lock){
			super.setCapacity(capacity); //TODO implement
		}
	}
	
	@Override
	public final int capacity() {
		synchronized(this.lock){
			return super.capacity(); //TODO implement
		}
	}

	@Override
	public final boolean isEmpty() {
		synchronized(this.lock){
			return super.isEmpty();
		}
	}

	@Override
	protected final void waitWhileEmpty(){
		synchronized(this.lock){
			while(this.isEmpty()){
				try {
					this.lock.wait();
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
		}
	}
	
	@Override
	public final boolean isFull() {
		synchronized(this.lock){
			return super.isFull();
		}
	}
	
	@Override
	protected final synchronized void waitWhileFull(){
		synchronized(this.lock){
			while(this.isFull()){
				try {
					this.lock.wait();
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
		}
	}
	
	@Override
	public final int size() {
		synchronized(this.lock){
			return super.size();
		}
	}
	
	@Override
	public int remainingCapacity() {
		synchronized(this.lock){
			return super.remainingCapacity(); //TODO implement
		}
	}
	
	@Override
	protected final void waitForSize(final int size) {
		synchronized(this.lock){
			while(this.size() < size){
				try {
					this.lock.wait();
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
		}
	}
	
	@Override
	protected final void signalAdd() {
		synchronized(this.lock){
			this.externalSignalAdd();
			this.lock.notifyAll();
		}
	}
	
	protected void externalSignalAdd(){
		
	}
	
	@Override
	public final void add(final Type value) {
		synchronized(this.lock){
			super.add(value);
		}
	}
	
	@Override
	public final void addFirst(final Type value) {
		synchronized(this.lock){
			super.addFirst(value);
		}
	}
	
	@Override
	public final void addLast(final Type value) {
		synchronized(this.lock){
			super.addLast(value);
		}
	}
	
	@Override
	public final boolean has(final int index) {
		synchronized(this.lock){
			return super.has(index);
		}
	}
	
	@Override
	public final boolean hasFirst() {
		synchronized(this.lock){
			return super.hasFirst();
		}
	}
	
	@Override
	public final boolean hasLast() {
		synchronized(this.lock){
			return super.hasLast();
		}
	}
	
	@Override
	public final Type get(final int index) {
		synchronized(this.lock){
			return super.get(index);
		}
	}
	
	@Override
	public final Type first() {
		synchronized(this.lock){
			return super.first();
		}
	}
	
	@Override
	public final Type last() {
		synchronized(this.lock){
			return super.last();
		}
	}
	
	@Override
	protected final void signalRemove() {
		synchronized(this.lock){
			this.externalSignalRemove();
			this.lock.notifyAll();
		}
	}
	
	protected void externalSignalRemove(){
		
	}
	
	@Override
	public final void remove() {
		synchronized(this.lock){
			super.remove();
		}
	}
	
	@Override
	public final void removeFirst() {
		synchronized(this.lock){
			super.removeFirst();
		}
	}
	
	@Override
	public final void removeLast() {
		synchronized(this.lock){
			super.removeLast();
		}
	}
	
}
