package de.lmu.ifi.pms.java.util.collection.array;

import static de.lmu.ifi.pms.java.util.collection.array.Arrays.getArray;

import java.util.ListIterator;

import de.lmu.ifi.pms.java.lang.iteration.AbstractIterator;
import de.lmu.ifi.pms.java.util.collection.AbstractCollection;


public abstract class AbstractArray<Type> extends AbstractCollection<Type>{
	
	public AbstractArray(){
		
	}

	public abstract Class<?> ensuredType();

	@Override
	public abstract int size();
	
	@Override
	public Object[] toArray(){
		Object[] result = getArray(this.ensuredType(), this.size());
		this.copyTo(0, result, 0, this.size());
		return result;
	}
	
	@Override
	public <T> T[] toArray(final T[] a) {
		T[] result;
		if(a.length >= this.size()){
			result = a;
		}
		else{
			result = getArray(a, this.size());			
		}
		this.copyTo(0, result, 0, this.size());
		return result;
	}
	
	public int[] toArray(final int[] a) {
		int[] result;
		if(a.length >= this.size()){
			result = a;
		}
		else{
			result = new int[this.size()];			
		}
		this.copyTo(0, result, 0, this.size());
		return result;
	}
	
	public <T> void copyTo(final T[] destination, final int destinationStart){
		this.copyTo(0, destination, destinationStart, this.size());
	}
	
	public abstract <T> void copyTo(int sourceStart, T[] destination, int destinationStart, int count);

	public abstract void copyTo(int sourceStart, int[] destination, int destinationStart, int count);
	
	@Override
	public ArrayIterator<Type, ? extends AbstractArray<Type>> iterator() {
		return new ArrayIterator<Type, AbstractArray<Type>>(this);
	}
	
	public static class ArrayIterator<Type, ArrayType extends AbstractArray<Type>> extends AbstractIterator<Type> implements ListIterator<Type>{
		
		protected int position;
		protected final int firstIndexInclusive;
		protected final int lastIndexExclusive;
		protected final ArrayType array;

		public ArrayIterator(final ArrayType array, final int firstIndexInclusive, final int lastIndexExclusive, final int position) {
			super();
			this.array = array;
			this.position = firstIndexInclusive + position;
			this.firstIndexInclusive = firstIndexInclusive;
			this.lastIndexExclusive = lastIndexExclusive;
		}

		public ArrayIterator(final ArrayType array) {
			this(array, array.minIndexInc(), array.maxIndexExc(), 0);
		}
		
		public int position(){
			return this.position;
		}

		@Override
		public boolean hasNext() {
			return this.position < this.lastIndexExclusive;
		}

		@Override
		public Type next() {
			return this.array.get(this.position++);
		}

		@Override
		public int nextInt(){
			return this.array.getInt(this.position++);
		}

		@Override
		public long nextLong(){
			return this.array.getLong(this.position++);
		}

		@Override
		public double nextDouble(){
			return this.array.getDouble(this.position++);
		}

		@Override
		public float nextFloat(){
			return this.array.getFloat(this.position++);
		}

		@Override
		public byte nextByte(){
			return this.array.getByte(this.position++);
		}

		@Override
		public short nextShort(){
			return this.array.getShort(this.position++);
		}

		@Override
		public char nextChar(){
			return this.array.getChar(this.position++);
		}

		@Override
		public boolean nextBoolean(){
			return this.array.getBoolean(this.position++);
		}

		@Override
		public boolean hasPrevious() {
			return this.position - 2 >= this.firstIndexInclusive;
		}

		@Override
		public final int nextIndex() {
			return this.position();
		}

		@Override
		public final int previousIndex() {
			return this.nextIndex() - 2;
		}

		@Override
		public Type previous() {
			return this.array.get((--this.position) - 1);
		}

		
		@Override
		@Deprecated
		public void add(final Type e) {
			 throw new UnsupportedOperationException();
		}

		@Override
		@Deprecated
		public void set(final Type e) {
			throw new UnsupportedOperationException();
		}
		
	}
	
	public final boolean has(final int index){
		return this.minIndexInc() <= index && index <= this.maxIndexExc();
	}
	
	public final void checkIndex(final int index) throws ArrayIndexOutOfBoundsException{
		if(index < this.minIndexInc() || this.maxIndexExc() <= index){
			throw new ArrayIndexOutOfBoundsException();
		}
	}
	
	public abstract Type get(int index);
	
	public int getInt(final int index){
		return toInt(this.get(index));
	}
	
	public long getLong(final int index){
		return toLong(this.get(index));
	}
	
	public float getFloat(final int index){
		return toFloat(this.get(index));
	}
	
	public double getDouble(final int index){
		return toDouble(this.get(index));
	}
	
	public byte getByte(final int index){
		return toByte(this.get(index));
	}
	
	public short getShort(final int index){
		return toShort(this.get(index));
	}
	
	public char getChar(final int index){
		return toChar(this.get(index));
	}
	
	public boolean getBoolean(final int index){
		return toBoolean(this.get(index));
	}
	
	public abstract int minIndexInc();
	
	public Type first(){
		return this.get(0);
	}
	
	public Type last(){
		return this.get(this.size() - 1);
	}
	
	public int maxIndexExc(){
		return this.minIndexInc() + this.size(); 
	}
	
	public abstract AbstractArray<Type> subArray(int offset, int length);
	
}
