package de.lmu.ifi.pms.java.util.collection.array;

import java.io.Serializable;

import de.lmu.ifi.pms.java.lang.iteration.FixedIterator;
import de.lmu.ifi.pms.java.util.collection.AbstractCollection;



@SuppressWarnings("serial")
public class Array<Type> extends AbstractCollection<Type> implements Serializable{
	
	protected final Type[] array;
	private int offset;
	private int length;
	
	public Array(final Class<Type> type, final int size){
		this(Arrays.getArray(type, size));
	}
	
	public Array(final Type... values) {
		this(0,values.length,values);
	}
	
	public Array(final int offset, final int length, final Type... values) {
		super();
		this.array = values;
		this.offset = offset;
		this.length = length;
	}

	@Override
	public int size() {
		return this.length;
	}
	
	@Override
	public FixedIterator<Type> iterator() {
		return new ArrayIterator<Type>(this.array, this.offset(), this.size());
	}
	
	protected final int offset(){
		return this.offset;
	}
	
	public final boolean has(final int index){
		return this.minIndex() <= index && index < this.maxIndex();
	}
	
	protected final boolean primitiveHas(final int index){
		return this.offset() <= index && index < this.size(); 
	}
	
	public Type get(final int index){
		if(this.primitiveHas(index)){
			return this.primitiveGet(index);
		}
		else{
			throw new IndexOutOfBoundsException();//TODO better exception
		}
	}
	
	protected Type primitiveGet(final int index){
		return this.array[index];
	}
	
	public void set(final int index, final Type value){
		if(this.primitiveHas(index)){
			this.primitiveSet(index, value);
		}
		else{
			throw new IndexOutOfBoundsException();//TODO better exception
		}
	}
	
	protected void primitiveSet(final int index, final Type value){
		this.array[index] = value;
	}
	
	public int minIndex(){
		return 0;
	}
	
	public int maxIndex(){
		return this.size()-1;
	}
	
	public Array<Type> subArray(final int offset, final int length){
		return new Array<Type>(this.offset() + offset, length, this.array);
	}
	
	public void removeFirst(){
		this.removeFirst(1);
	}
	
	public void removeFirst(final int n){
		this.offset = this.offset + n;//TODO unchecked
		this.length = this.length - n;
	}
	
	public void removeLast(){
		this.removeLast(1);
	}
	
	public void removeLast(final int n){
		this.length = this.length - n;
	}
	
}
