package de.lmu.ifi.pms.java.util.collection.array;

import static de.lmu.ifi.pms.java.util.collection.array.Arrays.*;

public class ModifyableArray<Type> extends AbstractModifyableArray<Type>{

	private final Object array;
	private int offset;
	private int size;
	
	public ModifyableArray(final int size, final Class<? super Type> type){
		this(getUntypedInitializedArray(type, size), 0, size, true);
	}
	
	@SuppressWarnings("unchecked")
	public ModifyableArray(final Class<Type> type, final Type... values) {
		this(getArray(type, values.length), 0, values.length);
		copy(values, 0, (Type[])this.array, 0, values.length);
	}
	
	public ModifyableArray(final Type[] values, final int offset, final int length) {
		super();
		this.array = values;
		this.offset = offset;
		this.size = length;
	}
	
	public ModifyableArray(final boolean... values) {
		this(values, 0, values.length, true);
	}
	
	public ModifyableArray(final boolean[] values, final int offset, final int length) {
		this(values, 0, values.length, true);
	}
	
	public ModifyableArray(final char... values) {
		this(values, 0, values.length, true);
	}
	
	public ModifyableArray(final char[] values, final int offset, final int length) {
		this(values, 0, values.length, true);
	}
	
	public ModifyableArray(final byte... values) {
		this(values, 0, values.length, true);
	}
	
	public ModifyableArray(final byte[] values, final int offset, final int length) {
		this(values, 0, values.length, true);
	}
	
	public ModifyableArray(final short... values) {
		this(values, 0, values.length, true);
	}
	
	public ModifyableArray(final short[] values, final int offset, final int length) {
		this(values, 0, values.length, true);
	}
	
	/*public ModifyableArray(final int... values) {
		this(values, 0, values.length, true);
	}*/
	
	public ModifyableArray(final int[] values, final int offset, final int length) {
		this(values, 0, values.length, true);
	}
	
	/*public ModifyableArray(final long... values) {
		this(values, 0, values.length, true);
	}*/
	
	public ModifyableArray(final long[] values, final int offset, final int length) {
		this(values, 0, values.length, true);
	}
	
	/*public ModifyableArray(final float... values) {
		this(values, 0, values.length, true);
	}*/
	
	public ModifyableArray(final float[] values, final int offset, final int length) {
		this(values, 0, values.length, true);
	}
	
	/*public ModifyableArray(final double... values) {
		this(values, 0, values.length, true);
	}*/
	
	public ModifyableArray(final double[] values, final int offset, final int length) {
		this(values, 0, values.length, true);
	}
	
	protected ModifyableArray(final Object array, final int offset, final int length, final boolean unchecked) {
		super();
		if(unchecked || isArray(array)){
			this.array = array;
		}
		else{
			throw new IllegalArgumentException();//TODO better exception
		}
		this.offset = offset;
		this.size = length;
	}
	
	/* (non-Javadoc)
	 * @see extension.java.util.collection.array.AbstractArray#ensuredType()
	 */
	@Override
	@SuppressWarnings("unchecked")
	public Class<? super Type> ensuredType() {
		return (Class<? super Type>) this.array.getClass().getComponentType();
	}

	@Override
	public int size() {
		return this.size;
	}
	
	@Override
	public ArrayIterator<Type> iterator() {
		return new ArrayIterator<Type>(this);
	}
	
	public static class ArrayIterator<Type> extends ModifyableArrayIterator<Type, ModifyableArray<Type>>{

		public ArrayIterator(final ModifyableArray<Type> array){
			this(array, 0);
		}
		
		public ArrayIterator(final ModifyableArray<Type> array, final int position) {
			super(array, array.offset(), array.offset() + array.size(), position);
		}
		
		/* (non-Javadoc)
		 * @see extension.java.util.collection.array.AbstractArray.ArrayIterator#position()
		 */
		@Override
		public int position() {
			return this.position - this.array.offset();
		}

		@Override
		public Type next() {
			return this.array.get(this.position++);
		}
		
		@Override
		public int nextInt(){
			return this.array.getInt(this.position++);
		}
		
		@Override
		public long nextLong(){
			return this.array.getLong(this.position++);
		}
		
		@Override
		public double nextDouble(){
			return this.array.getDouble(this.position++);
		}
		
		@Override
		public float nextFloat(){
			return this.array.getFloat(this.position++);
		}
		
		@Override
		public byte nextByte(){
			return this.array.getByte(this.position++);
		}
		
		@Override
		public short nextShort(){
			return this.array.getShort(this.position++);
		}
		
		@Override
		public char nextChar(){
			return this.array.getChar(this.position++);
		}
		
		@Override
		public boolean nextBoolean(){
			return this.array.getBoolean(this.position++);
		}
		
		@Override
		public Type previous() {
			return this.array.get((--this.position)-1);
		}
		
		@Override
		public void replace(final Type newValue) {
			if(this.position > this.array.offset()){
				this.array.set(this.position - 1, newValue);
			}
			else{
				throw new ArrayIndexOutOfBoundsException();
			}
		}
		
		
		
	}
	
	protected final int offset(){
		return this.offset;
	}
	
	protected int computePrimitivIndex(final int index){
		return index + this.offset();
	}
	
	@SuppressWarnings("unchecked")
	@Override
	public final Type get(final int index){
		this.checkIndex(index);
		return (Type) Arrays.get(this.computePrimitivIndex(index), this.array);
	}
	
	@Override	
	public final int getInt(final int index){
		this.checkIndex(index);
		return Arrays.getInt(this.computePrimitivIndex(index), this.array);
	}
	
	@Override	
	public final long getLong(final int index){
		this.checkIndex(index);
		return Arrays.getLong(this.computePrimitivIndex(index), this.array);
	}
	
	@Override	
	public final float getFloat(final int index){
		this.checkIndex(index);
		return Arrays.getFloat(this.computePrimitivIndex(index), this.array);
	}
	
	@Override	
	public final double getDouble(final int index){
		this.checkIndex(index);
		return Arrays.getDouble(this.computePrimitivIndex(index), this.array);
	}
	
	@Override	
	public final byte getByte(final int index){
		this.checkIndex(index);
		return Arrays.getByte(this.computePrimitivIndex(index), this.array);
	}
	
	@Override	
	public final short getShort(final int index){
		this.checkIndex(index);
		return Arrays.getShort(this.computePrimitivIndex(index), this.array);
	}
	
	@Override	
	public final char getChar(final int index){
		this.checkIndex(index);
		return Arrays.getChar(this.computePrimitivIndex(index), this.array);
	}
	
	@Override	
	public final boolean getBoolean(final int index){
		this.checkIndex(index);
		return Arrays.getBoolean(this.computePrimitivIndex(index), this.array);
	}
	
	@Override	
	public final void set(final int index, final Type value){
		this.checkIndex(index);
		Arrays.set(this.computePrimitivIndex(index), this.array, value);
	}
	
	@Override
	public final void set(final int index, final int value){
		this.checkIndex(index);
		Arrays.set(this.computePrimitivIndex(index), this.array, value);
	}
	
	@Override	
	public final void set(final int index, final long value){
		this.checkIndex(index);
		Arrays.set(this.computePrimitivIndex(index), this.array, value);
	}
	
	@Override	
	public final void set(final int index, final double value){
		this.checkIndex(index);
		Arrays.set(this.computePrimitivIndex(index), this.array, value);
	}
	
	@Override	
	public final void set(final int index, final float value){
		this.checkIndex(index);
		Arrays.set(this.computePrimitivIndex(index), this.array, value);
	}
	
	@Override	
	public final void set(final int index, final short value){
		this.checkIndex(index);
		Arrays.set(this.computePrimitivIndex(index), this.array, value);
	}
	
	@Override
	public final void set(final int index, final byte value){
		this.checkIndex(index);
		Arrays.set(this.computePrimitivIndex(index), this.array, value);
	}
	
	@Override
	public final void set(final int index, final char value){
		this.checkIndex(index);
		Arrays.set(this.computePrimitivIndex(index), this.array, value);
	}
	
	@Override
	public final void set(final int index, final boolean value){
		this.checkIndex(index);
		Arrays.set(this.computePrimitivIndex(index), this.array, value);
	}
	
	@Override
	public int minIndexInc(){
		return 0;
	}
	
	@Override
	public int maxIndexExc(){
		return this.size();
	}
	
	@Override
	public ModifyableArray<Type> subArray(final int offset, final int length){
		return new ModifyableArray<Type>(this.array, this.offset() + offset, length, true);
	}
	
	/* (non-Javadoc)
	 * @see extension.java.util.collection.array.AbstractModifyableArray#insertIndexDropLast(int, java.lang.Object)
	 */
	@Override
	public void insertIndexDropLast(final int index, final Type value) {
		Arrays.insertIndexDropLast(this.array, index, value);
	}

	@Override
	public <T> void copyTo(final int sourceStart, final T[] destination, final int destinationStart, final int count) {
		try{
			System.arraycopy(this.array, this.computePrimitivIndex(sourceStart), destination, destinationStart, count);
		} catch(ArrayStoreException e){
			for(int i_src = this.computePrimitivIndex(sourceStart), max_i_src = i_src + count, i_dest = destinationStart; i_src < max_i_src; ++i_src, ++i_dest){
				Arrays.set(i_dest, destination, Arrays.get(i_src, this.array));
			}
		}
	}
	
	@Override
	public void copyTo(final int sourceStart, final int[] destination,final int destinationStart, final int count) {
		System.arraycopy(this.array, sourceStart, destination, destinationStart, count);
	}
	
}
