package de.lmu.ifi.pms.java.util.collection.array.fixed;
 
import java.util.Comparator;

import org.simpleframework.xml.Element;
import org.simpleframework.xml.ElementList;
import org.simpleframework.xml.core.Persister;

import de.lmu.ifi.pms.java.lang.iteration.AbstractFinalListIterator;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.array.Arrays;
import de.lmu.ifi.pms.java.util.collection.fixed.empty.Empty;
import de.lmu.ifi.pms.java.util.collection.fixed.set.FixedNavigableSet;
import de.lmu.ifi.pms.java.util.collection.fixed.set.FixedSortedSet;
import de.lmu.ifi.pms.java.util.collection.javaWrapper.ToJavaNavigableSetWrapper;
import de.lmu.ifi.pms.java.util.collection.set.AbstractSortedSet.AbstractJavaNavigableSet;
import de.lmu.ifi.pms.java.util.collection.set.Set;
import de.lmu.ifi.pms.java.util.collection.set.SortedSet;
import de.lmu.ifi.pms.java.util.comparator.AbstractComparator;
import de.lmu.ifi.pms.java.util.filter.Filter;
import de.lmu.ifi.pms.java.util.string.StringBuilder;


public class FixedArrayNavigableSet<Type> extends FixedArraySet<Type> implements FixedNavigableSet<Type>{
	
	@Element(name="comparator", required=false)
	private Comparator<? super Type> comparator;
	
	
	@SuppressWarnings("unchecked")
	public FixedArrayNavigableSet(){
		super(false, false);
	}
	
	public FixedArrayNavigableSet(final Comparator<? super Type> comparator, final boolean copyArray, final boolean check, final Type... elements) {
		super(true);
		this.fillBody(copyArray, check, elements, comparator);
	}
	
	@SuppressWarnings("unchecked")
	public FixedArrayNavigableSet(final Type element){
		this(false, false, element);
	}
	
	public FixedArrayNavigableSet(final boolean copyArray, final boolean check, final Type... elements) {
		super(true);
		this.fillBody(copyArray, check, elements);
	}

	public FixedArrayNavigableSet(final boolean check, final Collection<Type> collection) {
		super(check, collection);
	}

	public FixedArrayNavigableSet(final boolean check, final FixedArray<Type> vector) {
		super(check, vector);
	}

	public FixedArrayNavigableSet(final boolean check, final Iterable<Type> collection, final int elementCount) {
		super(check, collection, elementCount);
	}

	public FixedArrayNavigableSet(final boolean copyArray, final Type... elements) {
		super(copyArray, elements);
	}

	public FixedArrayNavigableSet(final boolean bodyOnly) {
		super(bodyOnly);
	}

	public FixedArrayNavigableSet(final Collection<Type> collection) {
		super(collection);
	}
	
	public FixedArrayNavigableSet(@ElementList(entry="element",inline=true, required=false, empty=false) final java.util.Collection<Type> collection){
		super(collection);
	}

	public FixedArrayNavigableSet(final FixedArray<Type> vector) {
		super(vector);
	}

	public FixedArrayNavigableSet(final FixedArraySet<Type> vector) {
		super(vector);
	}

	public FixedArrayNavigableSet(final Iterable<Type> collection, final int elementCount) {
		super(collection, elementCount);
	}

	public FixedArrayNavigableSet(final Set<Type> collection) {
		super(collection);
	}

	public FixedArrayNavigableSet(final Type... elements) {
		super(elements);
	}

	public FixedArrayNavigableSet(final Collection<Type> collection, final Comparator<Type> comparator) {
		super(true);
		this.fillBody(collection, comparator);
	}
	
	public FixedArrayNavigableSet(final SortedSet<Type> collection) {
		super(true);
		this.fillBody(collection);
	}
	
	public FixedArrayNavigableSet(final FixedArrayNavigableSet<Type> vector) {
		super(true);
		this.fillBody(vector);
	}
	
	public void fillBody(final Collection<Type> collection, final Comparator<Type> comparator) {
		this.fillBody(true, collection, comparator);
	}
	
	@Override
	public void fillBody(final boolean check, final Collection<? extends Type> collection) {
		this.fillBody(check, collection, null);
	}
	
	@Override
	public void fillBody(final boolean check, final FixedArray<Type> vector){
		this.fillBody(check, vector, null);
	}
	
	@Override
	public void fillBody(final boolean copyArray, final boolean check, final Type... array) {
		this.fillBody(copyArray, check, array, null);
	}
	
	@SuppressWarnings("unchecked")
	public void fillBody(final boolean copyArray, final boolean check, Type[] array, final Comparator<? super Type> comparator) {
		this.comparator = comparator;
		if(copyArray){
			array = java.util.Arrays.copyOf(array, array.length);
		}
		if(check){
			if(array.length > 0){
				//java.util.Arrays.sort(array, this.comparator);
				Arrays.sort(array, null, false);
				
				//remove dublicates
				int readIndex = 1;
				int writeIndex = 1;
				int size = array.length;
				assert size > 0;
				Type last = array[0];
				while(readIndex < size){
					Type current = array[readIndex];
					if(this.compare(current, last) > 0){						
						array[writeIndex] = current;
						++writeIndex;
						last = current;
					}
					++readIndex;
				}
				
				Type[] set;
				if(writeIndex < readIndex){//eliminated dublicates
					set = java.util.Arrays.copyOf(array, writeIndex);
				}
				else{
					set = array;
				}
				
				
				//Besserer Array, mit passendem Typ?
				super.fillBody(false ,false, set);
			}
			else{
				assert array.length == 0;
				super.fillBody(false, false, (Type[])Arrays.getArray(array.getClass().getComponentType(), 0));
			}
		}
		else{
			super.fillBody(false, false, array);
		}
	}
	
	@SuppressWarnings("unchecked")
	public void fillBody(final boolean check, final FixedArray<Type> vector, final Comparator<Type> comparator) {
		this.fillBody(false, check, vector.toArray((Type[])Arrays.getArray(vector.ensuredType(), vector.size())), comparator);
	}
	
	public void fillBody(final boolean check, final Collection<? extends Type> collection, final Comparator<Type> comparator) {
		if(check){
			this.fillBody(check, new FixedArray<Type>(collection));
		}
		else{
			this.comparator = comparator;
			super.fillBody(check, collection);
		}
	}
	
	public void fillBody(final SortedSet<Type> collection) {
		this.fillBody(false,collection);
		this.comparator = collection.comparator();
	}
	
	public void fillBody(final FixedArrayNavigableSet<Type> vector) {
		super.fillBody(vector);
		this.comparator = vector.comparator;
	}
	
	@Override
	public final void fillBody(final FixedArraySet<Type> vector) {
		this.fillBody(true,vector);
	}
	
	@Override
	public void fillBody(final Set<Type> s) {
		this.fillBody(true, s);
	}
	
	public void fillBody(final boolean check, final Set<Type> s) {
		this.fillBody(check,(Collection<Type>)s);
	}
	
	public void fillBody(final boolean check, final FixedArraySet<Type> vector) {
		if(check){
			this.fillBody((Collection<Type>)vector);
		}
		else{
			super.fillBody(vector);
		}
	}

	@Override
	public Comparator<? super Type> comparator() {
		return this.comparator;
	}

	@Override
	public FixedSortedSet<Type> headSet(final Type toElement) {
		return this.headSet(toElement, false);
	}
	
	public int startIndex(final Type fromElement, final boolean inclusive){
		if(inclusive){
			return this.ceilingIndex(fromElement);
		}
		else{
			return this.higherIndex(fromElement);
		}
	}
	
	public int endIndex(final Type toElement, final boolean inclusive){
		if(inclusive){
			return this.floorIndex(toElement);
		}
		else{
			return this.lowerIndex(toElement);
		}
	}

	@Override
	public FixedArrayNavigableSet<Type> headSet(final Type toElement, final boolean inclusive) {
		return this.subList(0, this.endIndex(toElement, inclusive));
	}

	@Override
	public FixedSortedSet<Type> tailSet(final Type fromElement) {
		return this.tailSet(fromElement, true);
	}

	@Override
	public FixedArrayNavigableSet<Type> tailSet(final Type fromElement, final boolean inclusive) {
		return this.subList(this.startIndex(fromElement, inclusive), this.size());
	}

	@Override
	public FixedArrayNavigableSet<Type> subSet(final Type fromElement, final Type toElement) {
		return this.subSet(fromElement, true, toElement, false);
	}

	@Override
	public FixedArrayNavigableSet<Type> subSet(final Type fromElement, final boolean fromInclusive, final Type toElement, final boolean toInclusive) {
		return this.subList(this.startIndex(fromElement, fromInclusive), this.endIndex(toElement, toInclusive));
	}

	@Override
	public FixedArrayNavigableSet<Type> subList(final int fromIndex, final int toIndex) {
		return this.subArray(fromIndex, toIndex - fromIndex);
	}

	@Override
	public FixedArrayNavigableSet<Type> subArray(final int offset, final int length) {
		return new FinalArrayNavigableSubSet<Type>(this, offset, length);
	}

	@Override
	public AbstractFinalListIterator<Type> descendingIterator() {
		return null;//FIXME implement
	}

	@Override
	public FixedArrayNavigableSet<Type> descendingSet() {
		return null;//FIXME implement
	}
	
	@Override
	public final int indexOf(final Object o) {
		this.ensuredType().cast(o);
		@SuppressWarnings("unchecked")
		int search = this.search((Type)o);
		return search;
	}

	public final int ceilingIndex(final Type e){
		int search = this.search(e);
		if(search < 0){
			search = - search - 1;
		}
		return search;
	}

	@Override
	public Type ceiling(final Type e) {
		int search = this.ceilingIndex(e);
		if(!this.has(search)){
			return null;
		}
		else{
			return this.get(search);
		}
	}
	
	public final int floorIndex(final Type e){
		int search = this.search(e);
		if(search < 0){
			search = - search - 2;
		}
		return search;
	}

	@Override
	public Type floor(final Type e) {
		int search = this.floorIndex(e);
		if(!this.has(search)){
			return null;
		}
		else{
			return this.get(search);
		}
	}
	
	public final int higherIndex(final Type e){
		int search = this.search(e);
		if(search < 0){
			search = - search - 1;
		}
		else{
			++search;
		}
		return search;
	}

	@Override
	public Type higher(final Type e) {
		int search = this.higherIndex(e);
		if(!this.has(search)){
			return null;
		}
		else{
			return this.get(search);
		}
	}
	
	public final int lowerIndex(final Type e){
		int search = this.search(e);
		if(search < 0){
			search = - search - 2;
		}
		else{
			--search;
		}
		return search;
	}

	@Override
	public Type lower(final Type e) {
		int search = this.lowerIndex(e);
		if(!this.has(search)){
			return null;
		}
		else{
			return this.get(search);
		}
	}
	
	@SuppressWarnings("unchecked")
	@Override
	public boolean contains(final Object o) {
			return this.search((Type)o) >= 0;
	}
	
	protected int search(final Type e){
		return java.util.Arrays.binarySearch(this.vector, e, this.comparator);
		//return this.search(e, 0, this.size());
	}
	
/*	private final int search(final Type e, final int offset, final int length){
		int comp;
		int center;
		if(length == 0){
			return -offset-1;
		}
		else{
			center = offset + Math.div(length, 2);
			comp = this.compare(e, this.get(center));
			if(comp == 0){
				return center;
			}
			else if( comp > 0){
				return this.search(e, center + 1, length + offset - center - 1);
			}
			else{
				return this.search(e, offset, center - offset);
			}
		}
	}*/
	
	protected int compare(final Type e1, final Type e2){
		return AbstractComparator.compare(e1, e2, this.comparator);
	}
	
	public FixedArrayNavigableSet<Type> filter(final Filter<Type> filter){
		Type[] filteredElements = Arrays.getArray(this.vector, this.vector.length);
		int newIndex = 0;
		Type element;
		for(int index = 0; index < this.vector.length; ++index){
			element = this.vector[index];
			if(filter.accepts(element)){
				filteredElements[newIndex] = element;
				++newIndex;
			}
		}
		return new FixedArrayNavigableSet<Type>(false, java.util.Arrays.copyOf(filteredElements, newIndex));
	}
	
	@Override
	public AbstractJavaNavigableSet<Type> javaCollection() {
		return new ToJavaNavigableSetWrapper<Type>(this);
	}
	
	private static class FinalArrayNavigableSubSet<Type> extends FixedArrayNavigableSet<Type> {
		private final int offset;
		private final int length;

		public FinalArrayNavigableSubSet(final FixedArrayNavigableSet<Type> vector, final int offset, final int length) {
			super(vector);
			if(offset + length < vector.size()){
				this.offset = offset + vector.getOffset();
				this.length = length;
			}
			else{
				throw new IndexOutOfBoundsException();
			}
		}

		@Override
		public Type get(final int i) {
			return super.get(i+this.offset);
		}

		@Override
		public int size() {
			return this.length;
		}

		@Override
		protected int getOffset() {
			return this.offset;
		}
		
		//FIXME other methods, e.g. copyTo, filter
				
	}
	

	
	public static <Type> FixedArrayNavigableSet<Type> merge(final FixedArrayNavigableSet<Type> set1, final FixedArrayNavigableSet<Type> set2){
		return merge(set1, set2, null);
	}

	@SuppressWarnings("unchecked")
	public static <Type> FixedArrayNavigableSet<Type> merge(final FixedArrayNavigableSet<Type> set1, final FixedArrayNavigableSet<Type> set2, final int[] sourcePositionNew){
		return new FixedArrayNavigableSet<Type>(false, false, Arrays.merge((Type[])set1.toArray(), (Type[])set2.toArray(), set1.comparator, sourcePositionNew));
	}
	
	public static <Type> FixedArrayNavigableSet<Type> merge(FixedArrayNavigableSet<Type>... sets){
		if(sets.length == 0){
			return new FixedArrayNavigableSet<Type>(Empty.<Type>instance());
		}
		else{
			int length = sets.length;
			while(length > 1){//more than one set
				int length_div_2 = length >>> 1; //equivalent to length_div_2 = length div 2
				length = ++length >>> 1; // equivalent to length = (length div 2) + (length mod 2)
				@SuppressWarnings("unchecked")
				FixedArrayNavigableSet<Type>[] new_sets = new FixedArrayNavigableSet[length];
				for(int i=0, j = length_div_2 ; i < length_div_2; ++i, ++j){
					new_sets[i] = merge(sets[i],sets[j]);	
				}
				if(length_div_2 < length){//odd number of sets (=> maxIndex is even)
					new_sets[length_div_2] = sets[length_div_2 << 1]; 
				}
				sets = new_sets;
			}
			return sets[0];
		}
	}
	
	@SuppressWarnings("unchecked")
	public static <Type> FixedArrayNavigableSet<Type> merge(final Collection<? extends FixedArrayNavigableSet<Type>> sets){
		FixedArrayNavigableSet<?>[] setsArray = new FixedArrayNavigableSet<?>[sets.size()];
		sets.toArray((FixedArrayNavigableSet<Type>[])setsArray);
		return merge((FixedArrayNavigableSet<Type>[])setsArray);
	}
	
	public static <Type> FixedArrayNavigableSet<Type> intersect(final FixedArrayNavigableSet<Type> set1, final FixedArrayNavigableSet<Type> set2){
		@SuppressWarnings("unchecked")
		final Type[] set1array = (Type[])set1.toArray();
		@SuppressWarnings("unchecked")
		final Type[] set2array = (Type[])set2.toArray();
		final Type[] newSetArray;
		if(set1array.getClass().getComponentType().isAssignableFrom(set2array.getClass().getComponentType())){
			newSetArray = set2array;
		}
		else{
			newSetArray = set1array;
		}
		int newLength = Arrays.sortedIntersect(set1array, set2array, newSetArray, null);
		return new FixedArrayNavigableSet<Type>(false, false, java.util.Arrays.copyOf(newSetArray, newLength));
	}
	
	public static <Type> FixedArrayNavigableSet<Type> difference(final FixedArrayNavigableSet<Type> set1, final FixedArrayNavigableSet<Type> set2){
		@SuppressWarnings("unchecked")
		final Type[] set1array = (Type[])set1.toArray();
		@SuppressWarnings("unchecked")
		final Type[] set2array = (Type[])set2.toArray();
		final Type[] newSetArray = set1array;
		int newLength = Arrays.sortedDifference(set1array, set2array, newSetArray, null);
		return new FixedArrayNavigableSet<Type>(false, false, java.util.Arrays.copyOf(newSetArray, newLength));
	}
	
	public static <Type> FixedArrayNavigableSet<Type> symmetricDifference(final FixedArrayNavigableSet<Type> set1, final FixedArrayNavigableSet<Type> set2){
		@SuppressWarnings("unchecked")
		final Type[] set1array = (Type[])set1.toArray();
		@SuppressWarnings("unchecked")
		final Type[] set2array = (Type[])set2.toArray();
		
		return new FixedArrayNavigableSet<Type>(false, false, Arrays.sortedSymmetricDifference(set1array, set2array, null));
	}
	
	public static void main(final String[] args) throws Exception{
		FixedArrayNavigableSet<String> array = new FixedArrayNavigableSet<String>("a");
		Persister p = new Persister();
		StringBuilder sb = new StringBuilder();
		p.write(array, sb);
		String serialized_array = sb.toString();
		System.out.println(serialized_array);
		
		@SuppressWarnings("unchecked")
		FixedArrayNavigableSet<String> deserialized_array = p.read(array.getClass(), serialized_array);
		
		System.out.println(deserialized_array);
	}

}
