package de.lmu.ifi.pms.java.util.collection.array.fixed;

import org.simpleframework.xml.ElementList;

import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.fixed.set.FixedSet;
import de.lmu.ifi.pms.java.util.collection.fixed.wrapper.FixedIterableWrapper;
import de.lmu.ifi.pms.java.util.collection.javaWrapper.ToJavaSetWrapper;
import de.lmu.ifi.pms.java.util.collection.set.NormalSet;
import de.lmu.ifi.pms.java.util.collection.set.Set;


@SuppressWarnings("serial")
public class FixedArraySet<Type> extends FixedArray<Type> implements FixedSet<Type>{	
	
	public FixedArraySet(final boolean copyArray, final Type... elements) {
		super(copyArray, elements);
	}

	public FixedArraySet(final boolean bodyOnly) {
		super(bodyOnly);
	}

	public FixedArraySet(final FixedArray<Type> vector) {
		super(vector);
	}
	
	public FixedArraySet(final boolean check, final FixedArray<Type> vector) {
		super(true);
		this.fillBody(check, vector);
	}

	public FixedArraySet(final Type... elements) {
		super(elements);
	}

	public FixedArraySet(final boolean copyArray, final boolean check, final Type... elements) {
		super(true);
		this.fillBody(copyArray, check, elements);
	}
	
	

	public FixedArraySet(final Collection<Type> collection) {
		super(collection);
	}
	
	public FixedArraySet(@ElementList(entry="element",inline=true) final java.util.Collection<Type> collection){
		super(collection);
	}
	
	public FixedArraySet(final boolean check, final Collection<Type> collection) {
		super(true);
		this.fillBody(check, collection);
	}

	public FixedArraySet(final Iterable<Type> collection, final int elementCount) {
		super(collection, elementCount);
	}
	
	public FixedArraySet(final boolean check, final Iterable<Type> collection, final int elementCount) {
		super(true);
		this.fillBody(check, collection, elementCount);
	}

	public FixedArraySet(final Set<Type> collection) {
		super(true);
		this.fillBody(collection);
	}
	
	public FixedArraySet(final FixedArraySet<Type> vector) {
		super(false);
		this.fillBody(vector);
	}
	
	public void fillBody(final boolean copyArray, final boolean check, final Type... elements) {
		if(check){
			this.fillBody(new FixedArray<Type>(copyArray, elements));
		}
		else{
			super.fillBody(copyArray, elements);
		}
	}
	
	@Override
	public void fillBody(final boolean copyArray, final Type... elements) {
		this.fillBody(copyArray, true, elements);
	}
	
	@Override
	public final void fillBody(final Collection<? extends Type> collection) {
		this.fillBody(true, collection);
	}
	
	public void fillBody(final boolean check, final Collection<? extends Type> collection) {
		if(check){
			super.fillBody(NormalSet.eliminate(collection));
		}
		else{
			super.fillBody(collection);
		}
	}
	
	@Override
	public final void fillBody(final FixedArray<Type> vector) {
		this.fillBody(true, vector);
	}
	
	public void fillBody(final boolean check, final FixedArray<Type> vector) {
		if(vector instanceof FixedArraySet<?>){
			this.fillBody((FixedArraySet<Type>)vector);
		}
		else{
			if(check){
				this.fillBody((Collection<Type>)vector);
			}
			else{
				super.fillBody(vector);
			}
		}
	}
	
	public void fillBody(final Set<Type> s){
		this.fillBody(false, s);
	}
	
	public void fillBody(final FixedArraySet<Type> vector) {
		super.fillBody(vector);
	}
	
	@Override
	public void fillBody(final Iterable<Type> collection, final int elementCount) {
		super.fillBody(collection, elementCount); //TODO implement
	}
	
	public void fillBody(final boolean check, final Iterable<Type> collection, final int elementCount) {
		if(check){
			this.fillBody(new FixedIterableWrapper<Type>(collection, elementCount));
		}
		else{
			super.fillBody(collection, elementCount);
		}
	}
	
	@Override
	public java.util.AbstractSet<Type> javaCollection() {
		return new ToJavaSetWrapper<Type>(this);
	}

	@Override
	public FixedArraySet<Type> subList(final int fromIndex, final int toIndex) {
		return this.subArray(fromIndex, toIndex - fromIndex);
	}

	@Override
	public FixedArraySet<Type> subArray(final int offset, final int length) {
		return new FinalVectorSubSet<Type>(this, offset, length);
	}
	
	

	private static class FinalVectorSubSet<Type> extends FixedArraySet<Type> {
		private final int offset;
		private final int length;

		public FinalVectorSubSet(final FixedArraySet<Type> vector, final int offset, final int length) {
			super(vector);
			if(offset + length < vector.size()){
				this.offset = offset + vector.getOffset();
				this.length = length;
			}
			else{
				throw new IndexOutOfBoundsException();
			}
		}

		@Override
		public Type get(final int i) {
			return super.get(i+this.offset);
		}

		@Override
		public int size() {
			return this.length;
		}

		@Override
		protected int getOffset() {
			return this.offset;
		}	
				
	}
	
}
