package de.lmu.ifi.pms.java.util.collection.composition;

import de.lmu.ifi.pms.java.lang.iteration.ExtendibleComposedIterator;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.extendible.ExtendibleCollection;
import de.lmu.ifi.pms.java.util.collection.javaWrapper.FromJavaCollectionWrapper;

public class ExtendibleComposedCollection<Type> extends ComposedCollection<Type, ExtendibleCollection<Type>> implements ExtendibleCollection<Type>{

	public ExtendibleComposedCollection(final Collection<? extends ExtendibleCollection<Type>> collections) {
		super(collections);
	}

	public ExtendibleComposedCollection(final ExtendibleCollection<Type>... collections) {
		super(collections);
	}

	@Override
	public ExtendibleComposedIterator<Type> iterator() {
		return new ExtendibleComposedIterator<Type>(this.collections, true);
	}
	
	@Override
	public final boolean add(final Type e) {
		return this.collections.last().add(e);
	}
	
	@Override
	public boolean addAll(final java.util.Collection<? extends Type> c) {
		return this.addAll(new FromJavaCollectionWrapper<Type>(c));
	}

	@Override	
	public final boolean addAll(final Collection<? extends Type> c) {
		return this.collections.last().addAll(c);
	}

	@Override
	public final void clear() {
		for(ExtendibleCollection<Type> c: this.collections){
			c.clear();
		}
	}

	@Override
	public final boolean remove(final Object o) {
		for(ExtendibleCollection<Type> c: this.collections){
			c.remove(o);
		}
		return true;
	}
	
	@SuppressWarnings({ "unchecked", "rawtypes" })
	@Override
	public boolean removeAll(final java.util.Collection<?> c) {
		return this.removeAll(new FromJavaCollectionWrapper(c));
	}

	@Override
	public final boolean removeAll(final Collection<?> collection) {
		for(ExtendibleCollection<Type> c: this.collections){
			c.removeAll(collection);
		}
		return true;
	}
	
	@SuppressWarnings({ "unchecked", "rawtypes" })
	@Override
	public boolean retainAll(final java.util.Collection<?> c) {
		return this.retainAll(new FromJavaCollectionWrapper(c));
	}

	@Override
	public final boolean retainAll(final Collection<?> collection) {
		for(ExtendibleCollection<Type> c: this.collections){
			c.retainAll(collection);
		}
		return true;
	}

}
