package de.lmu.ifi.pms.java.util.collection.set;

import de.lmu.ifi.pms.java.lang.iteration.OrderedDiffIterator;
import de.lmu.ifi.pms.java.lang.iteration.OrderedIntersectionIterator;
import de.lmu.ifi.pms.java.lang.iteration.OrderedUnionIterator;
import de.lmu.ifi.pms.java.util.collection.array.extendible.IArrayQueue;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArraySet;


public class Operations {
	
	  public static <Type> FixedArraySet<Type> intersectionOf(final Set<? extends Type> s1, final Set<? extends Type> s2){
		  IArrayQueue<Type> a = new IArrayQueue<Type>(Object.class);
		  for(Type t1: s1){
			  if(s2.contains(t1)){
				  a.add(t1);
			  }
		  }
		  return new FixedArraySet<Type>(false, a);
	  }

	
	  public static <Type> FixedArraySet<Type> unionOf(final Set<? extends Type> s1, final Set<? extends Type> s2){
		  IArrayQueue<Type> a = new IArrayQueue<Type>(Object.class);
		  a.addAll(s1);
		  for(Type t2: s2){
			  if(!s1.contains(t2)){
				  a.add(t2);
			  }
		  }
		  return new FixedArraySet<Type>(false, a);
	  }
  
	  public static <Type> FixedArraySet<Type> diffOf(final Set<? extends Type> s1, final Set<? extends Type> s2){
		  IArrayQueue<Type> a = new IArrayQueue<Type>(Object.class);
		  for(Type t1: s1){
			  if(!s2.contains(t1)){
				  a.add(t1);
			  }
		  }
		  return new FixedArraySet<Type>(false, a);
	  }
	  
	  @SuppressWarnings("unchecked")
	public static <Type> FixedArrayNavigableSet<Type> intersectionOf(final SortedSet<? extends Type>... sets){
		  IArrayQueue<Type> a = new IArrayQueue<Type>(Object.class);
		  for(Type t1: new OrderedIntersectionIterator<Type>(new FixedArray<SortedSet<Type>>((SortedSet<Type>[])sets)).stream()){
			  a.add(t1);
		  }
		  return new FixedArrayNavigableSet<Type>(false, a);
	  }

	
	  @SuppressWarnings("unchecked")
	public static <Type> FixedArrayNavigableSet<Type> unionOf(final SortedSet<? extends Type>... sets){
		  IArrayQueue<Type> a = new IArrayQueue<Type>(Object.class);
		  for(Type t2: new OrderedUnionIterator<Type>(new FixedArray<SortedSet<Type>>((SortedSet<Type>[])sets)).stream()){
			  a.add(t2);
		  }
		  return new FixedArrayNavigableSet<Type>(false, a);
	  }
  
	  @SuppressWarnings("unchecked")
	public static <Type> FixedArrayNavigableSet<Type> diffOf(final SortedSet<? extends Type> s1, final SortedSet<? extends Type> s2){
		  IArrayQueue<Type> a = new IArrayQueue<Type>(Object.class);
		  for(Type t1: new OrderedDiffIterator<Type>((SortedSet<Type>)s1, (SortedSet<Type>)s2).stream()){
			  a.add(t1);
		  }
		  return new FixedArrayNavigableSet<Type>(false, a);
	  }
	 
}
