package de.lmu.ifi.pms.java.util.map;

import de.lmu.ifi.pms.java.lang.iteration.ExtendibleIterator;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.javaWrapper.FromJavaCollectionWrapper;
import de.lmu.ifi.pms.java.util.collection.set.Set;
import de.lmu.ifi.pms.java.util.map.javaWrapper.FromJavaMapFixedWrapper;

public abstract class AbstractExtendibleMap<KeyType, ValueType, EntryType extends ExtendibleEntry<KeyType, ValueType>> extends
		AbstractModifyableMap<KeyType, ValueType, EntryType> implements ExtendibleMap<KeyType, ValueType, EntryType>{
	
	@Override
	public abstract Set<KeyType> keys();

	@Override
	public abstract EntryType entry(KeyType key);

	@Override
	public abstract ValueType get(KeyType key);

	@Override
	public abstract boolean has(KeyType key);

	@Override
	public abstract Collection<ValueType> values();

	@Override
	public abstract ExtendibleIterator<EntryType> iterator();

	@Override
	public abstract int size();	
	
	@Override
	public void set(final KeyType key, final ValueType value){
		if(this.has(key)){
			this.add(key, value);
		}
		else{
			throw new IllegalArgumentException("Key not yet known");
		}
	}

	@Override
	public abstract void clear();
	
	@Override
	@Deprecated
	public final boolean put(final KeyType key, final ValueType value) {
		return this.add(key, value);
	}
	
	@Override
	public abstract boolean add(KeyType key, ValueType value);
	
	@Override
	public boolean add(final Entry<KeyType, ValueType> entry){
		return this.add(entry.key(), entry.value());
	}
	
	@Override
	public boolean add(final EntryType entry){
		return this.add((Entry<KeyType, ValueType>)entry);
	}
	
	@Override
	public boolean addAll(final java.util.Collection<? extends EntryType> c) {
		return this.addAll(new FromJavaCollectionWrapper<EntryType>(c));
	}
	
	@Override
	public boolean addAll(final Collection<? extends EntryType> c){
		boolean modified = false;
		for(EntryType entry: c){
			if(this.add(entry)){
				modified = true;
			}
		}
		return modified;
	}
	
	@Override
	public boolean add(final java.util.Map<KeyType, ValueType> map) {
		return this.add(new FromJavaMapFixedWrapper<KeyType, ValueType>(map));
	}
	
	@Deprecated
	@Override
	public final boolean putAll(final java.util.Map<KeyType, ValueType> map) {
		return this.add(map);
	}
	
	@Override
	public boolean add(final Map<KeyType, ValueType, ?> map){
		boolean modified = false;
		for(Entry<KeyType, ValueType> entry: map){
			if(this.add(entry)){
				modified = true;
			}
		}
		return modified;
	}
	
	@Deprecated
	@SuppressWarnings("unchecked")
	@Override
	public final boolean remove(final Object o){
		return this.removeKey((KeyType)o);
	}
	
	@SuppressWarnings({ "unchecked", "rawtypes" })
	@Override
	public boolean removeAll(final java.util.Collection<?> c) {
		return this.removeAll(new FromJavaCollectionWrapper(c));
	}
	
	@Override
	public boolean removeAll(final Collection<?> c) {
		boolean modified = false;
		ExtendibleIterator<EntryType> e = this.iterator();
		while (e.hasNext()) {
		    if (c.contains(e.next())) {
			e.remove();
			modified = true;
		    }
		}
		return modified;
	}
	
	@SuppressWarnings({ "unchecked", "rawtypes" })
	@Override
	public boolean retainAll(final java.util.Collection<?> c) {
		return this.retainAll(new FromJavaCollectionWrapper(c));
	}

	@Override
	public boolean retainAll(final Collection<?> c) {
		boolean modified = false;
		ExtendibleIterator<EntryType> e = this.iterator();
		while (e.hasNext()) {
		    if (!c.contains(e.next())) {
			e.remove();
			modified = true;
		    }
		}
		return modified;
	}
	
	@Override
	public abstract boolean remove(Entry<KeyType, ValueType> o);

	
	@Override
	public abstract boolean removeKey(KeyType key);
	

}
