package de.lmu.ifi.pms.java.util.map.fixed;


import java.io.Serializable;
import java.util.Comparator;

import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.adaption.CollectionAdapterImpl;
import de.lmu.ifi.pms.java.util.collection.AbstractCollection;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.array.Arrays;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArraySet;
import de.lmu.ifi.pms.java.util.collection.fixed.set.FixedNavigableSet;
import de.lmu.ifi.pms.java.util.collection.set.Set;
import de.lmu.ifi.pms.java.util.filter.Filter;
import de.lmu.ifi.pms.java.util.map.AbstractNavigableMap.AbstractJavaNavigableMap;
import de.lmu.ifi.pms.java.util.map.Map;
import de.lmu.ifi.pms.java.util.map.SortedMap;
import de.lmu.ifi.pms.java.util.map.javaWrapper.FromJavaMapFixedWrapper;
import de.lmu.ifi.pms.java.util.map.javaWrapper.ToJavaNavigableMapWrapper;


public class FixedArrayNavigableMap<KeyType, ValueType> extends AbstractFixedArrayMap<KeyType, ValueType, FixedArrayNavigableSet<KeyType>> implements FixedNavigableMap<KeyType, ValueType, AbstractFixedEntry<KeyType,ValueType>>,Serializable{
	
	public FixedArrayNavigableMap(){
		this(false);
	}
	
	public FixedArrayNavigableMap(final boolean emptyBody){
		super(emptyBody);
	}
	
	public FixedArrayNavigableMap(final KeyType key, final ValueType value){
		this(new FixedArrayNavigableSet<KeyType>(key), new FixedArray<ValueType>(value));
	}
	
	public FixedArrayNavigableMap(final Map<KeyType, ? extends ValueType, ?> map){
		super(map);
	}
	
	public FixedArrayNavigableMap(final java.util.Map<KeyType, ? extends ValueType> map){
		this(new FromJavaMapFixedWrapper<KeyType, ValueType>(map));
	}
	
	public FixedArrayNavigableMap(final SortedMap<KeyType, ValueType, ?> map){
		super(true);
		this.fillBody(map);
	}
	
	public FixedArrayNavigableMap(final Adapter<ValueType, KeyType> indexer, final Collection<ValueType> values) {
		super(true);
		this.fillBody(new FixedArray<KeyType>(new CollectionAdapterImpl<ValueType, KeyType>(values, indexer)), new FixedArray<ValueType>(values));
	}
	
	public FixedArrayNavigableMap(final Adapter<ValueType, KeyType> indexer, final FixedArray<ValueType> values) {
		super(true);
		this.fillBody(new FixedArray<KeyType>(new CollectionAdapterImpl<ValueType, KeyType>(values, indexer)), values);
	}
	
	public FixedArrayNavigableMap(final AbstractCollection<KeyType> keys, final AbstractCollection<? extends ValueType> values) {
		this(new FixedArray<KeyType>(keys), new FixedArray<ValueType>(values));
		//TODO more efficient
	}
	
	public FixedArrayNavigableMap(final FixedArray<KeyType> keys, final FixedArray<? extends ValueType> values) {
		super(true);
		this.fillBody(keys, values);
	}
	
	public FixedArrayNavigableMap(final FixedArraySet<KeyType> keys, final FixedArray<? extends ValueType> values) {
		this(new FixedArrayMap<KeyType, ValueType>(keys, values));
	}
	
	public FixedArrayNavigableMap(final FixedArrayNavigableSet<KeyType> keys, final FixedArray<? extends ValueType> values) {
		super(true);
		this.fillBody(keys, values);
	}
	
	public FixedArrayNavigableMap(final KeyType[] keys, final ValueType[] values) {
		super(true);
		this.fillBody(true, keys, values);
	}
	/*
	public FixedArrayNavigableMap(final Collection<KeyType> keys, final Collection<ValueType> values) {
		super(true);
		this.fillBody(false, keys.t, values);
	}*/
	
	@Override
	protected FixedArrayNavigableSet<KeyType> newKeySet(final Set<KeyType> keys) {
		return new FixedArrayNavigableSet<KeyType>(keys);
	}
	
	@SuppressWarnings("unchecked")
	public void fillBody(final FixedArray<KeyType> keys, final FixedArray<? extends ValueType> values){
		this.fillBody(false, (KeyType[])keys.toArray(), (ValueType[])values.toArray());
	}
	
	public void fillBody(final boolean copy, KeyType[] keys, final ValueType[] values){
		this.checkFillBody();
		if(copy){
			keys = java.util.Arrays.copyOf(keys, keys.length);
		}
		int[] sourcePositionsNew = new int[keys.length];
		final int length = Arrays.sort(keys, sourcePositionsNew, false);
		if(length < keys.length){
			keys = java.util.Arrays.copyOf(keys, length);
		}
		
		/*if(this.keys.size() < keys.size()){
			throw new IllegalArgumentException("Dublicate key!");
		}*/
		
		@SuppressWarnings("unchecked")
		ValueType[] reordered_values = (ValueType[]) new Object[length];
		for(int i= 0; i < length; ++i)
		{
			reordered_values[i] = values[sourcePositionsNew[i]];
		}
		this.fillBody(new FixedArrayNavigableSet<KeyType>(false, false, keys), new FixedArray<ValueType>(false, reordered_values));
	}
	
	@SuppressWarnings("unchecked")
	@Override
	public void fillBody(final Map<KeyType, ? extends ValueType, ?> map){
		this.checkFillBody();
		if(map instanceof SortedMap<?,?,?>){
			this.fillBody((SortedMap<KeyType, ValueType, ?>) map);
		}
		else
		{
			this.fillBody(false, (KeyType[])map.keys().toArray(), (ValueType[])map.values().toArray());
		}
	}
	
	@SuppressWarnings("unchecked")
	public void fillBody(final SortedMap<KeyType, ValueType, ?> map){
		this.checkFillBody();
		if(map instanceof FixedArrayNavigableMap<?, ?>){
			this.fillBody((FixedArrayNavigableMap<KeyType, ValueType>) map);
		}
		else
		{
			this.keys = new FixedArrayNavigableSet<KeyType>(map.keys());
			this.values = new FixedArray<ValueType>(map.values());
		}
	}
	
	public void fillBody(final FixedArrayNavigableMap<KeyType, ValueType> map){
		this.checkFillBody();
		this.keys = map.keys();
		this.values = map.values();
	}
	
	@Override
	public void fillBody(final FixedArrayNavigableSet<KeyType> keys, final FixedArray<? extends ValueType> values){
		this.checkFillBody();
		if(keys.size() == values.size()){
			super.fillBody(keys, values);
		}
		else{
			throw new IllegalArgumentException("Number of key and values does not match.");//TODO better exception
		}
	}	
	
	@Override
	public Comparator<? super KeyType> comparator() {
		return this.keys.comparator();
	}

	public KeyType ceilingKey(final KeyType key) {
		return this.keys.ceiling(key);
	}

	public KeyType higherKey(final KeyType key) {
		return this.keys.higher(key);
	}

	public KeyType lowerKey(final KeyType key) {
		return this.keys.lower(key);
	}

	@Override
	public KeyType floorKey(final KeyType key) {
		return this.keys.floor(key);
	}

	@Override
	public FixedEntry<KeyType, ValueType> ceilingEntry(final KeyType key) {
		int ceilingIndex = this.keys.ceilingIndex(key);
		return new FixedEntryImpl<KeyType, ValueType>(this.keys.get(ceilingIndex), this.values.get(ceilingIndex));
	}

	@Override
	public FixedEntry<KeyType, ValueType> higherEntry(final KeyType key) {
		int higherIndex = this.keys.higherIndex(key);
		return new FixedEntryImpl<KeyType, ValueType>(this.keys.get(higherIndex), this.values.get(higherIndex));
	}

	@Override
	public FixedEntry<KeyType, ValueType> lowerEntry(final KeyType key) {
		int lowerIndex = this.keys.lowerIndex(key);
		return new FixedEntryImpl<KeyType, ValueType>(this.keys.get(lowerIndex), this.values.get(lowerIndex));
	}

	@Override
	public FixedEntry<KeyType, ValueType> floorEntry(final KeyType key) {
		int floorIndex = this.keys.floorIndex(key);
		return new FixedEntryImpl<KeyType, ValueType>(this.keys.get(floorIndex), this.values.get(floorIndex));
	}

	@Deprecated
	@Override
	public FixedNavigableSet<KeyType> descendingKeySet() {
		return this.keys.descendingSet();
	}

	@Deprecated
	@Override
	public FixedArrayNavigableSet<KeyType> navigableKeySet() {
		return this.keys();
	}

	@Override
	public FixedNavigableMap<KeyType, ValueType, ?> headMap(final KeyType toKey, final boolean inclusive) {
		int endIndex = this.keys.endIndex(toKey, inclusive);
		return new FixedArrayNavigableMap<KeyType, ValueType>(this.keys.subList(0, endIndex), this.values.subList(0, endIndex));
	}

	@Override
	public FixedNavigableMap<KeyType, ValueType, ?> headMap(final KeyType toKey) {
		return this.headMap(toKey, false);
	}

	@Override
	public FixedNavigableMap<KeyType, ValueType, ?> subMap(final KeyType fromKey, final boolean fromInclusive, final KeyType toKey, final boolean toInclusive) {
		int startIndex = this.keys.startIndex(fromKey, fromInclusive);
		int endIndex = this.keys.endIndex(toKey, toInclusive);
		return new FixedArrayNavigableMap<KeyType, ValueType>(this.keys.subList(startIndex, endIndex), this.values.subList(startIndex, endIndex));
	}

	@Override
	public FixedNavigableMap<KeyType, ValueType, ?> subMap(final KeyType fromKey, final KeyType toKey) {
		return this.subMap(fromKey, true, toKey, false);
	}

	@Override
	public FixedNavigableMap<KeyType, ValueType, ?> tailMap(final KeyType fromKey, final boolean inclusive) {
		int startIndex = this.keys.startIndex(fromKey, inclusive);
		return new FixedArrayNavigableMap<KeyType, ValueType>(this.keys.subList(startIndex, this.size()), this.values.subList(startIndex, this.size()));
	}

	@Override
	public FixedNavigableMap<KeyType, ValueType, ?> tailMap(final KeyType fromKey) {
		return this.tailMap(fromKey, true);
	}

	@Override
	public FixedNavigableMap<KeyType, ValueType, ?> descendingMap() {
		return null; //TODO implement
	}

	@Override
	public AbstractJavaNavigableMap<KeyType, ValueType> javaMap() {
		return new ToJavaNavigableMapWrapper<KeyType, ValueType>(this);
	}
	
	public FixedArrayNavigableMap<KeyType, ValueType> filter(final Filter<KeyType> filter){
		//TODO more efficient?
		@SuppressWarnings("unchecked")
		KeyType[] filteredKeys = (KeyType[]) Arrays.getArray(this.keys().ensuredType(), this.keys().size());
		@SuppressWarnings("unchecked")
		ValueType[] filteredValues = (ValueType[]) Arrays.getArray(this.values().ensuredType(), filteredKeys.length);
		int newIndex = 0;
		KeyType element;
		for(int index = 0; index < filteredKeys.length; ++index){
			element = this.keys().get(index);
			if(filter.accepts(element)){
				filteredKeys[newIndex] = element;
				filteredValues[newIndex] = this.values().get(index);
				++newIndex;
			}
		}
		return new FixedArrayNavigableMap<KeyType, ValueType>(new FixedArrayNavigableSet<KeyType>(false, java.util.Arrays.copyOf(filteredKeys, newIndex)), new FixedArray<ValueType>(java.util.Arrays.copyOf(filteredValues, newIndex)));
	}
	
	public static <KeyType, ValueType> FixedArrayNavigableMap<KeyType, ValueType> merge(final FixedArrayNavigableMap<KeyType, ValueType> map1, final FixedArrayNavigableMap<KeyType, ValueType> map2){
		final int length1 = map1.size();
		final int length2 = map2.size();
		int[] sourcePositionsNew = new int[length1 + length2];
		FixedArrayNavigableSet<KeyType> mergedKeys = FixedArrayNavigableSet.merge(map1.keys(), map2.keys(), sourcePositionsNew);
		int newLength = mergedKeys.size();
		
		Class<?> ensuredType;
		if(map1.values().ensuredType().isAssignableFrom(map2.values().ensuredType())){
			ensuredType = map1.values().ensuredType();
		}
		else{
			ensuredType = map2.values().ensuredType();
		}
		
		final FixedArray<ValueType> values1 = map1.values();
		final FixedArray<ValueType> values2 = map2.values();
		@SuppressWarnings("unchecked")
		ValueType[] mergedValues = (ValueType[])Arrays.getArray(ensuredType, newLength);
		for(int entry = 0; entry < newLength; ++entry){
			int source = sourcePositionsNew[entry];
			if(source < length1){
				mergedValues[entry] = values1.get(source);
			}
			else{
				mergedValues[entry] = values2.get(source - length1);
			}
		}
		return new FixedArrayNavigableMap<KeyType, ValueType>(mergedKeys, new FixedArray<ValueType>(false, mergedValues));
	}
	
	public static <KeyType, ValueType> FixedArrayNavigableMap<KeyType, ValueType> merge(FixedArrayNavigableMap<KeyType, ValueType>... maps){
		if(maps.length == 0){
			return new FixedArrayNavigableMap<KeyType, ValueType>();
		}
		else{
			int length = maps.length;
			while(length > 1){//more than one set
				int length_div_2 = length >>> 1; //equivalent to length_div_2 = length div 2
				length = ++length >>> 1; // equivalent to length = (length div 2) + (length mod 2)
				@SuppressWarnings("unchecked")
				FixedArrayNavigableMap<KeyType, ValueType>[] new_sets = new FixedArrayNavigableMap[length];
				for(int i=0; i < length_div_2; ++i){
					new_sets[i] = merge(maps[i<<1],maps[(i<<1) + 1]);	
				}
				if(length_div_2 < length){//odd number of sets (=> maxIndex is even)
					new_sets[length_div_2] = maps[length_div_2 << 1]; 
				}
				maps = new_sets;
			}
			return maps[0];
		}
	}
	
//	public FixedArrayNavigableMap<ValueType, KeyType> invert(){
//		
//	}
	
	@SuppressWarnings("unchecked")
	public static <KeyType, ValueType> FixedArrayNavigableMap<KeyType, ValueType> merge(final Collection<? extends FixedArrayNavigableMap<KeyType, ValueType>> maps){
		FixedArrayNavigableMap<?,?>[] setsArray = new FixedArrayNavigableMap<?,?>[maps.size()];
		maps.toArray((FixedArrayNavigableMap<KeyType, ValueType>[])setsArray);
		return merge((FixedArrayNavigableMap<KeyType, ValueType>[])setsArray);
	}
	

	
	
}
