package de.lmu.ifi.pms.java.util.stream;

import java.io.Closeable;
import java.io.IOException;

import de.lmu.ifi.pms.java.lang.iteration.ExaminingIteratorWrapper;
import de.lmu.ifi.pms.java.util.stream.io.BlockingPipe;


public abstract class StreamSpliter<Type> implements Closeable{
	
	private final SourceStream<Type> stream;
	
	private final ExaminingIteratorWrapper<Type> streamIterator;
	
	private boolean sbReading = false;
	
	private final LocalBlockingPipe positiveStream;
	
	private final LocalBlockingPipe negativeStream;
	
	private class LocalBlockingPipe extends BlockingPipe<Type>{
		
		final boolean positive;
		
		public LocalBlockingPipe(final int bufferCapacity, final boolean positive) {
			super(bufferCapacity);
			this.positive = positive;
		}

		@Override
		protected void waitWhileEmpty() {
			synchronized(this.lock){
				while(this.buffer.isEmpty() && this.isOpen()){
					boolean read = false;
					synchronized (StreamSpliter.this.streamIterator) {
						if(!StreamSpliter.this.sbReading){
							StreamSpliter.this.sbReading = true;
							read = true;
						}
					}
					if(read){
						StreamSpliter.this.tryNext();
						synchronized (StreamSpliter.this.streamIterator) {
							StreamSpliter.this.sbReading = false;
							Object otherLock = (this.positive ? StreamSpliter.this.negativeStream.lock : StreamSpliter.this.positiveStream.lock);
							synchronized (otherLock) {
								otherLock.notifyAll();
							}
						}
					}
					else{
						try {
							this.buffer.wait();
						} catch (InterruptedException e) {

						}
					}
				}
			}
			
			synchronized(StreamSpliter.this.streamIterator){
				while(this.buffer.isEmpty() && this.isOpen()){
					StreamSpliter.this.tryNext();
				}
			}
		}
		
		@Override
		public void writeNext(final Type e){
			synchronized(this.buffer){
				this.waitWhileFull();
				if(this.isOpen()){
					this.buffer.addLast(e);
					this.buffer.notifyAll();
				}
				else{
					
				}
			}
		}
		
	}
	
	
	
	public StreamSpliter(final SourceStream<Type> stream) {
		super();
		this.stream = stream;
		this.streamIterator = new ExaminingIteratorWrapper<Type>(stream.iterator());
		this.positiveStream = new LocalBlockingPipe(256, true);
		this.negativeStream = new LocalBlockingPipe(256, false);
	}

	private void tryNext(){
		if(this.streamIterator.hasNext()){
			if(this.accepts(this.streamIterator.examineNext())){
				this.positiveStream.writeNext(this.streamIterator.next());
			}
			else{
				this.negativeStream.writeNext(this.streamIterator.next());
			}
		}
		else{
			try {
				this.close();
			} catch (final IOException e) {

			}
		}
	}
	
	public abstract boolean accepts(Type element);
	
	public SourceStream<Type> positiveStream(){
		return this.positiveStream;
	}
	
	public SourceStream<Type> negativStream(){
		return this.negativeStream;
	}
	
	@Override
	public void close() throws IOException {
		 this.positiveStream.close();
		 this.negativeStream.close();
		 this.stream.close();
	}

}
