package de.lmu.ifi.pms.java.util.stream.io;

import java.io.Closeable;
import java.io.IOException;
import java.io.InputStream;
import java.util.NoSuchElementException;

import de.lmu.ifi.pms.java.util.tupel.Single;



public class DistributedObjectInputStream<Type> extends ObjectInputStream<Type, Single<Closeable,java.io.ObjectInputStream>>{
	
	private final java.io.ObjectInputStream inputStream;
	
	private final Class<Type> type;
	
	private Type next = null;
	
	public DistributedObjectInputStream(final InputStream inputStream) throws IOException{
		this(new java.io.ObjectInputStream(inputStream));
	}
	
	public DistributedObjectInputStream(final Class<Type> type, final InputStream inputStream) throws IOException{
		this(type,new java.io.ObjectInputStream(inputStream));
	}
	
	@SuppressWarnings("unchecked")
	private DistributedObjectInputStream(final java.io.ObjectInputStream inputStream){
		this((Class<Type>)Object.class, inputStream);
	}
	
	public DistributedObjectInputStream(final Class<Type> type, final java.io.ObjectInputStream inputStream){
		super();
		this.inputStream = inputStream;
		this.type = type;
	}
	
	@Override
	public Single<Closeable,java.io.ObjectInputStream> resources() {
		return new Single<Closeable,java.io.ObjectInputStream>(this.inputStream);
	}
	
	@Override
	public Object closeLock() {
		return this;
	}
	
	@Override
	public Type readObject(){
		if(this.isOpen()){
			synchronized (this.inputStream) {
				if(this.hasObject()){
					final Type result = this.next;
					this.next = null;
					return result;
				}
				else{
					throw new NoSuchElementException();//TODO better exception;
				}
			}
		}
		else{
			//TODO better Exception
			throw new IllegalStateException();
		}
	}
	
	@Override
	public boolean hasObject(){
		synchronized (this.inputStream) {
			if(this.next == null){
				try {
					this.next = this.type.cast(this.inputStream.readObject());
				} catch (final IOException e) {
					if(this.isOpen()){
						// TODO ExceptionHandling
						throw new RuntimeException(e);
					}
				} catch (final ClassNotFoundException e) {
					if(this.isOpen()){
						// TODO ExceptionHandling
						throw new RuntimeException(e);
					}
				}
			}
		}
		return this.isOpen();
	}
	
}
