package de.lmu.ifi.pms.cep.unitTests;


import java.io.BufferedReader;
import java.io.InputStreamReader;

import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

import de.lmu.ifi.pms.cep.event_mill.console.MainThread;
import de.lmu.ifi.pms.cep.event_mill.console.StartThread;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.java.util.stream.io.BlockingPipe;


public class Engine_Test
{

	static final String prog_namespace = "testns";
	static final String prog_name = "test";
	static final String prog_instance = "first";
	static final String file = "src/unitTests/DuraC.example";
	
	static final String meta_schema = "test_meta";
	static final String input_schema = "test_input";
	static final String working_schema = "test_working";
	static final String output_schema = "test_output";
	static final String log_schema = "test_log";
		
	static BlockingPipe<String> commandPipe = new BlockingPipe<String>(64);
	
	static MainThread mainThread;
	
	@BeforeClass
	public static void setUpBeforeClass() throws Exception
	{
		StartThread startThread = new StartThread(Engine_Test.commandPipe.iterator(), new BufferedReader(new InputStreamReader(System.in)), System.out);
		startThread.start();
		Engine_Test.mainThread = startThread.getMainThread();
		Engine_Test.commandPipe.writeNext("clear progs");
		Engine_Test.commandPipe.writeNext("clear specs");
	}

	@AfterClass
	public static void tearDownAfterClass() throws Exception
	{
		Engine_Test.commandPipe.writeNext("quit");
		Engine_Test.commandPipe.writeNext("quit");
		Engine_Test.mainThread.join();
		Engine_Test.commandPipe.close();
	}

	@Before
	public void setUp() throws Exception
	{

	}
	
	@Test
	public void test1() throws Exception{
		this.compile();
		this.init();
		this.run();
		this.start();
		this.stop();
		this.quit();
	}
	
	public void compile() throws Exception{
		Engine_Test.commandPipe.writeNext(String.format("compile %s, %s", TSA_SyntaxPatterns.programFullName(Engine_Test.prog_namespace, Engine_Test.prog_name), Engine_Test.file));
	}
	
	public void init() throws Exception{
		Engine_Test.commandPipe.writeNext(String.format("init %s, %s, %s, %s, %s, %s", TSA_SyntaxPatterns.programInstFullName(Engine_Test.prog_namespace, Engine_Test.prog_name, Engine_Test.prog_instance), Engine_Test.meta_schema, Engine_Test.input_schema, Engine_Test.working_schema, Engine_Test.output_schema, Engine_Test.log_schema));
	}
	
	public void run() throws Exception{
		Engine_Test.commandPipe.writeNext(String.format("run %s", TSA_SyntaxPatterns.programInstFullName(Engine_Test.prog_namespace, Engine_Test.prog_name, Engine_Test.prog_instance)));
	}
	
	public void start() throws Exception{
		Engine_Test.commandPipe.writeNext("start");
		Thread.sleep(5000);
	}
	
	public void stop() throws Exception{
		Engine_Test.commandPipe.writeNext("stop");
	}
	
	public void quit() throws Exception{
		Engine_Test.commandPipe.writeNext("quit");
	}

	@After
	public void tearDown() throws Exception
	{
		
	}
	
	public static void main(final String[] args) throws Exception{
		setUpBeforeClass();
		new Engine_Test().test1();
		tearDownAfterClass();
	}
	
}
