package de.lmu.ifi.pms.cep.unitTests.time.constraints.tests;

import static org.junit.Assert.assertEquals;

import org.junit.Before;
import org.junit.Test;

import de.lmu.ifi.pms.cep.tsa.time.constraints.ConjunctiveTemporalRelationFormula;
import de.lmu.ifi.pms.cep.tsa.time.constraints.DisjunctiveTemporalRelationFormula;
import de.lmu.ifi.pms.cep.tsa.time.constraints.NegatedTemporalRelationFormula;
import de.lmu.ifi.pms.cep.tsa.time.constraints.SimpleTemporalFormula;
import de.lmu.ifi.pms.cep.tsa.time.constraints.TemporalMinMax;
import de.lmu.ifi.pms.cep.tsa.time.constraints.TemporalRelationFormula;
import de.lmu.ifi.pms.cep.tsa.time.constraints.enums.BinaryComparisonOperator;
import de.lmu.ifi.pms.cep.tsa.time.constraints.enums.MinMax;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;


public class TestNegation {
	SimpleTemporalFormula st1;
	SimpleTemporalFormula st2;
	SimpleTemporalFormula st3;
	SimpleTemporalFormula st4;
	SimpleTemporalFormula st5;
	SimpleTemporalFormula st6;
	SimpleTemporalFormula st7;
	SimpleTemporalFormula st8;
	SimpleTemporalFormula st9;
	TemporalRelationFormula td1;
	TemporalRelationFormula td2;
	TemporalRelationFormula mm;
	
	@Before
	public void init(){
		this.st1 = new SimpleTemporalFormula("v", BinaryComparisonOperator.LESS, "w",0);
		this.st2 = new SimpleTemporalFormula("x", BinaryComparisonOperator.LESS_EQUAL, "y",0);
		this.st3 = new SimpleTemporalFormula("m", BinaryComparisonOperator.EQUAL, "n",0);
		this.st4 = new SimpleTemporalFormula("v1", BinaryComparisonOperator.LESS, "w1",10);
		this.st5 = new SimpleTemporalFormula("x1", BinaryComparisonOperator.LESS_EQUAL, "y1",10);
		this.st6 = new SimpleTemporalFormula("m1", BinaryComparisonOperator.EQUAL, "n1",10);
		this.st7 = new SimpleTemporalFormula("v2", BinaryComparisonOperator.LESS, "w2",-10);
		this.st8 = new SimpleTemporalFormula("x2", BinaryComparisonOperator.LESS_EQUAL, "y2",-10);
		this.st9 = new SimpleTemporalFormula("m2", BinaryComparisonOperator.EQUAL, "n2",-10);
		this.mm = new TemporalMinMax("v",MinMax.MAX,"a","b");
		ArrayList<TemporalRelationFormula> trf1 = new ArrayList<TemporalRelationFormula>();
		ArrayList<TemporalRelationFormula> trf2 = new ArrayList<TemporalRelationFormula>();
		trf1.add(this.st1);
		trf1.add(this.st2);
		trf1.add(this.st3);
		trf2.add(this.st4);
		trf2.add(this.st5);
		trf2.add(this.st6);
		this.td1 = new DisjunctiveTemporalRelationFormula(trf1);
		this.td2 = new ConjunctiveTemporalRelationFormula(trf2);
		
	}
	
	@Test
	public void testDisjunctiveNegation(){
		TemporalRelationFormula neg = new NegatedTemporalRelationFormula(this.td1);
		assertEquals("(w <= v AND y <= x - e AND (m <= n - e OR n <= m - e))",neg.simpleForm().toString());
		TemporalRelationFormula neg2 = new NegatedTemporalRelationFormula(this.td2);
		assertEquals("(w1 <= v1-10 OR y1 <= x1-10 - e OR (m1 <= n1+10 - e OR n1 <= m1-10 - e))",neg2.simpleForm().toString());
		TemporalRelationFormula neg3 = new NegatedTemporalRelationFormula(new NegatedTemporalRelationFormula(this.td2));
		assertEquals("(v1 <= w1+10 - e AND x1 <= y1+10 AND (m1 <= n1+10 AND n1 <= m1-10))",neg3.simpleForm().toString());
		TemporalRelationFormula neg4 = new NegatedTemporalRelationFormula(new ConjunctiveTemporalRelationFormula(this.td1,this.td2));
		assertEquals("((w <= v AND y <= x - e AND (m <= n - e OR n <= m - e)) OR (w1 <= v1-10 OR y1 <= x1-10 - e OR (m1 <= n1+10 - e OR n1 <= m1-10 - e)))",neg4.simpleForm().toString());
		TemporalRelationFormula neg5 = new NegatedTemporalRelationFormula(this.mm);
		assertEquals("((v <= a - e OR v <= b - e) OR (a <= v - e AND b <= v - e))",neg5.simpleForm().toString());
	}
	
	
}
